/*
 * The Mops, a free 3D Tcl/Tk/OpenGL/RenderMan modeler
 *
 * The Mops is copyrighted (C) 1998-1999 by Randolf Schultz
 * (rschultz@informatik.uni-rostock.de)
 *
 * See the file License for details.
 *
 */

#include <limits.h>
#include <float.h>
#include "mops.h"

/* Bounding Box Calculation */

/* functions local to this module */
int mops_bb_light(mops_object *o, double *bbox);
int mops_bb_sphere(mops_sphere_object *o, double *bbox);
int mops_bb_box(mops_box_object *o, double *bbox);
int mops_bb_cone(mops_cone_object *o, double *bbox);
int mops_bb_disk(mops_disk_object *o, double *bbox);
int mops_bb_cylinder(mops_cylinder_object *o, double *bbox);
int mops_bb_paraboloid(mops_paraboloid_object *o, double *bbox);
int mops_bb_hyperboloid(mops_hyperboloid_object *o, double *bbox);
int mops_bb_torus(mops_torus_object *o, double *bbox);
int mops_bb_nurbcurve(mops_nurbcurve_object *o, double *bbox);
int mops_bb_nurbpatch(mops_nurbpatch_object *o, double *bbox);
int mops_bb_level(mops_object *o, double *bbox);

/* functions */
int
mops_get_bb(mops_object *o, double *bbox)
{
 int mops_status = MOPS_OK;
 double bb[24] = {0};
 int i, a;
 GLdouble m[16] = {0};
 GLfloat mr[4][4];
 mops_list_object *cl = mops_current_level;
 mops_custom_object *c = NULL;
 mops_object *p = NULL;

  if(!o)
    return MOPS_OK;

  /*
    when adding elements to this switch, do the same with
    the switch in the function below (mops_get_sub_bb())!
  */
  switch(o->type)
    {
    case MOPS_OTSPHERE:
      mops_status = mops_bb_sphere((mops_sphere_object*)(o->object), bb);
      break;
    case MOPS_OTBOX:
      mops_status = mops_bb_box((mops_box_object*)(o->object), bb);
      break;
    case MOPS_OTCONE:
      mops_status = mops_bb_cone((mops_cone_object*)(o->object), bb);
      break;
    case MOPS_OTCYLINDER:
      mops_status = mops_bb_cylinder((mops_cylinder_object*)(o->object), bb);
      break;
    case MOPS_OTDISK:
      mops_status = mops_bb_disk((mops_disk_object*)(o->object), bb);
      break;
    case MOPS_OTPARABOLOID:
      mops_status = mops_bb_paraboloid((mops_paraboloid_object*)(o->object),
				       bb);
      break;
    case MOPS_OTHYPERBOLOID:
      mops_status = mops_bb_hyperboloid((mops_hyperboloid_object*)(o->object),
					bb);
      break;
    case MOPS_OTTORUS:
      mops_status = mops_bb_torus((mops_torus_object*)(o->object), bb);
      break;
    case MOPS_OTNURBCURVE:
      mops_status = mops_bb_nurbcurve((mops_nurbcurve_object*)(o->object), bb);
      break;
    case MOPS_OTNURBPATCH:
      mops_status = mops_bb_nurbpatch((mops_nurbpatch_object*)(o->object), bb);
      break;
    case MOPS_OTLEVEL:
      mops_status = mops_bb_level(o, bb);
      break;
    case MOPS_OTLIGHT:
      mops_status = mops_bb_light(o, bb);
      break;
    case MOPS_OTCUSTOM:
      mops_status = mops_bb_custom(o, bb);

      if(mops_status == MOPS_HASH)
	{
	  mops_status = MOPS_OK;

	  c = (mops_custom_object *)o->object;
	  if(c->provides)
	    {
	      mops_get_object(o, c->provides, &p);

	      mops_get_bb(p, bb);

	      mops_delete_multi_objects(p);
	      memcpy(bbox, bb, 24*sizeof(double));
	      return mops_status; /* XXXX early exit! */
	    }

	}

      break;
    case MOPS_OTINSTANCE:
      mops_get_object(o, ((mops_object*)o->object)->type, &p);
      mops_get_bb(p, bb);
      mops_delete_object(p);
      memcpy(bbox, bb, 24*sizeof(double));
      return mops_status; /* XXXX early exit! */
      break;
   } /* switch */

  /* apply transformation */
  glMatrixMode (GL_MODELVIEW);
   glPushMatrix();
   glLoadIdentity();

   if(cl)
     mops_get_all_transform(cl->next);

   glTranslated(o->movx, o->movy, o->movz);
	
   mops_build_rotmatrix(mr, o->quat);
   glMultMatrixf(&mr[0][0]);
  
   glScaled (o->scalx, o->scaly, o->scalz);
   glGetDoublev(GL_MODELVIEW_MATRIX, m);
  glPopMatrix();

  a = 0;
  for(i = 0; i < 8; i++)
    {
      mops_apply_transform3(&(bb[a]), m);
      a += 3;
    }

  memcpy(bbox, bb, 24*sizeof(double));

 return mops_status;
}


int
mops_get_sub_bb(mops_object *o, double *bbox)
{
 int mops_status = MOPS_OK;
 double bb[24] = {0};
 int i, a;
 GLdouble m[16] = {0};
 GLfloat mr[4][4];
 mops_custom_object *c = NULL;
 mops_object *p = NULL;

  if(!o)
    return MOPS_OK;

  switch(o->type)
    {
    case MOPS_OTSPHERE:
      mops_status = mops_bb_sphere((mops_sphere_object*)(o->object), bb);
      break;
    case MOPS_OTBOX:
      mops_status = mops_bb_box((mops_box_object*)(o->object), bb);
      break;
    case MOPS_OTCONE:
      mops_status = mops_bb_cone((mops_cone_object*)(o->object), bb);
      break;
    case MOPS_OTCYLINDER:
      mops_status = mops_bb_cylinder((mops_cylinder_object*)(o->object), bb);
      break;
    case MOPS_OTDISK:
      mops_status = mops_bb_disk((mops_disk_object*)(o->object), bb);
      break;
    case MOPS_OTPARABOLOID:
      mops_status = mops_bb_paraboloid((mops_paraboloid_object*)(o->object),
				       bb);
      break;
    case MOPS_OTHYPERBOLOID:
      mops_status = mops_bb_hyperboloid((mops_hyperboloid_object*)(o->object),
					bb);
      break;
    case MOPS_OTTORUS:
      mops_status = mops_bb_torus((mops_torus_object*)(o->object), bb);
      break;
    case MOPS_OTNURBCURVE:
      mops_status = mops_bb_nurbcurve((mops_nurbcurve_object*)(o->object), bb);
      break;
    case MOPS_OTNURBPATCH:
      mops_status = mops_bb_nurbpatch((mops_nurbpatch_object*)(o->object), bb);
      break;
    case MOPS_OTLEVEL:
      mops_status = mops_bb_level(o, bb);
      break;
    case MOPS_OTLIGHT:
      mops_status = mops_bb_light(o, bb);
      break;
    case MOPS_OTCUSTOM:
      mops_status = mops_bb_custom(o, bb);

      if(mops_status == MOPS_HASH)
	{
	  mops_status = MOPS_OK;

	  c = (mops_custom_object *)o->object;
	  if(c->provides)
	    {
	      mops_get_object(o, c->provides, &p);

	      mops_get_bb(p, bb);

	      mops_delete_multi_objects(p);
	      memcpy(bbox, bb, 24*sizeof(double));
	      return mops_status; /* XXXX early exit! */
	    }

	}

      break;
    case MOPS_OTINSTANCE:
      mops_get_object(o, ((mops_object*)o->object)->type, &p);
      mops_get_bb(p, bb);
      mops_delete_object(p);
      memcpy(bbox, bb, 24*sizeof(double));
      return mops_status; /* XXXX early exit! */
      break;
   } /* switch */

  /* apply transformation */
  glMatrixMode (GL_MODELVIEW);
   glPushMatrix();
   glLoadIdentity();

   glTranslated(o->movx, o->movy, o->movz);
	
   mops_build_rotmatrix(mr, o->quat);
   glMultMatrixf(&mr[0][0]);
  
   glScaled (o->scalx, o->scaly, o->scalz);
   glGetDoublev(GL_MODELVIEW_MATRIX, m);
  glPopMatrix();

  a = 0;
  for(i = 0; i < 8; i++)
    {
      mops_apply_transform3(&(bb[a]), m);
      a += 3;
    }

  memcpy(bbox, bb, 24*sizeof(double));

 return mops_status;
}


int
mops_bb_sphere(mops_sphere_object *s, double *bbox)
{
 double r = 0.0, zmi = 0.0, zma = 0.0;
  if(!s)
    return MOPS_OK;

  r = s->radius;
  zmi = s->zmin;
  zma = s->zmax;

  /* XXXX does not take into account ThetaMax! */

  /* P1 */
  bbox[0] = -r; bbox[1] = r; bbox[2] = zma;
  /* P2 */
  bbox[3] = -r; bbox[4] = -r; bbox[5] = zma;
  /* P3 */
  bbox[6] = r; bbox[7] = -r; bbox[8] = zma;
  /* P4 */
  bbox[9] = r; bbox[10] = r; bbox[11] = zma;

  /* P5 */
  bbox[12] = -r; bbox[13] = r; bbox[14] = zmi;
  /* P6 */
  bbox[15] = -r; bbox[16] = -r; bbox[17] = zmi;
  /* P7 */
  bbox[18] = r; bbox[19] = -r; bbox[20] = zmi;
  /* P8 */
  bbox[21] = r; bbox[22] = r; bbox[23] = zmi;

 return MOPS_OK;
}

int
mops_bb_box(mops_box_object *b, double *bbox)
{
 double wh = 0.0, lh = 0.0, hh = 0.0;

  if(!b)
    return MOPS_OK;

  wh = b->width/2.0;
  lh = b->length/2.0;
  hh = b->height/2.0;

  /* P1 */
  bbox[0] = -wh; bbox[1] = hh; bbox[2] = lh;
  /* P2 */
  bbox[3] = -wh; bbox[4] = hh; bbox[5] = -lh;
  /* P3 */
  bbox[6] = wh; bbox[7] = hh; bbox[8] = -lh;
  /* P4 */
  bbox[9] = wh; bbox[10] = hh; bbox[11] = lh;

  /* P5 */
  bbox[12] = -wh; bbox[13] = -hh; bbox[14] = lh;
  /* P6 */
  bbox[15] = -wh; bbox[16] = -hh; bbox[17] = -lh;
  /* P7 */
  bbox[18] = wh; bbox[19] = -hh; bbox[20] = -lh;
  /* P8 */
  bbox[21] = wh; bbox[22] = -hh; bbox[23] = lh;

 return MOPS_OK;
}

int
mops_bb_cone(mops_cone_object *c, double *bbox)
{
 double r = 0.0, h = 0.0;

  if(!c)
    return MOPS_OK;

  r = c->radius;
  h = c->height;

  /* XXXX does not take into account ThetaMax! */

  /* P1 */
  bbox[0] = -r; bbox[1] = r; bbox[2] = h;
  /* P2 */
  bbox[3] = -r; bbox[4] = -r; bbox[5] = h;
  /* P3 */
  bbox[6] = r; bbox[7] = -r; bbox[8] = h;
  /* P4 */
  bbox[9] = r; bbox[10] = r; bbox[11] = h;

  /* P5 */
  bbox[12] = -r; bbox[13] = r; bbox[14] = 0.0;
  /* P6 */
  bbox[15] = -r; bbox[16] = -r; bbox[17] = 0.0;
  /* P7 */
  bbox[18] = r; bbox[19] = -r; bbox[20] = 0.0;
  /* P8 */
  bbox[21] = r; bbox[22] = r; bbox[23] = 0.0;

 return MOPS_OK;
}


int
mops_bb_disk(mops_disk_object *d, double *bbox)
{
 double r = 0.0, h = 0.0;

  if(!d)
    return MOPS_OK;

  r = d->radius;
  h = d->height;

  /* XXXX does not take into account ThetaMax! */

  /* P1 */
  bbox[0] = -r; bbox[1] = r; bbox[2] = h;
  /* P2 */
  bbox[3] = -r; bbox[4] = -r; bbox[5] = h;
  /* P3 */
  bbox[6] = r; bbox[7] = -r; bbox[8] = h;
  /* P4 */
  bbox[9] = r; bbox[10] = r; bbox[11] = h;

  /* P5 */
  bbox[12] = -r; bbox[13] = r; bbox[14] = h;
  /* P6 */
  bbox[15] = -r; bbox[16] = -r; bbox[17] = h;
  /* P7 */
  bbox[18] = r; bbox[19] = -r; bbox[20] = h;
  /* P8 */
  bbox[21] = r; bbox[22] = r; bbox[23] = h;

 return MOPS_OK;
}


int
mops_bb_cylinder(mops_cylinder_object *c, double *bbox)
{
 double r = 0.0, zmi = 0.0, zma = 0.0;

  if(!c)
    return MOPS_OK;

  r = c->radius;
  zmi = c->zmin;
  zma = c->zmax;

  /* XXXX does not take into account ThetaMax! */

  /* P1 */
  bbox[0] = -r; bbox[1] = r; bbox[2] = zma;
  /* P2 */
  bbox[3] = -r; bbox[4] = -r; bbox[5] = zma;
  /* P3 */
  bbox[6] = r; bbox[7] = -r; bbox[8] = zma;
  /* P4 */
  bbox[9] = r; bbox[10] = r; bbox[11] = zma;

  /* P5 */
  bbox[12] = -r; bbox[13] = r; bbox[14] = zmi;
  /* P6 */
  bbox[15] = -r; bbox[16] = -r; bbox[17] = zmi;
  /* P7 */
  bbox[18] = r; bbox[19] = -r; bbox[20] = zmi;
  /* P8 */
  bbox[21] = r; bbox[22] = r; bbox[23] = zmi;

 return MOPS_OK;
}


int
mops_bb_hyperboloid(mops_hyperboloid_object *h, double *bbox)
{
 double r = 0.0, rmi = 0.0, rma = 0.0, zmi = 0.0, zma = 0.0;

  if(!h)
    return MOPS_OK;

  rmi = sqrt((h->p1x*h->p1x)+(h->p1z*h->p1z));
  rma = sqrt((h->p2x*h->p2x)+(h->p2z*h->p2z));
  zmi = h->p1y;
  zma = h->p2y;

  r = rmi>rma?rmi:rma;

  /* XXXX does not take into account ThetaMax! */

  /* P1 */
  bbox[0] = -r; bbox[1] = r; bbox[2] = zma;
  /* P2 */
  bbox[3] = -r; bbox[4] = -r; bbox[5] = zma;
  /* P3 */
  bbox[6] = r; bbox[7] = -r; bbox[8] = zma;
  /* P4 */
  bbox[9] = r; bbox[10] = r; bbox[11] = zma;

  /* P5 */
  bbox[12] = -r; bbox[13] = r; bbox[14] = zmi;
  /* P6 */
  bbox[15] = -r; bbox[16] = -r; bbox[17] = zmi;
  /* P7 */
  bbox[18] = r; bbox[19] = -r; bbox[20] = zmi;
  /* P8 */
  bbox[21] = r; bbox[22] = r; bbox[23] = zmi;

 return MOPS_OK;
}


int
mops_bb_paraboloid(mops_paraboloid_object *p, double *bbox)
{
 double r = 0.0, zmi = 0.0, zma = 0.0;

  if(!p)
    return MOPS_OK;

  r = p->rmax;
  zmi = p->zmin;
  zma = p->zmax;

  /* XXXX does not take into account ThetaMax! */

  /* P1 */
  bbox[0] = -r; bbox[1] = r; bbox[2] = zma;
  /* P2 */
  bbox[3] = -r; bbox[4] = -r; bbox[5] = zma;
  /* P3 */
  bbox[6] = r; bbox[7] = -r; bbox[8] = zma;
  /* P4 */
  bbox[9] = r; bbox[10] = r; bbox[11] = zma;

  /* P5 */
  bbox[12] = -r; bbox[13] = r; bbox[14] = zmi;
  /* P6 */
  bbox[15] = -r; bbox[16] = -r; bbox[17] = zmi;
  /* P7 */
  bbox[18] = r; bbox[19] = -r; bbox[20] = zmi;
  /* P8 */
  bbox[21] = r; bbox[22] = r; bbox[23] = zmi;

 return MOPS_OK;
}


int
mops_bb_torus(mops_torus_object *t, double *bbox)
{
 double r = 0.0, zmi = 0.0, zma = 0.0;

  if(!t)
    return MOPS_OK;

  r = t->majorrad+t->minorrad;
  zmi = -(t->minorrad);
  zma = t->minorrad;

  /* XXXX does not take into account ThetaMax, PhiMin and PhiMax! */

  /* P1 */
  bbox[0] = -r; bbox[1] = r; bbox[2] = zma;
  /* P2 */
  bbox[3] = -r; bbox[4] = -r; bbox[5] = zma;
  /* P3 */
  bbox[6] = r; bbox[7] = -r; bbox[8] = zma;
  /* P4 */
  bbox[9] = r; bbox[10] = r; bbox[11] = zma;

  /* P5 */
  bbox[12] = -r; bbox[13] = r; bbox[14] = zmi;
  /* P6 */
  bbox[15] = -r; bbox[16] = -r; bbox[17] = zmi;
  /* P7 */
  bbox[18] = r; bbox[19] = -r; bbox[20] = zmi;
  /* P8 */
  bbox[21] = r; bbox[22] = r; bbox[23] = zmi;

 return MOPS_OK;
}


int
mops_bb_nurbcurve(mops_nurbcurve_object *c, double *bbox)
{
 double xmin, xmax, ymin, ymax, zmin, zmax;
 double *controlv = NULL;
 int stride, i, a;

  if(!c)
    return MOPS_OK;

  controlv = c->controlv;

  xmin = controlv[0];
  xmax = xmin;
  ymin = controlv[1];
  ymax = ymin;
  zmin = controlv[2];
  zmax = zmin;

  stride = c->has_weights?4:3;

  a = 0;
  for(i = 0; i <= c->length; i++)
    {
      if(controlv[a] < xmin)
	xmin = controlv[a];
      if(controlv[a] > xmax)
	xmax = controlv[a];

      if(controlv[a+1] < ymin)
	ymin = controlv[a+1];
      if(controlv[a+1] > ymax)
	ymax = controlv[a+1];

      if(controlv[a+2] < zmin)
	zmin = controlv[a+2];
      if(controlv[a+2] > zmax)
	zmax = controlv[a+2];

      a += stride;
    }

  /* P1 */
  bbox[0] = xmin; bbox[1] = ymax; bbox[2] = zmax;
  /* P2 */
  bbox[3] = xmin; bbox[4] = ymax; bbox[5] = zmin;
  /* P3 */
  bbox[6] = xmax; bbox[7] = ymax; bbox[8] = zmin;
  /* P4 */
  bbox[9] = xmax; bbox[10] = ymax; bbox[11] = zmax;

  /* P5 */
  bbox[12] = xmin; bbox[13] = ymin; bbox[14] = zmax;
  /* P6 */
  bbox[15] = xmin; bbox[16] = ymin; bbox[17] = zmin;
  /* P7 */
  bbox[18] = xmax; bbox[19] = ymin; bbox[20] = zmin;
  /* P8 */
  bbox[21] = xmax; bbox[22] = ymin; bbox[23] = zmax;

 return MOPS_OK;
}


int
mops_bb_nurbpatch(mops_nurbpatch_object *p, double *bbox)
{
 double xmin, xmax, ymin, ymax, zmin, zmax;
 double *controlv = NULL;
 int stride, i, a;

  if(!p)
    return MOPS_OK;

  controlv = p->controlv;

  xmin = controlv[0];
  xmax = xmin;
  ymin = controlv[1];
  ymax = ymin;
  zmin = controlv[2];
  zmax = zmin;

  stride = p->has_weights?4:3;

  a = 0;
  for(i = 0; i < (p->width*p->height); i++)
    {
      if(controlv[a] < xmin)
	xmin = controlv[a];
      if(controlv[a] > xmax)
	xmax = controlv[a];

      if(controlv[a+1] < ymin)
	ymin = controlv[a+1];
      if(controlv[a+1] > ymax)
	ymax = controlv[a+1];

      if(controlv[a+2] < zmin)
	zmin = controlv[a+2];
      if(controlv[a+2] > zmax)
	zmax = controlv[a+2];

      a += stride;
    }

  /* P1 */
  bbox[0] = xmin; bbox[1] = ymax; bbox[2] = zmax;
  /* P2 */
  bbox[3] = xmin; bbox[4] = ymax; bbox[5] = zmin;
  /* P3 */
  bbox[6] = xmax; bbox[7] = ymax; bbox[8] = zmin;
  /* P4 */
  bbox[9] = xmax; bbox[10] = ymax; bbox[11] = zmax;

  /* P5 */
  bbox[12] = xmin; bbox[13] = ymin; bbox[14] = zmax;
  /* P6 */
  bbox[15] = xmin; bbox[16] = ymin; bbox[17] = zmin;
  /* P7 */
  bbox[18] = xmax; bbox[19] = ymin; bbox[20] = zmin;
  /* P8 */
  bbox[21] = xmax; bbox[22] = ymin; bbox[23] = zmax;

 return MOPS_OK;
}

int
mops_bb_light(mops_object *o, double *bbox)
{
 int mops_status = MOPS_OK;
 mops_object *d = NULL;
 mops_light_object *l = NULL;
 mops_shader *shader = NULL;
 mops_shader_arg *sarg = NULL;
 double xmin = DBL_MAX, xmax = -DBL_MAX, ymin = DBL_MAX;
 double ymax = -DBL_MAX, zmin = DBL_MAX, zmax = -DBL_MAX;
 double fromx = 0.0, fromy = 0.0, fromz = 0.0;
 double tox = 0.0, toy = 0.0, toz = 0.0;
 double bbt[24] = {0};
 int i, a, has_from = MOPS_FALSE, has_to = MOPS_FALSE;
 
  if(!o)
    return MOPS_OK;


  l = (mops_light_object*)(o->object);
  switch(l->type)
    {
    case MOPS_LITPOINT:
      has_from = MOPS_TRUE;
      fromx = 0.0;
      fromy = 0.0;
      fromz = 0.0;
      break;
    case MOPS_LITSPOT:
      has_from = MOPS_TRUE;
      fromx = 0.0;
      fromy = 0.0;
      fromz = 0.0;
      has_to = MOPS_TRUE;
      tox = 0.0;
      toy = 0.0;
      toz = -10.0;
      break;
    case MOPS_LITDIST:
      has_from = MOPS_TRUE;
      fromx = 0.0;
      fromy = 0.0;
      fromz = 0.0;
      has_to = MOPS_TRUE;
      tox = 0.0;
      toy = 0.0;
      toz = -10.0;
      break;
    case MOPS_LITCUSTOM:
      shader = o->shaders;
      while(shader)
	{
	  if(shader->type == MOPS_STLIGHT)
	    {
	      sarg = shader->arg;
	      while(sarg)
		{
		  if(!strcmp(sarg->name,"from"))
		    {
		      has_from = MOPS_TRUE;
		      fromx = sarg->val.point[0];
		      fromy = sarg->val.point[1];
		      fromz = sarg->val.point[2];
		    }
		  if(!strcmp(sarg->name,"to"))
		    {
		      has_to = MOPS_TRUE;
		      tox = sarg->val.point[0];
		      toy = sarg->val.point[1];
		      toz = sarg->val.point[2];

		    }
		  sarg = sarg->next;	      
		}/* while */
	    } /* if */
	  shader = shader->next;
	} /* while */
      break;

    } /* switch */

  if(has_from)
    {
      if(fromx < xmin)
	xmin = fromx;
      if(fromx > xmax)
	xmax = fromx;

      if(fromy < ymin)
	ymin = fromy;
      if(fromy > ymax)
	ymax = fromy;

      if(fromz < zmin)
	zmin = fromz;
      if(fromz > zmax)
	zmax = fromz;
    }

  if(has_to)
    {
      if(tox < xmin)
	xmin = tox;
      if(tox > xmax)
	xmax = tox;

      if(toy < ymin)
	ymin = toy;
      if(toy > ymax)
	ymax = toy;

      if(toz < zmin)
	zmin = toz;
      if(toz > zmax)
	zmax = toz;
    }

  d = o->down;
  while(d)
    {
      mops_status = mops_get_sub_bb(d, bbt);

      a = 0;
      for(i = 0; i < 8; i++)
	{
	  if(bbt[a] < xmin)
	    xmin = bbt[a];
	  if(bbt[a] > xmax)
	    xmax = bbt[a];
	  a += 3;
	}

      a = 1;
      for(i = 0; i < 8; i++)
	{
	  if(bbt[a] < ymin)
	    ymin = bbt[a];
	  if(bbt[a] > ymax)
	    ymax = bbt[a];
	  a += 3;
	}

      a = 2;
      for(i = 0; i < 8; i++)
	{
	  if(bbt[a] < zmin)
	    zmin = bbt[a];
	  if(bbt[a] > zmax)
	    zmax = bbt[a];
	  a += 3;
	}

      d = d->next;
    } /* while */


  /* P1 */
  bbox[0] = xmin; bbox[1] = ymax; bbox[2] = zmax;
  /* P2 */
  bbox[3] = xmin; bbox[4] = ymax; bbox[5] = zmin;
  /* P3 */
  bbox[6] = xmax; bbox[7] = ymax; bbox[8] = zmin;
  /* P4 */
  bbox[9] = xmax; bbox[10] = ymax; bbox[11] = zmax;

  /* P5 */
  bbox[12] = xmin; bbox[13] = ymin; bbox[14] = zmax;
  /* P6 */
  bbox[15] = xmin; bbox[16] = ymin; bbox[17] = zmin;
  /* P7 */
  bbox[18] = xmax; bbox[19] = ymin; bbox[20] = zmin;
  /* P8 */
  bbox[21] = xmax; bbox[22] = ymin; bbox[23] = zmax;

 return MOPS_OK;
}


int
mops_bb_level(mops_object *o, double *bbox)
{
 int mops_status = MOPS_OK;
 mops_object *d = NULL; 
 double xmin = DBL_MAX, xmax = -DBL_MAX, ymin = DBL_MAX;
 double ymax = -DBL_MAX, zmin = DBL_MAX, zmax = -DBL_MAX;
 double bbt[24] = {0};
 int i, a;
 
  if(!o)
    return MOPS_OK;

  if(!o->down)
    return MOPS_OK;

  d = o->down;
  
  while(d)
    {
      mops_status = mops_get_sub_bb(d, bbt);

      a = 0;
      for(i = 0; i < 8; i++)
	{
	  if(bbt[a] < xmin)
	    xmin = bbt[a];
	  if(bbt[a] > xmax)
	    xmax = bbt[a];
	  a += 3;
	}

      a = 1;
      for(i = 0; i < 8; i++)
	{
	  if(bbt[a] < ymin)
	    ymin = bbt[a];
	  if(bbt[a] > ymax)
	    ymax = bbt[a];
	  a += 3;
	}

      a = 2;
      for(i = 0; i < 8; i++)
	{
	  if(bbt[a] < zmin)
	    zmin = bbt[a];
	  if(bbt[a] > zmax)
	    zmax = bbt[a];
	  a += 3;
	}

      d = d->next;
    } /* while */

  /* P1 */
  bbox[0] = xmin; bbox[1] = ymax; bbox[2] = zmax;
  /* P2 */
  bbox[3] = xmin; bbox[4] = ymax; bbox[5] = zmin;
  /* P3 */
  bbox[6] = xmax; bbox[7] = ymax; bbox[8] = zmin;
  /* P4 */
  bbox[9] = xmax; bbox[10] = ymax; bbox[11] = zmax;

  /* P5 */
  bbox[12] = xmin; bbox[13] = ymin; bbox[14] = zmax;
  /* P6 */
  bbox[15] = xmin; bbox[16] = ymin; bbox[17] = zmin;
  /* P7 */
  bbox[18] = xmax; bbox[19] = ymin; bbox[20] = zmin;
  /* P8 */
  bbox[21] = xmax; bbox[22] = ymin; bbox[23] = zmax;

 return MOPS_OK;
}

