/*
 * The Mops, a free 3D Tcl/Tk/OpenGL/RenderMan modeler
 *
 * The Mops is copyrighted (C) 1998-1999 by Randolf Schultz
 * (rschultz@informatik.uni-rostock.de)
 *
 * See the file License for details.
 *
 */

#include "mops.h"

/* CSphere, an example custom object for Mops */

/* The Name */
static char *CSphere_name = "CSphere";

/* prototypes */
typedef struct CSphere_s
{
 double radius;
 double zmin, zmax;
 double thetamax;
} CSphere;

int CSphere_create_cb(Tcl_Interp *interp, int argc, char *argv[], void **c);
int CSphere_delete_cb(void *c);
int CSphere_copy_cb(void *source, void **dest);
int CSphere_draww_cb(struct Togl *togl, mops_object *o);
int CSphere_draws_cb(struct Togl *togl, mops_object *o);
int CSphere_wrib_cb(mops_object *o);
int CSphere_read_cb(FILE *fileptr, mops_custom_object *custom);
int CSphere_write_cb(FILE *fileptr, mops_custom_object *custom);
int CSphere_getprop_cb(Tcl_Interp *interp);
int CSphere_set_attrprop_tcmd(ClientData clientData, Tcl_Interp *interp,
			     int argc, char *argv[]);
int CSphere_get_attrprop_tcmd(ClientData clientData, Tcl_Interp *interp,
			     int argc, char *argv[]);


char CSphere_mops_version_ma[] = MOPS_VERSIONSTR;
char CSphere_mops_version_mi[] = MOPS_VERSIONSTRMI;


/* functions */
int
Csphere_Init(Tcl_Interp *interp)
{
 int mops_status = MOPS_OK;
 char fname[] = "Csphere_Init";
 char success_cmd[] =
   "puts stdout \"CustomObject \\\"CSphere\\\" successfully loaded.\"\n";

  /* check versions */
  if(strcmp(mops_version_ma, CSphere_mops_version_ma))
    {
      mops_error(interp, MOPS_ERROR, fname,
		  "Plugin has been compiled for a different Mops!");
      mops_error(interp, MOPS_ERROR, fname,
		  "It is unsafe to continue! Bailing out...");
      return TCL_OK;
    }
  if(strcmp(mops_version_mi, CSphere_mops_version_mi))
    {
      mops_error(interp, MOPS_ERROR, fname,
		  "Plugin has been compiled for a different Mops!");
      mops_error(interp, MOPS_ERROR, fname,
		  "However, it is probably safe to continue...");
    }

  mops_status = mops_register_custom_objecttype(
				     CSphere_name,
				     CSphere_create_cb,
				     CSphere_draws_cb, 
				     NULL, /* drawh */ /* no handles! */
				     CSphere_draww_cb,
				     CSphere_delete_cb,
				     CSphere_copy_cb,
				     CSphere_getprop_cb,
				     NULL, /* getpoint */ /* no single point
					                     editing! */
				     CSphere_wrib_cb,
				     CSphere_read_cb,
				     CSphere_write_cb,
				     NULL /* no notification */
				     );

  if(mops_status)
    {
      mops_error(interp, mops_status, fname, NULL);
      return TCL_ERROR;
    }

  /* register some C-functions as Tcl-Commands */
  /* they deal with the CSphere-Attributes Property GUI */
  Tcl_CreateCommand (interp, "csphere_getCSphereAttr",
		     CSphere_get_attrprop_tcmd,
		     (ClientData) NULL, (Tcl_CmdDeleteProc *) NULL);
  Tcl_CreateCommand (interp, "csphere_setCSphereAttr",
		     CSphere_set_attrprop_tcmd,
		     (ClientData) NULL, (Tcl_CmdDeleteProc *) NULL);

  /* source csphere.tcl, it contains a Tcl-procedure to build
     the CSphere-Attributes Property GUI */
  if((Tcl_EvalFile(interp, "csphere.tcl")) != TCL_OK)
     {
       mops_error(interp, MOPS_ERROR, fname,
		  "Error while sourcing \\\"csphere.tcl\\\"!");
       return TCL_ERROR;
     }

  Tcl_Eval(interp, success_cmd);

 return TCL_OK;
}


int
CSphere_create_cb(Tcl_Interp *interp, int argc, char *argv[], void **c)
{
 int mops_status = MOPS_OK;
 CSphere *new = NULL;
 char fname[] = "CSphere_create";

  if(!(new = calloc(1,sizeof(CSphere))))
    {
      mops_error(interp,MOPS_OUTOFMEM,fname,NULL);
      return TCL_OK;
    }

  new->radius = 1.0;
  new->zmin = -1.0;
  new->zmax = 1.0;
  new->thetamax = 360.0;

  *c = new;

 return mops_status;
}


int
CSphere_delete_cb(void *c)
{
 int mops_status = MOPS_OK;

  if(!c)
    return MOPS_OK;

  free(c);

 return mops_status;
}


int
CSphere_copy_cb(void *source, void **dest)
{
 int mops_status = MOPS_OK;
 CSphere *new = NULL;

  /* copy custom object */
  if(!(new = calloc(1,sizeof(CSphere))))
    return MOPS_OUTOFMEM;
  
  *dest = new;

  memcpy(new, source, sizeof(CSphere));

 return mops_status;
}


int
CSphere_draww_cb(struct Togl *togl, mops_object *o)
{
 int mops_status = MOPS_OK;
 int instance = MOPS_FALSE;
 GLfloat color[4] = {0.8, 0.75, 0.65, 1.0};
 CSphere *CSphere = NULL;
 mops_custom_object *custom = NULL;
 GLdouble radius = 0.0;
 GLdouble w = 0.0;
 int reset_color = 0;

  if(!o)
    return MOPS_OK;

  if(o->type == MOPS_OTINSTANCE)
    {
      o = (mops_object *)o->object;

      instance = MOPS_TRUE;
    }

  if(!o->editor)
    return MOPS_OK;

  custom = o->object;
  if(!custom)
    return MOPS_OK;    

  CSphere = custom->object;
  if(!CSphere)
    return MOPS_OK;

  radius = (GLdouble)(CSphere->radius);
  w = (GLdouble)(sqrt(2.0)*0.5);

  glPushMatrix();
  if(!instance)
    {
      mops_render_attributes(o);

      if(o->selected)
	{
	  glColor3f MOPS_CLSEL;
	  reset_color = 1;
	}
    }
  else
    {
      if(o->colr != -1 && o->colg != -1 && o->colb != -1)
	{
	  color[0] = o->colr/255.0;
	  color[1] = o->colg/255.0;
	  color[2] = o->colb/255.0;
	  color[3] = o->cola/255.0; 
	  glMaterialfv (GL_FRONT_AND_BACK, GL_AMBIENT_AND_DIFFUSE, color);
	}
    }
 
  /* draw */
  /* xy-plane-ring */
  glBegin(GL_LINE_STRIP);
  glVertex3d(radius,0.0,0.0);
  glVertex3d(radius*w,-radius*w,0.0);
  glVertex3d(0.0,-radius,0.0);
  glVertex3d(-radius*w,-radius*w,0.0);
  glVertex3d(-radius,0.0,0.0);
  glVertex3d(-radius*w,radius*w,0.0);
  glVertex3d(0.0,radius,0.0);
  glVertex3d(radius*w,radius*w,0.0);
  glVertex3d(radius,0.0,0.0);
  glEnd();
  /* xz-plane-ring */
  glBegin(GL_LINE_STRIP);
  glVertex3d(radius,0.0,0.0);
  glVertex3d(radius*w,0.0,-radius*w);
  glVertex3d(0.0,0.0,-radius);
  glVertex3d(-radius*w,0.0,-radius*w);
  glVertex3d(-radius,0.0,0.0);
  glVertex3d(-radius*w,0.0,radius*w);
  glVertex3d(0.0,0.0,radius);
  glVertex3d(radius*w,0.0,radius*w);
  glVertex3d(radius,0.0,0.0);
  glEnd();
  /* yz-plane-ring */
  glBegin(GL_LINE_STRIP);
  glVertex3d(0.0,radius,0.0);
  glVertex3d(0.0,radius*w,-radius*w);
  glVertex3d(0.0,0.0,-radius);
  glVertex3d(0.0,-radius*w,-radius*w);
  glVertex3d(0.0,-radius,0.0);
  glVertex3d(0.0,-radius*w,radius*w);
  glVertex3d(0.0,0.0,radius);
  glVertex3d(0.0,radius*w,radius*w);
  glVertex3d(0.0,radius,0.0);
  glEnd();

  glPopMatrix();

 return mops_status;
}


int
CSphere_draws_cb(struct Togl *togl, mops_object *o)
{
 int mops_status = MOPS_OK;
 int instance = MOPS_FALSE;
 GLfloat color[4] = {0.8, 0.75, 0.65, 1.0};
 CSphere *CSphere = NULL;
 mops_custom_object *custom = NULL;
 GLdouble radius = 0.0;
 GLUquadricObj *qobj = NULL;

  if(!o)
    return MOPS_OK;

  if(o->type == MOPS_OTINSTANCE)
    {
      o = (mops_object *)o->object;

      instance = MOPS_TRUE;
    }

  if(!o->editor)
    return MOPS_OK;

  custom = o->object;
  if(!custom)
    return MOPS_OK;    

  CSphere = custom->object;
  if(!CSphere)
    return MOPS_OK;

  if(!(qobj = gluNewQuadric()))
    return MOPS_OUTOFMEM;

  radius = (GLdouble)(CSphere->radius);

  glPushMatrix();

  if(!instance)
    mops_render_attributes(o);
  else
    {
      if(o->colr != -1 && o->colg != -1 && o->colb != -1)
	{
	  color[0] = o->colr/255.0;
	  color[1] = o->colg/255.0;
	  color[2] = o->colb/255.0;
	  color[3] = o->cola/255.0; 
	  glMaterialfv (GL_FRONT_AND_BACK, GL_AMBIENT_AND_DIFFUSE, color);
	}
    }
 
  /* draw */
  gluSphere(qobj,radius,10,10);

  glPopMatrix();

 return mops_status;
}


int
CSphere_wrib_cb(mops_object *o)
{
 int mops_status = MOPS_OK;
 CSphere *CSphere = NULL;
 mops_custom_object *custom = NULL;

  if(!o)
    return MOPS_OK;

  custom = o->object;
  if(!custom)
    return MOPS_OK;    

  CSphere = custom->object;
  if(!CSphere)
    return MOPS_OK;

      RiAttributeBegin();
      RiTransformBegin();
       mops_wrib_attributes(o);
       mops_wrib_shaders(o);

       RiSphere((RtFloat)CSphere->radius,
		(RtFloat)CSphere->zmin,
		(RtFloat)CSphere->zmax,
		(RtFloat)CSphere->thetamax,
		NULL);

      RiTransformEnd();
      RiAttributeEnd(); 

 return mops_status;
}


int
CSphere_write_cb(FILE *fileptr, mops_custom_object *custom)
{
 int mops_status = MOPS_OK;
 CSphere *sphere = NULL;

  if(!custom)
    return MOPS_OK;    
  sphere = (CSphere *)(custom->object);
  if(!sphere)
    return MOPS_OK;

  fprintf(fileptr,"%g\n", sphere->radius);
  fprintf(fileptr,"%g\n", sphere->zmin);
  fprintf(fileptr,"%g\n", sphere->zmax);
  fprintf(fileptr,"%g\n", sphere->thetamax);

 return mops_status;
}

int
CSphere_read_cb(FILE *fileptr, mops_custom_object *custom)
{
 int mops_status = MOPS_OK;
 CSphere *sphere = NULL;

  if(!custom)
    return MOPS_OK;
    
  if(!(custom->object = calloc(1,sizeof(CSphere))))
    return MOPS_OUTOFMEM;

  sphere = (CSphere *)(custom->object);

  fscanf(fileptr,"%lg\n",&sphere->radius);
  fscanf(fileptr,"%lg\n",&sphere->zmin);
  fscanf(fileptr,"%lg\n",&sphere->zmax);
  fscanf(fileptr,"%lg\n",&sphere->thetamax);

 return mops_status;
}


/************************/
/************************/
/************************/

int
CSphere_getprop_cb(Tcl_Interp *interp)
{
 int mops_status = MOPS_OK;

  /* Surface-Shader */
  mops_status = mops_add_surfshader_property(interp);

  /* Displacement-Shader */
  mops_status = mops_add_dispshader_property(interp);

  /* Interior-Shader */
  mops_status = mops_add_intshader_property(interp);

  /* Exterior-Shader */
  mops_status = mops_add_extshader_property(interp);

  /* CSphere-Parameters */
  Tcl_SetVar(interp,"propertyList","CSphereAttr", TCL_APPEND_VALUE |\
	     TCL_LIST_ELEMENT | TCL_LEAVE_ERR_MSG | TCL_GLOBAL_ONLY);

  Tcl_SetVar(interp,"propertyProcList","csphere_buildCSphereAttrGUI",
	     TCL_APPEND_VALUE | TCL_LIST_ELEMENT |\
	     TCL_LEAVE_ERR_MSG | TCL_GLOBAL_ONLY);

  Tcl_SetVar(interp,"propertyGetProcList","csphere_getCSphereAttr",
	     TCL_APPEND_VALUE | TCL_LIST_ELEMENT |\
	     TCL_LEAVE_ERR_MSG | TCL_GLOBAL_ONLY);

  Tcl_SetVar(interp,"propertySetProcList","csphere_setCSphereAttr",
	     TCL_APPEND_VALUE | TCL_LIST_ELEMENT |\
	     TCL_LEAVE_ERR_MSG | TCL_GLOBAL_ONLY);

  Tcl_SetVar(interp,"propertyDataList","CSphereAttrData",
	     TCL_APPEND_VALUE | TCL_LIST_ELEMENT |\
	     TCL_LEAVE_ERR_MSG | TCL_GLOBAL_ONLY);

 return TCL_OK;
}


/* Tcl -> C! */
int
CSphere_set_attrprop_tcmd(ClientData clientData, Tcl_Interp *interp,
			 int argc, char *argv[])
{
 mops_object *o = NULL;
 mops_list_object *sel = mops_current_selection;
 mops_custom_object *custom = NULL;
 CSphere *sphere = NULL;
 char *n1="CSphereAttrData";
 Tcl_Obj *to = NULL, *toa = NULL, *ton = NULL;

  if(!sel) /* oops? */
    return TCL_OK;

  o = sel->object; 
  if(!o)
    return TCL_OK;

  custom = o->object;
  if(!custom)
    return TCL_OK;

  sphere = (CSphere *) custom->object;

  toa = Tcl_NewStringObj(n1,-1);
  ton = Tcl_NewStringObj("Radius",-1);
  to = Tcl_ObjGetVar2(interp,toa,ton,TCL_LEAVE_ERR_MSG | TCL_GLOBAL_ONLY);
  Tcl_GetDoubleFromObj(interp,to, &sphere->radius);

  Tcl_SetStringObj(ton,"ZMin",-1);
  to = Tcl_ObjGetVar2(interp,toa,ton,TCL_LEAVE_ERR_MSG | TCL_GLOBAL_ONLY);
  Tcl_GetDoubleFromObj(interp,to, &sphere->zmin);

  Tcl_SetStringObj(ton,"ZMax",-1);
  to = Tcl_ObjGetVar2(interp,toa,ton,TCL_LEAVE_ERR_MSG | TCL_GLOBAL_ONLY);
  Tcl_GetDoubleFromObj(interp,to, &sphere->zmax);

  Tcl_SetStringObj(ton,"ThetaMax",-1);
  to = Tcl_ObjGetVar2(interp,toa,ton,TCL_LEAVE_ERR_MSG | TCL_GLOBAL_ONLY);
  Tcl_GetDoubleFromObj(interp,to, &sphere->thetamax);

  Tcl_IncrRefCount(toa);Tcl_DecrRefCount(toa);
  Tcl_IncrRefCount(ton);Tcl_DecrRefCount(ton);

 return TCL_OK;
}

/* C -> Tcl! */
int
CSphere_get_attrprop_tcmd(ClientData clientData, Tcl_Interp *interp,
			 int argc, char *argv[])
{
 mops_object *o = NULL;
 mops_list_object *sel = mops_current_selection;
 mops_custom_object *custom = NULL;
 CSphere *sphere = NULL;
 char *n1="CSphereAttrData";
 Tcl_Obj *to = NULL, *toa = NULL, *ton = NULL;

  if(!sel) /* oops? */
    return TCL_OK;

  o = sel->object; 
  if(!o)
    return TCL_OK;

  custom = o->object;
  if(!custom)
    return TCL_OK;

  sphere = (CSphere *) custom->object;

  toa = Tcl_NewStringObj(n1,-1);
  ton = Tcl_NewStringObj("Radius",-1);
  to = Tcl_NewDoubleObj(sphere->radius);
  Tcl_ObjSetVar2(interp,toa,ton,to,TCL_LEAVE_ERR_MSG |
		 TCL_GLOBAL_ONLY);

  Tcl_SetStringObj(ton,"ZMin",-1);
  to = Tcl_NewDoubleObj(sphere->zmin);
  Tcl_ObjSetVar2(interp,toa,ton,to,TCL_LEAVE_ERR_MSG |
		 TCL_GLOBAL_ONLY);

  Tcl_SetStringObj(ton,"ZMax",-1);
  to = Tcl_NewDoubleObj(sphere->zmax);
  Tcl_ObjSetVar2(interp,toa,ton,to,TCL_LEAVE_ERR_MSG |
		 TCL_GLOBAL_ONLY);

  Tcl_SetStringObj(ton,"ThetaMax",-1);
  to = Tcl_NewDoubleObj(sphere->thetamax);
  Tcl_ObjSetVar2(interp,toa,ton,to,TCL_LEAVE_ERR_MSG |
		 TCL_GLOBAL_ONLY);

  Tcl_IncrRefCount(toa);Tcl_DecrRefCount(toa);
  Tcl_IncrRefCount(ton);Tcl_DecrRefCount(ton);

 return TCL_OK;
}

