/*
 * The Mops, a free 3D Tcl/Tk/OpenGL/RenderMan modeler
 *
 * The Mops is copyrighted (C) 1998-1999 by Randolf Schultz
 * (rschultz@informatik.uni-rostock.de)
 *
 * See the file License for details.
 *
 */

#include "mops.h"

#ifdef MOPSIDR
#include "idr.h"
#endif

/* Extrude */

/* The Name */
static char *Extrude_name = "Extrude";

static char *Extrude_customid = NULL;

/* prototypes */
typedef struct Extrude_s
{
 double height;
 mops_object *extrusions;
 int has_upper_cap;
 mops_object *upper_cap;
 int has_lower_cap;
 mops_object *lower_cap;
 int has_upper_bevel;
 mops_object *upper_bevels;
 int has_lower_bevel;
 mops_object *lower_bevels;
 int bevel_type; /* 0 no, 1 round, 2 ridge */
 double bevel_radius;
 double glu_sampling_tolerance;
 int glu_display_mode;
} Extrude;

int Extrude_create_cb(Tcl_Interp *interp, int argc, char *argv[], void **c);
int Extrude_delete_cb(void *c);
int Extrude_copy_cb(void *source, void **dest);
int Extrude_draww_cb(struct Togl *togl, mops_object *o);
int Extrude_draws_cb(struct Togl *togl, mops_object *o);
int Extrude_wrib_cb(mops_object *o);
int Extrude_read_cb(FILE *fileptr, mops_custom_object *custom);
int Extrude_write_cb(FILE *fileptr, mops_custom_object *custom);
int Extrude_getprop_cb(Tcl_Interp *interp);
int Extrude_set_attrprop_tcmd(ClientData clientData, Tcl_Interp *interp,
			     int argc, char *argv[]);
int Extrude_get_attrprop_tcmd(ClientData clientData, Tcl_Interp *interp,
			     int argc, char *argv[]);
int Extrude_notify_cb(mops_object *o);
int Extrude_convert(Tcl_Interp *interp, mops_object *o);
int Extrude_convertNP_tcmd(ClientData clientData, Tcl_Interp *interp,
			 int argc, char *argv[]);
void Extrude_apply_rsNC(mops_object *o);

int Extrude_provide_cb(mops_object *o, void **target);

#ifdef MOPSIDR
void Extrude_countbmms_cb(mops_object *o);
#endif

/* functions */
int
Extrude_Init(Tcl_Interp *interp)
{
 int mops_status = MOPS_OK;
 Tcl_HashEntry *entry;
 char fname[] = "Extrude_Init";

  mops_status = mops_register_custom_objecttype(
				     Extrude_name,
				     Extrude_create_cb,
				     Extrude_draws_cb, 
				     NULL,
				     Extrude_draww_cb,
				     Extrude_delete_cb,
				     Extrude_copy_cb,
				     Extrude_getprop_cb,
				     NULL, /* getpoint */ /* no single point
					                     editing! */
				     Extrude_wrib_cb,
				     Extrude_read_cb,
				     Extrude_write_cb,
				     Extrude_notify_cb
				     );

  if(mops_status)
    {
      mops_error(interp, mops_status, fname, NULL);
      return TCL_ERROR;
    }

  /* get custom-id from name */
  entry = Tcl_FindHashEntry(&mops_custom_id_ht, Extrude_name);

  if (entry != NULL)
    {
      Extrude_customid = Tcl_GetHashValue(entry);
    }
  else
    {
      mops_error(interp, MOPS_HASH, fname, NULL);
      return TCL_OK;
    }

  /* register converter */
  mops_status = mops_register_converter(interp,Extrude_name,Extrude_convert);

  /* register some C-functions as Tcl-Commands */
  /* they deal with the Extrude-Attributes Property GUI */
  Tcl_CreateCommand (interp, "Extrude_getExtrudeAttr",
		     Extrude_get_attrprop_tcmd,
		     (ClientData) NULL, (Tcl_CmdDeleteProc *) NULL);
  Tcl_CreateCommand (interp, "Extrude_setExtrudeAttr",
		     Extrude_set_attrprop_tcmd,
		     (ClientData) NULL, (Tcl_CmdDeleteProc *) NULL);

  /* another command, convert to a NURBPatch */
  Tcl_CreateCommand (interp, "extrudeCNP", Extrude_convertNP_tcmd,
		     (ClientData) NULL, (Tcl_CmdDeleteProc *) NULL);

  /* register provider */
  mops_status = mops_register_provider(interp, Extrude_name,
				       MOPS_OTNURBPATCH, Extrude_provide_cb);



#ifdef MOPSIDR
  mops_status = idr_register_countbmmcb(Extrude_customid,
					Extrude_countbmms_cb);
#endif

 return TCL_OK;
}


int
Extrude_create_cb(Tcl_Interp *interp, int argc, char *argv[], void **c)
{
 int mops_status = MOPS_OK;
 mops_list_object *sel = mops_current_selection;
 mops_object *extrusion = NULL;
 mops_object *curve = NULL;
 Extrude *new = NULL;
 int got_object = MOPS_FALSE;
 char fname[] = "Extrude_create";

  if(!sel)
    {
      return MOPS_NOSEL;
    }

  if(sel->object->type != MOPS_OTNURBCURVE)
    {
      mops_status = mops_get_object(sel->object, MOPS_OTNURBCURVE, &curve);
      if(curve)
	got_object = MOPS_TRUE;
    }
  else
    {
      curve = sel->object;
    }

  if(!curve)
    {
      mops_error(interp, MOPS_ERROR, fname, "object is not a NURBCurve");
      return MOPS_ERROR;
    }

  if(!(new = calloc(1, sizeof(Extrude))))
    {
      return MOPS_OUTOFMEM;
    }

  if(!(extrusion = calloc(1, sizeof(mops_object))))
    {
      free(new); return MOPS_OUTOFMEM;
    }

  mops_default_attributes(extrusion);
  extrusion->type = MOPS_OTNURBPATCH;

  new->height = 1.0;

  /* create no caps */
  new->has_upper_cap = 0;
  new->has_lower_cap = 0;

  /* create no bevels */
  new->has_upper_bevel = 0;
  new->has_lower_bevel = 0;

  /* more defaults */
  new->bevel_radius = 0.1;
  new->glu_sampling_tolerance = 30.0;

  new->extrusions = extrusion;
  
  mops_status = mops_extrude_nurb(1.0, curve, (mops_nurbpatch_object **)
				  &(extrusion->object));

  if(mops_status)
    {
      free(new);
      return mops_status;
    }

  *c = new;

  if(got_object)
    mops_delete_object(curve);

 return mops_status;
}


int
Extrude_delete_cb(void *c)
{
 int mops_status = MOPS_OK;
 Extrude *del = NULL;
 mops_object *bevel = NULL, *nb = NULL;
 mops_object *extrusion = NULL, *ne = NULL;

  if(!c)
    return MOPS_OK;

  del = (Extrude *)c;

  /* free extrusions */
  extrusion = del->extrusions;
  while(extrusion)
    {
      ne = extrusion->next;
      mops_delete_object(extrusion);
      extrusion = ne;
    }
  
  /* free bevels */
  bevel = del->upper_bevels;
  while(bevel)
    {
      nb = bevel->next;
      mops_delete_object(bevel);
      bevel = nb;
    }
  bevel = del->lower_bevels;
  while(bevel)
    {
      nb = bevel->next;
      mops_delete_object(bevel);
      bevel = nb;
    }

  if(del->upper_cap)
    mops_delete_object(del->upper_cap);

  if(del->lower_cap)
    mops_delete_object(del->lower_cap);

  free(c);

 return mops_status;
}


int
Extrude_copy_cb(void *source, void **dest)
{
 int mops_status = MOPS_OK;
 Extrude *new = NULL;
 

  /* copy custom object */
  if(!(new = calloc(1,sizeof(Extrude))))
    return MOPS_OUTOFMEM;
  
  memcpy(new, source, sizeof(Extrude));
  /* danger, links point to original hierarchy! */


  new->extrusions = NULL;
  new->upper_cap = NULL;
  new->lower_cap = NULL;
  new->upper_bevels = NULL;
  new->lower_bevels = NULL;

  *dest = new;

 return mops_status;
}


int
Extrude_draww_cb(struct Togl *togl, mops_object *o)
{
 int mops_status = MOPS_OK;
 int instance = MOPS_FALSE;
 GLfloat color[4] = {0.8, 0.75, 0.65, 1.0};
 GLfloat m[4][4];
 Extrude *ext = NULL;
 mops_custom_object *custom = NULL;
 mops_object *extrusion = NULL, *bevel = NULL;
 int reset_color = 0;


  if(!o)
    return MOPS_OK;

  if(o->type == MOPS_OTINSTANCE)
    {
      o = (mops_object *)o->object;

      instance = MOPS_TRUE;
    }

  if(!o->editor)
    return MOPS_OK;

  custom = o->object;
  if(!custom)
    return MOPS_OK;    

  ext = custom->object;
  if(!ext)
    return MOPS_OK;

  glPushMatrix();

  if(!instance)
    {
      mops_render_attributes(o);

      if(o->selected)
	{
	  glColor3f MOPS_CLSEL;
	  reset_color = 1;
	}
    }
  else
    {
      if(o->colr != -1 && o->colg != -1 && o->colb != -1)
	{
	  color[0] = o->colr/255.0;
	  color[1] = o->colg/255.0;
	  color[2] = o->colb/255.0;
	  color[3] = o->cola/255.0; 
	  glMaterialfv (GL_FRONT_AND_BACK, GL_AMBIENT_AND_DIFFUSE, color);
	}
    }

  /* draw extrusions */
  extrusion = ext->extrusions;
  while(extrusion)
    {
      mops_status = mops_draw_nurbpatch_c(togl, extrusion);
      extrusion = extrusion->next;
    }

  /* draw bevels */
  bevel = ext->upper_bevels;
  while(bevel)
    {
      glPushMatrix();
       glTranslated(bevel->movx, bevel->movy, bevel->movz);
       mops_build_rotmatrix(m, bevel->quat);
       glMultMatrixf(&m[0][0]);
       glScaled (bevel->scalx, bevel->scaly, bevel->scalz);
       mops_status = mops_draw_nurbpatch_c(togl, bevel);
      glPopMatrix();
      bevel = bevel->next;
    }
  bevel = ext->lower_bevels;
  while(bevel)
    {
      glPushMatrix();
       glTranslated(bevel->movx, bevel->movy, bevel->movz);
       mops_build_rotmatrix(m, bevel->quat);
       glMultMatrixf(&m[0][0]);
       glScaled (bevel->scalx, bevel->scaly, bevel->scalz);
       mops_status = mops_draw_nurbpatch_c(togl, bevel);
      glPopMatrix();
      bevel = bevel->next;
    }

  /* draw caps */
  if(ext->upper_cap)
    {
      glPushMatrix();
      glTranslated(ext->upper_cap->movx, ext->upper_cap->movy,
		   ext->upper_cap->movz);
      mops_build_rotmatrix(m, ext->upper_cap->quat);
      glMultMatrixf(&m[0][0]);
      glScaled (ext->upper_cap->scalx, ext->upper_cap->scaly,
		 ext->upper_cap->scalz);

      mops_status = mops_draw_nurbpatch_c(togl, ext->upper_cap);
      glPopMatrix();
    }

  if(ext->lower_cap)
    {
      glPushMatrix();
      glTranslated(ext->lower_cap->movx, ext->lower_cap->movy,
		   ext->lower_cap->movz);
      mops_build_rotmatrix(m, ext->lower_cap->quat);
      glMultMatrixf(&m[0][0]);
      glScaled (ext->lower_cap->scalx, ext->lower_cap->scaly,
		 ext->lower_cap->scalz);

      mops_status = mops_draw_nurbpatch_c(togl, ext->lower_cap);
      glPopMatrix();
    }

  if(reset_color)
    glColor3f MOPS_CLUSEL;

  glPopMatrix();

 return mops_status;
}


int
Extrude_draws_cb(struct Togl *togl, mops_object *o)
{
 int mops_status = MOPS_OK;
 int instance = MOPS_FALSE;
 GLfloat color[4] = {0.8, 0.75, 0.65, 1.0};
 GLfloat m[4][4];
 Extrude *ext = NULL;
 mops_custom_object *custom = NULL;
 mops_object *extrusion = NULL, *bevel = NULL;

  if(!o)
    return MOPS_OK;

  if(o->type == MOPS_OTINSTANCE)
    {
      o = (mops_object *)o->object;

      instance = MOPS_TRUE;
    }

  if(!o->editor)
    return MOPS_OK;

  custom = o->object;
  if(!custom)
    return MOPS_OK;    

  ext = custom->object;
  if(!ext)
    return MOPS_OK;


  glPushMatrix();

  if(!instance)
    mops_render_attributes(o);
  else
    {
      if(o->colr != -1 && o->colg != -1 && o->colb != -1)
	{
	  color[0] = o->colr/255.0;
	  color[1] = o->colg/255.0;
	  color[2] = o->colb/255.0;
	  color[3] = o->cola/255.0; 
	  glMaterialfv (GL_FRONT_AND_BACK, GL_AMBIENT_AND_DIFFUSE, color);
	}
    }

  /* draw extrusions */
  extrusion = ext->extrusions;
  while(extrusion)
    {
      mops_status = mops_draw_nurbpatch(togl, extrusion);
      extrusion = extrusion->next;
    }

  /* draw bevels */
  bevel = ext->upper_bevels;
  while(bevel)
    {
      glPushMatrix();
       glTranslated(bevel->movx, bevel->movy, bevel->movz);
       mops_build_rotmatrix(m, bevel->quat);
       glMultMatrixf(&m[0][0]);
       glScaled (bevel->scalx, bevel->scaly, bevel->scalz);
       mops_status = mops_draw_nurbpatch(togl, bevel);
      glPopMatrix();
      bevel = bevel->next;
    }

  /* lower bevels have a negative scalz, thats why we need
     to reverse the normals here, to draw them */
  glFrontFace(GL_CW);
  bevel = ext->lower_bevels;
  while(bevel)
    {
      glPushMatrix();
       glTranslated(bevel->movx, bevel->movy, bevel->movz);
       mops_build_rotmatrix(m, bevel->quat);
       glMultMatrixf(&m[0][0]);
       glScaled (bevel->scalx, bevel->scaly, bevel->scalz);
       mops_status = mops_draw_nurbpatch(togl, bevel);
      glPopMatrix();
      bevel = bevel->next;
    }
  glFrontFace(GL_CCW);

  /* draw caps */
  if(ext->upper_cap)
    {
      glPushMatrix();
      glTranslated(ext->upper_cap->movx, ext->upper_cap->movy,
		   ext->upper_cap->movz);
      mops_build_rotmatrix(m, ext->upper_cap->quat);
      glMultMatrixf(&m[0][0]);
      glScaled (ext->upper_cap->scalx, ext->upper_cap->scaly,
		 ext->upper_cap->scalz);

      mops_status = mops_draw_nurbpatch(togl, ext->upper_cap);
      glPopMatrix();
    }

  if(ext->lower_cap)
    {
      glPushMatrix();
      glTranslated(ext->lower_cap->movx, ext->lower_cap->movy,
		   ext->lower_cap->movz);
      mops_build_rotmatrix(m, ext->lower_cap->quat);
      glMultMatrixf(&m[0][0]);
      glScaled (ext->lower_cap->scalx, ext->lower_cap->scaly,
		 ext->lower_cap->scalz);

      mops_status = mops_draw_nurbpatch(togl, ext->lower_cap);
      glPopMatrix();
    }

  glPopMatrix();

 return mops_status;
}


int
Extrude_wrib_cb(mops_object *o)
{
 int mops_status = MOPS_OK;
 Extrude *ext = NULL;
 mops_custom_object *custom = NULL;
 mops_object *extrusion = NULL;
 mops_object *bevel = NULL;

  if(!o)
    return MOPS_OK;

  custom = o->object;
  if(!custom)
    return MOPS_OK;    

  ext = custom->object;
  if(!ext)
    return MOPS_OK;

  RiAttributeBegin();
  RiTransformBegin();
   mops_wrib_attributes(o);
   mops_wrib_shaders(o);
       
   extrusion = ext->extrusions;
   while(extrusion)
     {
       mops_wrib_npatch(extrusion);
       extrusion = extrusion->next;
     }

   if(ext->upper_bevels)
     {
       bevel = ext->upper_bevels;
       while(bevel)
	 {
	   RiTransformBegin();
	   mops_wrib_trafos(bevel);
	   mops_wrib_npatch(bevel);
	   RiTransformEnd();
	   bevel = bevel->next;
	 }
     }

   if(ext->lower_bevels)
     {
       bevel = ext->lower_bevels;
       while(bevel)
	 {
	   RiTransformBegin();
	   mops_wrib_trafos(bevel);
	   mops_wrib_npatch(bevel);
	   RiTransformEnd();
	   bevel = bevel->next;
	 }
     }
   
   if(ext->upper_cap)
     {
       RiTransformBegin();
       mops_wrib_trafos(ext->upper_cap);
       mops_wrib_trimcurves(ext->upper_cap);
       mops_wrib_npatch(ext->upper_cap);
       RiTransformEnd();
     }
   if(ext->lower_cap)
     {
       RiTransformBegin();
       mops_wrib_trafos(ext->lower_cap);
       mops_wrib_trimcurves(ext->lower_cap);
       mops_wrib_npatch(ext->lower_cap);
       RiTransformEnd();
     }

  RiTransformEnd();
  RiAttributeEnd(); 

 return mops_status;
}


int
Extrude_write_cb(FILE *fileptr, mops_custom_object *custom)
{
 int mops_status = MOPS_OK;
 Extrude *ext = NULL;

  if(!custom)
    return MOPS_OK;    
  ext = (Extrude *)(custom->object);
  if(!ext)
    return MOPS_OK;

  fprintf(fileptr,"%d\n",ext->has_upper_cap);
  fprintf(fileptr,"%d\n",ext->has_lower_cap);
  fprintf(fileptr,"%g\n",ext->height);
  fprintf(fileptr,"%d\n",ext->has_upper_bevel);
  fprintf(fileptr,"%d\n",ext->has_lower_bevel);
  fprintf(fileptr,"%g\n",ext->bevel_radius);
  fprintf(fileptr,"%d\n",ext->bevel_type);
  fprintf(fileptr,"%g\n",ext->glu_sampling_tolerance);
  fprintf(fileptr,"%d\n",ext->glu_display_mode);

 return mops_status;
}

int
Extrude_read_cb(FILE *fileptr, mops_custom_object *custom)
{
 int mops_status = MOPS_OK;
 Extrude *ext = NULL;

  if(!custom)
    return MOPS_OK;
    
  if(!(custom->object = calloc(1,sizeof(Extrude))))
    return MOPS_OUTOFMEM;

  ext = (Extrude *)(custom->object);

  ext->height = 1.0;
  ext->glu_sampling_tolerance = 30.0;
  ext->bevel_radius = 0.1;

  fscanf(fileptr,"%d\n",&ext->has_upper_cap);
  fscanf(fileptr,"%d\n",&ext->has_lower_cap);
  if(mops_read_version > 4)
    {
      fscanf(fileptr,"%lg\n",&ext->height);
      fscanf(fileptr,"%d\n",&ext->has_upper_bevel);
      fscanf(fileptr,"%d\n",&ext->has_lower_bevel);
      fscanf(fileptr,"%lg\n",&ext->bevel_radius);
      fscanf(fileptr,"%d\n",&ext->bevel_type);
      fscanf(fileptr,"%lg\n",&ext->glu_sampling_tolerance);
      fscanf(fileptr,"%d\n",&ext->glu_display_mode);
    }

 return mops_status;
}

int
Extrude_notify_cb(mops_object *o)
{
 int mops_status = MOPS_OK;
 Extrude *ext = NULL;
 mops_object *down = NULL, *c = NULL, *cap = NULL, *trim = NULL;
 mops_object *bevel = NULL, *nb = NULL;
 mops_object *extrusion = NULL, *ne = NULL, *ot = NULL;
 mops_nurbcurve_object *curve = NULL;
 mops_nurbpatch_object *patch = NULL;
 mops_custom_object *custom = NULL;
 double tolerance = 30.0;
 int display_mode = 0, got_object = MOPS_FALSE;
 double uminx, umaxx, uminy, umaxy;
 double lminx, lmaxx, lminy, lmaxy;
 double firstmovex = 0.0, firstmovey = 0.0;
 double trimmx = 0.0, trimmy = 0.0;
 double angle = 0.0, z = 0.0;

  if(!o)
    return MOPS_OK;

  custom = o->object;
  if(!custom)
    return MOPS_OK;    

  ext = custom->object;
  if(!ext)
    return MOPS_OK;

  /* ok, something changed below, we have to rebuild the extrusions */

  tolerance = ext->glu_sampling_tolerance;
  display_mode = ext->glu_display_mode;

  /* clear old extrusions, caps and bevels */
  extrusion = ext->extrusions;
  while(extrusion)
    {
      ne = extrusion->next;
      mops_delete_object(extrusion);
      extrusion = ne;
    }
  ext->extrusions = NULL;

  if(ext->upper_cap)
    mops_delete_object(ext->upper_cap);
  ext->upper_cap = NULL;

  if(ext->lower_cap)
    mops_delete_object(ext->lower_cap);
  ext->lower_cap = NULL;

  bevel = ext->upper_bevels;
  while(bevel)
    {
      nb = bevel->next;
      mops_delete_object(bevel);
      bevel = nb;
    }
  ext->upper_bevels = NULL;

  bevel = ext->lower_bevels;
  while(bevel)
    {
      nb = bevel->next;
      mops_delete_object(bevel);
      bevel = nb;
    }
  ext->lower_bevels = NULL;

  /* create new extrusions, caps and bevels */
  down = o->down;
  while(down)
    {
      got_object = MOPS_FALSE;
      c = NULL;
      if(down->type != MOPS_OTNURBCURVE)
	{
	  mops_status = mops_get_object(down, MOPS_OTNURBCURVE, &c);
	  if(c)
	    got_object = MOPS_TRUE;
	}
      else
	{
	  c = down;
	}

      /* normally, we do this at the end of the while,
         but this is _so_ far away... */
      down = down->next;

      if(c)
	{
	  extrusion = ext->extrusions;
	  if(!(ext->extrusions = calloc(1, sizeof(mops_object))))
	    return MOPS_OUTOFMEM;
	  mops_copy_attributes(o, ext->extrusions);
	  /*	  mops_default_attributes(ext->extrusions);*/
	  mops_extrude_nurb(ext->height, c, (mops_nurbpatch_object **)
			    &(ext->extrusions->object));
	  ((mops_nurbpatch_object *)ext->extrusions->object)->
	    glu_display_mode = display_mode;
	  ((mops_nurbpatch_object *)ext->extrusions->object)->
	    glu_sampling_tolerance = tolerance;
	  ext->extrusions->type = MOPS_OTNURBPATCH;
	  ext->extrusions->next = extrusion;

	  /* create bevel */
	  if(ext->has_upper_bevel || ext->has_lower_bevel)
	    {
	      bevel = calloc(1, sizeof(mops_object));
	      bevel->type = MOPS_OTNURBPATCH;
	      mops_default_attributes(bevel);

	      mops_status = mops_bevel_nurb(ext->bevel_type,
					    ext->bevel_radius,
					    c,
				(mops_nurbpatch_object **)&(bevel->object));

	      ((mops_nurbpatch_object *)bevel->object)->
		glu_display_mode = display_mode;
	      ((mops_nurbpatch_object *)bevel->object)->
		glu_sampling_tolerance = tolerance;
	    }
	  /* link bevel */
	  if(ext->has_upper_bevel)
	    {
	      bevel->next = ext->upper_bevels;
	      ext->upper_bevels = bevel;
	      bevel->movz = ext->height;
	      
	      if(ext->has_lower_bevel)
		{
		  nb = ext->lower_bevels;
		  mops_status = mops_copy_object(bevel,
						 &(ext->lower_bevels));
		  ext->lower_bevels->movz = 0.0;
		  
		  ext->lower_bevels->scalz = -1;
		  ext->lower_bevels->next = nb;
		}
	    }
	  else
	    {
	      if(ext->has_lower_bevel)
		{
		  bevel->next = ext->lower_bevels;
		  ext->lower_bevels = bevel;
		  bevel->movz = 0.0;
		  bevel->scalz = -1;
		  
		}
	    }

	  /* create caps */
	  if(ext->has_upper_cap)
	    {
	      mops_status = mops_copy_object(c,&trim);
	      if(mops_status)
		return mops_status;

	      /* work around error with rotated curves */
	      Extrude_apply_rsNC(trim);

	      curve = (mops_nurbcurve_object *)trim->object;

	      /* first c? -> trimcurve must be clockwise */
	      if(!ext->upper_cap)
		{
		  firstmovex = trim->movx;
		  firstmovey = trim->movy;

		  z = ext->height;
		  /* 
		     if we have bevels, we need a new control vector
		     copied directly from the beveling patch,
		     this is the same vector as the vector from the
		     curve, but displaced by bevel_radius...
		   */
		  if(ext->upper_bevels)
		    {
		      bevel = ext->upper_bevels;
		      patch = (mops_nurbpatch_object*)bevel->object;
		      curve->has_weights = MOPS_TRUE;
		      free(curve->controlv);
		      curve->controlv = NULL;
		      if(!(curve->controlv = 
			   calloc(4*curve->length, sizeof(double))))
			return MOPS_OUTOFMEM;
		      memcpy(curve->controlv,&(patch->controlv[
				(patch->width-1)*4*patch->height]),
			     4*curve->length*sizeof(double));
		      z += curve->controlv[2];
		      mops_default_attributes(trim);
		    }

		  if(!(cap = calloc(1,sizeof(mops_object))))
		    return MOPS_OUTOFMEM;
		  cap->type = MOPS_OTNURBPATCH;
		  mops_default_attributes(cap);
		  mops_create_cap(z,curve,&uminx,&umaxx,&uminy,&umaxy,&angle,
				  (mops_nurbpatch_object **)&(cap->object));

		  umaxx *= trim->scalx;
		  uminx *= trim->scalx;
		  umaxy *= trim->scaly;
		  uminy *= trim->scaly;

		  ((mops_nurbpatch_object *)cap->object)->
		    glu_display_mode = display_mode;
		  ((mops_nurbpatch_object *)cap->object)->
		    glu_sampling_tolerance = tolerance;

		  if(!ext->upper_bevels)
		    {
		      mops_status = mops_copy_transform(trim,cap);
		    }

		  ext->upper_cap = cap;

		  mops_get_nc_orientation((mops_nurbcurve_object *)
					  curve, &angle);

		  if(angle<0.0)
		    mops_revert_ncurve(trim->object);

		  trim->scalx /= fabs(umaxx-uminx);
		  trim->scaly /= fabs(umaxy-uminy);
		  trim->movx = -(uminx + (fabs(umaxx-uminx)/2.0))*trim->scalx;
		  trim->movy = -(uminy + (fabs(umaxy-uminy)/2.0))*trim->scaly;
	    
		  trim->movx += 0.5;
		  trim->movy += 0.5;
		  
		}
	      else
		{ /* not first curve  */

		  if(ext->upper_bevels)
		    {
		      bevel = ext->upper_bevels;
		      patch = (mops_nurbpatch_object*)bevel->object;
		      curve->has_weights = MOPS_TRUE;
		      free(curve->controlv);
		      curve->controlv = NULL;
		      if(!(curve->controlv = 
			   calloc(4*curve->length, sizeof(double))))
			return MOPS_OUTOFMEM;
		      memcpy(curve->controlv,&(patch->controlv[
					(patch->width-1)*4*patch->height]),
			     4*curve->length*sizeof(double));
		      mops_default_attributes(trim);
		    }

		  mops_get_nc_orientation((mops_nurbcurve_object *)
					  trim->object, &angle);

		  if(angle>0.0)
		    mops_revert_ncurve(trim->object);

		  if(!ext->upper_bevels)
		    {
		      trimmx = trim->movx-firstmovex;
		      trimmy = trim->movy-firstmovey;
		    }
		  else
		    {
		      trimmx = 0.0;
		      trimmy = 0.0;
		    }

		  trim->scalx /= fabs(umaxx-uminx);
		  trim->scaly /= fabs(umaxy-uminy);
		    
		  trim->movx = -(uminx + (fabs(umaxx-uminx)/2.0))*trim->scalx;
		  trim->movy = -(uminy + (fabs(umaxy-uminy)/2.0))*trim->scaly;

		  trim->movx += 0.5+(trimmx/fabs(umaxx-uminx));
		  trim->movy += 0.5+(trimmy/fabs(umaxy-uminy));

		}

	      ot = ext->upper_cap->down;
	      ext->upper_cap->down = trim;
	      trim->next = ot;

	    }

	  if(ext->has_lower_cap)
	    {
	      mops_status = mops_copy_object(c,&trim);
	      if(mops_status)
		return mops_status;

	      /* work around error with rotated curves */
	      Extrude_apply_rsNC(trim);

	      curve = (mops_nurbcurve_object *)trim->object;

	      /* first c? -> trimcurve must be cw */
	      if(!ext->lower_cap)
		{
		  if(!ext->has_upper_cap)
		    {
		      firstmovex = trim->movx;
		      firstmovey = trim->movy;
		    }

		  z = 0.0;
		  /* 
		     if we have bevels, we need a new control vector
		     copied directly from the beveling patch,
		     this is the same vector as the vector from the
		     curve, but displaced by bevel_radius...
		   */
		  if(ext->lower_bevels)
		    {
		      bevel = ext->lower_bevels;
		      patch = (mops_nurbpatch_object*)bevel->object;
		      curve->has_weights = MOPS_TRUE;
		      free(curve->controlv);
		      curve->controlv = NULL;
		      if(!(curve->controlv = 
			   calloc(4*curve->length, sizeof(double))))
			return MOPS_OUTOFMEM;
		      memcpy(curve->controlv,&(patch->controlv[
					(patch->width-1)*4*patch->height]),
			     4*curve->length*sizeof(double));
		      z -= curve->controlv[2];
		      mops_default_attributes(trim);       
		    }

		  if(!(cap = calloc(1,sizeof(mops_object))))
		    return MOPS_OUTOFMEM;
		  cap->type = MOPS_OTNURBPATCH;
		  mops_default_attributes(cap);
		  mops_create_cap(z,curve,&lminx,&lmaxx,&lminy,&lmaxy,&angle,
				  (mops_nurbpatch_object **)&(cap->object));

		  lmaxx *= trim->scalx;
		  lminx *= trim->scalx;
		  lmaxy *= trim->scaly;
		  lminy *= trim->scaly;

		  ((mops_nurbpatch_object *)cap->object)->
		    glu_display_mode = display_mode;
		  ((mops_nurbpatch_object *)cap->object)->
		    glu_sampling_tolerance = tolerance;

		  if(!ext->lower_bevels)
		    {
		      mops_status = mops_copy_transform(trim,cap);
		    }

		  ext->lower_cap = cap;

		  mops_get_nc_orientation((mops_nurbcurve_object *)
					  curve, &angle);

		  if(angle<0.0)
		    mops_revert_ncurve(trim->object);
		  
		  trim->scalx /= fabs(lmaxx-lminx);
		  trim->scaly /= fabs(lmaxy-lminy);

		  trim->movx = -(lminx + (fabs(lmaxx-lminx)/2.0))*trim->scalx;
		  trim->movy = -(lminy + (fabs(lmaxy-lminy)/2.0))*trim->scaly;
	    
		  trim->movx += 0.5;
		  trim->movy += 0.5;
		}
	      else
		{
		  if(ext->lower_bevels)
		    {
		      bevel = ext->lower_bevels;
		      patch = (mops_nurbpatch_object*)bevel->object;
		      curve->has_weights = MOPS_TRUE;
		      free(curve->controlv);
		      curve->controlv = NULL;
		      if(!(curve->controlv = 
			   calloc(4*curve->length, sizeof(double))))
			return MOPS_OUTOFMEM;
		      memcpy(curve->controlv,&(patch->controlv[
					(patch->width-1)*4*patch->height]),
			     4*curve->length*sizeof(double));
		      mops_default_attributes(trim);
		    }

		  mops_get_nc_orientation((mops_nurbcurve_object *)
					  trim->object, &angle);
		    
		  if(angle>0.0)
		    mops_revert_ncurve(trim->object);

		  if(!ext->lower_bevels)
		    {
		      trimmx = trim->movx-firstmovex;
		      trimmy = trim->movy-firstmovey;
		    }
		  else
		    {
		      trimmx = 0.0;
		      trimmy = 0.0;
		    }

		  trim->scalx /= fabs(lmaxx-lminx);
		  trim->scaly /= fabs(lmaxy-lminy);

		  trim->movx = -(lminx + (fabs(lmaxx-lminx)/2.0))*trim->scalx;
		  trim->movy = -(lminy + (fabs(lmaxy-lminy)/2.0))*trim->scaly;
		    
		  trim->movx += 0.5+(trimmx/fabs(lmaxx-lminx));
		  trim->movy += 0.5+(trimmy/fabs(lmaxy-lminy));

		}

	      ot = ext->lower_cap->down;
	      ext->lower_cap->down = trim;
	      trim->next = ot;
	      
	    }

	  if(got_object)
	    mops_delete_object(c);

	} /* if c */

    } /* while down */

 return mops_status;
}


/************************/
/************************/
/************************/

int
Extrude_getprop_cb(Tcl_Interp *interp)
{
 int mops_status = MOPS_OK;

  /* Surface-Shader */
  mops_status = mops_add_surfshader_property(interp);

  /* Displacement-Shader */
  mops_status = mops_add_dispshader_property(interp);

  /* Interior-Shader */
  mops_status = mops_add_intshader_property(interp);

  /* Exterior-Shader */
  mops_status = mops_add_extshader_property(interp);

  /* Extrude-Parameters */
  Tcl_SetVar(interp,"propertyList","ExtrudeAttr", TCL_APPEND_VALUE |\
	     TCL_LIST_ELEMENT | TCL_LEAVE_ERR_MSG | TCL_GLOBAL_ONLY);

  Tcl_SetVar(interp,"propertyProcList","Extrude_buildExtrudeAttrGUI",
	     TCL_APPEND_VALUE | TCL_LIST_ELEMENT |\
	     TCL_LEAVE_ERR_MSG | TCL_GLOBAL_ONLY);

  Tcl_SetVar(interp,"propertyGetProcList","Extrude_getExtrudeAttr",
	     TCL_APPEND_VALUE | TCL_LIST_ELEMENT |\
	     TCL_LEAVE_ERR_MSG | TCL_GLOBAL_ONLY);

  Tcl_SetVar(interp,"propertySetProcList","Extrude_setExtrudeAttr",
	     TCL_APPEND_VALUE | TCL_LIST_ELEMENT |\
	     TCL_LEAVE_ERR_MSG | TCL_GLOBAL_ONLY);

  Tcl_SetVar(interp,"propertyDataList","ExtrudeAttrData",
	     TCL_APPEND_VALUE | TCL_LIST_ELEMENT |\
	     TCL_LEAVE_ERR_MSG | TCL_GLOBAL_ONLY);

 return TCL_OK;
}


/* Tcl -> C! */
int
Extrude_set_attrprop_tcmd(ClientData clientData, Tcl_Interp *interp,
			 int argc, char *argv[])
{
 mops_object *o = NULL;
 mops_list_object *sel = mops_current_selection;
 mops_custom_object *custom = NULL;
 Extrude *ext = NULL;
 char *n1="ExtrudeAttrData";
 Tcl_Obj *to = NULL, *toa = NULL, *ton = NULL;

  if(!sel) /* oops? */
    return TCL_OK;

  o = sel->object; 
  if(!o)
    return TCL_OK;

  custom = o->object;
  if(!custom)
    return TCL_OK;

  ext = (Extrude *) custom->object;

  toa = Tcl_NewStringObj(n1,-1);
  ton = Tcl_NewStringObj("HasUpperCap",-1);
  to = Tcl_ObjGetVar2(interp,toa,ton,TCL_LEAVE_ERR_MSG | TCL_GLOBAL_ONLY);
  Tcl_GetIntFromObj(interp,to, &ext->has_upper_cap);

  Tcl_SetStringObj(ton,"HasLowerCap",-1);
  to = Tcl_ObjGetVar2(interp,toa,ton,TCL_LEAVE_ERR_MSG | TCL_GLOBAL_ONLY);
  Tcl_GetIntFromObj(interp,to, &ext->has_lower_cap);

  Tcl_SetStringObj(ton,"UpperBevel",-1);
  to = Tcl_ObjGetVar2(interp,toa,ton,TCL_LEAVE_ERR_MSG | TCL_GLOBAL_ONLY);
  Tcl_GetIntFromObj(interp,to, &ext->has_upper_bevel);

  Tcl_SetStringObj(ton,"LowerBevel",-1);
  to = Tcl_ObjGetVar2(interp,toa,ton,TCL_LEAVE_ERR_MSG | TCL_GLOBAL_ONLY);
  Tcl_GetIntFromObj(interp,to, &ext->has_lower_bevel);

  Tcl_SetStringObj(ton,"BevelType",-1);
  to = Tcl_ObjGetVar2(interp,toa,ton,TCL_LEAVE_ERR_MSG | TCL_GLOBAL_ONLY);
  Tcl_GetIntFromObj(interp,to, &ext->bevel_type);

  Tcl_SetStringObj(ton,"BevelRadius",-1);
  to = Tcl_ObjGetVar2(interp,toa,ton,TCL_LEAVE_ERR_MSG | TCL_GLOBAL_ONLY);
  Tcl_GetDoubleFromObj(interp,to, &(ext->bevel_radius));

  Tcl_SetStringObj(ton,"Tolerance",-1);
  to = Tcl_ObjGetVar2(interp,toa,ton,TCL_LEAVE_ERR_MSG | TCL_GLOBAL_ONLY);
  Tcl_GetDoubleFromObj(interp,to, &(ext->glu_sampling_tolerance));

  Tcl_SetStringObj(ton,"Display",-1);
  to = Tcl_ObjGetVar2(interp,toa,ton,TCL_LEAVE_ERR_MSG | TCL_GLOBAL_ONLY);
  Tcl_GetIntFromObj(interp,to, &(ext->glu_display_mode));

  Tcl_SetStringObj(ton,"Height",-1);
  to = Tcl_ObjGetVar2(interp,toa,ton,TCL_LEAVE_ERR_MSG | TCL_GLOBAL_ONLY);
  Tcl_GetDoubleFromObj(interp,to, &(ext->height));

  Tcl_IncrRefCount(toa);Tcl_DecrRefCount(toa);
  Tcl_IncrRefCount(ton);Tcl_DecrRefCount(ton);

  Extrude_notify_cb(o);

 return TCL_OK;
}

/* C -> Tcl! */
int
Extrude_get_attrprop_tcmd(ClientData clientData, Tcl_Interp *interp,
			 int argc, char *argv[])
{
 mops_object *o = NULL;
 mops_list_object *sel = mops_current_selection;
 mops_custom_object *custom = NULL;
 Extrude *ext = NULL;
 char *n1="ExtrudeAttrData";
 Tcl_Obj *to = NULL, *toa = NULL, *ton = NULL;

  if(!sel) /* oops? */
    return TCL_OK;

  o = sel->object; 
  if(!o)
    return TCL_OK;

  custom = o->object;
  if(!custom)
    return TCL_OK;

  ext = (Extrude *) custom->object;

  toa = Tcl_NewStringObj(n1,-1);
  ton = Tcl_NewStringObj("HasUpperCap",-1);
  to = Tcl_NewIntObj(ext->has_upper_cap);
  Tcl_ObjSetVar2(interp,toa,ton,to,TCL_LEAVE_ERR_MSG |
		 TCL_GLOBAL_ONLY);

  Tcl_SetStringObj(ton,"HasLowerCap",-1);
  to = Tcl_NewIntObj(ext->has_lower_cap);
  Tcl_ObjSetVar2(interp,toa,ton,to,TCL_LEAVE_ERR_MSG |
		 TCL_GLOBAL_ONLY);

  Tcl_SetStringObj(ton,"UpperBevel",-1);
  to = Tcl_NewIntObj(ext->has_upper_bevel);
  Tcl_ObjSetVar2(interp,toa,ton,to,TCL_LEAVE_ERR_MSG |
		 TCL_GLOBAL_ONLY);

  Tcl_SetStringObj(ton,"LowerBevel",-1);
  to = Tcl_NewIntObj(ext->has_lower_bevel);
  Tcl_ObjSetVar2(interp,toa,ton,to,TCL_LEAVE_ERR_MSG |
		 TCL_GLOBAL_ONLY);

  Tcl_SetStringObj(ton,"BevelType",-1);
  to = Tcl_NewIntObj(ext->bevel_type);
  Tcl_ObjSetVar2(interp,toa,ton,to,TCL_LEAVE_ERR_MSG |
		 TCL_GLOBAL_ONLY);

  Tcl_SetStringObj(ton,"BevelRadius",-1);
  to = Tcl_NewDoubleObj(ext->bevel_radius);
  Tcl_ObjSetVar2(interp,toa,ton,to,TCL_LEAVE_ERR_MSG | TCL_GLOBAL_ONLY);

  Tcl_SetStringObj(ton,"Tolerance",-1);
  to = Tcl_NewDoubleObj(ext->glu_sampling_tolerance);
  Tcl_ObjSetVar2(interp,toa,ton,to,TCL_LEAVE_ERR_MSG | TCL_GLOBAL_ONLY);

  Tcl_SetStringObj(ton,"Display",-1);
  to = Tcl_NewIntObj(ext->glu_display_mode);
  Tcl_ObjSetVar2(interp,toa,ton,to,TCL_LEAVE_ERR_MSG | TCL_GLOBAL_ONLY);

  Tcl_SetStringObj(ton,"Height",-1);
  to = Tcl_NewDoubleObj(ext->height);
  Tcl_ObjSetVar2(interp,toa,ton,to,TCL_LEAVE_ERR_MSG | TCL_GLOBAL_ONLY);

  Tcl_IncrRefCount(toa);Tcl_DecrRefCount(toa);
  Tcl_IncrRefCount(ton);Tcl_DecrRefCount(ton);

 return TCL_OK;
}

int
Extrude_convert(Tcl_Interp *interp, mops_object *o)
{
 int mops_status = MOPS_OK;
 mops_custom_object *custom = NULL;
 Extrude *ex = NULL;
 mops_object *bevel = NULL;
 mops_object *extrusion = NULL;
 mops_object *new = NULL;
 mops_level_object *newl = NULL;

 char fname[] = "Extrude_convert";

  custom = (mops_custom_object*)o->object;
  if(!custom)
    return MOPS_ERROR;

  if(custom->type != Extrude_customid)
    return MOPS_ERROR;

  ex = (Extrude *) custom->object;

  /* create parent-level */
  if(!(new = calloc(1,sizeof(mops_object))))
    return MOPS_OUTOFMEM;
  new->type = MOPS_OTLEVEL;
  if(!(newl = calloc(1,sizeof(mops_level_object))))
    return MOPS_OUTOFMEM;
  newl->type = MOPS_LTLEVEL;
  new->object = (void *)newl;
  mops_default_attributes(new);
  mops_copy_attributes(o, new);
  mops_status = mops_link_new_object(new);
  MOPS_CRERROR

  /* enter new level */
  mops_next_object = &(new->down);
  mops_status = mops_add_level_object(new);
  MOPS_CRERROR
  mops_status = mops_add_level_object(NULL);
  MOPS_CRERROR

  new = NULL;
  /* convert extrusions */
  extrusion = ex->extrusions;
  while(extrusion)
    {
      mops_status = mops_copy_object(extrusion, &new);
      MOPS_CRERROR
      mops_status = mops_link_new_object(new);
      MOPS_CRERROR
      extrusion = extrusion->next;
    }

  /* convert bevels */
  bevel = ex->upper_bevels;
  while(bevel)
    {
      mops_status = mops_copy_object(bevel, &new);
      MOPS_CRERROR
	/*      new->movz += ex->height;*/
      mops_status = mops_link_new_object(new);
      MOPS_CRERROR
      bevel = bevel->next;
    }
  bevel = ex->lower_bevels;
  while(bevel)
    {
      mops_status = mops_copy_object(bevel, &new);
      MOPS_CRERROR
      mops_status = mops_link_new_object(new);
      MOPS_CRERROR
      bevel = bevel->next;
    }
  
  /* convert upper cap */
  if(ex->upper_cap)
    {
      mops_status = mops_copy_object(ex->upper_cap, &new);

      if(mops_status)
	{
	  mops_error(interp,mops_status,fname,NULL);
	  return mops_status;
	}
      mops_status = mops_default_attributes(new);
      /*      mops_status = mops_copy_transform(o, new);*/
      mops_status = mops_link_new_object(new);
    }
  /* convert lower cap */
  if(ex->lower_cap)
    {
      mops_status = mops_copy_object(ex->lower_cap, &new);

      if(mops_status)
	{
	  mops_error(interp,mops_status,fname,NULL);
	  return mops_status;
	}

      mops_status = mops_default_attributes(new);
      /*      mops_status = mops_copy_transform(o, new);*/
      mops_status = mops_link_new_object(new);
    }

  /* go up again */
  mops_status = mops_del_level_object();

  mops_status = mops_del_level_object();

 return mops_status;
}


int
Extrude_convertNP_tcmd(ClientData clientData, Tcl_Interp *interp,
		     int argc, char *argv[])
{
 int mops_status = MOPS_OK;
 mops_object *o = NULL;
 mops_list_object *sel = mops_current_selection;
 mops_custom_object *custom = NULL;
 char fname[] = "extrude_convertNP";

  if(!sel) /* oops? */
    return TCL_OK;

  o = sel->object; 
  if(!o)
    return TCL_OK;

  if(o->type != MOPS_OTCUSTOM)
    return TCL_OK;

  custom = (mops_custom_object*)o->object;
  if(!custom)
    return TCL_OK;

  if(custom->type != Extrude_customid)
    return TCL_OK;

  mops_status = Extrude_convert(interp, o);
  if(mops_status)
    {
      mops_error(interp,MOPS_ERROR,fname,"converter failed");
      return TCL_OK;
    }

 return TCL_OK;
}

/* apply rotate & scale to a NURB curves controlv */
void
Extrude_apply_rsNC(mops_object *o)
{
 mops_nurbcurve_object *curve;
 double *cv = NULL, x = 0, y = 0, z = 0, w = 0;
 GLdouble m[16] = {0};
 GLfloat mr[4][4];
 int i = 0, a = 0;

  curve = (mops_nurbcurve_object *)o->object;
  cv = curve->controlv;

  glMatrixMode (GL_MODELVIEW);
  glPushMatrix();

   mops_build_rotmatrix(mr, o->quat);
   glMultMatrixf(&mr[0][0]);
   glScaled(o->scalx,o->scaly,o->scalz);
   glGetDoublev(GL_MODELVIEW_MATRIX, m);
  glPopMatrix();


  if(curve->has_weights)
    {
      for(i = 0; i < curve->length; i++)
	{
	  x = cv[a]; y = cv[a+1]; z = cv[a+2]; w = cv[a+3];

	  cv[a]   = m[0]*x + m[4]*y + m[8]*z + m[12]*w;
	  cv[a+1] = m[1]*x + m[5]*y + m[9]*z + m[13]*w;
	  cv[a+2] = m[2]*x + m[6]*y + m[10]*z + m[14]*w;
	  cv[a+3] = m[3]*x + m[7]*y + m[11]*z + m[15]*w;

	  a += 4;
	}
    }
  else
    {
      w = 1.0;
      for(i = 0; i < curve->length; i++)
	{
	  x = cv[a]; y = cv[a+1]; z = cv[a+2];

	  cv[a]   = m[0]*x + m[4]*y + m[8]*z + m[12]*w;
	  cv[a+1] = m[1]*x + m[5]*y + m[9]*z + m[13]*w;
	  cv[a+2] = m[2]*x + m[6]*y + m[10]*z + m[14]*w;

	  a += 3;
	}
    }

  o->rotx = 0.0;
  o->roty = 0.0;
  o->rotz = 0.0;

  o->quat[0] = 0.0;
  o->quat[1] = 0.0;
  o->quat[2] = 0.0;
  o->quat[3] = 1.0;

  o->scalx = 1.0;
  o->scaly = 1.0;
  o->scalz = 1.0;
  
 return;
}


int
Extrude_provide_cb(mops_object *o, void **target)
{
 int mops_status = MOPS_OK;
 Extrude *ex = NULL;
 mops_custom_object *custom = NULL;

  custom = (mops_custom_object *)o->object;

  ex = (Extrude *)custom->object;

  mops_status = Extrude_notify_cb(o);

  if(mops_status)
    return mops_status;

  /* mops_copy_attributes(o, ex->extrusions); */

  *target = ex->extrusions;

 return mops_status;
}



#ifdef MOPSIDR
void
Extrude_countbmms_cb(mops_object *o)
{
 mops_custom_object *custom = NULL;
 Extrude *ex = NULL;

 if(!o)
   return;

 custom = (mops_custom_object *)o->object;

 if(!custom)
   return;

 ex = (Extrude *)custom->object;

 if(!ex)
   return;

 if((ex->has_upper_cap) || (ex->has_lower_cap))
   idr_bmm.trim++;

 return;
}
#endif
