/*
 * The Mops, a free 3D Tcl/Tk/OpenGL/RenderMan modeler
 *
 * The Mops is copyrighted (C) 1998-1999 by Randolf Schultz
 * (rschultz@informatik.uni-rostock.de)
 *
 * See the file License for details.
 *
 */

#include "mops.h"

/* Hierarchy */

int
mops_link_new_object(mops_object *new_object)
{
 int mops_status = MOPS_OK;
 mops_list_object *lev = mops_current_level;
 mops_object *clevel = NULL;

 clevel = lev->object;

 mops_status = mops_del_level_object();

 /* Are we inside a nurbpatch? */
 if(mops_current_level->object->type == MOPS_OTNURBPATCH)
   { /* Check type of new object*/
     mops_status = mops_add_level_object(clevel);
     if((new_object->type != MOPS_OTNURBCURVE) &&
	(new_object->type != MOPS_OTLEVEL))
       { /* user tried to link some other objecttype
	    than a nurbcurve inside a nurbpatch, tsstss 
	 */
	 return MOPS_HIER;
       }
   }
 else
   if(clevel != mops_root)
     mops_status = mops_add_level_object(clevel);

 if(!clevel)
   { /* first object of current level? */
     mops_status = mops_del_level_object();
     mops_status = mops_add_level_object(new_object);
   }

 /* getclevel() should have set mops_next_object for us        */
 /* or we have done this, when we were linking the last object */

  *mops_next_object = new_object;

  mops_next_object = &(new_object->next);

 return mops_status;
}


int
mops_unlink_object(mops_object *unlink_object)
{
 mops_list_object *lev = mops_current_level;
 mops_object *o = NULL;

  if(lev)
    {
      /* unlink first object of current level? */
      if(lev->object == unlink_object)
	{
	  mops_del_level_object();

	  mops_current_level->object->down = unlink_object->next;

	  /* was it the last object remaining in this level? */
	  if(mops_next_object == &(unlink_object->next))
	    {
	      mops_next_object = &(mops_current_level->object->down);
	    }

	  mops_add_level_object(unlink_object->next);
	  return MOPS_OK;
	}

      /* unlink any other object in current level */
      o = lev->object;
      while(o)
	{
	  if(o->next == unlink_object)
	  {
	    o->next = unlink_object->next;

	    /* happens only when object is the last in current level */
	    if(mops_next_object == &(unlink_object->next))
	    {
	      mops_next_object = &(o->next);
	    }

	    o = NULL;
	  }
	  else
	    o = o->next;
	}
    }

 return MOPS_OK;
}


int
mops_delete_object(mops_object *del)
{
 int mops_status = MOPS_OK; 
 mops_object *target = NULL;
 mops_object *sub = NULL, *sub2 = NULL;
 mops_shader *shader = NULL, *nextshader = NULL;
 mops_tag_object *tag = NULL, *nexttag = NULL;

  if(del)
    {

      if(del->refcount)
	{
	  return MOPS_REFCOUNT;
	}
      if(del->down)
	{
	  sub = del->down;
	  while(sub)
	    {
	      sub2 = sub->next;
	      mops_delete_object(sub);
	      sub = sub2;
	    }
	}

      /* delete the real object & attributes */
      switch(del->type)
	{
	case MOPS_OTCUSTOM:
	  mops_status = mops_free_custom((mops_custom_object *)del->object);
	  break;
	case MOPS_OTNURBPATCH:
	  mops_status =
	    mops_free_nurbpatch((mops_nurbpatch_object *)del->object);
	  break;
	case MOPS_OTNURBCURVE:
	  mops_status =
	    mops_free_nurbcurve((mops_nurbcurve_object *)del->object);
	  break;
	case MOPS_OTLIGHT:
	  mops_status = mops_free_light((mops_light_object *)del->object);
	  break;
	case MOPS_OTLEVEL:
	  mops_status = mops_free_level((mops_level_object *)del->object);
	  break;
	case MOPS_OTSPHERE:
	case MOPS_OTCYLINDER:
	case MOPS_OTDISK:
	case MOPS_OTCONE:
	case MOPS_OTTORUS:
	case MOPS_OTHYPERBOLOID:
	case MOPS_OTPARABOLOID:
	  mops_status = mops_free_quadric(del->object, del->type);
	  break;
	case MOPS_OTBOX:
	  mops_status = mops_free_box((mops_box_object *)del->object);
	  break;
	case MOPS_OTINSTANCE:
	  target = (mops_object *)del->object;
	  target->refcount--;
	  break;
	}
      
      /* delete selected points */
      if(del->selp)
	mops_clear_point_selection(del);

      if(del->name)
	free(del->name);

      /* delete shaders */
      shader = del->shaders;
      while(shader)
	{
	  nextshader = shader->next;
	  mops_free_shader(shader);
	  shader = nextshader;
	}
      /* delete tags */
      tag = del->tags;
      while(tag)
	{
	  nexttag = tag->next;
	  mops_free_tag(tag);
	  tag = nexttag;
	}
      /* finally, delete the generic object */
      free(del);
    }
 
 return mops_status;
}


int
mops_delete_multi_objects(mops_object *del)
{
 int mops_status = MOPS_OK; 
 mops_object *tmp = NULL;

  while(del)
    {
      tmp = del;
      del = del->next;
      mops_status = mops_delete_object(tmp);
      if(mops_status)
	return mops_status;
    }
 
 return mops_status;
}


int
mops_clear_scene()
{
 mops_object *o = mops_root, *o2;

  o = o->next;

  while(o)
    {
      o2 = o->next;
      mops_delete_object(o);    
      o = o2;
    }
  o = mops_root;

  mops_next_object = &(o->next);

  o->next = NULL;

  mops_del_all_level_objects();

  mops_clear_undo_info();

  if(mops_current_selection)
    mops_free_selection();

 return MOPS_OK;
}


int
mops_copy_object(mops_object *source, mops_object **dest)
{
 int mops_status = MOPS_OK; 
 mops_object *sub = NULL, *sub2 = NULL, *temp = NULL, *new = NULL;
 mops_shader *shader = NULL, **newshaderptr = NULL;
 mops_tag_object *tag = NULL, **newtagptr = NULL;

 if(!source)
  return MOPS_OK;

 /* copy generic object */
 if(!(new = calloc(1,sizeof(mops_object))))
   return MOPS_OUTOFMEM;

 *dest = new;

 memcpy(new, source, sizeof(mops_object));
 /* danger! links point to original hierarchy */

 new->next = NULL;
 new->down = NULL;
 new->shaders = NULL;
 new->selp = NULL;
 new->tags = NULL;

 new->refcount = 0;

 /* copy name */
 if(source->name)
   {
     if(!(new->name = calloc(strlen(source->name)+1,sizeof(char))))
       {
	 free(new);
	 return MOPS_OUTOFMEM;
       }

     strcpy(new->name,source->name);
   }

 /* copy shaders */
 shader = source->shaders;
 newshaderptr = &(new->shaders);
 while(shader)
   {
     mops_status = mops_copy_shader(shader,newshaderptr);
     newshaderptr = &((*newshaderptr)->next);
     *newshaderptr = NULL;
     shader = shader->next;
   }

 /* copy tags */
 tag = source->tags;
 newtagptr = &(new->tags);
 while(tag)
   {
     mops_status = mops_copy_tag(tag,newtagptr);
     newtagptr = &((*newtagptr)->next);
     *newtagptr = NULL;
     tag = tag->next;
   }

 /* copy the real object */
 switch(source->type)
   {
   case MOPS_OTCUSTOM:
     mops_status = mops_copy_custom
       ((mops_custom_object *)source->object,
	(mops_custom_object **)&new->object);
     break;
   case MOPS_OTNURBPATCH:
     mops_status = mops_copy_nurbpatch
       ((mops_nurbpatch_object *)source->object,
	(mops_nurbpatch_object **)&new->object);
     break;
   case MOPS_OTNURBCURVE:
     mops_status = mops_copy_nurbcurve
       ((mops_nurbcurve_object *)source->object,
	(mops_nurbcurve_object **)&new->object);
     break;
   case MOPS_OTLIGHT:
     mops_status = mops_copy_light
       ((mops_light_object *)source->object,
	(mops_light_object **)&new->object);
     break;
   case MOPS_OTLEVEL:
     mops_status = mops_copy_level
       ((mops_level_object *)source->object,
	(mops_level_object **)&new->object);
     break;
   case MOPS_OTSPHERE:
   case MOPS_OTCYLINDER:
   case MOPS_OTDISK:
   case MOPS_OTCONE:
   case MOPS_OTTORUS:
   case MOPS_OTHYPERBOLOID:
   case MOPS_OTPARABOLOID:
     mops_status = mops_copy_quadric(source->object,
				     &new->object,
				     source->type);
     break;
   case MOPS_OTBOX:
     mops_status = mops_copy_box
       ((mops_box_object *)source->object,
	(mops_box_object **)&new->object);
     break;
   case MOPS_OTINSTANCE:
       new->object = source->object;
       ((mops_object *)new->object)->refcount++;
     break;
   }
 
 if(source->down)
   {
     sub = source->down;
     
     if((mops_status = mops_copy_object(sub, &new->down)))
       return mops_status;
     temp = new->down;
     sub = sub->next;
     
     while(sub)
       {
	 sub2 = sub->next;
	 if((mops_status = mops_copy_object(sub, &temp->next)))
	   return mops_status;
	 temp = temp->next;
	 
	 sub = sub2;
       } /* while */

     if(source->type == MOPS_OTCUSTOM)
       {
	 mops_notify_custom(new);
       }

   } /* if */

 if(source->type == MOPS_OTCUSTOM)
   {
     new->modified = MOPS_TRUE;
   }

 return mops_status;
}


int
mops_add_level_object(mops_object *new_object)
{
 mops_list_object *lev = NULL;

  if(!(lev = calloc(1,sizeof(mops_list_object))))
    return MOPS_OUTOFMEM;

  lev->object = new_object;

  if(mops_current_level)
    lev->next = mops_current_level;

  mops_current_level = lev;  

 return MOPS_OK;
}


int
mops_del_level_object()
{
 mops_list_object *lev = mops_current_level;

  if(lev->next)
  {
    mops_current_level = lev->next;
    free(lev);
  }

 return MOPS_OK;
}


int
mops_del_all_level_objects()
{
 mops_list_object *lev = mops_current_level;

  if(lev)
    while(lev->next)
      {
	mops_current_level = lev->next;
	free(lev);
	lev = mops_current_level;
      }

 return MOPS_OK;
}


int
mops_set_oname_tcmd(ClientData clientData, Tcl_Interp * interp,
		    int argc, char *argv[])
{
 mops_list_object *sel = mops_current_selection;
 mops_object *o = NULL;
 char fname[] = "set_oname";

 if(!sel)
   return TCL_OK;
 
 if(!(o = sel->object))
   return TCL_OK;

 if((o->type == MOPS_OTROOT) || (o->type == MOPS_OTVIEW))
   return TCL_OK;

 if(o->name)
   {
     free(o->name);
     o->name = NULL;
   }

 if(argc>1)
   {
     if(!(o->name = calloc(strlen(argv[1])+1,sizeof(char))))
       {
	 mops_error(interp,MOPS_OUTOFMEM,fname,NULL);
	 return TCL_OK;
       }
     strcpy(o->name,argv[1]);
   }

 return TCL_OK;
}

int
mops_get_objectname(mops_object *o, int typeonly, char **result)
{
 int mops_status = MOPS_OK;
 mops_level_object *level = NULL;
 Tcl_DString ds;
 char *temp = NULL;
 
 Tcl_DStringInit(&ds);

  if((typeonly != MOPS_TRUE) && o->name)
    {
      Tcl_DStringAppend(&ds, o->name,-1);
      Tcl_DStringAppend(&ds, " ",-1);
      switch(o->type)
	{
	case MOPS_OTCUSTOM:
	  Tcl_DStringAppend(&ds, "(CU)",-1);
	  break;
	case MOPS_OTNURBPATCH:
	  Tcl_DStringAppend(&ds, "(NP)",-1);
	  break;
	case MOPS_OTNURBCURVE:
	  Tcl_DStringAppend(&ds, "(NC)",-1);
	  break;
	case MOPS_OTLEVEL:
	  level = (mops_level_object*)o->object;
	  switch(level->type)
	    {
	    case MOPS_LTUNION:
	      Tcl_DStringAppend(&ds, "(UN)",-1);
	      break;
	    case MOPS_LTINTERSECTION:
	      Tcl_DStringAppend(&ds, "(IN)",-1);
	      break;
	    case MOPS_LTDIFFERENCE:
	      Tcl_DStringAppend(&ds, "(DI)",-1);
	      break;
	    case MOPS_LTPRIMITIVE:
	      Tcl_DStringAppend(&ds, "(PR)",-1);
	      break;
	    case MOPS_LTLEVEL:
	      Tcl_DStringAppend(&ds, "(LE)",-1);
	      break;
	    case MOPS_LTTRIMLOOP:
	      Tcl_DStringAppend(&ds, "(TL)",-1);
	      break;
	    }
	  break;
	case MOPS_OTLIGHT:
	  Tcl_DStringAppend(&ds, "(LI)",-1);
	  break;
	case MOPS_OTSPHERE:
	  Tcl_DStringAppend(&ds, "(SP)",-1);
	  break;
	case MOPS_OTCONE:
	  Tcl_DStringAppend(&ds, "(CO)",-1);
	  break;
	case MOPS_OTCYLINDER:
	  Tcl_DStringAppend(&ds, "(CY)",-1);
	  break;
	case MOPS_OTDISK:
	  Tcl_DStringAppend(&ds, "(DI)",-1);
	  break;
	case MOPS_OTTORUS:
	  Tcl_DStringAppend(&ds, "(TO)",-1);
	  break;
	case MOPS_OTPARABOLOID:
	  Tcl_DStringAppend(&ds, "(PA)",-1);
	  break;
	case MOPS_OTHYPERBOLOID:
	  Tcl_DStringAppend(&ds, "(HY)",-1);
	  break;
	case MOPS_OTBOX:
	  Tcl_DStringAppend(&ds, "(BX)",-1);
	  break;
	case MOPS_OTVIEW:
	  Tcl_DStringAppend(&ds, " (VI)",-1);
	  break;
	case MOPS_OTINSTANCE:
	  Tcl_DStringAppend(&ds, " (IN)",-1);
	  break;
	}
    }
  else
    {
      switch(o->type)
	{
	case MOPS_OTCUSTOM:
	  mops_status = mops_get_custom_objectname(o, &temp);
	  if(mops_status) /* umm? */
	    Tcl_DStringAppend(&ds, "Custom",-1);
	  else
	    Tcl_DStringAppend(&ds, temp,-1);
	  break;
	case MOPS_OTNURBPATCH:
	  Tcl_DStringAppend(&ds, "NURBPatch",-1);
	  break;
	case MOPS_OTNURBCURVE:
	  Tcl_DStringAppend(&ds, "NURBCurve",-1);
	  break;
	case MOPS_OTLEVEL:
	  level = (mops_level_object*)o->object;
	  switch(level->type)
	    {
	    case MOPS_LTUNION:
	      Tcl_DStringAppend(&ds, "Union",-1);
	      break;
	    case MOPS_LTINTERSECTION:
	      Tcl_DStringAppend(&ds, "Intersection",-1);
	      break;
	    case MOPS_LTDIFFERENCE:
	      Tcl_DStringAppend(&ds, "Difference",-1);
	      break;
	    case MOPS_LTPRIMITIVE:
	      Tcl_DStringAppend(&ds, "Primitive",-1);
	      break;
	    case MOPS_LTLEVEL:
	      Tcl_DStringAppend(&ds, "Level",-1);
	      break;
	    case MOPS_LTTRIMLOOP:
	      Tcl_DStringAppend(&ds, "TrimLoop",-1);
	      break;
	    }
	  break;
	case MOPS_OTLIGHT:
	  Tcl_DStringAppend(&ds, "Light",-1);
	  break;
	case MOPS_OTSPHERE:
	  Tcl_DStringAppend(&ds, "Sphere",-1);
	  break;
	case MOPS_OTCONE:
	  Tcl_DStringAppend(&ds, "Cone",-1);
	  break;
	case MOPS_OTCYLINDER:
	  Tcl_DStringAppend(&ds, "Cylinder",-1);
	  break;
	case MOPS_OTDISK:
	  Tcl_DStringAppend(&ds, "Disk",-1);
	  break;
	case MOPS_OTTORUS:
	  Tcl_DStringAppend(&ds, "Torus",-1);
	  break;
	case MOPS_OTPARABOLOID:
	  Tcl_DStringAppend(&ds, "Paraboloid",-1);
	  break;
	case MOPS_OTHYPERBOLOID:
	  Tcl_DStringAppend(&ds, "Hyperboloid",-1);
	  break;
	case MOPS_OTBOX:
	  Tcl_DStringAppend(&ds, "Box",-1);
	  break;
	case MOPS_OTINSTANCE:
	  Tcl_DStringAppend(&ds, "Instance",-1);
	  break;
	case MOPS_OTVIEW:
	  Tcl_DStringAppend(&ds,
	   Togl_Ident(((mops_view_info*)o->object)->togl),-1);
	  break;

	}

    }

  if(!(*result = calloc(Tcl_DStringLength(&ds)+1,sizeof(char))))
    return MOPS_OUTOFMEM;

  strcpy(*result,Tcl_DStringValue(&ds));

  Tcl_DStringFree(&ds);

 return MOPS_OK;
}


int
mops_get_clevel_tcmd(ClientData clientData, Tcl_Interp * interp,
		  int argc, char *argv[])
{
 int mops_status = MOPS_OK;
 mops_object *o = mops_current_level->object;
 char *name = NULL;
 char fname[] = "get_clevel";

  /* check args */
  if(argc != 2)
    {
      mops_error(interp, MOPS_ARGS, fname, "varname");
      return TCL_OK;
    }
  
  Tcl_UnsetVar(interp,argv[1], TCL_LEAVE_ERR_MSG);

  Tcl_SetVar(interp, argv[1], "",
	     TCL_APPEND_VALUE | TCL_LEAVE_ERR_MSG);

  if(mops_current_level->object != mops_root)
    Tcl_SetVar(interp, argv[1], "..",
	       TCL_APPEND_VALUE | TCL_LEAVE_ERR_MSG);
  else
    Tcl_SetVar(interp, argv[1], "Scene",
	       TCL_APPEND_VALUE | TCL_LEAVE_ERR_MSG);

  while(o)
    {
      if(o != mops_root)
	{
	  name = NULL;
	  mops_status = mops_get_objectname(o, MOPS_FALSE, &name);

	  if(!mops_status)
	    Tcl_SetVar(interp, argv[1], name, TCL_APPEND_VALUE |
		       TCL_LIST_ELEMENT | TCL_LEAVE_ERR_MSG);

	  if(name)
	    free(name);
	}

      mops_next_object = &(o->next);
      o = o->next;  
    }


 return TCL_OK;
}


int
mops_godown_tcmd(ClientData clientData, Tcl_Interp * interp,
		  int argc, char *argv[])
{
  int mops_status = MOPS_OK;
  int i = 0, j = 1, argvi = 0, start = 1;
  mops_object *o = mops_current_level->object;
  char fname[] = "go_down";

  /* check args */
  if(argc != 2)
    {
      mops_error(interp, MOPS_ARGS, fname, "index");
      return TCL_OK;
    }

  /* sublevel? */
  if(mops_root != o)
    {
      /* ".." selected? -> go up instead! */
      if(atoi(argv[1]) == 0)
	{

	  mops_status = mops_del_level_object();
	  mops_status = mops_del_level_object();
	  o = mops_current_level->object;
	  if(o)
	    {
	      while(o->next)
		o = o->next;
	      mops_next_object = &(o->next);
	    }
	
	  return TCL_OK;
	}
    }
  else
    {
      if(atoi(argv[1]) == 0)
	{
	  if(mops_root->down)
	    { /* dive into scene object, but do not set mops_next_object! */
	      mops_status = mops_add_level_object(mops_root);
	      MOPS_CRERROR
		mops_status = mops_add_level_object(mops_root->down);
	      MOPS_CRERROR
		return TCL_OK;
	    }
	  else
	    return TCL_OK;
	}
      
      if(o->next)
	o = o->next;
    }

  for(i=start;i<argc;i++)
    {
      argvi = atoi(argv[i]);
      while(j != argvi)
	{
	  j++;

	  if(o)
	    o = o->next;

	  if(!o)
	    break;
	  /* no check for EndOfList, I know;
	     we _have to_ be fast here!      */
	}

      if(!o)
	break;

      /* found a (the first) selected object: lets dive in! */
      if(o->down)
	{
	  mops_status = mops_add_level_object(o);
	  MOPS_CRERROR
	  mops_status = mops_add_level_object(o->down);
	  MOPS_CRERROR
	  o = o->down;
	  while(o->next)
	    o = o->next;
	  mops_next_object = &(o->next);

	  return TCL_OK;
	}
      else /* Entering an empty Level! */
	if(o->type == MOPS_OTLEVEL || o->type == MOPS_OTNURBPATCH ||
	   o->type == MOPS_OTLIGHT || o->type == MOPS_OTCUSTOM)
	{
	  mops_next_object = &(o->down);
	  mops_status = mops_add_level_object(o);
	  MOPS_CRERROR
	  mops_status = mops_add_level_object(NULL);
	  MOPS_CRERROR
	  return TCL_OK;
	}

    }

 return TCL_OK;
}


int
mops_goup_tcmd(ClientData clientData, Tcl_Interp * interp,
		  int argc, char *argv[])
{
 int mops_status = MOPS_OK;

 mops_status = mops_del_level_object();

 mops_status = mops_del_level_object();

 return TCL_OK;
}


int
mops_gotop_tcmd(ClientData clientData, Tcl_Interp * interp,
		  int argc, char *argv[])
{
 int mops_status = MOPS_OK;

  mops_status = mops_del_all_level_objects();

 return TCL_OK;
}

int
mops_swapob_tcmd(ClientData clientData, Tcl_Interp * interp,
	       int argc, char *argv[])
{
 mops_list_object *sel = mops_current_selection;
 mops_list_object *lev = mops_current_level;
 mops_object *o=NULL, *o2 = NULL, *t = NULL;
 mops_object **p1 = NULL, **p2 = NULL, **p3 = NULL, **p4 = NULL;
 mops_object *temp = NULL;
 char fname[] = "swapOb";

  if((!sel) || (!sel->next))
    {
      mops_error(interp, MOPS_ERROR, fname, "need two selected objects!");
      return TCL_OK;
    }
  
  o = sel->object;
  o2 = sel->next->object;

  if(!o || !o2)
    {
      mops_error(interp, MOPS_ERROR, fname, "need two selected objects!");
      return TCL_OK;
    }

  if((o == mops_root) || (o2 == mops_root))
    {
      mops_error(interp, MOPS_ERROR, fname, "cannot swap root!");
      return TCL_OK;
    }

  /* find pointers */
  p2 = &(o->next);
  p4 = &(o2->next);

  if(lev && lev->object == o)
    {
      p1 = &(lev->next->object->down);
    }
  else
    {
      t = lev->object;
      while(t)
	{
	  if(t->next == o)
	    {
	      p1 = &(t->next);
	      break;
	    }
	  t = t->next;
	}
    }

  if(lev && lev->object == o2)
    {
      p3 = &(lev->next->object->down);
    }
  else
    {
      t = lev->object;
      while(t)
	{
	  if(t->next == o2)
	    {
	      p3 = &(t->next);
	      break;
	    }
	  t = t->next;
	}
    }

  /* swap */
  temp = *p1;
  *p1 = *p3;
  *p3 = temp;

  temp = *p2;
  *p2 = *p4;
  *p4 = temp;

  if(mops_current_level->object == o)
    {
      mops_del_level_object();
      mops_add_level_object(o2);
    }


  mops_notify_parent();

  mops_free_selection();

 return TCL_OK;
}
