/*
 * The Mops, a free 3D Tcl/Tk/OpenGL/RenderMan modeler
 *
 * The Mops is copyrighted (C) 1998-1999 by Randolf Schultz
 * (rschultz@informatik.uni-rostock.de)
 *
 * See the file License for details.
 *
 */

#include <limits.h>
#include <float.h>
#include "mops.h"

/* ICurve - Interpolating C2 Cubic B-Spline Curves */

/* The Name */
static char *ICurve_name = "ICurve";

static char *ICurve_customid = NULL;

/* prototypes */
typedef struct ICurve_s
{
 int closed;
 int imode;
 int iorder;
 double iparam;
 mops_object *curve;
 double glu_sampling_tolerance;
 /* legacy */
 double *controlv;
 int length;
} ICurve;

int ICurve_create_cb(Tcl_Interp *interp, int argc, char *argv[], void **c);
int ICurve_delete_cb(void *c);
int ICurve_copy_cb(void *source, void **dest);
int ICurve_draww_cb(struct Togl *togl, mops_object *o);
int ICurve_drawh_cb(struct Togl *togl, mops_object *o);
int ICurve_draws_cb(struct Togl *togl, mops_object *o);
int ICurve_getpoint_cb(mops_object *o, double objX, double objY, double objZ);
int ICurve_read_cb(FILE *fileptr, mops_custom_object *custom);
int ICurve_write_cb(FILE *fileptr, mops_custom_object *custom);
int ICurve_getprop_cb(Tcl_Interp *interp);
int ICurve_set_attrprop_tcmd(ClientData clientData, Tcl_Interp *interp,
			     int argc, char *argv[]);
int ICurve_get_attrprop_tcmd(ClientData clientData, Tcl_Interp *interp,
			     int argc, char *argv[]);

int ICurve_notify_cb(mops_object *o);
int ICurve_convert_length(double newlength, ICurve *ic);
int ICurve_convert(Tcl_Interp *interp, mops_object *o);
int ICurve_interpolateC2C(double iparam, int closed, int length,
			  double *controlv, mops_nurbcurve_object **c);
int ICurve_interpolateG4D(int iorder, int length,
			  double *controlv, mops_nurbcurve_object **c);
int ICurve_interpolateG4DC(int iorder, int length, double iparam,
			  double *controlv, mops_nurbcurve_object **c);
int ICurve_convertNC_tcmd(ClientData clientData, Tcl_Interp *interp,
			 int argc, char *argv[]);
int ICurve_provide_cb(mops_object *o, void **target);

/* functions */
int
ICurve_Init(Tcl_Interp *interp)
{
 int mops_status = MOPS_OK;
 Tcl_HashEntry *entry;
 char fname[] = "ICurve_Init";

  mops_status = mops_register_custom_objecttype(
				     ICurve_name,
				     ICurve_create_cb,
				     ICurve_draws_cb, 
				     ICurve_drawh_cb,
				     ICurve_draww_cb,
				     ICurve_delete_cb,
				     ICurve_copy_cb,
				     ICurve_getprop_cb,
				     ICurve_getpoint_cb,
				     NULL, /* does not appear in RIBs */
				     ICurve_read_cb,
				     ICurve_write_cb,
				     ICurve_notify_cb
				     );

  if(mops_status)
    {
      mops_error(interp, mops_status, fname, NULL);
      return TCL_ERROR;
    }

  /* get custom-id from name */
  entry = Tcl_FindHashEntry(&mops_custom_id_ht, ICurve_name);

  if (entry != NULL)
    {
      ICurve_customid = Tcl_GetHashValue(entry);
    }
  else
    {
      mops_error(interp, MOPS_HASH, fname, NULL);
      return TCL_OK;
    }

  /* register converter */
  mops_status = mops_register_converter(interp,ICurve_name,ICurve_convert);

  /* register some C-functions as Tcl-Commands */
  /* they deal with the ICurve-Attributes Property GUI */
  Tcl_CreateCommand (interp, "ICurve_getICurveAttr",
		     ICurve_get_attrprop_tcmd,
		     (ClientData) NULL, (Tcl_CmdDeleteProc *) NULL);
  Tcl_CreateCommand (interp, "ICurve_setICurveAttr",
		     ICurve_set_attrprop_tcmd,
		     (ClientData) NULL, (Tcl_CmdDeleteProc *) NULL);

  /* another command, convert to a NURBCurve */
  Tcl_CreateCommand (interp, "ICurveCNC", ICurve_convertNC_tcmd,
		     (ClientData) NULL, (Tcl_CmdDeleteProc *) NULL);

  /* register provider */
  mops_status = mops_register_provider(interp, ICurve_name,
				       MOPS_OTNURBCURVE, ICurve_provide_cb);


 return TCL_OK;
}


int
ICurve_create_cb(Tcl_Interp *interp, int argc, char *argv[], void **c)
{
 int mops_status = MOPS_OK;
 int a, b, i, stride, got_object = MOPS_FALSE;
 double *controlv;
 mops_list_object *sel = mops_current_selection;
 mops_object *curve = NULL, *ncurve = NULL;
 mops_nurbcurve_object *co = NULL;

 ICurve *new = NULL;
 char fname[] = "ICurve_create";

  if(!sel)
    {
      return MOPS_NOSEL;
    }

  if(sel->object->type != MOPS_OTNURBCURVE)
    {
      mops_status = mops_get_object(sel->object, MOPS_OTNURBCURVE, &curve);
      if(curve)
	got_object = MOPS_TRUE;
    }
  else
    {
      curve = sel->object;
    }

  if(!curve)
    {
      mops_error(interp, MOPS_ERROR, fname, "object is not a NURBCurve");
      return MOPS_ERROR;
    }

  co = (mops_nurbcurve_object *)curve->object;

  if(!(new = calloc(1, sizeof(ICurve))))
    {
      mops_error(interp, MOPS_OUTOFMEM, fname, NULL);
      return TCL_ERROR;
    }

  if(!(controlv = calloc(1, co->length*3*sizeof(double))))
    {
      free(new);
      mops_error(interp, MOPS_OUTOFMEM, fname, NULL);
      return TCL_ERROR;
    }

  if(!(ncurve = calloc(1, sizeof(mops_object))))
    {
      free(controlv); free(new);
      mops_error(interp, MOPS_OUTOFMEM, fname, NULL);
      return TCL_ERROR;
    }

  mops_default_attributes(ncurve);
  ncurve->type = MOPS_OTNURBCURVE;

  stride = co->has_weights?4:3;
  a = 0; b = 0;
  for(i = 0; i < co->length; i++)
    {
      controlv[a] = co->controlv[b];
      controlv[a+1] = co->controlv[b+1];
      controlv[a+2] = co->controlv[b+2];
      a += 3;
      b += stride;
    }

  new->glu_sampling_tolerance = co->glu_sampling_tolerance;
  new->iparam = 1.0/8.0;
  new->iorder = 4;
  new->closed = co->closed;

  mops_status = ICurve_interpolateC2C(new->iparam, new->closed,
				      co->length,
				      controlv,
				      (mops_nurbcurve_object **)(&(ncurve->object)));
  new->curve = ncurve;

  ((mops_nurbcurve_object *)ncurve->object)->glu_sampling_tolerance =
    new->glu_sampling_tolerance;

  if(mops_status)
    {
      free(new->curve); free(controlv); free(new);
      mops_error(interp, mops_status, fname, NULL);
      return TCL_ERROR;
    }

  *c = new;

  if(got_object)
    mops_delete_object(curve);

  free(controlv);

 return mops_status;
}


int
ICurve_delete_cb(void *c)
{
 int mops_status = MOPS_OK;
 ICurve *del = NULL;


  if(!c)
    return MOPS_OK;

  del = (ICurve *)c;

  if(del->curve)
    mops_delete_object(del->curve);

  free(c);

 return mops_status;
}


int
ICurve_copy_cb(void *source, void **dest)
{
 int mops_status = MOPS_OK;
 ICurve *new = NULL, *old = NULL;
 

  old = (ICurve *)source;

  /* copy custom object */
  if(!(new = calloc(1,sizeof(ICurve))))
    return MOPS_OUTOFMEM;
  
  memcpy(new, source, sizeof(ICurve));
  /* danger, links point to original hierarchy! */

  new->curve = NULL;

  *dest = new;

 return mops_status;
}


int
ICurve_interpolateC2C(double iparam, int closed, int length,
		      double *controlv, mops_nurbcurve_object **c)
{
 int mops_status = MOPS_OK;
 int a,b,i,j;
 int nlength;
 double v[3], vlen = 0.0;
 double *ncontrolv = NULL, *ccontrolv = NULL;
 double *knotv, knot = 0.0, *lengths, chordlength = 0.0;
 mops_nurbcurve_object *new = NULL;

  nlength = length+2;
  
  if(closed)
    {
      nlength++;
      if(!(ccontrolv = calloc((length+1)*3, sizeof(double))))
	return MOPS_OUTOFMEM;

      memcpy(&(ccontrolv[0]),&(controlv[0]),length*3*sizeof(double));
      memcpy(&(ccontrolv[length*3]),&(controlv[0]),3*sizeof(double));

    }
  
  if(!(ncontrolv = calloc(nlength*3, sizeof(double))))
    return MOPS_OUTOFMEM;

  /* create knotvector */
  if(!(knotv = calloc(nlength+4, sizeof(double))))
    return MOPS_OUTOFMEM;

  a = 0;
  if(!closed)
    {
      if(!(lengths = calloc(length-1, sizeof(double))))
	return MOPS_OUTOFMEM;
      for(i = 0; i < (length-1); i++)
	{
      
	  lengths[i] = MOPS_VLEN((controlv[a+3] - controlv[a]),
				 (controlv[a+4] - controlv[a+1]),
				 (controlv[a+5] - controlv[a+2]));

	  chordlength += lengths[i];
	  a += 3;
	}
    }
  else
    {
      if(!(lengths = calloc(length, sizeof(double))))
	return MOPS_OUTOFMEM;
      for(i = 0; i < length; i++)
	{
      
	  lengths[i] = MOPS_VLEN((ccontrolv[a+3] - ccontrolv[a]),
				 (ccontrolv[a+4] - ccontrolv[a+1]),
				 (ccontrolv[a+5] - ccontrolv[a+2]));

	  chordlength += lengths[i];
	  a += 3;
	}
    }

  for(i = 0; i < 4; i++)
    knotv[i] = 0.0;
  j = 0;
  knot = 0.0;
  for(i = 4; i < nlength; i++)
    {
      knot += lengths[j]/chordlength/*1.0/(nlength-3)*/;
      knotv[i] = knot;
      j++;
    }
  for(i = nlength; i < nlength+4; i++)
    knotv[i] = 1.0;

  free(lengths);

  /* create first two and last two controls */
  if(!closed)
    {
      ncontrolv[0] = controlv[0];
      ncontrolv[1] = controlv[1];
      ncontrolv[2] = controlv[2];
      
      v[0] = controlv[3]-controlv[0];
      v[1] = controlv[4]-controlv[1];
      v[2] = controlv[5]-controlv[2];

      vlen = MOPS_V3LEN(v);
      vlen *= iparam;

      MOPS_V3SCAL(v, vlen)

      ncontrolv[3] = controlv[0]+v[0];
      ncontrolv[4] = controlv[1]+v[1];
      ncontrolv[5] = controlv[2]+v[2];

      a = (nlength-2)*3;
      b = (length-2)*3;

      v[0] = controlv[b+3]-controlv[b];
      v[1] = controlv[b+4]-controlv[b+1];
      v[2] = controlv[b+5]-controlv[b+2];

      vlen = MOPS_V3LEN(v);
      vlen *= iparam;

      MOPS_V3SCAL(v, vlen)

      ncontrolv[a]   = controlv[b+3]-v[0];
      ncontrolv[a+1] = controlv[b+4]-v[1];
      ncontrolv[a+2] = controlv[b+5]-v[2];

      a = (nlength-1)*3;
      b = (length-1)*3;
      ncontrolv[a]   = controlv[b];
      ncontrolv[a+1] = controlv[b+1];
      ncontrolv[a+2] = controlv[b+2];
    }
  else
    {
      ncontrolv[0] = controlv[0];
      ncontrolv[1] = controlv[1];
      ncontrolv[2] = controlv[2];

      b = (length-1)*3;
      v[0] = controlv[3]-controlv[b];
      v[1] = controlv[4]-controlv[b+1];
      v[2] = controlv[5]-controlv[b+2];

      vlen = MOPS_V3LEN(v);
      vlen *= iparam;

      MOPS_V3SCAL(v, vlen)

      ncontrolv[3] = controlv[0]+v[0];
      ncontrolv[4] = controlv[1]+v[1];
      ncontrolv[5] = controlv[2]+v[2];

      a = (nlength-2)*3;
      ncontrolv[a]   = controlv[0]-v[0];
      ncontrolv[a+1] = controlv[1]-v[1];
      ncontrolv[a+2] = controlv[2]-v[2];

      a = (nlength-1)*3;
      ncontrolv[a]   = controlv[0];
      ncontrolv[a+1] = controlv[1];
      ncontrolv[a+2] = controlv[2];
    }

  if(!closed)
    {
      mops_status = mops_nb_SolveTridiagonal(length-1, controlv,
					     knotv, ncontrolv);
    }
  else
    {
      mops_status = mops_nb_SolveTridiagonal(length, ccontrolv,
					     knotv, ncontrolv);
      free(ccontrolv);
    }

  mops_status = mops_create_nurbcurve(4, nlength, MOPS_KTCUSTOM,
				      MOPS_FALSE, ncontrolv, knotv, &new);

  if(!mops_status)
    *c = new;

 return mops_status;
}


int
ICurve_interpolateG4D(int iorder, int length,
		      double *controlv, mops_nurbcurve_object **c)
{
 int mops_status = MOPS_OK;
 int a,i,j,k,order/*,iN,index*/;
 int nlength;
 double *ncontrolv = NULL;
 double *knotv, knot = 0.0, *lengths, chordlength = 0.0, *vk = NULL;
 mops_nurbcurve_object *new = NULL;


  nlength = length;
  
  if(iorder > nlength)
    order = nlength;
  else
    order = iorder;

  if(!(ncontrolv = calloc(nlength*4, sizeof(double))))
    return MOPS_OUTOFMEM;

  /* create knotvector */
  if(!(knotv = calloc(nlength+order, sizeof(double))))
    return MOPS_OUTOFMEM;

  a = 0;
  if(!(lengths = calloc(length-1, sizeof(double))))
    return MOPS_OUTOFMEM;
  for(i = 0; i < (length-1); i++)
    {
      
      lengths[i] = MOPS_VLEN((controlv[a+3] - controlv[a]),
			     (controlv[a+4] - controlv[a+1]),
			     (controlv[a+5] - controlv[a+2]));

      chordlength += lengths[i];

      a += 3;
    }

  if(!(vk = calloc(length, sizeof(double))))
    return MOPS_OUTOFMEM;

  vk[0] = 0.0;
  j = 0;
  knot = 0.0;
  for(i = 1; i < (length-1); i++)
    {
      knot += lengths[j]/chordlength;
      vk[i] = knot;
      j++;
    }
  vk[length-1] = 1.0;
  
  for(i = 1; i < nlength-(order-1); i++)
    {
      knotv[i+(order-1)] = 0.0;
      for(k = i; k < (i+(order-1)); k++)
	{
	  knotv[i+(order-1)] += vk[k];
	}
      knotv[i+(order-1)] /= (order-1);
    }

  for(i = 0; i < order; i++)
    knotv[i] = 0.0;
  for(i = nlength; i < nlength+order; i++)
    knotv[i] = 1.0;

  free(lengths);

  for(i = 0; i < length; i++)
    {
      memcpy(&(ncontrolv[i*4]),&(controlv[i*3]),3*sizeof(double));
      ncontrolv[i*4+3] = 1.0;
    }

  mops_status = mops_nb_GlobalInterpolation4D(nlength-1, ncontrolv, vk,
					      knotv, order-1);

  if(mops_status)
    { free(vk); free(ncontrolv); free(knotv); return mops_status; }

  mops_status = mops_create_nurbcurve(order, nlength, MOPS_KTCUSTOM,
				      MOPS_TRUE, ncontrolv, knotv, &new);
  
  if(!mops_status)
    *c = new;

  free(vk);

 return mops_status;
}

/* Global 4D Closed */
int
ICurve_interpolateG4DC(int iorder, int length, double iparam,
		       double *controlv, mops_nurbcurve_object **c)
{
 int mops_status = MOPS_OK;
 int a, i, j, order = 0, iN, index;
 int nlength;
 double *ncontrolv = NULL;
 double *knotv, knot = 0.0, *lengths, chordlength = 0.0, *vk = NULL;
 double D[3] = {0};
 mops_nurbcurve_object *new = NULL;

  if(iorder > (length+3))
    order = length + 3;
  else
    order = iorder;

  nlength = length+3;

  if(!(ncontrolv = calloc(nlength*4, sizeof(double))))
    return MOPS_OUTOFMEM;

  if(!(knotv = calloc(nlength+order, sizeof(double))))
    return MOPS_OUTOFMEM;

  if(!(lengths = calloc(length, sizeof(double))))
    return MOPS_OUTOFMEM;

  a = 0;
  for(i = 0; i < length-1; i++)
    {
      
      lengths[i] = MOPS_VLEN((controlv[a+3] - controlv[a]),
			     (controlv[a+4] - controlv[a+1]),
			     (controlv[a+5] - controlv[a+2]));

      chordlength += lengths[i];

      a += 3;
    }
  a = (length-1)*3;
  lengths[length-1] = MOPS_VLEN((controlv[0] - controlv[a]),
				(controlv[1] - controlv[a+1]),
				(controlv[2] - controlv[a+2]));

  chordlength += lengths[length-1];

  /* derivative */
  D[0] = controlv[3] - controlv[a];
  D[1] = controlv[4] - controlv[a+1];
  D[2] = controlv[5] - controlv[a+2];

  MOPS_V3SCAL(D,iparam)

  /* calc parametric values */
  if(!(vk = calloc(length+1, sizeof(double))))
    return MOPS_OUTOFMEM;

  vk[0] = 0.0;
  j = 0;
  knot = 0.0;
  for(i = 1; i < (length+1); i++)
    {
      knot += lengths[j]/chordlength;
      vk[i] = knot;
      j++;
    }
  vk[length] = 1.0;

  iN = length+1-(order-1)+1;
  for(j = 0; j <= iN; j++)
    {
      index = j+order;
      knotv[index] = 0.0;
	for(i = j; i <= j+(order-1)-1; i++)
	  {
	    knotv[index] += vk[i];
	  }
      knotv[index] /= (order - 1);
    }

  for(i = 0; i < order; i++)
    knotv[i] = 0.0;
  for(i = nlength; i < nlength+order; i++)
    knotv[i] = 1.0;

  free(lengths);

  /* first point */
  memcpy(ncontrolv,controlv,3*sizeof(double));
  ncontrolv[3] = 1.0;
  ncontrolv[7] = 1.0;
  /* inner points */
  j=1;
  for(i = 2; i <= length; i++)
    {
      memcpy(&(ncontrolv[i*4]),&(controlv[j*3]),3*sizeof(double));
      ncontrolv[i*4+3] = 1.0;
      j++;
    }
  ncontrolv[((nlength-1)*4)-1] = 1.0;
  /* last point */
  memcpy(&(ncontrolv[(nlength-1)*4]),&(controlv[0]),3*sizeof(double));
  ncontrolv[((nlength-1)*4)+3] = 1.0;

  mops_status = mops_nb_GlobalInterpolation4DD(length, ncontrolv, vk,
					       knotv, order-1, D, D);

  if(mops_status)
    { free(vk); free(ncontrolv); free(knotv); return mops_status; }

  mops_status = mops_create_nurbcurve(order, nlength, MOPS_KTCUSTOM,
				      MOPS_TRUE, ncontrolv, knotv, &new);
  
  if(!mops_status)
    *c = new;

  free(vk);

 return mops_status;
}

/* ICurve_dumb_interpolate, left here for sentimental
   reasons, it was my first attempt, but it is just C1 and
   generates simply too much De-Boor-points */
/* 
int
ICurve_dumb_interpolate(double iparam, int closed, int length,
                        double *controlv,
		        mops_nurbcurve_object **c)
{
 int mops_status = MOPS_OK;
 int a,b,i,j;
 int nlength;
 double v[3], vlen = 0.0;
 double *ncontrolv = NULL;
 double *knotv, knot = 0.0, *lengths, chordlength = 0.0;
 mops_nurbcurve_object *new = NULL;


  if(closed)
    {
      nlength = length + length*4 - 1;
    }
  else
    {
      nlength = ((length-2)*3)+((length-1)*2)+2;
    }

  if(!(ncontrolv = calloc(nlength*3, sizeof(double))))
    return MOPS_OUTOFMEM;

  if(!closed)
    {
      a = 6; b = 3;
      for(i = 1; i < (length-1); i++)
	{
	  for(j = 0; j < 5; j++)
	    {
	      memcpy(&(ncontrolv[a]), &(controlv[b]), 3*sizeof(double));
	      a += 3;
	    }
	  
	  a -= 15;

	  v[0] = controlv[b+3]-controlv[b-3];
	  v[1] = controlv[b+4]-controlv[b-2];
	  v[2] = controlv[b+5]-controlv[b-1];

	  vlen = MOPS_V3LEN(v);
	  vlen *= iparam;

	  MOPS_V3SCAL(v, vlen)

	  ncontrolv[a]   -= v[0];
	  ncontrolv[a+1] -= v[1];
	  ncontrolv[a+2] -= v[2];

	  a += 12;
	  ncontrolv[a]   += v[0];
	  ncontrolv[a+1] += v[1];
	  ncontrolv[a+2] += v[2];

	  a += 3;
	  b += 3;
	}

      memcpy(&(ncontrolv[0]), &(controlv[0]), 3*sizeof(double));

      v[0] = ncontrolv[6] - ncontrolv[0];
      v[1] = ncontrolv[7] - ncontrolv[1];
      v[2] = ncontrolv[8] - ncontrolv[2];
      vlen = MOPS_V3LEN(v);
      vlen *= (iparam);

      MOPS_V3SCAL(v, vlen)

      ncontrolv[3] = ncontrolv[0] + v[0];
      ncontrolv[4] = ncontrolv[1] + v[1];
      ncontrolv[5] = ncontrolv[2] + v[2];

      a = (nlength-1)*3;
      b = (length-1)*3;
      memcpy(&(ncontrolv[a]), &(controlv[b]), 3*sizeof(double));

      v[0] = ncontrolv[a] - ncontrolv[a-6];
      v[1] = ncontrolv[a+1] - ncontrolv[a-5];
      v[2] = ncontrolv[a+2] - ncontrolv[a-4];
      vlen = MOPS_V3LEN(v);
      vlen *= (iparam);

      MOPS_V3SCAL(v, vlen)

      ncontrolv[a-3] = ncontrolv[a] - v[0];
      ncontrolv[a-2] = ncontrolv[a+1] - v[1];
      ncontrolv[a-1] = ncontrolv[a+2] - v[2];
    }
  else
    {
      memcpy(&(ncontrolv[0]), &(controlv[0]), 3*sizeof(double));
      b = length*3;
      v[0] = controlv[b-3]-controlv[3];
      v[1] = controlv[b-2]-controlv[4];
      v[2] = controlv[b-1]-controlv[5];
      vlen = MOPS_V3LEN(v);
      vlen *= iparam;

      MOPS_V3SCAL(v, vlen)

      ncontrolv[3] = controlv[0] - v[0];
      ncontrolv[4] = controlv[1] - v[1];
      ncontrolv[5] = controlv[2] - v[2];

      a = nlength*3;
      ncontrolv[a-6] = controlv[0] + v[0];
      ncontrolv[a-5] = controlv[1] + v[1];
      ncontrolv[a-4] = controlv[2] + v[2];

      memcpy(&(ncontrolv[(nlength*3)-3]), &(controlv[0]), 3*sizeof(double));

      a = 6; b = 3;
      for(i = 1; i < length; i++)
	{
	  for(j = 0; j < 5; j++)
	    {
	      memcpy(&(ncontrolv[a]), &(controlv[b]), 3*sizeof(double));
	      a += 3;
	    }
	  
	  a -= 15;

	  if(i == (length - 1))
	    {
	      v[0] = controlv[0] - controlv[b-3];
	      v[1] = controlv[1] - controlv[b-2];
	      v[2] = controlv[2] - controlv[b-1];
	    }
	  else
	    {
	      v[0] = controlv[b+3] - controlv[b-3];
	      v[1] = controlv[b+4] - controlv[b-2];
	      v[2] = controlv[b+5] - controlv[b-1];
	    }


	  vlen = MOPS_V3LEN(v);
	  vlen *= iparam;

	  MOPS_V3SCAL(v, vlen)

	  ncontrolv[a]   -= v[0];
	  ncontrolv[a+1] -= v[1];
	  ncontrolv[a+2] -= v[2];

	  a += 12;
	  ncontrolv[a]   += v[0];
	  ncontrolv[a+1] += v[1];
	  ncontrolv[a+2] += v[2];

	  a += 3;
	  b += 3;
	}

    }


  if(!(lengths = calloc(nlength-1, sizeof(double))))
    return MOPS_OUTOFMEM;

  if(!(knotv = calloc(nlength+4, sizeof(double))))
    return MOPS_OUTOFMEM;

  a = 0;
  for(i = 0; i < (nlength-1); i++)
    {
      
      lengths[i] = MOPS_VLEN((ncontrolv[a+3] - ncontrolv[a]),
			     (ncontrolv[a+4] - ncontrolv[a+1]),
			     (ncontrolv[a+5] - ncontrolv[a+2]));

      chordlength += lengths[i];

      a += 3;
    }

  knotv[0] = 0.0;
  knotv[1] = 0.0;
  knotv[2] = 0.0;
  knotv[3] = 0.0;
  if(closed)
    j = 1;
  else
    j = 0;
  for(i = 4; i < (nlength+2); i++)
    {
      knot += lengths[j]/chordlength;
      knotv[i] = knot;
      j++;
    }
  knotv[nlength] = 1.0;
  knotv[nlength+1] = 1.0;
  knotv[nlength+2] = 1.0;
  knotv[nlength+3] = 1.0;

  free(lengths);

  mops_status = mops_create_nurbcurve(4, nlength, MOPS_KTCUSTOM, MOPS_FALSE,
				      ncontrolv, knotv, &new);

  if(!mops_status)
    *c = new;

 return mops_status;
}
*/

int
ICurve_draww_cb(struct Togl *togl, mops_object *o)
{
 int mops_status = MOPS_OK;
 int instance = MOPS_FALSE;
 int a, i, stride;
 GLfloat color[4] = {0.8, 0.75, 0.65, 1.0};
 ICurve *ic = NULL;
 mops_custom_object *custom = NULL;
 mops_nurbcurve_object *c = NULL;
 int reset_color = 0;

 /* XXXX this way the wireframe of the ICurve gets drawn twice  */
 /* but disabling this function will lead to deselected drawing */
 /* of the wireframe, when the ICurve is selected               */

  if(!o)
    return MOPS_OK;

  if(o->type == MOPS_OTINSTANCE)
    {
      o = (mops_object *)o->object;

      instance = MOPS_TRUE;
    }

  if(!o->editor)
    return MOPS_OK;

  custom = o->object;
  if(!custom)
    return MOPS_OK;    

  ic = custom->object;
  if(!ic)
    return MOPS_OK;

  if(o->down)
    {
      if(o->down->type == MOPS_OTNURBCURVE)
	{
	  c = (mops_nurbcurve_object *)(o->down->object);
	}
    }

  if(!c)
    return MOPS_OK;

  glPushMatrix();

  if(!instance)
    {
      mops_render_attributes(o);

      if(o->selected)
	{
	  glColor3f MOPS_CLSEL;
	  reset_color = 1;
	}
    }
  else
    {
      if(o->colr != -1 && o->colg != -1 && o->colb != -1)
	{
	  color[0] = o->colr/255.0;
	  color[1] = o->colg/255.0;
	  color[2] = o->colb/255.0;
	  color[3] = o->cola/255.0; 
	  glMaterialfv (GL_FRONT_AND_BACK, GL_AMBIENT_AND_DIFFUSE, color);
	}
    }

  glBegin(GL_LINE_STRIP);
  a = 0;
  stride = c->has_weights?4:3;
   for(i = 0; i < c->length; i++)
     {
       glVertex3dv((GLdouble *)&(c->controlv[a]));
       a += stride;
     }

   if(ic->closed)
     glVertex3dv((GLdouble *)&(c->controlv[0]));
  glEnd();


  if(reset_color)
    glColor3f MOPS_CLUSEL;

  glPopMatrix();

 return mops_status;
}

int
ICurve_drawh_cb(struct Togl *togl, mops_object *o)
{
 int mops_status = MOPS_OK;
 int a, i;
 ICurve *ic = NULL;
 mops_nurbcurve_object *c = NULL;
 mops_custom_object *custom = NULL;
 int stride = 3;

  if(!o)
    return MOPS_OK;

  if(o->type == MOPS_OTINSTANCE)
    return MOPS_OK;

  if(!o->editor)
    return MOPS_OK;

  custom = o->object;
  if(!custom)
    return MOPS_OK;    

  ic = custom->object;
  if(!ic)
    return MOPS_OK;

  if(o->down)
    {
      if(o->down->type == MOPS_OTNURBCURVE)
	{
	  c = (mops_nurbcurve_object *)(o->down->object);
	}
    }

  if(!c)
    return MOPS_OK;

  glPushMatrix();

  mops_render_attributes(o);


  glColor3f MOPS_CLSEL;


  glPointSize(6.0);
  glBegin(GL_POINTS);
   a = 0;
   stride = c->has_weights?4:3;
   for(i = 0; i < c->length; i++)
     {
       glVertex3dv((GLdouble *)&(c->controlv[a]));
       a += stride;
     }
  glEnd();

  glPopMatrix();

 return mops_status;
}


int
ICurve_draws_cb(struct Togl *togl, mops_object *o)
{
 int mops_status = MOPS_OK;
 int instance = MOPS_FALSE;
 GLfloat color[4] = {0.8, 0.75, 0.65, 1.0};
 int reset_color = 0;
 ICurve *ic = NULL;
 mops_custom_object *custom = NULL;


  if(!o)
    return MOPS_OK;

  if(o->type == MOPS_OTINSTANCE)
    {
      o = (mops_object *)o->object;

      instance = MOPS_TRUE;
    }

  if(!o->editor)
    return MOPS_OK;

  custom = o->object;
  if(!custom)
    return MOPS_OK;    

  ic = custom->object;
  if(!ic)
    return MOPS_OK;

  if(o->modified)
    {
      mops_status = ICurve_notify_cb(o);

      o->modified = MOPS_FALSE;
    }

  glPushMatrix();

  if(!instance)
    {
      mops_render_attributes(o);

      if(o->selected)
	{
	  glColor3f MOPS_CLSEL;
	  reset_color = 1;
	}
    }
  else
    {
      if(o->colr != -1 && o->colg != -1 && o->colb != -1)
	{
	  color[0] = o->colr/255.0;
	  color[1] = o->colg/255.0;
	  color[2] = o->colb/255.0;
	  color[3] = o->cola/255.0; 
	  glMaterialfv (GL_FRONT_AND_BACK, GL_AMBIENT_AND_DIFFUSE, color);
	}
    }

  if(ic->curve)
    mops_draw_nurbcurve(togl, ic->curve);

  if(reset_color)
    glColor3f MOPS_CLUSEL;

  glPopMatrix();

 return mops_status;
}



int
ICurve_write_cb(FILE *fileptr, mops_custom_object *custom)
{
 int mops_status = MOPS_OK;
 ICurve *ic = NULL;

  if(!custom)
    return MOPS_OK;    
  ic = (ICurve *)(custom->object);
  if(!ic)
    return MOPS_OK;

  fprintf(fileptr,"%g\n",ic->glu_sampling_tolerance);
  fprintf(fileptr,"%g\n",ic->iparam);
  fprintf(fileptr,"%d\n",ic->closed);
  fprintf(fileptr,"%d\n",ic->imode);
  fprintf(fileptr,"%d\n",ic->iorder);

 return mops_status;
}


int
ICurve_read_cb(FILE *fileptr, mops_custom_object *custom)
{
 int mops_status = MOPS_OK;
 ICurve *ic = NULL;
 double *controlv = NULL;
 int a = 0, i = 0, dlen;

  if(!custom)
    return MOPS_OK;
    
  if(!(ic = calloc(1,sizeof(ICurve))))
    return MOPS_OUTOFMEM;

  fscanf(fileptr,"%lg\n",&ic->glu_sampling_tolerance);
  fscanf(fileptr,"%lg\n",&ic->iparam);
  fscanf(fileptr,"%d\n",&ic->closed);
  if(mops_read_version < 7)
    {
      fscanf(fileptr,"%d\n",&dlen);

      if(!(controlv = calloc(dlen*3, sizeof(double))))
	return MOPS_OUTOFMEM;
      
      a = 0;
      for(i = 0; i < dlen; i++)
	{
	  fscanf(fileptr,"%lg\n",&controlv[a]);
	  fscanf(fileptr,"%lg\n",&controlv[a+1]);
	  fscanf(fileptr,"%lg\n",&controlv[a+2]);
	  a+=3;
	}

      ic->controlv = controlv;
      ic->length = dlen;
    }

  if(mops_read_version > 5)
    {
      fscanf(fileptr,"%d\n",&ic->imode);
      fscanf(fileptr,"%d\n",&ic->iorder);
    }

  custom->object = ic;

 return mops_status;
}


int
ICurve_getpoint_cb(mops_object *o, double objX, double objY, double objZ)
{
 int mops_status = MOPS_OK;
 mops_custom_object *custom = NULL;
 mops_nurbcurve_object *c = NULL;
 ICurve *ic = NULL;
 double *control = NULL, min_distance = mops_prefs.pick_epsilon,
   distance = 0.0, v[3];
 int i, j, stride;
 double *pecoords = NULL;

  if(!o)
    return MOPS_OK;

  custom = o->object;
  if(!custom)
    return MOPS_OK;    

  ic = custom->object;
  if(!ic)
    return MOPS_OK;

  if(o->down)
    {
      if(o->down->type == MOPS_OTNURBCURVE)
	{
	  c = (mops_nurbcurve_object *)(o->down->object);
	}
    }

  if(!c)
    return MOPS_OK;

 if(min_distance == 0.0)
    min_distance = DBL_MAX;

  if(mops_point_edit_coords)
    free(mops_point_edit_coords);

  mops_point_edit_coords = NULL;

  control = c->controlv;
  stride = c->has_weights?4:3;

  j = 0;
  for(i = 0; i < c->length; i++)
    {

      v[0] = objX - control[j];
      v[1] = objY - control[j+1];
      v[2] = objZ - control[j+2];
      distance = MOPS_V3LEN(v);

      if(distance < min_distance)
	{
	  pecoords = &(control[j]);
	  min_distance = distance;
	}

      j += stride;
    }

  if(!pecoords)
    return MOPS_OK; /* XXXX should this return a 'MOPS_FAILEDPICK' ? */

  mops_point_edit_coords_homogenous = c->has_weights;

  if(!(mops_point_edit_coords = calloc(1,sizeof(double*))))
    return MOPS_OUTOFMEM;
  mops_point_edit_coords_number = 1;
  mops_point_edit_coords[0] = pecoords;

 return mops_status;
}


int
ICurve_notify_cb(mops_object *o)
{
 int mops_status = MOPS_OK;
 ICurve *ic = NULL;
 mops_custom_object *custom = NULL;
 mops_nurbcurve_object *c = NULL, *new = NULL;
 double *controlv = NULL;
 int a = 0, i;

  if(!o)
    return MOPS_OK;

  custom = o->object;
  if(!custom)
    return MOPS_OK;    

  ic = custom->object;
  if(!ic)
    return MOPS_OK;

  if(ic->curve)
    mops_delete_object(ic->curve);
  ic->curve = NULL;

  if(o->down)
    {
      if(o->down->type == MOPS_OTNURBCURVE)
	{
	  c = (mops_nurbcurve_object *)(o->down->object);
	  c->hide_curve = MOPS_TRUE;
	}
    }
  else
    {
      /* legacy support for loading pre 0.42d scene files */
      if(ic->controlv)
	{
	        
	  mops_status = mops_create_nurbcurve(4, ic->length, MOPS_KTNURB,
					      MOPS_FALSE, ic->controlv, NULL,
					      &new);
	  if(!(o->down= calloc(1, sizeof(mops_object))))
	    return MOPS_OUTOFMEM;
	  new->hide_curve = MOPS_TRUE;
	  mops_default_attributes(o->down);
	  o->down->type = MOPS_OTNURBCURVE;
	  o->down->object = new;
	  ic->controlv = NULL;
	  c = new;
	}
    }

  if(!c)
    {
      mops_clear_point_selection(o);
      return MOPS_OK;
    }


  if(!(ic->curve = calloc(1, sizeof(mops_object))))
    return MOPS_OUTOFMEM;
  mops_default_attributes(ic->curve);

  ic->curve->type = MOPS_OTNURBCURVE;

  if(c->has_weights)
    {
      if(!(controlv = calloc(c->length*3, sizeof(double))))
	return MOPS_OUTOFMEM;

      a = 0;
      for(i = 0; i < c->length; i++)
	{
	  controlv[i*3] = c->controlv[a];
	  controlv[i*3+1] = c->controlv[a+1];
	  controlv[i*3+2] = c->controlv[a+2];

	  a += 4;
	}

    }
  else
    {
      controlv = c->controlv;
    }

  if(ic->imode)
    {
      if(ic->closed)
	{
	  mops_status = ICurve_interpolateG4DC(ic->iorder, c->length,
					       ic->iparam, controlv,
			   (mops_nurbcurve_object **)&(ic->curve->object));
	}
      else
	{
	  mops_status = ICurve_interpolateG4D(ic->iorder, c->length,
					      controlv,
			   (mops_nurbcurve_object **)&(ic->curve->object));
	}

    }
  else
    {
      mops_status = ICurve_interpolateC2C(ic->iparam, ic->closed, c->length,
					  controlv,
			   (mops_nurbcurve_object **)&(ic->curve->object));
    }

  if(!mops_status)
    ((mops_nurbcurve_object *)ic->curve->object)->glu_sampling_tolerance =
      ic->glu_sampling_tolerance;

  o->modified = MOPS_FALSE;

  if(c->has_weights)
    free(controlv);

 return mops_status;
}


/************************/
/************************/
/************************/

int
ICurve_getprop_cb(Tcl_Interp *interp)
{

  /* ICurve-Parameters */
  Tcl_SetVar(interp,"propertyList","ICurveAttr", TCL_APPEND_VALUE |\
	     TCL_LIST_ELEMENT | TCL_LEAVE_ERR_MSG | TCL_GLOBAL_ONLY);

  Tcl_SetVar(interp,"propertyProcList","ICurve_buildICurveAttrGUI",
	     TCL_APPEND_VALUE | TCL_LIST_ELEMENT |\
	     TCL_LEAVE_ERR_MSG | TCL_GLOBAL_ONLY);

  Tcl_SetVar(interp,"propertyGetProcList","ICurve_getICurveAttr",
	     TCL_APPEND_VALUE | TCL_LIST_ELEMENT |\
	     TCL_LEAVE_ERR_MSG | TCL_GLOBAL_ONLY);

  Tcl_SetVar(interp,"propertySetProcList","ICurve_setICurveAttr",
	     TCL_APPEND_VALUE | TCL_LIST_ELEMENT |\
	     TCL_LEAVE_ERR_MSG | TCL_GLOBAL_ONLY);

  Tcl_SetVar(interp,"propertyDataList","ICurveAttrData",
	     TCL_APPEND_VALUE | TCL_LIST_ELEMENT |\
	     TCL_LEAVE_ERR_MSG | TCL_GLOBAL_ONLY);

 return TCL_OK;
}


/* Tcl -> C! */
int
ICurve_set_attrprop_tcmd(ClientData clientData, Tcl_Interp *interp,
			 int argc, char *argv[])
{
 mops_object *o = NULL;
 mops_list_object *sel = mops_current_selection;
 mops_custom_object *custom = NULL;
 ICurve *ic = NULL;
 char *n1="ICurveAttrData";
 Tcl_Obj *to = NULL, *toa = NULL, *ton = NULL;

  if(!sel) /* oops? */
    return TCL_OK;

  o = sel->object; 
  if(!o)
    return TCL_OK;

  mops_clear_point_selection(o);

  custom = o->object;
  if(!custom)
    return TCL_OK;

  ic = (ICurve *) custom->object;

  toa = Tcl_NewStringObj(n1,-1);
  ton = Tcl_NewStringObj("Closed",-1);

  to = Tcl_ObjGetVar2(interp,toa,ton,TCL_LEAVE_ERR_MSG | TCL_GLOBAL_ONLY);
  Tcl_GetIntFromObj(interp,to, &ic->closed);

  Tcl_SetStringObj(ton,"Mode",-1);
  to = Tcl_ObjGetVar2(interp,toa,ton,TCL_LEAVE_ERR_MSG | TCL_GLOBAL_ONLY);
  Tcl_GetIntFromObj(interp,to, &ic->imode);

  Tcl_SetStringObj(ton,"Order",-1);
  to = Tcl_ObjGetVar2(interp,toa,ton,TCL_LEAVE_ERR_MSG | TCL_GLOBAL_ONLY);
  Tcl_GetIntFromObj(interp,to, &ic->iorder);

  Tcl_SetStringObj(ton,"IParam",-1);
  to = Tcl_ObjGetVar2(interp,toa,ton,TCL_LEAVE_ERR_MSG | TCL_GLOBAL_ONLY);
  Tcl_GetDoubleFromObj(interp,to, &ic->iparam);

  Tcl_SetStringObj(ton,"Tolerance",-1);
  to = Tcl_ObjGetVar2(interp,toa,ton,TCL_LEAVE_ERR_MSG | TCL_GLOBAL_ONLY);
  Tcl_GetDoubleFromObj(interp,to, &ic->glu_sampling_tolerance);


  Tcl_IncrRefCount(toa);Tcl_DecrRefCount(toa);
  Tcl_IncrRefCount(ton);Tcl_DecrRefCount(ton);
  
  ICurve_notify_cb(o);

  mops_notify_parent();

 return TCL_OK;
}

/* C -> Tcl! */
int
ICurve_get_attrprop_tcmd(ClientData clientData, Tcl_Interp *interp,
			 int argc, char *argv[])
{
 mops_object *o = NULL;
 mops_list_object *sel = mops_current_selection;
 mops_custom_object *custom = NULL;
 ICurve *ic = NULL;
 char *n1="ICurveAttrData";
 Tcl_Obj *to = NULL, *toa = NULL, *ton = NULL;

  if(!sel) /* oops? */
    return TCL_OK;

  o = sel->object; 
  if(!o)
    return TCL_OK;

  custom = o->object;
  if(!custom)
    return TCL_OK;

  ic = (ICurve *) custom->object;

  toa = Tcl_NewStringObj(n1,-1);
  ton = Tcl_NewStringObj("Length",-1);

  ton = Tcl_NewStringObj("Closed",-1);
  to = Tcl_NewIntObj(ic->closed);
  Tcl_ObjSetVar2(interp,toa,ton,to,TCL_LEAVE_ERR_MSG |
		 TCL_GLOBAL_ONLY);

  Tcl_SetStringObj(ton,"Mode",-1);
  to = Tcl_NewIntObj(ic->imode);
  Tcl_ObjSetVar2(interp,toa,ton,to,TCL_LEAVE_ERR_MSG |
		 TCL_GLOBAL_ONLY);

  Tcl_SetStringObj(ton,"Order",-1);
  to = Tcl_NewIntObj(ic->iorder);
  Tcl_ObjSetVar2(interp,toa,ton,to,TCL_LEAVE_ERR_MSG |
		 TCL_GLOBAL_ONLY);

  Tcl_SetStringObj(ton,"IParam",-1);
  to = Tcl_NewDoubleObj(ic->iparam);
  Tcl_ObjSetVar2(interp,toa,ton,to,TCL_LEAVE_ERR_MSG |
		 TCL_GLOBAL_ONLY);

  Tcl_SetStringObj(ton,"Tolerance",-1);
  to = Tcl_NewDoubleObj(ic->glu_sampling_tolerance);
  Tcl_ObjSetVar2(interp,toa,ton,to,TCL_LEAVE_ERR_MSG |
		 TCL_GLOBAL_ONLY);

  Tcl_IncrRefCount(toa);Tcl_DecrRefCount(toa);
  Tcl_IncrRefCount(ton);Tcl_DecrRefCount(ton);

 return TCL_OK;
}

int
ICurve_convert(Tcl_Interp *interp, mops_object *o)
{
 int mops_status = MOPS_OK;
 mops_custom_object *custom = NULL;
 ICurve *ic = NULL;
 mops_object *new = NULL;
 char fname[] = "ICurve_convert";

  custom = (mops_custom_object*)o->object;
  if(!custom)
    return MOPS_ERROR;

  if(custom->type != ICurve_customid)
    return MOPS_ERROR;

  ic = (ICurve *) custom->object;

  if(ic->curve)
    {
      mops_status = mops_copy_object(ic->curve, &new);
      MOPS_CRERROR
      mops_status = mops_copy_transform(o, new);
      mops_status = mops_link_new_object(new);
      MOPS_CRERROR
    }

 return mops_status;
}


int
ICurve_convertNC_tcmd(ClientData clientData, Tcl_Interp *interp,
		     int argc, char *argv[])
{
 int mops_status = MOPS_OK;
 mops_object *o = NULL;
 mops_list_object *sel = mops_current_selection;
 mops_custom_object *custom = NULL;
 char fname[] = "ICurve_convertNC";

  if(!sel) /* oops? */
    return TCL_OK;

  o = sel->object; 
  if(!o)
    return TCL_OK;

  if(o->type != MOPS_OTCUSTOM)
    return TCL_OK;

  custom = (mops_custom_object*)o->object;
  if(!custom)
    return TCL_OK;

  if(custom->type != ICurve_customid)
    return TCL_OK;

  mops_status = ICurve_convert(interp, o);

  if(mops_status)
    {
      mops_error(interp,MOPS_ERROR,fname,"converter failed");
      return TCL_OK;
    }

 return TCL_OK;
}


int
ICurve_provide_cb(mops_object *o, void **target)
{
 int mops_status = MOPS_OK;
 ICurve *ic = NULL;
 mops_custom_object *custom = NULL;

  custom = (mops_custom_object *)o->object;

  ic = (ICurve *)custom->object;

  mops_status = ICurve_notify_cb(o);

  if(mops_status)
    return mops_status;

  if(ic->curve)
  {
    ic->curve->movx = o->movx;
    ic->curve->movy = o->movy;
    ic->curve->movz = o->movz;

    ic->curve->scalx = o->scalx;
    ic->curve->scaly = o->scaly;
    ic->curve->scalz = o->scalz;

    ic->curve->rotx = o->rotx;
    ic->curve->roty = o->roty;
    ic->curve->rotz = o->rotz;
    memcpy(ic->curve->quat, o->quat, 4*sizeof(float));
  }

  *target = ic->curve;

 return mops_status;
}
