/*
 * The Mops, a free 3D Tcl/Tk/OpenGL/RenderMan modeler
 *
 * The Mops is copyrighted (C) 1998-1999 by Randolf Schultz
 * (rschultz@informatik.uni-rostock.de)
 *
 * See the file License for details.
 *
 */

#include "mops.h"

/* Mfio - A 3DMF Import/Export Plugin */

#include <mf3d.h>
#include <mferrors.h>

/* functions local to this module */
int mfio_read_nurbcurve(MF3DVoidObjPtr object);

int mfio_read_nurbpatch(MF3DVoidObjPtr object);

int mfio_read_cntr(MF3DVoidObjPtr object);

int mfio_read_ecntr(MF3DVoidObjPtr object);

int mfio_read_scal(MF3DVoidObjPtr object);

int mfio_read_tran(MF3DVoidObjPtr object);

int mfio_read_rot(MF3DVoidObjPtr object);

int mfio_read_dcol(MF3DVoidObjPtr object);

int mfio_read_tcol(MF3DVoidObjPtr object);

int mfio_read_object(MF3DVoidObjPtr object);

int mfio_read_scene(Tcl_Interp *interp, char *filename);

int mfio_write_nurbcurve(MF3D_FilePtr fileptr, mops_object *o);

int mfio_write_trimcurve(MF3D_FilePtr fileptr, mops_object *o);

int mfio_write_nurbpatch(MF3D_FilePtr fileptr, mops_object *o);

int mfio_write_sphere(MF3D_FilePtr fileptr, mops_object *o);

int mfio_write_attributes(MF3D_FilePtr fileptr, mops_object *o);

int mfio_write_cntr(MF3D_FilePtr fileptr);

int mfio_write_ecntr(MF3D_FilePtr fileptr);

int mfio_write_object(MF3D_FilePtr fileptr, mops_object *object);

int mfio_write_scene(Tcl_Interp *interp, char *filename);


/* globals */

static mops_object *mfio_last_read_object;

static MF3DErr mfio_mf3d_errno;

static MF3DDataFormatEnum mfio_mf3d_data_format;

char mfio_mops_version_ma[] = MOPS_VERSIONSTR;
char mfio_mops_version_mi[] = MOPS_VERSIONSTRMI;

/********************************************/
int
mfio_read_nurbpatch(MF3DVoidObjPtr object)
{
 int mops_status = MOPS_OK;
 int width, height, i, j, a, b;
 double *controlv = NULL;
 double *uknotv = NULL, *vknotv = NULL;
 MF3DNurbPatchObjPtr o = (MF3DNurbPatchObjPtr) object;
 mops_nurbpatch_object *patch = NULL;
 mops_object *newo = NULL;

  /* get some info about the patch */
  width = o->numColumns;
  height = o->numRows;

  /* alloc new (safe) memory */
  if(!(controlv = calloc(width*height*4, sizeof(double))))
    return MOPS_OUTOFMEM; 
  if(!(uknotv = calloc(width+o->uOrder, sizeof(double))))
    { free(controlv); return MOPS_OUTOFMEM; }
  if(!(vknotv = calloc(height+o->vOrder, sizeof(double))))
    { free(controlv); free(uknotv); return MOPS_OUTOFMEM; }

  /* copy data into new (safe) memory */
  a = 0; b = 0;
  for(i=0;i<width;i++)
    for(j=0;j<height;j++)
      {
	controlv[b++] = (o->points)[a].x;
	controlv[b++] = (o->points)[a].y;
	controlv[b++] = (o->points)[a].z;
	controlv[b++] = (o->points)[a].w;
	a++;
      }
  a = 0;
  for(i=0;i<width+o->uOrder;i++)
    {
      uknotv[a] = (o->uKnots)[a];
      a++;
    }
  a = 0;
  for(i=0;i<height+o->vOrder;i++)
    {
      vknotv[a] = (o->vKnots)[a];
      a++;
    }

  /* now create a MOPS NURBPatch */
  mops_status = mops_create_nurbpatch(o->uOrder, o->vOrder, width, height,
				      MOPS_KTCUSTOM, 
				      MOPS_KTCUSTOM, MOPS_TRUE,
				      controlv, uknotv, vknotv,
				      &patch);

  if(mops_status)
    { free(controlv); free(uknotv); free(vknotv); return mops_status; }

  /* link the new patch into the hierarchy */
  if(!(newo = calloc(1, sizeof(mops_object))))
    { free(controlv); free(uknotv); free(vknotv); return MOPS_OUTOFMEM; }

  newo->type = MOPS_OTNURBPATCH;
  newo->object = patch;

  mops_status = mops_link_new_object(newo);

  if(mops_status)
    mops_status = mops_delete_object(newo);

  mfio_last_read_object = newo;

  mops_status = mops_default_attributes(newo);

 return mops_status;
}


int
mfio_read_nurbcurve(MF3DVoidObjPtr object)
{
 int mops_status = MOPS_OK;
 int length, i, a, b;
 double *controlv = NULL;
 double *knotv = NULL;
 MF3DNURBCurveObjPtr o = (MF3DNURBCurveObjPtr) object;
 mops_nurbcurve_object *curve = NULL;
 mops_object *newo = NULL;

  /* get some info about the curve */
  length = o->nPoints;

  /* alloc new (safe) memory */
  if(!(controlv = calloc(length*4, sizeof(double))))
    return MOPS_OUTOFMEM; 
  if(!(knotv = calloc(length+o->order, sizeof(double))))
    { free(controlv); return MOPS_OUTOFMEM; }

  /* copy data into new (safe) memory */
  a = 0; b = 0;
  for(i=0;i<length;i++)
    {
      controlv[b++] = (o->points)[a].x;
      controlv[b++] = (o->points)[a].y;
      controlv[b++] = (o->points)[a].z;
      controlv[b++] = (o->points)[a].w;
      a++;
    }
  a = 0;
  for(i=0;i<length+o->order;i++)
    {
      knotv[a] = (o->knots)[a];
      a++;
    }

  /* now create a MOPS NURBCurve */
  mops_status = mops_create_nurbcurve(o->order, length,
				      MOPS_KTCUSTOM, MOPS_TRUE,
				      controlv, knotv,
				      &curve);

  if(mops_status)
    { free(controlv); free(knotv); return mops_status; }  

  /* link the new curve into the hierarchy */
  if(!(newo = calloc(1, sizeof(mops_object))))
    { free(controlv); free(knotv); return MOPS_OUTOFMEM; }

  newo->type = MOPS_OTNURBCURVE;
  newo->object = curve;

  mops_status = mops_link_new_object(newo);

  if(mops_status)
    {
      mops_delete_object(newo);
      return mops_status;
    }

  mfio_last_read_object = newo;

  mops_status = mops_default_attributes(newo);

 return mops_status;
}


int
mfio_read_box(MF3DVoidObjPtr object)
{
 int mops_status = MOPS_OK;
 MF3DBoxObjPtr o = (MF3DBoxObjPtr) object;
 MF3DBoxDataPtr ed;
 mops_box_object *box = NULL;
 mops_object *newo = NULL;

  if(!(box = calloc(1,sizeof(mops_box_object))))
    return MOPS_OUTOFMEM;

  box->width = 1.0;
  box->height = 1.0;
  box->length = 1.0;

  if(!(newo = calloc(1, sizeof(mops_object))))
    { free(box); return MOPS_OUTOFMEM; }

  newo->type = MOPS_OTBOX;
  newo->object = box;

  mops_status = mops_link_new_object(newo);

  if(mops_status)
    {
      mops_delete_object(newo);
      return mops_status;
    }

  mfio_last_read_object = newo;

  mops_status = mops_default_attributes(newo);

  newo->movx = 0.5;
  newo->movy = 0.5;
  newo->movz = 0.5;

  if(o->box)
    {
      ed = o->box;
      newo->movx += ed->origin.x;
      newo->movy += ed->origin.y;
      newo->movz += ed->origin.z;

      /* XXXX minor & major Axis? */

    }

 return mops_status;
}


int
mfio_read_ellipsoid(MF3DVoidObjPtr object)
{
 int mops_status = MOPS_OK;
 MF3DEllipsoidObjPtr o = (MF3DEllipsoidObjPtr) object;
 MF3DEllipsoidDataPtr ed;
 mops_sphere_object *sphere = NULL;
 mops_object *newo = NULL;

  if(!(sphere = calloc(1,sizeof(mops_sphere_object))))
    return MOPS_OUTOFMEM;

  sphere->closed = MOPS_FALSE;
  sphere->radius = 1.0;
  sphere->zmin = -1.0;
  sphere->zmax = 1.0;
  sphere->thetamax = 360.0;

  if(!(newo = calloc(1, sizeof(mops_object))))
    { free(sphere); return MOPS_OUTOFMEM; }

  newo->type = MOPS_OTSPHERE;
  newo->object = sphere;

  mops_status = mops_link_new_object(newo);

  if(mops_status)
    {
      mops_delete_object(newo);
      return mops_status;
    }

  mfio_last_read_object = newo;

  mops_status = mops_default_attributes(newo);

  if(o->ellipsoid)
    {
      ed = o->ellipsoid;
      newo->movx = ed->origin.x;
      newo->movy = ed->origin.y;
      newo->movz = ed->origin.z;

      /* XXXX minor & major Radius? */

    }

 return mops_status;
}


int
mfio_read_cylinder(MF3DVoidObjPtr object)
{
 int mops_status = MOPS_OK;
 MF3DCylinderObjPtr o = (MF3DCylinderObjPtr) object;
 MF3DCylinderDataPtr cd;
 mops_cylinder_object *cylinder = NULL;
 mops_object *newo = NULL;

  if(!(cylinder = calloc(1,sizeof(mops_cylinder_object))))
    return MOPS_OUTOFMEM;

  cylinder->closed = MOPS_FALSE;
  cylinder->radius = 1.0;
  cylinder->zmin = -1.0;
  cylinder->zmax = 1.0;
  cylinder->thetamax = 360.0;

  if(!(newo = calloc(1, sizeof(mops_object))))
    { free(cylinder); return MOPS_OUTOFMEM; }

  newo->type = MOPS_OTCYLINDER;
  newo->object = cylinder;

  mops_status = mops_link_new_object(newo);

  if(mops_status)
    {
      mops_delete_object(newo);
      return mops_status;
    }

  mfio_last_read_object = newo;

  mops_status = mops_default_attributes(newo);

  if(o->cylinder)
    {
      cd = o->cylinder;
      newo->movx = cd->origin.x;
      newo->movy = cd->origin.y;
      newo->movz = cd->origin.z;

      /* XXXX orientation & minor & major Radius? */

    }

 return mops_status;
}


int
mfio_read_cone(MF3DVoidObjPtr object)
{
 int mops_status = MOPS_OK;
 MF3DConeObjPtr o = (MF3DConeObjPtr) object;
 MF3DConeDataPtr cd;
 mops_cone_object *cone = NULL;
 mops_object *newo = NULL;

  if(!(cone = calloc(1,sizeof(mops_cone_object))))
    return MOPS_OUTOFMEM;

  cone->closed = MOPS_FALSE;
  cone->radius = 1.0;
  cone->height = 1.0;
  cone->thetamax = 360.0;

  if(!(newo = calloc(1, sizeof(mops_object))))
    { free(cone); return MOPS_OUTOFMEM; }

  newo->type = MOPS_OTCONE;
  newo->object = cone;

  mops_status = mops_link_new_object(newo);

  if(mops_status)
    {
      mops_delete_object(newo);
      return mops_status;
    }

  mfio_last_read_object = newo;

  mops_status = mops_default_attributes(newo);

  if(o->cone)
    {
      cd = o->cone;
      newo->movx = cd->origin.x;
      newo->movy = cd->origin.y;
      newo->movz = cd->origin.z;

      /* XXXX orientation & minor & major Radius? */

    }

 return mops_status;
}


int
mfio_read_disk(MF3DVoidObjPtr object)
{
 int mops_status = MOPS_OK;
 MF3DDiskObjPtr o = (MF3DDiskObjPtr) object;
 MF3DDiskDataPtr cd;
 mops_disk_object *disk = NULL;
 mops_object *newo = NULL;

  if(!(disk = calloc(1,sizeof(mops_disk_object))))
    return MOPS_OUTOFMEM;

  disk->radius = 1.0;
  disk->height = 1.0;
  disk->thetamax = 360.0;

  if(!(newo = calloc(1, sizeof(mops_object))))
    { free(disk); return MOPS_OUTOFMEM; }

  newo->type = MOPS_OTDISK;
  newo->object = disk;

  mops_status = mops_link_new_object(newo);

  if(mops_status)
    {
      mops_delete_object(newo);
      return mops_status;
    }

  mfio_last_read_object = newo;

  mops_status = mops_default_attributes(newo);

  if(o->disk)
    {
      cd = o->disk;
      newo->movx = cd->origin.x;
      newo->movy = cd->origin.y;
      newo->movz = cd->origin.z;

      /* XXXX orientation & minor & major Radius? */

    }

 return mops_status;
}


int
mfio_read_torus(MF3DVoidObjPtr object)
{
 int mops_status = MOPS_OK;
 MF3DTorusObjPtr o = (MF3DTorusObjPtr) object;
 MF3DTorusDataPtr cd;
 mops_torus_object *torus = NULL;
 mops_object *newo = NULL;

  if(!(torus = calloc(1,sizeof(mops_torus_object))))
    return MOPS_OUTOFMEM;

  torus->closed = MOPS_FALSE;
  torus->majorrad = 0.75;
  torus->minorrad = 0.25;
  torus->phimin = 0.0;
  torus->phimax = 360.0;
  torus->thetamax = 360.0;

  if(!(newo = calloc(1, sizeof(mops_object))))
    { free(torus); return MOPS_OUTOFMEM; }

  newo->type = MOPS_OTTORUS;
  newo->object = torus;

  mops_status = mops_link_new_object(newo);

  if(mops_status)
    {
      mops_delete_object(newo);
      return mops_status;
    }

  mfio_last_read_object = newo;

  mops_status = mops_default_attributes(newo);

  if(o->torus)
    {
      cd = o->torus;
      newo->movx = cd->origin.x;
      newo->movy = cd->origin.y;
      newo->movz = cd->origin.z;

      /* XXXX orientation & minor & major Radius? */

    }

 return mops_status;
}


int
mfio_read_scal(MF3DVoidObjPtr object)
{
 int mops_status = MOPS_OK;
 mops_object *o = mops_current_level->object;
 MF3DScaleObjPtr t = (MF3DScaleObjPtr) object;

 if(!o)
   return MOPS_OK;

 o->scalx *= t->scale.x;
 o->scaly *= t->scale.y;
 o->scalz *= t->scale.z;

 return mops_status;
}

int
mfio_read_tran(MF3DVoidObjPtr object)
{
 int mops_status = MOPS_OK;
 mops_object *o = mops_current_level->object;
 MF3DTranslateObjPtr t = (MF3DTranslateObjPtr) object;

 if(!o)
   return MOPS_OK;

 o->movx += t->translate.x;
 o->movy += t->translate.y;
 o->movz += t->translate.z;

 return mops_status;
}


int
mfio_read_rot(MF3DVoidObjPtr object)
{
 int mops_status = MOPS_OK;
 mops_object *o = mops_current_level->object;
 MF3DRotateObjPtr t = (MF3DRotateObjPtr) object;
 float xaxis[3]={1.0,0.0,0.0};
 float yaxis[3]={0.0,1.0,0.0};
 float zaxis[3]={0.0,0.0,1.0};
 float quat[4], drot;

 if(!o)
   return MOPS_OK;

 drot = MOPS_R2D(t->radians);

 if(drot != 0.0)
   { 
     if(t->axis == MF3DAxisX)
       {
	 o->rotx += drot;
	 mops_axis_to_quat(xaxis,t->radians,quat);
	 mops_add_quat(quat,o->quat,o->quat);
       }
     else
     if(t->axis == MF3DAxisY)
       {
	 o->roty += drot;
	 mops_axis_to_quat(yaxis,t->radians,quat);
	 mops_add_quat(quat,o->quat,o->quat);
       }
     else
     if(t->axis == MF3DAxisZ)
       {
	 o->rotz += drot;
	 mops_axis_to_quat(zaxis,t->radians,quat);
	 mops_add_quat(quat,o->quat,o->quat);
       }
   }

 return mops_status;
}

int
mfio_read_dcol(MF3DVoidObjPtr object)
{
 int mops_status = MOPS_OK;
 mops_object *o = mops_current_level->object;
 MF3DDiffuseColorObjPtr c = (MF3DDiffuseColorObjPtr) object;

 if(!o)
   return MOPS_OK;

 o->colr = c->diffuseColor.red;
 o->colg = c->diffuseColor.green;
 o->colb = c->diffuseColor.blue;

 return mops_status;
}


int
mfio_read_tcol(MF3DVoidObjPtr object)
{
 int mops_status = MOPS_OK;
 mops_object *o = mops_current_level->object;
 MF3DTransparencyColorObjPtr c = (MF3DTransparencyColorObjPtr) object;

 if(!o)
   return MOPS_OK;

 o->opr = c->transparency.red;
 o->opg = c->transparency.green;
 o->opb = c->transparency.blue;

 return mops_status;
}


int
mfio_read_cntr(MF3DVoidObjPtr object)
{
 int mops_status = MOPS_OK;
 mops_level_object *newl = NULL;
 mops_object *newo = NULL;

  if(!(newl = calloc(1, sizeof(mops_level_object))))
    { return MOPS_OUTOFMEM; }
 
  if(!(newo = calloc(1, sizeof(mops_object))))
    { free(newl); return MOPS_OUTOFMEM; }
 
  newo->object = newl;

  newo->type = MOPS_OTLEVEL;

  mfio_last_read_object = newo;

  mops_status = mops_link_new_object(newo);

  mops_add_level_object(newo);

  mops_next_object = &(newo->down);

  mops_status = mops_default_attributes(newo);

 return mops_status;
}


int
mfio_read_ecntr(MF3DVoidObjPtr object)
{
 int mops_status = MOPS_OK;

 mops_next_object = &(mops_current_level->object->next);

 mops_last_read_object = mops_current_level->object;
 
 mops_del_level_object();

 return mops_status;
}


int
mfio_read_object(MF3DVoidObjPtr object)
{
 int mops_status = MOPS_OK;

  switch(object->objectType)
    {
    case kMF3DObjNURBPatch:
      mops_status = mfio_read_nurbpatch(object);
      break;
    case kMF3DObjNURBCurve:
      mops_status = mfio_read_nurbcurve(object);
      break;
    case kMF3DObjBox:
      mops_status = mfio_read_box(object);
      break;
    case kMF3DObjContainer:
      mops_status = mfio_read_cntr(object);
      break;
    case kMF3DObjEndContainer:
      mops_status = mfio_read_ecntr(object);
      break;
    case kMF3DObjTranslateTransform:
      mops_status = mfio_read_tran(object);
      break;
    case kMF3DObjScaleTransform:
      mops_status = mfio_read_scal(object);
      break;
    case kMF3DObjRotateTransform:
      mops_status = mfio_read_rot(object);
      break;
    case kMF3DObjDiffuseColor:
      mops_status = mfio_read_dcol(object);
      break;
    case kMF3DObjTransparencyColor:
      mops_status = mfio_read_tcol(object);
      break;
    }

 return mops_status;
}


int
mfio_read_scene(Tcl_Interp *interp, char *filename)
{
 MF3D_FilePtr metafilePtr;	/* MF3D internal file pointer   */
 MF3DVoidObjPtr object;	        /* object returned by read      */
 MF3DVoidObjPtr extStorageObj = NULL;
 MF3DBoolean nextObjIsRoot;	/* next obj read will be root   */
 MF3DErr status = kMF3DNoErr;	/* temporary result code        */
 int mops_status = MOPS_OK;

  /* Open the metafile */
  status = MF3DOpenInputStdCFile (filename, &metafilePtr);
  if (status != kMF3DNoErr)
    return MOPS_OPENFILE;


  nextObjIsRoot = kMF3DBooleanFalse;

  while (status == kMF3DNoErr)
    {				
      /* Read an object from the metafile */
      status = MF3DReadAnObject (metafilePtr, &object);

      if (status != kMF3DNoErr)
	{
	  if (status != kMF3DNoMoreObjects)
	    { 
	      mfio_mf3d_errno = status;
	      return MOPS_ERROR;
	    }
	}

      if ((status == kMF3DNoErr) && object->objectType == kMF3DObjReference)
	{	/* If we have just read a Reference object, resolve it.
		 * There are two types of references: local and external.
		 * An external reference will be the root object of a container
		 * and will have a subobject specifying the path
		 * to an external file.
		 */
	  extStorageObj = NULL;
	  if (nextObjIsRoot == kMF3DBooleanTrue)
	    {	/* We are in a container which must mean we have an
		 * external reference.
		 */
	      status = MF3DReadAnObject (metafilePtr, &extStorageObj);

	      if (status != kMF3DNoErr)
		{ 
		  mfio_mf3d_errno = status;
		  return MOPS_ERROR;
		}

	    }

	  /* Resolve the reference. This causes metafilePtr to point
	   * to the referenced object, so that the next call to
	   * MF3DReadAnObject will read the object which is being
	   * referenced. MF3DReadAnObject will then restore the correct
	   * file position after the referenced object has been read.
	   *
	   * If the reference is local, then extStorageObj is NULL.
	   * If the reference is external, then the file specified by
	   * extStorageObj will be opened automatically (and closed
	   * automatically after the reference has been resolved).
	   */
	  if(status == kMF3DNoErr)
	    status = MF3DResolveReference (metafilePtr,
					 (MF3DReferenceObjPtr) object,
					 (MF3DStorageObjPtr) extStorageObj);

	  /* Dispose the external storage object, if there is one.
	   * The Reference object will be disposed later.
	   */
	  MF3DDisposeObject (extStorageObj);
	}

      /* If the object we just read is a BeginContainer object, the next
       * object read will be the root object of the container.
       */
      if (status == kMF3DNoErr && object->objectType == kMF3DObjContainer)
	nextObjIsRoot = kMF3DBooleanTrue;
      else
	nextObjIsRoot = kMF3DBooleanFalse;

      if (status == kMF3DNoErr)
	{
	  mops_status = mfio_read_object(object);
	  if(mops_status)
	    return mops_status;
	}
      /* Dispose the object */
      if (status == kMF3DNoErr)
	MF3DDisposeObject(object);

    }

  /* Close the file */
  status = MF3DClose (metafilePtr);
  if (status != kMF3DNoErr)
    return MOPS_CLOSEFILE;

  return MOPS_OK;
}


int 
mfio_write_trimcurve(MF3D_FilePtr fileptr, mops_object *o)
{
 int mops_status = MOPS_OK;
 int i = 0, a = 0, b = 0;
 MF3DNURBCurve2DObj mf3do = {0};
 MF3DErr status = kMF3DNoErr;	/* temporary result code */  
 mops_nurbcurve_object *curve = (mops_nurbcurve_object*)(o->object);
 GLdouble m[16] = {0};
 GLfloat mr[4][4];
 double x,y,z,w;

  /* we fill the MF3D-object now */
  mf3do.objectType = kMF3DObjNURBCurve2D;
  mf3do.nPoints = curve->length;
  mf3do.order = curve->order;

  if(!(mf3do.points = calloc(curve->length, 
			    sizeof(MF3DRationalPoint3D))))
    return MOPS_OUTOFMEM;
  if(!(mf3do.knots = calloc(curve->length+curve->order,
			    sizeof(MF3DKnot))))
    { free(mf3do.points); return MOPS_OUTOFMEM; }

  /* get curves transformation-matrix */
  glMatrixMode (GL_MODELVIEW);
  glPushMatrix();
  glLoadIdentity ();
  glTranslated(o->movx, o->movy, o->movz);

  mops_build_rotmatrix(mr, o->quat);
  glMultMatrixf(&mr[0][0]);

  glScaled (o->scalx, o->scaly, o->scalz);
  glGetDoublev(GL_MODELVIEW_MATRIX, m);
  glPopMatrix();

  a = 0; b = 0;
  for(i=0; i<curve->length; i++)
    {
      x = curve->controlv[b];
      b++;
      y = curve->controlv[b];
      b++;
      z = curve->controlv[b];
      b++;
      if(curve->has_weights)
	{
	  w = curve->controlv[b];
	  b++;
	}
      else
	w = 1.0;

      (mf3do.points)[a].x = m[0]*x + m[4]*y + m[8]*z + m[12]*w;
      (mf3do.points)[a].y = m[1]*x + m[5]*y + m[9]*z + m[13]*w;

      if(curve->has_weights)
	(mf3do.points)[a].w = w;
      else
	(mf3do.points)[a].w = 1.0;

      a++;
    }

  a = 0;
  for(i=0; i<curve->length+curve->order; i++)
    {
      (mf3do.knots)[a] = curve->knotv[a]; 
      a++;
    }

  /* we write out the curve now */
  status = MF3DWriteAnObject (fileptr, (MF3DVoidObjPtr)&mf3do);
  if(status != kMF3DNoErr)
    { free(mf3do.points); free(mf3do.knots); mfio_mf3d_errno = status; 
      return MOPS_ERROR; }

  /* delete the object */
  free(mf3do.points);
  free(mf3do.knots);

 return mops_status;
}


int 
mfio_write_nurbpatch(MF3D_FilePtr fileptr, mops_object *o)
{
 int mops_status = MOPS_OK;
 int i = 0, j = 0, a = 0, b = 0;
 MF3DNurbPatchObj mf3do = {0};
 MF3DTrimCurvesObj tco = {0};
 MF3DErr status = kMF3DNoErr;	/* temporary result code */  
 mops_nurbpatch_object *patch = (mops_nurbpatch_object*)(o->object);
 mops_object *trim;

  /* write enclosing container */
  mops_status = mfio_write_cntr(fileptr);
  if(mops_status)
    { return mops_status; }
 
  /* we fill the MF3D-object now */
  mf3do.objectType = kMF3DObjNURBPatch;
  mf3do.numColumns = patch->width;
  mf3do.numRows = patch->height;
  mf3do.uOrder = patch->uorder;
  mf3do.vOrder = patch->vorder;

  if(!(mf3do.points = calloc(patch->width*patch->height, 
			    sizeof(MF3DRationalPoint4D))))
    return MOPS_OUTOFMEM;
  if(!(mf3do.uKnots = calloc(patch->width+patch->uorder,
			    sizeof(MF3DKnot))))
    { free(mf3do.points); return MOPS_OUTOFMEM; }
  if(!(mf3do.vKnots = calloc(patch->height+patch->vorder,
			    sizeof(MF3DKnot))))
    { free(mf3do.points); free(mf3do.vKnots); return MOPS_OUTOFMEM; }

  a = 0; b = 0;
  for(i=0; i<patch->width; i++)
    for(j=0; j<patch->height; j++)
      {
	(mf3do.points)[a].x = patch->controlv[b++];
	(mf3do.points)[a].y = patch->controlv[b++];
	(mf3do.points)[a].z = patch->controlv[b++];
	if(patch->has_weights)
	  (mf3do.points)[a].w = patch->controlv[b++];
	else
	  (mf3do.points)[a].w = 1.0;
	a++;
      }
  a = 0;
  for(i=0; i<patch->width+patch->uorder; i++)
    {
      (mf3do.uKnots)[a] = patch->uknotv[a]; 
      a++;
    }
  a = 0;
  for(i=0; i<patch->height+patch->vorder; i++)
    {
      (mf3do.vKnots)[a] = patch->vknotv[a];
      a++;
    }

  /* we write out the patch now */
  status = MF3DWriteAnObject (fileptr, (MF3DVoidObjPtr)&mf3do);
  if(status != kMF3DNoErr)
    { free(mf3do.points); free(mf3do.vKnots); free(mf3do.uKnots);
      mfio_mf3d_errno = status; return MOPS_ERROR; }

  /* write attributes */
  mops_status = mfio_write_attributes(fileptr, o);
  if(mops_status)
    { free(mf3do.points); free(mf3do.vKnots); free(mf3do.uKnots);
      return mops_status; }

  /* write trimcurves */
  trim = o->down;
  if(trim)
    {

      /* write enclosing container */
      mops_status = mfio_write_cntr(fileptr);
      if(mops_status)
	{ free(mf3do.points); free(mf3do.vKnots); free(mf3do.uKnots);
          return mops_status; }

      tco.objectType = kMF3DObjTrimCurves;
      /* we write out the trimcurve-object now */
      status = MF3DWriteAnObject (fileptr, (MF3DVoidObjPtr)&tco);
      if(status != kMF3DNoErr)
	{ free(mf3do.points); free(mf3do.vKnots); free(mf3do.uKnots);
	mfio_mf3d_errno = status; return MOPS_ERROR; }

      if(trim->type == MOPS_OTNURBCURVE)       /* single trimcurve? */
	{
	  mops_status = mfio_write_trimcurve(fileptr, trim);
	  if(mops_status)
	    { free(mf3do.points); free(mf3do.vKnots); free(mf3do.uKnots);
	      return mops_status; }
	}
      else if(trim->type == MOPS_OTLEVEL)      /* trimloop? */
	{
	  trim = trim->down;
	  while(trim)
	    {
	      mops_status = mfio_write_trimcurve(fileptr, trim);
	      if(mops_status)
		{ free(mf3do.points); free(mf3do.vKnots); free(mf3do.uKnots);
		return mops_status; }

	      trim = trim->next;
	    } /* while */

	} /* if */

      /* close container */
      mops_status = mfio_write_ecntr(fileptr);
      if(mops_status)
	{ free(mf3do.points); free(mf3do.vKnots); free(mf3do.uKnots);
	  return mops_status; }

    } /* if */
    
  /* delete the object */
  free(mf3do.points);
  free(mf3do.vKnots);
  free(mf3do.uKnots);

  /* close container */
  mops_status = mfio_write_ecntr(fileptr);

 return mops_status;
}


int 
mfio_write_nurbcurve(MF3D_FilePtr fileptr, mops_object *o)
{
 int mops_status = MOPS_OK;
 int i = 0, a = 0, b = 0;
 MF3DNURBCurveObj mf3do = {0};
 MF3DErr status = kMF3DNoErr;	/* temporary result code */  
 mops_nurbcurve_object *curve = (mops_nurbcurve_object*)(o->object);

  /* write enclosing container */
  mops_status = mfio_write_cntr(fileptr);
  if(mops_status)
    { return mops_status; }

 
  /* we fill the MF3D-object now */
  mf3do.objectType = kMF3DObjNURBCurve;
  mf3do.nPoints = curve->length;
  mf3do.order = curve->order;

  if(!(mf3do.points = calloc(curve->length, 
			    sizeof(MF3DRationalPoint4D))))
    return MOPS_OUTOFMEM;
  if(!(mf3do.knots = calloc(curve->length+curve->order,
			    sizeof(MF3DKnot))))
    { free(mf3do.points); return MOPS_OUTOFMEM; }

  a = 0; b = 0;
  for(i=0; i<curve->length; i++)
    {
	(mf3do.points)[a].x = curve->controlv[b++];
	(mf3do.points)[a].y = curve->controlv[b++];
	(mf3do.points)[a].z = curve->controlv[b++];
	if(curve->has_weights)
	  (mf3do.points)[a].w = curve->controlv[b++];
	else
	  (mf3do.points)[a].w = 1.0;

	a++;
      }
  a = 0;
  for(i=0; i<curve->length+curve->order; i++)
    {
      (mf3do.knots)[a] = curve->knotv[a]; 
      a++;
    }

  /* we write out the curve now */
  status = MF3DWriteAnObject (fileptr, (MF3DVoidObjPtr)&mf3do);
  if(status != kMF3DNoErr)
    { mfio_mf3d_errno = status; return MOPS_ERROR; }

  /* write attributes */
  mops_status = mfio_write_attributes(fileptr, o);
  if(mops_status)
    { free(mf3do.points); free(mf3do.knots); return mops_status; }

  /* delete the object */
  free(mf3do.points);
  free(mf3do.knots);

  /* close container */
  mops_status = mfio_write_ecntr(fileptr);

 return mops_status;
}


int 
mfio_write_box(MF3D_FilePtr fileptr, mops_object *o)
{
 int mops_status = MOPS_OK;
 MF3DBoxObj mf3do = {0};
 MF3DBoxData bd;
 MF3DErr status = kMF3DNoErr;	/* temporary result code */  
 mops_box_object *box = (mops_box_object*)(o->object);

  /* write enclosing container */
  mops_status = mfio_write_cntr(fileptr);
  if(mops_status)
    { return mops_status; }

  /* we fill the MF3D-object now */
  bd.orientation.x = 0.0;
  bd.orientation.y = 0.0;
  bd.orientation.z = box->length;

  bd.majorAxis.x = 0.0;
  bd.majorAxis.y = box->height;
  bd.majorAxis.z = 0.0;

  bd.minorAxis.x = box->width;
  bd.minorAxis.y = 0.0;
  bd.minorAxis.z = 0.0;

  bd.origin.x = -box->width/2.0;
  bd.origin.y = -box->height/2.0;
  bd.origin.z = -box->length/2.0;

  mf3do.objectType = kMF3DObjBox;
  mf3do.box = &bd;

  /* we write out the box now */
  status = MF3DWriteAnObject (fileptr, (MF3DVoidObjPtr)&mf3do);
  if(status != kMF3DNoErr)
    { mfio_mf3d_errno = status; return MOPS_ERROR; }

  /* write attributes */
  mops_status = mfio_write_attributes(fileptr, o);
  if(mops_status)
    return mops_status;

  /* close container */
  mops_status = mfio_write_ecntr(fileptr);

 return mops_status;
}


int 
mfio_write_sphere(MF3D_FilePtr fileptr, mops_object *o)
{
 int mops_status = MOPS_OK;
 MF3DEllipsoidObj mf3do = {0};
 MF3DEllipsoidData ed;
 MF3DErr status = kMF3DNoErr;	/* temporary result code */  
 mops_sphere_object *sphere = (mops_sphere_object*)(o->object);

  /* write enclosing container */
  mops_status = mfio_write_cntr(fileptr);
  if(mops_status)
    { return mops_status; }

  /* we fill the MF3D-object now */
  ed.orientation.x = 0.0;
  ed.orientation.y = 0.0;
  ed.orientation.z = sphere->radius;

  ed.majorRadius.x = 0.0;
  ed.majorRadius.y = sphere->radius;
  ed.majorRadius.z = 0.0;

  ed.minorRadius.x = sphere->radius;
  ed.minorRadius.y = 0.0;
  ed.minorRadius.z = 0.0;

  ed.origin.x = 0.0;
  ed.origin.y = 0.0;
  ed.origin.z = 0.0;

  mf3do.objectType = kMF3DObjEllipsoid;
  mf3do.ellipsoid = &ed;

  /* we write out the sphere now */
  status = MF3DWriteAnObject (fileptr, (MF3DVoidObjPtr)&mf3do);
  if(status != kMF3DNoErr)
    { mfio_mf3d_errno = status; return MOPS_ERROR; }

  /* write attributes */
  mops_status = mfio_write_attributes(fileptr, o);
  if(mops_status)
    return mops_status;

  /* close container */
  mops_status = mfio_write_ecntr(fileptr);

 return mops_status;
}


int 
mfio_write_cylinder(MF3D_FilePtr fileptr, mops_object *o)
{
 int mops_status = MOPS_OK;
 MF3DCylinderObj mf3do = {0};
 MF3DCylinderData cd;
 MF3DCapsObj cp;
 MF3DErr status = kMF3DNoErr;	/* temporary result code */  
 mops_cylinder_object *cylinder = (mops_cylinder_object*)(o->object);

  /* write enclosing container */
  mops_status = mfio_write_cntr(fileptr);
  if(mops_status)
    { return mops_status; }

  /* we fill the MF3D-object now */
  cd.orientation.x = 0.0;
  cd.orientation.y = 0.0;
  cd.orientation.z = cylinder->zmax - cylinder->zmin;

  cd.majorAxis.x = 0.0;
  cd.majorAxis.y = cylinder->radius;
  cd.majorAxis.z = 0.0;

  cd.minorAxis.x = cylinder->radius;
  cd.minorAxis.y = 0.0;
  cd.minorAxis.z = 0.0;

  cd.origin.x = 0.0;
  cd.origin.y = 0.0;
  cd.origin.z = cylinder->zmin;

  mf3do.objectType = kMF3DObjCylinder;
  mf3do.cylinder = &cd;

  /* we write out the cylinder now */
  status = MF3DWriteAnObject (fileptr, (MF3DVoidObjPtr)&mf3do);
  if(status != kMF3DNoErr)
    { mfio_mf3d_errno = status; return MOPS_ERROR; }

  /* closed? */
  if(cylinder->closed)
    {
      cp.objectType = kMF3DObjCaps;
      cp.caps = kCapsBottom | kCapsTop;

      status = MF3DWriteAnObject (fileptr, (MF3DVoidObjPtr)&cp);
      if(status != kMF3DNoErr)
	{ mfio_mf3d_errno = status; return MOPS_ERROR; }
    }

  /* write attributes */
  mops_status = mfio_write_attributes(fileptr, o);
  if(mops_status)
    return mops_status;

  /* close container */
  mops_status = mfio_write_ecntr(fileptr);

 return mops_status;
}


int 
mfio_write_cone(MF3D_FilePtr fileptr, mops_object *o)
{
 int mops_status = MOPS_OK;
 MF3DConeObj mf3do = {0};
 MF3DConeData cd;
 MF3DCapsObj cp;
 MF3DErr status = kMF3DNoErr;	/* temporary result code */  
 mops_cone_object *cone = (mops_cone_object*)(o->object);

  /* write enclosing container */
  mops_status = mfio_write_cntr(fileptr);
  if(mops_status)
    { return mops_status; }

  /* we fill the MF3D-object now */
  cd.orientation.x = 0.0;
  cd.orientation.y = 0.0;
  cd.orientation.z = cone->height;

  cd.majorAxis.x = 0.0;
  cd.majorAxis.y = cone->radius;
  cd.majorAxis.z = 0.0;

  cd.minorAxis.x = cone->radius;
  cd.minorAxis.y = 0.0;
  cd.minorAxis.z = 0.0;

  cd.origin.x = 0.0;
  cd.origin.y = 0.0;
  cd.origin.z = 0.0;

  mf3do.objectType = kMF3DObjCone;
  mf3do.cone = &cd;

  /* we write out the cone now */
  status = MF3DWriteAnObject (fileptr, (MF3DVoidObjPtr)&mf3do);
  if(status != kMF3DNoErr)
    { mfio_mf3d_errno = status; return MOPS_ERROR; }

  /* closed? */
  if(cone->closed)
    {
      cp.objectType = kMF3DObjCaps;
      cp.caps = kCapsBottom;

      status = MF3DWriteAnObject (fileptr, (MF3DVoidObjPtr)&cp);
      if(status != kMF3DNoErr)
	{ mfio_mf3d_errno = status; return MOPS_ERROR; }
    }

  /* write attributes */
  mops_status = mfio_write_attributes(fileptr, o);
  if(mops_status)
    return mops_status;

  /* close container */
  mops_status = mfio_write_ecntr(fileptr);

 return mops_status;
}


int 
mfio_write_disk(MF3D_FilePtr fileptr, mops_object *o)
{
 int mops_status = MOPS_OK;
 MF3DDiskObj mf3do = {0};
 MF3DDiskData dd;
 MF3DErr status = kMF3DNoErr;	/* temporary result code */  
 mops_disk_object *disk = (mops_disk_object*)(o->object);

  /* write enclosing container */
  mops_status = mfio_write_cntr(fileptr);
  if(mops_status)
    { return mops_status; }

  /* we fill the MF3D-object now */
  dd.majorRadius.x = 0.0;
  dd.majorRadius.y = disk->radius;
  dd.majorRadius.z = 0.0;

  dd.minorRadius.x = disk->radius;
  dd.minorRadius.y = 0.0;
  dd.minorRadius.z = 0.0;

  dd.origin.x = 0.0;
  dd.origin.y = 0.0;
  dd.origin.z = disk->height;

  mf3do.objectType = kMF3DObjDisk;
  mf3do.disk = &dd;

  /* we write out the disk now */
  status = MF3DWriteAnObject (fileptr, (MF3DVoidObjPtr)&mf3do);
  if(status != kMF3DNoErr)
    { mfio_mf3d_errno = status; return MOPS_ERROR; }

  /* write attributes */
  mops_status = mfio_write_attributes(fileptr, o);
  if(mops_status)
    return mops_status;

  /* close container */
  mops_status = mfio_write_ecntr(fileptr);

 return mops_status;
}


int 
mfio_write_torus(MF3D_FilePtr fileptr, mops_object *o)
{
 int mops_status = MOPS_OK;
 MF3DTorusObj mf3do = {0};
 MF3DTorusData td;
 MF3DErr status = kMF3DNoErr;	/* temporary result code */  
 mops_torus_object *torus = (mops_torus_object*)(o->object);

  /* write enclosing container */
  mops_status = mfio_write_cntr(fileptr);
  if(mops_status)
    { return mops_status; }

  /* we fill the MF3D-object now */
  td.orientation.x = 0.0;
  td.orientation.y = 0.0;
  td.orientation.z = torus->minorrad;

  td.majorAxis.x = 0.0;
  td.majorAxis.y = torus->majorrad;
  td.majorAxis.z = 0.0;

  td.minorAxis.x = torus->majorrad;
  td.minorAxis.y = 0.0;
  td.minorAxis.z = 0.0;

  td.ratio = 1.0;

  td.origin.x = 0.0;
  td.origin.y = 0.0;
  td.origin.z = 0.0;

  mf3do.objectType = kMF3DObjTorus;
  mf3do.torus = &td;

  /* we write out the torus now */
  status = MF3DWriteAnObject (fileptr, (MF3DVoidObjPtr)&mf3do);
  if(status != kMF3DNoErr)
    { mfio_mf3d_errno = status; return MOPS_ERROR; }

  /* write attributes */
  mops_status = mfio_write_attributes(fileptr, o);
  if(mops_status)
    return mops_status;

  /* close container */
  mops_status = mfio_write_ecntr(fileptr);

 return mops_status;
}


int 
mfio_write_attributes(MF3D_FilePtr fileptr, mops_object *o)
{
 int mops_status = MOPS_OK;
 float euler[3] = {0};
 MF3DTranslateObj to = {0};
 MF3DScaleObj so = {0};
 MF3DRotateObj ro = {0};
 MF3DDiffuseColorObj dco = {0};
 MF3DTransparencyColorObj tco = {0};
 MF3DErr status = kMF3DNoErr;	/* temporary result code */  

  to.objectType = kMF3DObjTranslateTransform;
  to.translate.x = (MF3DFloat32)o->movx;
  to.translate.y = (MF3DFloat32)o->movy;
  to.translate.z = (MF3DFloat32)o->movz;

  if((o->movx != 0.0) || (o->movy != 0.0) || (o->movz != 0.0))
    status = MF3DWriteAnObject (fileptr, (MF3DVoidObjPtr)&to);
  if(status != kMF3DNoErr)
    { mfio_mf3d_errno = status; return MOPS_ERROR; }

  mops_quat_to_euler(o->quat,euler);
  ro.objectType = kMF3DObjRotateTransform;
  ro.axis = MF3DAxisZ;
  ro.radians = -euler[0];
  if(euler[0] != 0.0)
    status = MF3DWriteAnObject (fileptr, (MF3DVoidObjPtr)&ro);
  if(status != kMF3DNoErr)
    { mfio_mf3d_errno = status; return MOPS_ERROR; }

  ro.axis = MF3DAxisY;
  ro.radians = -euler[1];
  if(euler[1] != 0.0)
    status = MF3DWriteAnObject (fileptr, (MF3DVoidObjPtr)&ro);
  if(status != kMF3DNoErr)
    { mfio_mf3d_errno = status; return MOPS_ERROR; }

  ro.axis = MF3DAxisX;
  ro.radians = -euler[2];
  if(euler[2] != 0.0)
    status = MF3DWriteAnObject (fileptr, (MF3DVoidObjPtr)&ro);
  if(status != kMF3DNoErr)
    { mfio_mf3d_errno = status; return MOPS_ERROR; }

  so.objectType = kMF3DObjScaleTransform;
  so.scale.x = o->scalx;
  so.scale.y = o->scaly;
  so.scale.z = o->scalz;

  if((o->scalx != 1.0) || (o->scaly != 1.0) || (o->scalz != 1.0))
    status = MF3DWriteAnObject (fileptr, (MF3DVoidObjPtr)&so);
  if(status != kMF3DNoErr)
    { mfio_mf3d_errno = status; return MOPS_ERROR; }

  dco.objectType = kMF3DObjDiffuseColor;
  dco.diffuseColor.red = o->colr;
  dco.diffuseColor.green = o->colg;
  dco.diffuseColor.blue = o->colb;
  status = MF3DWriteAnObject (fileptr, (MF3DVoidObjPtr)&dco);
  if(status != kMF3DNoErr)
    { mfio_mf3d_errno = status; return MOPS_ERROR; }

  tco.objectType = kMF3DObjTransparencyColor;
  tco.transparency.red = o->opr;
  tco.transparency.green = o->opg;
  tco.transparency.blue = o->opb;
  status = MF3DWriteAnObject (fileptr, (MF3DVoidObjPtr)&tco);
  if(status != kMF3DNoErr)
    { mfio_mf3d_errno = status; return MOPS_ERROR; }

 return mops_status;
}


int 
mfio_write_level(MF3D_FilePtr fileptr, mops_object *o)
{
 int mops_status = MOPS_OK;
 mops_object *level = o;

  if(!o)
    return MOPS_OK;

  o = o->down;
  mops_status = mfio_write_cntr(fileptr);
  mops_status = mfio_write_attributes(fileptr, level);
  while(o)
    {
      mops_status = mfio_write_object(fileptr, o);
      o = o->next;
    }

  mops_status = mfio_write_ecntr(fileptr);

 return mops_status;
}


int 
mfio_write_cntr(MF3D_FilePtr fileptr)
{
 int mops_status = MOPS_OK;
 MF3DContainerObj mf3do = {0};
 MF3DErr status = kMF3DNoErr;	/* temporary result code */  
  
  mf3do.objectType = kMF3DObjContainer;

  /* we write out a container now */
  status = MF3DWriteAnObject (fileptr, (MF3DVoidObjPtr)&mf3do);
  if(status != kMF3DNoErr)
    { mfio_mf3d_errno = status; return MOPS_ERROR; }

 return mops_status;
}


int 
mfio_write_ecntr(MF3D_FilePtr fileptr)
{
 int mops_status = MOPS_OK;
 MF3DEndContainerObj mf3do = {0};
 MF3DErr status = kMF3DNoErr;	/* temporary result code */  
  
  mf3do.objectType = kMF3DObjEndContainer;
  status = MF3DWriteAnObject (fileptr, (MF3DVoidObjPtr)&mf3do);
  if(status != kMF3DNoErr)
    { mfio_mf3d_errno = status; return MOPS_ERROR; }

 return mops_status;
}


int 
mfio_write_object(MF3D_FilePtr fileptr, mops_object *object)
{
 int mops_status = MOPS_OK;
 mops_custom_object *custom = NULL;
 mops_object *po = NULL, *po2;

  switch(object->type)
    {
    case MOPS_OTNURBPATCH:
      mops_status = mfio_write_nurbpatch(fileptr, object);
      if(mops_status)
	return mops_status;
      break;

    case MOPS_OTNURBCURVE:
      mops_status = mfio_write_nurbcurve(fileptr, object);
      if(mops_status)
	return mops_status;
      break;

    case MOPS_OTLEVEL:
      if(object->down)
	{
	  mops_status = mfio_write_level(fileptr, object);
	  if(mops_status)
	    return mops_status;
	}
      break;
    case MOPS_OTLIGHT:
/*      mops_status = mfio_write_light(fileptr, object);*/
      if(mops_status)
	return mops_status;
      break;
    case MOPS_OTBOX:
      mops_status = mfio_write_box(fileptr, object);
      if(mops_status)
	return mops_status;
      break;
    case MOPS_OTSPHERE:
      mops_status = mfio_write_sphere(fileptr, object);
      if(mops_status)
	return mops_status;
      break;
    case MOPS_OTDISK:
      mops_status = mfio_write_disk(fileptr, object);
      if(mops_status)
	return mops_status;
      break;
    case MOPS_OTCONE:
      mops_status = mfio_write_cone(fileptr, object);
      if(mops_status)
	return mops_status;
      break;
    case MOPS_OTCYLINDER:
      mops_status = mfio_write_cylinder(fileptr, object);
      if(mops_status)
	return mops_status;
      break;
    case MOPS_OTTORUS:
      mops_status = mfio_write_torus(fileptr, object);
      if(mops_status)
	return mops_status;
      break;
      /*    case MOPS_OTPARABOLOID:
      mops_status = mfio_write_paraboloid(fileptr, object);
      if(mops_status)
	return mops_status;
      break;
    case MOPS_OTHYPERBOLOID:
      mops_status = mfio_write_hyperboloid(fileptr, object);
      if(mops_status)
	return mops_status;
      break;*/

    case MOPS_OTINSTANCE:
      mops_status = mfio_write_object(fileptr, (mops_object *)object->object);
      break;

      /* if we got so far there is probably no exporter for this object,
	 but maybe we are lucky, and we can convert the object
       */
    case MOPS_OTCUSTOM:
      custom = (mops_custom_object*)object->object;

      if(custom->provides)
	mops_status = mops_get_object(object, custom->provides, &po);

      if(po)
	{
	  po2 = po;
	  while(po2)
	    {
	      mfio_write_object(fileptr, po2);
	      po2 = po2->next;
	    }

	  mops_delete_multi_objects(po);

	}
      break;

    }

 return mops_status;
}

int
mfio_write_scene(Tcl_Interp *interp, char *filename)
{
 int mops_status = MOPS_OK;
 mops_object *co = mops_root->next;
 MF3D_FilePtr metafilePtr;	/* MF3D internal file pointer   */
 MF3DErr status = kMF3DNoErr;	/* temporary result code        */
 MF3DMetafileObj mfo = {0};

  mfo.objectType = kMF3DObjMetafile;
  mfo.dataFormat = mfio_mf3d_data_format;
  mfo.majorVersion = (MF3DUns16)1;
  mfo.minorVersion = (MF3DUns16)1;
  mfo.flags = kMF3DFormatStream;
  mfo.refSeed = kMF3DMaximumRefSeed;
  mfo.typeSeed = kMF3DMinimumTypeSeed;

  /* Open the metafile */
  status = MF3DOpenOutputStdCFile (mfio_mf3d_data_format, filename,
				   &metafilePtr);
  if (status != kMF3DNoErr)
    return MOPS_OPENFILE;

  /* write the metafile-object */
  status = MF3DWriteAnObject (metafilePtr, (MF3DVoidObjPtr)&mfo);
  if(status != kMF3DNoErr)
    { mfio_mf3d_errno = status; return MOPS_ERROR; }

  while(co)
    {
      mops_status = mfio_write_object(metafilePtr, co);
      if (mops_status)
	return mops_status;

      co = co->next;
    }

  /* Close the file */
  status = MF3DClose (metafilePtr);
  if (status != kMF3DNoErr)
    return MOPS_CLOSEFILE;

 return MOPS_OK;
}


/************************************/
/************************************/
/************************************/

int
mfio_import_scene_tcmd(ClientData clientData, Tcl_Interp * interp,
		      int argc, char *argv[])
{
 int mops_status = MOPS_OK;
 char fname[] = "import_mf3d_scene";
 char done_cmd[] = "puts stdout \"Done importing scene from: $mfilename\"";

 /* check args */
 if(argc != 2)
   {
     mops_error(interp, MOPS_ARGS, fname, "filename");
     return TCL_OK;
   }

 mfio_last_read_object = NULL;

 mops_status = mfio_read_scene(interp, argv[1]);

 MOPS_CRERROR

 Tcl_Eval(interp, done_cmd);

 return TCL_OK;
}


int
mfio_export_scene_tcmd(ClientData clientData, Tcl_Interp *interp,
		       int argc, char *argv[])
{
 int mops_status = MOPS_OK;
 char fname[] = "export_mf3d_scene";
 char done_cmd[] = "puts stdout \"Done exporting scene to: $mfilename\"";

  /* check args */
  if(argc != 2)
    {
      mops_error(interp, MOPS_ARGS, fname, "filename");
      return TCL_OK;
    }

  mops_status = mfio_write_scene(interp, argv[1]);
  
  MOPS_CRERROR

  Tcl_Eval(interp, done_cmd);

 return TCL_OK;
}


int
Mfio_Init(Tcl_Interp *interp)
{
 char fname[] = "Mfio_Init";
 char success_cmd[] =
   "puts stdout \"Mfio (3DMF) plugin successfully loaded.\"";

  /* check versions */
  if(strcmp(mops_version_ma, mfio_mops_version_ma))
    {
      mops_error(interp, MOPS_ERROR, fname,
		  "Plugin has been compiled for a different Mops!");
      mops_error(interp, MOPS_ERROR, fname,
		  "It is unsafe to continue! Bailing out...");
      return TCL_OK;
    }
  if(strcmp(mops_version_mi, mfio_mops_version_mi))
    {
      mops_error(interp, MOPS_ERROR, fname,
		  "Plugin has been compiled for a different Mops!");
      mops_error(interp, MOPS_ERROR, fname,
		  "However, it is probably safe to continue...");
    }

 
  mfio_mf3d_data_format = kMF3DFormatText /*kMF3DFormatBinary*/;

  /* register some C-functions as Tcl-Commands */
  Tcl_CreateCommand (interp, "mfio_import",
		     mfio_import_scene_tcmd,
		     (ClientData) NULL, (Tcl_CmdDeleteProc *) NULL);
  Tcl_CreateCommand (interp, "mfio_export",
		     mfio_export_scene_tcmd,
		     (ClientData) NULL, (Tcl_CmdDeleteProc *) NULL);

  /* source mfio.tcl, it contains some Tcl-procedures that
     attach to the custom menu */
  if((Tcl_EvalFile(interp, "mfio.tcl")) != TCL_OK)
     {
       mops_error(interp, MOPS_ERROR, fname,
		  "Error while sourcing \\\"mfio.tcl\\\"!");
       return TCL_OK;
     }

  Tcl_Eval(interp, success_cmd);

 return TCL_OK;
}
