/*
 * The Mops, a free 3D Tcl/Tk/OpenGL/RenderMan modeler
 *
 * The Mops is copyrighted (C) 1998-1999 by Randolf Schultz
 * (rschultz@informatik.uni-rostock.de)
 *
 * See the file License for details.
 *
 */

#include <limits.h>
#include <float.h>
#include "mops.h"

extern Tcl_Interp *interp;

#ifdef MOPSIDR
#include "idr.h"
#endif

/* Main Module */

/* Global Variables */

Tcl_Interp *mops_interp;

mops_prefs_object mops_prefs;
mops_riopt_object mops_riopt;

mops_object *mops_last_read_object;

int mops_read_version;

/* where do we link the next object that will be created? */
mops_object **mops_next_object;

/* the root of the object hierarchy */
mops_object *mops_root;

/* selected objects */
mops_list_object *mops_current_selection;

mops_list_object *mops_last_selection;


/* a pointer to the current level in the hierarchy */
mops_list_object *mops_current_level;

/* the clipboard */
mops_object *mops_object_clipboard;

/* undo information */
mops_object *mops_undo_objects;
mops_list_object *mops_undo_list;

/* the following is needed for single point editing */
double **mops_point_edit_coords;
int mops_point_edit_coords_homogenous;
int mops_point_edit_coords_number;
mops_object *mops_point_edit_object;

/* this is not the hashtable for the creation of
   custom objects, but the table for the crtOb
   command */
Tcl_HashTable mops_createcb_ht;

Tcl_HashTable mops_tagtypes_ht;

Tcl_HashTable mops_oidptr_ht;

Tcl_HashTable mops_converter_ht;

Tcl_HashTable mops_provider_ht;

char *mops_riattr_tagtype;
char *mops_oi_tagtype;

char mops_version_ma[] = MOPS_VERSIONSTR;
char mops_version_mi[] = MOPS_VERSIONSTRMI;

/* Functions */

/********************************************/
/*mops_init:
  initialize Mops (hashtables and global
  variables)                                */
/********************************************/
int
mops_init(Tcl_Interp *interp)
{
 int mops_status = MOPS_OK;
 mops_object *o = NULL;

  mops_interp = interp;

  Tcl_InitHashTable(&mops_createcb_ht, TCL_STRING_KEYS);

  /* register object creation callbacks */
  mops_register_crt(interp, "npatch", mops_crtnpatch_tcmd);
  mops_register_crt(interp, "ncurve", mops_crtncurve_tcmd);
  mops_register_crt(interp, "light", mops_crtlight_tcmd);
  mops_register_crt(interp, "nsphere", mops_crtnsphere_tcmd);
  mops_register_crt(interp, "ncircle", mops_crtncircle_tcmd);
  mops_register_crt(interp, "level", mops_crtlevel_tcmd);
  mops_register_crt(interp, "revolve", mops_revolve_tcmd);
  mops_register_crt(interp, "extrude", mops_extrude_tcmd);
  mops_register_crt(interp, "sweep", mops_sweep_tcmd);
  mops_register_crt(interp, "quadric", mops_crtquad_tcmd);
  mops_register_crt(interp, "custom", mops_custom_create_tcmd);
  mops_register_crt(interp, "box", mops_crtbox_tcmd);
  mops_register_crt(interp, "trimrect", mops_crt_trimrect_tcmd);
  mops_register_crt(interp, "cbsp", mops_crt_closedbsp_tcmd);
  mops_register_crt(interp, "concatnc", mops_concat_nc_tcmd);
  mops_register_crt(interp, "instance", mops_crtinstance_tcmd);

  /* init tag system */
  Tcl_InitHashTable(&mops_tagtypes_ht, TCL_STRING_KEYS);

  /* register tag types */
  mops_register_tagtype(interp, "RiAttrib", &mops_riattr_tagtype);
  mops_register_tagtype(interp, "OI", &mops_oi_tagtype);

  /* hash table for id -> original object pointers */
  Tcl_InitHashTable(&mops_oidptr_ht, TCL_STRING_KEYS);

  /* hash table for object converters */
  Tcl_InitHashTable(&mops_converter_ht, TCL_STRING_KEYS);

  /* register converters */
  mops_register_converter(interp,"Instance", mops_resolve_instance);

#ifdef MOPSIDR
  mops_status = mops_idr_init(interp);
#endif


  /* initialize hashtables for custom objects */
  mops_status = mops_init_custom_ht();

  /* hash table for object providers */
  Tcl_InitHashTable(&mops_provider_ht, TCL_STRING_KEYS);

  /* init custom objects, that are statically linked into The Mops */
  Revolve_Init(interp);
  Extrude_Init(interp);
  Sweep_Init(interp);
  ICurve_Init(interp);
  Skin_Init(interp);

#ifdef WIN32
  Mfio_Init(interp);
#endif

  /* create root object */
  if(!(o = calloc(1,sizeof(mops_object))))
    return(MOPS_OUTOFMEM);

  mops_root = o;
  o->type = MOPS_OTROOT;

  /* global vars */
  mops_next_object = &(o->next);
  
  mops_current_selection = NULL;
  mops_last_selection = NULL;

  mops_current_level = NULL;
  mops_add_level_object(o);

  mops_object_clipboard = NULL;

  mops_undo_objects = NULL;
  mops_undo_list = NULL;

  /* RiOptions */
  mops_riopt.width = 0;
  mops_riopt.height = 0;
  mops_riopt.Variance = 0.0;
  mops_riopt.Samples_X = 2;
  mops_riopt.Samples_Y = 2;
  mops_riopt.FilterFunc = 0;
  mops_riopt.FilterWidth = 2;
  mops_riopt.FilterHeight = 2;
  mops_riopt.ExpGain = 1.0;
  mops_riopt.ExpGamma = 1.0;
  mops_riopt.RGBA_ONE = 255;
  mops_riopt.RGBA_MIN = 0;
  mops_riopt.RGBA_MAX = 255;
  mops_riopt.RGBA_Dither = 0.5;

  mops_riopt.MinSamples = 4;
  mops_riopt.MaxSamples = 64;
  mops_riopt.MaxRayLevel = 4;
  mops_riopt.ShadowBias = 0.01;
  mops_riopt.PRManSpec = 1;

  mops_riopt.RadSteps = 0;
  mops_riopt.PatchSamples = 2;
  mops_riopt.texturemem = 10000;
  mops_riopt.geommem = 40000;

  /* prefs */
  mops_prefs.pick_epsilon = MOPS_DEFAULT_PICK_EPSILON;

 return MOPS_OK;
} /* mops_init */


/********************************************/
/* set standard attributes to render a
   mops_object with OpenGL                  */
/********************************************/
void
mops_render_attributes(mops_object *o)
{
 GLfloat color[4] = {0.8, 0.75, 0.65, 1.0};
 GLfloat m[4][4];

  glTranslated(o->movx, o->movy, o->movz);

  mops_build_rotmatrix(m, o->quat);
  glMultMatrixf(&m[0][0]);

  glScaled (o->scalx, o->scaly, o->scalz);
 
  if(o->colr != -1 && o->colg != -1 && o->colb != -1)
    {
      color[0] = o->colr/255.0;
      color[1] = o->colg/255.0;
      color[2] = o->colb/255.0;
      color[3] = o->cola/255.0; 
      glMaterialfv (GL_FRONT_AND_BACK, GL_AMBIENT_AND_DIFFUSE, color);
    }

 return;
} /* mops_render_attributes */


/********************************************/
/* setup the projection matrix of view
   *togl, to reflect it's "from" and "to"
   camera settings                          */
/********************************************/
void
mops_setup_projection(struct Togl *togl)
{
 mops_view_info *view = Togl_GetClientData(togl);
 int width = Togl_Width (togl);
 int height = Togl_Height (togl);
 float aspect = (float) width / (float) height;  
 GLfloat light_posf[4] = {0.0, 0.0, 10.0, 1.0};
 GLfloat light_poss[4] = {10.0, 0.0, 0.0, 1.0};
 GLfloat light_post[4] = {0.0, 10.0, 0.0, 1.0};

  glViewport (0, 0, width, height);

  glMatrixMode(GL_PROJECTION);

  glLoadIdentity();
  if(view->type == MOPS_VTPERSP)
    glFrustum(-aspect*view->zoom, aspect*view->zoom,
	       -1.0*view->zoom, 1.0*view->zoom,
	       1, 1000.0);
  else
    glOrtho(-aspect*view->zoom, aspect*view->zoom,
	    -1.0*view->zoom, 1.0*view->zoom,
	    -100.0, 100.0);

  if(view->roll != 0.0)
    glRotated(view->roll, 0.0, 0.0, 1.0);
  if(view->type != MOPS_VTTOP)
    gluLookAt(view->from_x, view->from_y, view->from_z,
	      view->to_x, view->to_y, view->to_z,
	      view->up_x, view->up_y, view->up_z);
  else
    gluLookAt(view->from_x, view->from_y, view->from_z,
	      view->to_x, view->to_y, view->to_z,
	      view->up_x, view->up_y, view->up_z);

  glMatrixMode(GL_MODELVIEW);

  /* XXXX set light0 at cam-pos? */
  glPushMatrix();
  /*  glLoadIdentity();*/
  glRotated(view->rotz, 0.0, 0.0, 1.0);
  glRotated(view->roty, 0.0, 1.0, 0.0);
  glRotated(view->rotx, 1.0, 0.0, 0.0);
  if((view->type == MOPS_VTPERSP) || (view->type == MOPS_VTFRONT))
    glLightfv (GL_LIGHT0, GL_POSITION, light_posf);
  if(view->type == MOPS_VTSIDE)
    glLightfv (GL_LIGHT0, GL_POSITION, light_poss);
  if(view->type == MOPS_VTTOP)
    glLightfv (GL_LIGHT0, GL_POSITION, light_post);
  glPopMatrix();

 return;
} /* mops_setup_projection */


/************************************************/
/* draw the current grid                        */
/************************************************/
void
mops_draw_grid(struct Togl *togl)
{
 mops_view_info *view = Togl_GetClientData(togl);
 int width = Togl_Width (togl);
 int height = Togl_Height (togl);
 double aspect = (double) width / (double)height;
 double grid = view->grid;
 double j;
 double minwinx=0.0, minwiny=0.0, maxwinx=0.0, maxwiny=0.0;

  if(grid == 0.0)
    return;

  switch (view->type)
    {
    case MOPS_VTFRONT:
    case MOPS_VTTRIM:
      if(aspect > 1.0)
	{
	  minwinx = floor(((-aspect * view->zoom) - fabs(view->from_x))/
			  (view->grid<1.0?view->grid:1.0/view->grid));
	  minwiny = floor(((-view->zoom) - fabs(view->from_y))/
			  (view->grid<1.0?view->grid:1.0/view->grid));
	}
      else
	{
	  minwiny = floor(((view->zoom/-aspect) - fabs(view->from_y))/
			  (view->grid<1.0?view->grid:1.0/view->grid));
	  minwinx = floor(((-view->zoom) - fabs(view->from_x))/
			  (view->grid<1.0?view->grid:1.0/view->grid));
	}
      break;
    case MOPS_VTSIDE:
      if(aspect > 1.0)
	{
	  minwinx = floor(((-aspect * view->zoom) - fabs(view->from_z))/
			  (view->grid<1.0?view->grid:1.0/view->grid));
	  minwiny = floor(((-view->zoom) - fabs(view->from_y))/
			  (view->grid<1.0?view->grid:1.0/view->grid));
	}
      else
	{
	  minwiny = floor(((view->zoom/-aspect) - fabs(view->from_y))/
			  (view->grid<1.0?view->grid:1.0/view->grid));
	  minwinx = floor(((-view->zoom) - fabs(view->from_z))/
			  (view->grid<1.0?view->grid:1.0/view->grid));
	}
      break;
    case MOPS_VTTOP:
      if(aspect > 1.0)
	{
	  minwinx = floor(((-aspect * view->zoom) - fabs(view->from_x))/
			  (view->grid<1.0?view->grid:1.0/view->grid));
	  minwiny = floor(((-view->zoom) - fabs(view->from_z))/
			  (view->grid<1.0?view->grid:1.0/view->grid));
	}
      else
	{
	  minwiny = floor(((view->zoom/-aspect) - fabs(view->from_z))/
			  (view->grid<1.0?view->grid:1.0/view->grid));
	  minwinx = floor(((-view->zoom) - fabs(view->from_x))/
			  (view->grid<1.0?view->grid:1.0/view->grid));
	}
      break;
    }

  maxwinx = -minwinx;
  maxwiny = -minwiny;
  glColor3f MOPS_CLGRID;
  glBegin(GL_LINES);

  switch (view->type)
    {
    case MOPS_VTFRONT:
    case MOPS_VTTRIM:
      /* vertical lines */
      for(j=0.0;j<maxwinx;j+=view->grid)
	{
	  glVertex3d(j,minwiny,0.0);
	  glVertex3d(j,maxwiny,0.0);
	}
      for(j=0.0-view->grid;j>minwinx;j-=view->grid)
	{
	  glVertex3d(j,minwiny,0.0);
	  glVertex3d(j,maxwiny,0.0);
	}

      /* horizontal lines */
      for(j=0.0;j<maxwiny;j+=view->grid)
	{
	  glVertex3d(minwinx,j,0.0);
	  glVertex3d(maxwinx,j,0.0);
	}
      for(j=0.0-view->grid;j>minwiny;j-=view->grid)
	{
	  glVertex3d(minwinx,j,0.0);
	  glVertex3d(maxwinx,j,0.0);
	}
      break;
    case MOPS_VTSIDE:
      /* vertical lines */
      for(j=0.0;j<maxwinx;j+=view->grid)
	{
	  glVertex3d(0.0,minwiny,j);
	  glVertex3d(0.0,maxwiny,j);
	}
      for(j=0.0-view->grid;j>minwinx;j-=view->grid)
	{
	  glVertex3d(0.0,minwiny,j);
	  glVertex3d(0.0,maxwiny,j);
	}

      /* horizontal lines */
      for(j=0.0;j<maxwiny;j+=view->grid)
	{
	  glVertex3d(0.0,j,minwinx);
	  glVertex3d(0.0,j,maxwinx);
	}
      for(j=0.0-view->grid;j>minwiny;j-=view->grid)
	{
	  glVertex3d(0.0,j,minwinx);
	  glVertex3d(0.0,j,maxwinx);
	}
      break;
    case MOPS_VTTOP:
      /* vertical lines */
      for(j=0.0;j<maxwinx;j+=view->grid)
	{
	  glVertex3d(j,0.0,minwiny);
	  glVertex3d(j,0.0,maxwiny);
	}
      for(j=0.0-view->grid;j>minwinx;j-=view->grid)
	{
	  glVertex3d(j,0.0,minwiny);
	  glVertex3d(j,0.0,maxwiny);
	}

      /* horizontal lines */
      for(j=0.0;j<maxwiny;j+=view->grid)
	{
	  glVertex3d(minwinx,0.0,j);
	  glVertex3d(maxwinx,0.0,j);
	}
      for(j=0.0-view->grid;j>minwiny;j-=view->grid)
	{
	  glVertex3d(minwinx,0.0,j);
	  glVertex3d(maxwinx,0.0,j);
	}
      break;
    case MOPS_VTPERSP:
      minwinx = -(25.0 * view->grid);
      minwiny = -(25.0 * view->grid);
      maxwinx = (25.0 * view->grid);
      maxwiny = (25.0 * view->grid);
      /* vertical lines */
      for(j=minwinx; j<maxwinx; j+=view->grid)
	{
	  glVertex3d(j,0.0,minwiny);
	  glVertex3d(j,0.0,maxwiny);
	}

      /* horizontal lines */
      for(j=minwiny; j<maxwiny; j+=view->grid)
	{
	  glVertex3d(minwinx,0.0,j);
	  glVertex3d(maxwinx,0.0,j);
	}
      break;
    }

  glEnd();
  
  glClear(GL_DEPTH_BUFFER_BIT);

 return;
} /* mops_draw_grid */


void
mops_draw_selp(mops_object *o)
{  /* now draw selected points (if there are any) */
 mops_point_object *point = NULL;


  if(o->selp)
    {
      glDisable(GL_DEPTH_TEST);

      glColor3f MOPS_CLSELP;
      point = o->selp;
      glBegin(GL_POINTS);
      while(point)
	{
	  glVertex3d(point->point[0],point->point[1],point->point[2]);
	  
	  point = point->next;
	}

      glEnd();
      glColor3f MOPS_CLSEL;
      glEnable(GL_DEPTH_TEST);
    }



return;
}

/********************************************/
/* draw trimcurves and the bounds of the
   nurbpatchs parameter space, if 
   a) the current level is a nurbpatch,
      or a trimloop-level in a nurbpatch
   b) there are any trimcurves              */
/********************************************/
void
mops_draw_trimview(struct Togl *togl)
{
 mops_view_info *view = Togl_GetClientData(togl);
 mops_object *o = NULL, *tmp = NULL, *tmp2 = NULL, *tmp3 = NULL;
 mops_object *trim = NULL, *loop = NULL;
 GLfloat minx=0.0, maxx=0.0;
 GLfloat miny=0.0, maxy=0.0;
 mops_nurbpatch_object *patch = NULL;
 int reset_color = 0;

 if(!mops_current_level)
   return;
 
 tmp = mops_current_level->object;

 mops_del_level_object();

 o = mops_current_level->object;

 if(o->type != MOPS_OTNURBPATCH)
   {
     if(o->type == MOPS_OTLEVEL)
       {
	 if(((mops_level_object *)(o->object))->type == MOPS_LTTRIMLOOP)
	 {

	   tmp2 = mops_current_level->object;
	   mops_del_level_object();
	   tmp3 = mops_current_level->object;
	   mops_del_level_object();
	   if(mops_current_level->object->type != MOPS_OTNURBPATCH)
	     {
	       mops_add_level_object(tmp3);
	       mops_add_level_object(tmp2);
	       mops_add_level_object(tmp);
	       return;
	     }
	   patch = (mops_nurbpatch_object *)
	     (mops_current_level->object->object);
	   trim = mops_current_level->object->down;
	   mops_add_level_object(tmp3);
	   mops_add_level_object(tmp2);
	   mops_add_level_object(tmp);
	 }
	 else
	 {
	   mops_add_level_object(tmp);
	   return;
	 }
       }
     else
       {
	 mops_add_level_object(tmp);
	 return;
       }
   }
 else
   {
     patch = (mops_nurbpatch_object *) o->object;
     trim = o->down;
     mops_add_level_object(tmp);
   }

  /* draw the bounds for the trimcurves as rectangle */
  minx = (GLfloat)((patch->uknotv)[0]);
  miny = (GLfloat)((patch->vknotv)[0]);
 
  maxx = (GLfloat)((patch->uknotv)[patch->width + patch->uorder - 1]);
  maxy = (GLfloat)((patch->vknotv)[patch->height + patch->vorder - 1]);
  glDisable (GL_LIGHTING);
  glColor3f(1.0,1.0,1.0);
  glBegin(GL_LINE_STRIP);
  glVertex3f(minx,miny,0.0);
  glVertex3f(maxx,miny,0.0);
  glVertex3f(maxx,maxy,0.0);
  glVertex3f(minx,maxy,0.0);
  glVertex3f(minx,miny,0.0);
  glEnd();
  glColor3f MOPS_CLUSEL;

  /* now draw the trimcurves */
  while(trim)
    {
      if((trim->type == MOPS_OTNURBCURVE) && (trim->editor))
	{
	  glPushMatrix();
	  mops_render_attributes(trim);
	  if(trim->selected)
	    {
	      glColor3f MOPS_CLSEL;
	      reset_color = 1;
	    }
	  
	  if(view->drawsel)
	    {
	      if(trim->selected)
		{
		  mops_draw_nurbcurve_h(togl, trim);
		  if(view->drawwire)
		    mops_draw_nurbcurve_c(togl, trim);
		  if(view->drawsurf)
		    mops_draw_nurbcurve(togl, trim);
		  if(trim->selp)
		    mops_draw_selp(trim);
		}
	    }
	  else
	    {
	      if(view->drawwire)
		mops_draw_nurbcurve_c(togl, trim);
	      if(view->drawsurf)
		mops_draw_nurbcurve(togl, trim);
	      if(trim->selected)
		{
		  mops_draw_nurbcurve_h(togl, trim);
		  if(trim->selp)
		    mops_draw_selp(trim);
		}
	    }
	  glPopMatrix();
	  
	  if(reset_color)
	    glColor3f MOPS_CLUSEL;
	}
      else if(trim->down)
	{
	  loop = trim->down;
	  while(loop)
	    {
	      if(loop->type == MOPS_OTNURBCURVE && loop->editor)
		{
		  glPushMatrix();
		  mops_render_attributes(loop);
		  if(loop->selected)
		    {
		      glColor3f MOPS_CLSEL;
		      reset_color = 1;
		    }
		 
		  if(view->drawsel)
		    {
		      if(loop->selected)
			{
			  mops_draw_nurbcurve_h(togl, loop);
			  if(view->drawwire)
			    mops_draw_nurbcurve_c(togl, loop);
			  if(view->drawsurf)
			    mops_draw_nurbcurve(togl, loop);
			  if(loop->selp)
			    mops_draw_selp(loop);

			}
		    }
		  else
		    {
		      if(view->drawwire)
			mops_draw_nurbcurve_c(togl, loop);
		      if(view->drawsurf)
			mops_draw_nurbcurve(togl, loop);
		      if(loop->selected)
			{
			  mops_draw_nurbcurve_h(togl, loop);
			  if(loop->selp)
			    mops_draw_selp(loop);
			}
		    }
		  glPopMatrix();
		}
	      loop = loop->next;

	      if(reset_color)
		glColor3f MOPS_CLUSEL;

	    }
	}

      trim = trim->next;
    }

 return;
}


/********************************************/
/*mops_draw_object:
 */
/********************************************/
void
mops_draw_object(struct Togl *togl, mops_object *o)
{
 mops_view_info *view = Togl_GetClientData(togl);
 mops_object *down = NULL;
 int reset_color = 0;
 GLfloat m[4][4];
 GLfloat color[4] = {0.8, 0.75, 0.65, 1.0};

  /* custom objects decide for themself wether to use the
     standard attributes */
  if(o->type != MOPS_OTCUSTOM)
    {
      if(!o->editor)
	return;

      glPushMatrix();
      glTranslated(o->movx, o->movy, o->movz);

      mops_build_rotmatrix(m, o->quat);
      glMultMatrixf(&m[0][0]);

      glScaled (o->scalx, o->scaly, o->scalz);

      
      if(o->type != MOPS_OTINSTANCE)
	if(o->colr != -1 && o->colg != -1 && o->colb != -1)
	  {
	    color[0] = o->colr/255.0;
	    color[1] = o->colg/255.0;
	    color[2] = o->colb/255.0;
	    color[3] = o->cola/255.0; 
	    glMaterialfv (GL_FRONT_AND_BACK, GL_AMBIENT_AND_DIFFUSE, color);
	  }

      if(o->selected)
	{
	  glColor3f MOPS_CLSEL;
	  reset_color = 1;
	}
    }

  switch(o->type)
    {
    case MOPS_OTCUSTOM:
      if(view->drawwire)
	mops_draw_custom_w(togl, o);
      if(view->drawsurf)
	mops_draw_custom_s(togl, o);
      break;

    case MOPS_OTLEVEL:
      down = o->down;
      while(down)
	{
	  mops_draw_object(togl, down);
	  down = down->next;
	}
      break;

    case MOPS_OTNURBPATCH: 
      if(view->drawwire)
	mops_draw_nurbpatch_c(togl, o);
      if(view->drawsurf)
	mops_draw_nurbpatch(togl, o);
      break;
    case MOPS_OTNURBCURVE: 
      if(view->drawwire)
	mops_draw_nurbcurve_c(togl, o);
      if(view->drawsurf)
	mops_draw_nurbcurve(togl, o);
      break;
    case MOPS_OTLIGHT: 
      mops_draw_light(togl, o);
      if(o->selected)
	mops_draw_light_h(togl,o);
      break;
    case MOPS_OTSPHERE: 
      if(view->drawwire)
	mops_draw_sphere_c(togl, o);
      if(view->drawsurf)
	mops_draw_sphere(togl, o);
      break;
    case MOPS_OTDISK: 
      if(view->drawwire)
	mops_draw_disk_c(togl, o);
      if(view->drawsurf)
	mops_draw_disk(togl, o);
      break;
    case MOPS_OTCONE: 
      if(view->drawwire)
	mops_draw_cone_c(togl, o);
      if(view->drawsurf)
	mops_draw_cone(togl, o);
      break;
    case MOPS_OTCYLINDER: 
      if(view->drawwire)
	mops_draw_cylinder_c(togl, o);
      if(view->drawsurf)
	mops_draw_cylinder(togl, o);
      break;
    case MOPS_OTTORUS: 
      if(view->drawwire)
	mops_draw_torus_c(togl, o);
      if(view->drawsurf)
	mops_draw_torus(togl, o);
      break;
    case MOPS_OTPARABOLOID: 
      if(view->drawwire)
	mops_draw_paraboloid_c(togl, o);
      if(view->drawsurf)
	mops_draw_paraboloid(togl, o);
      break;
    case MOPS_OTHYPERBOLOID: 
      if(view->drawwire)
	mops_draw_hyperboloid_c(togl, o);
      if(view->drawsurf)
	mops_draw_hyperboloid(togl, o);
      break;
    case MOPS_OTBOX: 
      if(view->drawwire)
	mops_draw_box_c(togl, o);
      if(view->drawsurf)
	mops_draw_box(togl, o);
      break;
    case MOPS_OTINSTANCE: 
      mops_draw_instance(togl, o);
      break;

    }

  if(o->type != MOPS_OTCUSTOM)
    glPopMatrix();

  if((o->editor) && (o->down) &&
     (o->type != MOPS_OTLEVEL) && (o->type != MOPS_OTNURBPATCH))
    {  
      glPushMatrix();
      mops_render_attributes(o);
      if(reset_color)
	glColor3f MOPS_CLSEL;
      down = o->down;
      while(down)
	{
	  mops_draw_object(togl, down);
	  down = down->next;
	}

      glPopMatrix();
    }


  if(reset_color)
    glColor3f MOPS_CLUSEL;

 return;
} /* mops_draw_object */


/********************************************/
/*mops_draw_instance:
 */
/********************************************/
void
mops_draw_instance(struct Togl *togl, mops_object *i)
{
 mops_view_info *view = Togl_GetClientData(togl);
 mops_object *down = NULL, *o = NULL;
 int reset_color = 0;
 GLfloat color[4] = {0.8, 0.75, 0.65, 1.0};

  o = (mops_object *)i->object;

  /* custom objects decide for themself wether to use the
     standard attributes */
  if(o->type != MOPS_OTCUSTOM)
    {
      if(!o->editor)
	return;

      if(o->colr != -1 && o->colg != -1 && o->colb != -1)
	{
	  color[0] = o->colr/255.0;
	  color[1] = o->colg/255.0;
	  color[2] = o->colb/255.0;
	  color[3] = o->cola/255.0; 
	  glMaterialfv (GL_FRONT_AND_BACK, GL_AMBIENT_AND_DIFFUSE, color);
	}

    }

  switch(o->type)
    {
    case MOPS_OTCUSTOM:
      if(view->drawwire)
	mops_draw_custom_w(togl, i);
      if(view->drawsurf)
	mops_draw_custom_s(togl, i);
      break;

    case MOPS_OTLEVEL:
      down = o->down;
      while(down)
	{
	  mops_draw_object(togl, down);
	  down = down->next;
	}
      break;

    case MOPS_OTNURBPATCH: 
      if(view->drawwire)
	mops_draw_nurbpatch_c(togl, o);
      if(view->drawsurf)
	mops_draw_nurbpatch(togl, o);
      break;
    case MOPS_OTNURBCURVE: 
      if(view->drawwire)
	mops_draw_nurbcurve_c(togl, o);
      if(view->drawsurf)
	mops_draw_nurbcurve(togl, o);
      break;
    case MOPS_OTLIGHT: 
      mops_draw_light(togl, o);
      if(o->selected)
	mops_draw_light_h(togl,o);
      break;
    case MOPS_OTSPHERE: 
      if(view->drawwire)
	mops_draw_sphere_c(togl, o);
      if(view->drawsurf)
	mops_draw_sphere(togl, o);
      break;
    case MOPS_OTDISK: 
      if(view->drawwire)
	mops_draw_disk_c(togl, o);
      if(view->drawsurf)
	mops_draw_disk(togl, o);
      break;
    case MOPS_OTCONE: 
      if(view->drawwire)
	mops_draw_cone_c(togl, o);
      if(view->drawsurf)
	mops_draw_cone(togl, o);
      break;
    case MOPS_OTCYLINDER: 
      if(view->drawwire)
	mops_draw_cylinder_c(togl, o);
      if(view->drawsurf)
	mops_draw_cylinder(togl, o);
      break;
    case MOPS_OTTORUS: 
      if(view->drawwire)
	mops_draw_torus_c(togl, o);
      if(view->drawsurf)
	mops_draw_torus(togl, o);
      break;
    case MOPS_OTPARABOLOID: 
      if(view->drawwire)
	mops_draw_paraboloid_c(togl, o);
      if(view->drawsurf)
	mops_draw_paraboloid(togl, o);
      break;
    case MOPS_OTHYPERBOLOID: 
      if(view->drawwire)
	mops_draw_hyperboloid_c(togl, o);
      if(view->drawsurf)
	mops_draw_hyperboloid(togl, o);
      break;
    case MOPS_OTBOX: 
      if(view->drawwire)
	mops_draw_box_c(togl, o);
      if(view->drawsurf)
	mops_draw_box(togl, o);
      break;
    }


  /*
    if((o->down) && (o->type != MOPS_OTLEVEL) && (o->type != MOPS_OTNURBPATCH))
    {  
      glPushMatrix();

      if(reset_color)
	glColor3f MOPS_CLSEL;
      down = o->down;
      while(down)
	{
	  mops_draw_object(togl, down);
	  down = down->next;
	}

      glPopMatrix();
    }

  */

  if(reset_color)
    glColor3f MOPS_CLUSEL;


 return;
} /* mops_draw_instance */

/********************************************/
/*mops_draw_objecth:
  draw just handles */
/********************************************/
void
mops_draw_objecth(struct Togl *togl, mops_object *o)
{
 mops_list_object *level = mops_current_level;
 mops_point_object *point = NULL;

  /* are we inside a NURBPatch? */
  if(level)
    level = level->next;

  if(level)
  {
    if(level->object)
      {
	if(level->object->type == MOPS_OTNURBPATCH)
	  {
	    return;
	  }
	else
	  {
	    if(level->object->type == MOPS_OTLEVEL)
	      {
		if(((mops_level_object *)(level->object->object))->type ==
		   MOPS_LTTRIMLOOP)
		  return;
	      }
	  }
      }
  }

  /* custom objects decide for themself wether to use the
     standard attributes */
  if(o->type != MOPS_OTCUSTOM)
    {
      if(!o->editor)
	return;
      glPushMatrix();
      mops_render_attributes(o);
    }

  /* only first-level selection objects can be edited, thus
     levels are not followed */
  switch(o->type)
    {
    case MOPS_OTCUSTOM:
	mops_draw_custom_h(togl, o);
      break;

    case MOPS_OTNURBPATCH:
	mops_draw_nurbpatch_h(togl, o);
      break;

    case MOPS_OTNURBCURVE: 
	mops_draw_nurbcurve_h(togl, o);
      break;
    case MOPS_OTLIGHT: 
	mops_draw_light_h(togl, o);
      break;
    }

  /* now draw selected points (if there are any) */
  if(o->selp)
    {
      glDisable(GL_DEPTH_TEST);

      glColor3f MOPS_CLSELP;
      point = o->selp;
      glBegin(GL_POINTS);
      while(point)
	{
	  glVertex3d(point->point[0],point->point[1],point->point[2]);
	  
	  point = point->next;
	}

      glEnd();
      glColor3f MOPS_CLSEL;
      glEnable(GL_DEPTH_TEST);
    }

  if(o->type != MOPS_OTCUSTOM)
    glPopMatrix();

 return;
} /* mops_draw_objecth */


/********************************************/
/*mops_draw_scene:
  traverse through scene and draw everything
  that should be drawn                      */
/********************************************/
void
mops_draw_scene(struct Togl *togl)
{
 mops_object *o = mops_root->next;
 mops_view_info *view = Togl_GetClientData(togl);
 mops_list_object *sel = mops_current_selection;
 int width = Togl_Width (togl);
 int height = Togl_Height (togl);
 GLfloat m[4][4];
 
  glDisable (GL_LIGHTING);
  
  if(view->drawgrid)
    mops_draw_grid(togl);

  if(view->type == MOPS_VTTRIM)
    {
      mops_draw_trimview(togl);
    }
  else
    {
      /* if(view->drawcross) {*/
      glBegin (GL_LINES);
      glColor3f (1.0, 0.0, 0.0);
      glVertex3f (0.0, 0.0, 0.0);
      glVertex3f (1.0, 0.0, 0.0);
      glColor3f (0.0, 1.0, 0.0);
      glVertex3f (0.0, 0.0, 0.0);
      glVertex3f (0.0, 1.0, 0.0);
      glColor3f (0.0, 0.0, 1.0);
      glVertex3f (0.0, 0.0, 0.0);
      glVertex3f (0.0, 0.0, 1.0);
      glEnd ();
      /*}*/

      if(view->lighting)
	glEnable(GL_LIGHTING);

      glColor3f MOPS_CLUSEL;

      if(!view->drawsel) /* draw just the selection ? */
	{ /* No */

	  while(o)
	    {
	      mops_draw_object(togl,o);
	      o = o->next;
	    }
	  if(view->type != MOPS_VTPERSP && view->drawhandles)
	    {
	      glClear(GL_DEPTH_BUFFER_BIT);
	      glDisable(GL_LIGHTING);
	      glColor3f MOPS_CLSEL;
	      glPushMatrix();
	      if(mops_current_level->next)
		mops_get_all_transform(mops_current_level->next);
	      
	      while(sel)
		{
		  mops_draw_objecth(togl,sel->object);
		  sel = sel->next;
		}

	      glPopMatrix();

	      glColor3f MOPS_CLUSEL;
	    }

	}
      else /* Yes */
	{
	  glPushMatrix();
	  if(mops_current_level->next)
	    mops_get_all_transform(mops_current_level->next);
	  
	  while(sel)
	    {
	      o = sel->object;
	      mops_draw_object(togl,o);
	      sel = sel->next;
	    }
      
	  if(view->type != MOPS_VTPERSP && view->drawhandles)
	    {
	      glClear(GL_DEPTH_BUFFER_BIT);
	      glDisable(GL_LIGHTING);
	      glColor3f MOPS_CLSEL;
	      sel = mops_current_selection;
	      while(sel)
		{
		  mops_draw_objecth(togl, sel->object);
		  sel = sel->next;
		}
	      glColor3f MOPS_CLUSEL;
	    }
	  
	  glPopMatrix();

	}
    } /* if type == trim */

  sel = mops_current_selection;
  if(sel)
    {
      if(!sel->next)
	{
	  o = sel->object;
	  glPushMatrix();
	  if(mops_current_level->next)
	    mops_get_all_transform(mops_current_level->next);

	  glTranslated(o->movx, o->movy, o->movz);

	  mops_build_rotmatrix(m, o->quat);
	  glMultMatrixf(&m[0][0]);

	  glBegin (GL_LINES);
	  glColor3f (1.0, 0.0, 0.0);
	  glVertex3f (0.0, 0.0, 0.0);
	  glVertex3f (0.5, 0.0, 0.0);
	  glColor3f (0.0, 1.0, 0.0);
	  glVertex3f (0.0, 0.0, 0.0);
	  glVertex3f (0.0, 0.5, 0.0);
	  glColor3f (0.0, 0.0, 1.0);
	  glVertex3f (0.0, 0.0, 0.0);
	  glVertex3f (0.0, 0.0, 0.5);
	  glEnd ();
	  glPopMatrix();

	}
    }
 

  if(view->draw_clip)
    {

      glMatrixMode(GL_PROJECTION);
      glPushMatrix();
      glLoadIdentity();
      glOrtho(0, width, 0, height, -100.0, 100.0);
      glMatrixMode(GL_MODELVIEW);
      glPushMatrix();
      glLoadIdentity();
      glBegin(GL_LINES);

      glVertex3d(view->clip_xmin, height-view->clip_ymin, 0.0);
      glVertex3d(view->clip_xmax, height-view->clip_ymin, 0.0);

      glVertex3d(view->clip_xmax, height-view->clip_ymin, 0.0);
      glVertex3d(view->clip_xmax, height-view->clip_ymax, 0.0);

      glVertex3d(view->clip_xmax, height-view->clip_ymax, 0.0);
      glVertex3d(view->clip_xmin, height-view->clip_ymax, 0.0);

      glVertex3d(view->clip_xmin, height-view->clip_ymax, 0.0);
      glVertex3d(view->clip_xmin, height-view->clip_ymin, 0.0);

      glEnd();
      glPopMatrix();
      glMatrixMode(GL_PROJECTION);
      glPopMatrix();
    }


 return;
} /* mops_draw_scene */


/********************************************/
/*create_cb:
 Togl Callback, called on creation of a 
 view                                       */
/********************************************/
void
create_cb (struct Togl *togl)
{
  mops_object *o = mops_root, *newo = NULL;
  mops_view_info *view_info = NULL;
  GLfloat color[4] = {0.8, 0.75, 0.65, 1.0};

  if((view_info = calloc(1,sizeof(mops_view_info))) == NULL)
    return;
  /* do _not_ forget to free this on destruction of a view! */

  /* setup some defaults */
  view_info->bgr = 0.5;
  view_info->bgg = 0.5;
  view_info->bgb = 0.5;
  view_info->bga = 0.0;

  view_info->togl = togl;
  view_info->drawhandles = MOPS_TRUE;

  /* add view info to togl-object */
  Togl_SetClientData(togl, view_info);

  /* link the same view_info object as sub object of the root */
  if(!(newo = calloc(1,sizeof(mops_object))))
    return;
  newo->type = MOPS_OTVIEW;
  newo->object = view_info;
  if(o->down)
    {
      o = o->down;
      while(o->next) o = o->next;
      o->next = newo;
    }
  else
    { /* first view */
      o->down = newo;
    }

  /* vital gl-defaults */
  glEnable(GL_DEPTH_TEST);
  glEnable(GL_AUTO_NORMAL);
  glEnable(GL_NORMALIZE);
  glDisable(GL_DITHER);
  glLightModelf (GL_LIGHT_MODEL_TWO_SIDE, 1.0);
  glLightModelf (GL_LIGHT_MODEL_LOCAL_VIEWER, GL_TRUE);
  glEnable (GL_LIGHT0);
  glMaterialfv (GL_FRONT_AND_BACK, GL_AMBIENT_AND_DIFFUSE, color);
 
  return;
} /* create_cb */


/********************************************/
/*destroy_cb:
 Togl Callback, called on destruction of a 
 view                                       */
/********************************************/
void
destroy_cb (struct Togl *togl)
{
 Tcl_Interp *interp = Togl_Interp(togl);
 mops_object *root = mops_root, *o, *o2;
 mops_view_info *view_info = Togl_GetClientData(togl);
 char cmd[] = "goTop";
  
  if((mops_view_info *)root->down->object == view_info)
    {
      o2 = root->down;
      root->down = root->down->next;
      free(o2);
    }
  else
    {
      o = root->down;
      while(o)
	{
	  if(o->next)
	    {
	      if((mops_view_info *)o->next->object == view_info)
		{
		  o2 = o->next;
		  o->next = o->next->next;
		  free(o2);
		  o = NULL;
		}
	      else
		o = o->next;
	    } /* if */
	} /* while */
    } /* if */

  if(view_info)
    free(view_info);

  /* if this was the last view, we go to the top */
  if(!root->down)
    {
      Tcl_Eval(interp, cmd);
    }

  return;
} /* destroy_cb */


/********************************************/
/*reshape_cb:
 Togl Callback, called when view is 
 exposed or reshaped                        */
/********************************************/
void
reshape_cb (struct Togl *togl)
{
  mops_view_info *view = Togl_GetClientData(togl);
  double aspect = 0.0, width = (double) Togl_Width(togl),
    height = (double) Togl_Height(togl);

  mops_setup_projection(togl);

  aspect = width/height;

  view->conv_x = (aspect * 2.0 / width) * view->zoom;
  view->conv_y = 2.0/height * view->zoom;

  /*  glMatrixMode (GL_MODELVIEW);*/

  return;
} /* reshape_cb */


/********************************************/
/*display_cb:
  redraw view *togl                         */
/********************************************/
void
display_cb (struct Togl *togl)
{
  mops_view_info *view_info = Togl_GetClientData(togl);

  if(!view_info->redraw)
    return;
  
  glClearColor (view_info->bgr, view_info->bgg,
		view_info->bgb, view_info->bga);
  glClear (GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

  /* draw scene */
  mops_draw_scene(togl);
 
  glFlush();    

  Togl_SwapBuffers (togl);

  return;
} /* display_cb */


/********************************************/
/*rotate_viewer:
  calculate new "from" and "to" values from
  rotx, roty and rotz                       */
/********************************************/
void
rotate_viewer(mops_view_info *view, double rotx, double roty, double rotz)
{
  GLdouble m[16] = {0};
  double x,y,z;

  glMatrixMode(GL_MODELVIEW);
  glPushMatrix();
  glLoadIdentity();

  if(rotx != 0.0)
    {
      glRotated(view->roty, 0.0, 1.0, 0.0);
      glRotated(rotx, 1.0, 0.0, 0.0);
      glRotated(-view->roty, 0.0, 1.0, 0.0);
    }
  
  if(rotz != 0.0)
    {
      glRotated(view->roty, 0.0, 1.0, 0.0);
      glRotated(rotz, 0.0, 0.0, 1.0);
      glRotated(-view->roty, 0.0, 1.0, 0.0);
    }
  
  if(roty != 0.0)
    glRotated(roty, 0.0, 1.0, 0.0);


  glGetDoublev(GL_MODELVIEW_MATRIX, m);
  glPopMatrix();


  x = view->to_x-view->from_x;
  y = view->to_y-view->from_y;
  z = view->to_z-view->from_z;

  view->from_x = view->to_x+(m[0]*x + m[4]*y + m[8]*z + m[12]);
  view->from_y = view->to_y+(m[1]*x + m[5]*y + m[9]*z + m[13]);
  view->from_z = view->to_z+(m[2]*x + m[6]*y + m[10]*z + m[14]);

  x = view->up_x;
  y = view->up_y;
  z = view->up_z;

  view->up_x = (m[0]*x + m[4]*y + m[8]*z + m[12]);
  view->up_y = (m[1]*x + m[5]*y + m[9]*z + m[13]);
  view->up_z = (m[2]*x + m[6]*y + m[10]*z + m[14]);

 return;
} /* rotate_viewer */


/********************************************/
/*setViewConfig_cb
  configure view *togl from Tcl             */
/********************************************/
int
setViewConfig_cb(struct Togl *togl, int argc, char *argv[])
{
 Tcl_Interp *interp = Togl_Interp (togl);
 mops_view_info *view = Togl_GetClientData(togl);
 int argi = 0;
 double argd = 0.0, rotx = 0.0, roty = 0.0, rotz = 0.0;
 double temp[3] = {0};
 int i = 2;
 char *n1 = "view";
 Tcl_Obj *to = NULL, *toa = NULL, *ton = NULL;
  
  toa = Tcl_NewStringObj(n1,-1);
  ton = Tcl_NewStringObj("From_X",-1);

  while(i+1 < argc)
  {
  rotx = 0.0; roty = 0.0; rotz = 0.0;
  Tcl_GetDouble(interp,argv[i+1],&argd);

  if(!strcmp(argv[i],"-fromx"))
    {
      view->from_x = argd;
      to = Tcl_NewDoubleObj(view->from_x);
      Tcl_SetStringObj(ton,"From_X",-1);
      Tcl_ObjSetVar2(interp,toa,ton,to,TCL_LEAVE_ERR_MSG | TCL_GLOBAL_ONLY);
    }
  else 
  if(!strcmp(argv[i],"-fromy"))
    {
      view->from_y = argd;
      to = Tcl_NewDoubleObj(view->from_y);
      Tcl_SetStringObj(ton,"From_Y",-1);
      Tcl_ObjSetVar2(interp,toa,ton,to,TCL_LEAVE_ERR_MSG | TCL_GLOBAL_ONLY);
    }
  else 
  if(!strcmp(argv[i],"-fromz"))
    {
      view->from_z = argd;
      to = Tcl_NewDoubleObj(view->from_z);
      Tcl_SetStringObj(ton,"From_Z",-1);
      Tcl_ObjSetVar2(interp,toa,ton,to,TCL_LEAVE_ERR_MSG | TCL_GLOBAL_ONLY);
    }
  else
  if(!strcmp(argv[i],"-tox"))
    {
      view->to_x = argd;
      to = Tcl_NewDoubleObj(view->to_x);
      Tcl_SetStringObj(ton,"To_X",-1);
      Tcl_ObjSetVar2(interp,toa,ton,to,TCL_LEAVE_ERR_MSG | TCL_GLOBAL_ONLY);
    }
  else 
  if(!strcmp(argv[i],"-toy"))
    {
      view->to_y = argd;
      to = Tcl_NewDoubleObj(view->to_y);
      Tcl_SetStringObj(ton,"To_Y",-1);
      Tcl_ObjSetVar2(interp,toa,ton,to,TCL_LEAVE_ERR_MSG | TCL_GLOBAL_ONLY);
    }
  else 
  if(!strcmp(argv[i],"-toz"))
    {
      view->to_z = argd;
      to = Tcl_NewDoubleObj(view->to_z);
      Tcl_SetStringObj(ton,"To_Z",-1);
      Tcl_ObjSetVar2(interp,toa,ton,to,TCL_LEAVE_ERR_MSG | TCL_GLOBAL_ONLY);
    }
  else
  if(!strcmp(argv[i],"-dfromx"))
    {
      view->from_x += argd;
      to = Tcl_NewDoubleObj(view->from_x);
      Tcl_SetStringObj(ton,"From_X",-1);
      Tcl_ObjSetVar2(interp,toa,ton,to,TCL_LEAVE_ERR_MSG | TCL_GLOBAL_ONLY);
    }
  else 
  if(!strcmp(argv[i],"-dfromy"))
    {
      view->from_y += argd;
      to = Tcl_NewDoubleObj(view->from_y);
      Tcl_SetStringObj(ton,"From_Y",-1);
      Tcl_ObjSetVar2(interp,toa,ton,to,TCL_LEAVE_ERR_MSG | TCL_GLOBAL_ONLY);
    }
  else 
  if(!strcmp(argv[i],"-dfromz"))
    {
      view->from_z += argd;
      to = Tcl_NewDoubleObj(view->from_z);
      Tcl_SetStringObj(ton,"From_Z",-1);
      Tcl_ObjSetVar2(interp,toa,ton,to,TCL_LEAVE_ERR_MSG | TCL_GLOBAL_ONLY);
    }
  else
  if(!strcmp(argv[i],"-dtox"))
    {
      view->to_x += argd;
      to = Tcl_NewDoubleObj(view->to_x);
      Tcl_SetStringObj(ton,"To_X",-1);
      Tcl_ObjSetVar2(interp,toa,ton,to,TCL_LEAVE_ERR_MSG | TCL_GLOBAL_ONLY);
    }
  else 
  if(!strcmp(argv[i],"-dtoy"))
    {
      view->to_y += argd;
      to = Tcl_NewDoubleObj(view->to_y);
      Tcl_SetStringObj(ton,"To_Y",-1);
      Tcl_ObjSetVar2(interp,toa,ton,to,TCL_LEAVE_ERR_MSG | TCL_GLOBAL_ONLY);
    }
  else 
  if(!strcmp(argv[i],"-dtoz"))
    {
      view->to_z += argd;
      to = Tcl_NewDoubleObj(view->to_z);
      Tcl_SetStringObj(ton,"To_Z",-1);
      Tcl_ObjSetVar2(interp,toa,ton,to,TCL_LEAVE_ERR_MSG | TCL_GLOBAL_ONLY);
    }
  else
  if(!strcmp(argv[i],"-upx"))
    {
      view->up_x = argd;
    }
  else 
  if(!strcmp(argv[i],"-upy"))
    {
      view->up_y = argd;
    }
  else 
  if(!strcmp(argv[i],"-upz"))
    {
      view->up_z = argd;      
    }
  else  
  if(!strcmp(argv[i],"-updroty"))
    {
      /* reset rotx,rotz recalc roty */
      view->rotx = 0.0;
      view->roty = 0.0;
      view->rotz = 0.0;

      temp[0] = view->to_x-view->from_x;
      temp[1] = view->to_z-view->from_z;
      view->roty = 180+MOPS_R2D(atan2(temp[0],temp[1]));
    }
  else  
  if(!strcmp(argv[i],"-drotx"))
    {
      roty = argd;
      view->roty += argd;

      rotate_viewer(view, 0.0, roty, 0.0);

      to = Tcl_NewDoubleObj(view->from_x);
      Tcl_SetStringObj(ton,"From_X",-1);
      Tcl_ObjSetVar2(interp,toa,ton,to,TCL_LEAVE_ERR_MSG | TCL_GLOBAL_ONLY);

      to = Tcl_NewDoubleObj(view->from_y);
      Tcl_SetStringObj(ton,"From_Y",-1);
      Tcl_ObjSetVar2(interp,toa,ton,to,TCL_LEAVE_ERR_MSG | TCL_GLOBAL_ONLY);

      to = Tcl_NewDoubleObj(view->from_z);
      Tcl_SetStringObj(ton,"From_Z",-1);
      Tcl_ObjSetVar2(interp,toa,ton,to,TCL_LEAVE_ERR_MSG | TCL_GLOBAL_ONLY);
    }
  else
  if(!strcmp(argv[i],"-droty"))
    {
      rotz=0.0;
      rotx=0.0;
      if(view->type == MOPS_VTSIDE)
	{
	  rotz = -argd;
	  view->rotz -= argd;
	}
      else
	{
	  rotx = argd;
	  view->rotx += argd;
	}

      rotate_viewer(view, rotx, 0.0, rotz);

      to = Tcl_NewDoubleObj(view->from_x);
      Tcl_SetStringObj(ton,"From_X",-1);
      Tcl_ObjSetVar2(interp,toa,ton,to,TCL_LEAVE_ERR_MSG | TCL_GLOBAL_ONLY);

      to = Tcl_NewDoubleObj(view->from_y);
      Tcl_SetStringObj(ton,"From_Y",-1);
      Tcl_ObjSetVar2(interp,toa,ton,to,TCL_LEAVE_ERR_MSG | TCL_GLOBAL_ONLY);

      to = Tcl_NewDoubleObj(view->from_z);
      Tcl_SetStringObj(ton,"From_Z",-1);
      Tcl_ObjSetVar2(interp,toa,ton,to,TCL_LEAVE_ERR_MSG | TCL_GLOBAL_ONLY);
    }
  else
  if(!strcmp(argv[i],"-droll"))
    {
      view->roll = view->roll + argd;
      if(view->roll > 180.0)
	view->roll = view->roll - 180.0;
      else
	if (view->roll < -180.0)
	  view->roll = view->roll + 180.0;

      to = Tcl_NewDoubleObj(view->roll);
      Tcl_SetStringObj(ton,"Roll",-1);
      Tcl_ObjSetVar2(interp,toa,ton,to,TCL_LEAVE_ERR_MSG | TCL_GLOBAL_ONLY);
    }
  else
  if(!strcmp(argv[i],"-roll"))
    {
      view->roll = argd;
      to = Tcl_NewDoubleObj(view->roll);
      Tcl_SetStringObj(ton,"Roll",-1);
      Tcl_ObjSetVar2(interp,toa,ton,to,TCL_LEAVE_ERR_MSG | TCL_GLOBAL_ONLY);
    }
  else
  if(!strcmp(argv[i],"-zoomd"))
    {
      view->zoom *= argd;

      to = Tcl_NewDoubleObj(view->zoom);
      Tcl_SetStringObj(ton,"Zoom",-1);
      Tcl_ObjSetVar2(interp,toa,ton,to,TCL_LEAVE_ERR_MSG | TCL_GLOBAL_ONLY);
    }
  else
  if(!strcmp(argv[i],"-zoom"))
    {
      view->zoom = argd;

      to = Tcl_NewDoubleObj(view->zoom);
      Tcl_SetStringObj(ton,"Zoom",-1);
      Tcl_ObjSetVar2(interp,toa,ton,to,TCL_LEAVE_ERR_MSG | TCL_GLOBAL_ONLY);
    }
  else
  if(!strcmp(argv[i],"-grid"))
    {
      if(argd >= 0.0)
	view->grid = argd;
      else
	view->grid = 0.0;

      to = Tcl_NewDoubleObj(view->grid);
      Tcl_SetStringObj(ton,"Grid",-1);
      Tcl_ObjSetVar2(interp,toa,ton,to,TCL_LEAVE_ERR_MSG | TCL_GLOBAL_ONLY);
    }
  else
  if(!strcmp(argv[i],"-draww"))
    {
      Tcl_GetInt(interp, argv[i+1], &argi);
      view->drawwire = argi;

      to = Tcl_NewIntObj(view->drawwire);
      Tcl_SetStringObj(ton,"DrawWire",-1);
      Tcl_ObjSetVar2(interp,toa,ton,to,TCL_LEAVE_ERR_MSG | TCL_GLOBAL_ONLY);
    }
  else
  if(!strcmp(argv[i],"-draws"))
    {
      Tcl_GetInt(interp, argv[i+1], &argi);
      view->drawsurf = argi;

      to = Tcl_NewIntObj(view->drawsurf);
      Tcl_SetStringObj(ton,"DrawSurf",-1);
      Tcl_ObjSetVar2(interp,toa,ton,to,TCL_LEAVE_ERR_MSG | TCL_GLOBAL_ONLY);
    }
  else
  if(!strcmp(argv[i],"-drawg"))
    {
      Tcl_GetInt(interp, argv[i+1], &argi);
      view->drawgrid = argi;

      to = Tcl_NewIntObj(view->drawgrid);
      Tcl_SetStringObj(ton,"DrawGrid",-1);
      Tcl_ObjSetVar2(interp,toa,ton,to,TCL_LEAVE_ERR_MSG | TCL_GLOBAL_ONLY);
    }
  else
  if(!strcmp(argv[i],"-dsel"))
    {
      Tcl_GetInt(interp, argv[i+1], &argi);
      view->drawsel = argi;

      to = Tcl_NewIntObj(view->drawsel);
      Tcl_SetStringObj(ton,"DrawSel",-1);
      Tcl_ObjSetVar2(interp,toa,ton,to,TCL_LEAVE_ERR_MSG | TCL_GLOBAL_ONLY);
    }
  else
  if(!strcmp(argv[i],"-draw"))
    {
      Tcl_GetInt(interp, argv[i+1], &argi);
      view->redraw = argi;

      to = Tcl_NewIntObj(view->redraw);
      Tcl_SetStringObj(ton,"Redraw",-1);
      Tcl_ObjSetVar2(interp,toa,ton,to,TCL_LEAVE_ERR_MSG | TCL_GLOBAL_ONLY);
    }
  else
  if(!strcmp(argv[i],"-light"))
    {
      Tcl_GetInt(interp, argv[i+1], &argi);
      view->lighting = argi;

      to = Tcl_NewIntObj(view->lighting);
      Tcl_SetStringObj(ton,"Lighting",-1);
      Tcl_ObjSetVar2(interp,toa,ton,to,TCL_LEAVE_ERR_MSG | TCL_GLOBAL_ONLY);
    }
  else
  if(!strcmp(argv[i],"-type"))
    {
      Tcl_GetInt(interp, argv[i+1], &argi);
      view->type = argi;

      view->rotx = 0.0;
      view->roty = 0.0;
      view->rotz = 0.0;

      to = Tcl_NewIntObj(view->type);
      Tcl_SetStringObj(ton,"Type",-1);
      Tcl_ObjSetVar2(interp,toa,ton,to,TCL_LEAVE_ERR_MSG | TCL_GLOBAL_ONLY);
    }
  else
  if(!strcmp(argv[i],"-clip"))
    {
      Tcl_GetDouble(interp, argv[i+1], &view->clip_xmin);
      Tcl_GetDouble(interp, argv[i+2], &view->clip_ymin);
      Tcl_GetDouble(interp, argv[i+3], &view->clip_xmax);
      Tcl_GetDouble(interp, argv[i+4], &view->clip_ymax);

      Tcl_GetInt(interp, argv[i+5], &view->draw_clip);

    }

  i+=2;
  }
  
  Tcl_IncrRefCount(toa); Tcl_DecrRefCount(toa);
  Tcl_IncrRefCount(ton); Tcl_DecrRefCount(ton);

  reshape_cb (togl);
  display_cb (togl);

  return TCL_OK;
}


/********************************************/
/* copy all information about the view *togl
   to the global view() array (Tcl)         */
/********************************************/
int
getViewConfig_cb(struct Togl *togl, int argc, char *argv[])
{
  Tcl_Interp *interp = Togl_Interp (togl);
  mops_view_info *view = Togl_GetClientData(togl);
  int width = Togl_Width (togl);
  int height = Togl_Height (togl);
  char *n1 = "view";
  Tcl_Obj *to = NULL, *toa = NULL, *ton = NULL;

  toa = Tcl_NewStringObj(n1,-1);

  to = Tcl_NewIntObj(view->type);

  ton = Tcl_NewStringObj("Type",-1);
  Tcl_ObjSetVar2(interp,toa,ton,to,TCL_LEAVE_ERR_MSG | TCL_GLOBAL_ONLY);
  Tcl_SetStringObj(ton,"OrigType",-1);
  Tcl_ObjSetVar2(interp,toa,ton,to,TCL_LEAVE_ERR_MSG | TCL_GLOBAL_ONLY);

  Tcl_SetStringObj(ton,"TypeStr",-1);
  switch(view->type)
    {
    case MOPS_VTFRONT:
      to = Tcl_NewStringObj("Front",-1);
      Tcl_ObjSetVar2(interp,toa,ton,to,TCL_LEAVE_ERR_MSG | TCL_GLOBAL_ONLY);
      break;
    case MOPS_VTSIDE:
      to = Tcl_NewStringObj("Side",-1);
      Tcl_ObjSetVar2(interp,toa,ton,to,TCL_LEAVE_ERR_MSG | TCL_GLOBAL_ONLY);
      break;
    case MOPS_VTTOP:
      to = Tcl_NewStringObj("Top",-1);
      Tcl_ObjSetVar2(interp,toa,ton,to,TCL_LEAVE_ERR_MSG | TCL_GLOBAL_ONLY);
      break;
    case MOPS_VTPERSP:
      to = Tcl_NewStringObj("Persp",-1);
      Tcl_ObjSetVar2(interp,toa,ton,to,TCL_LEAVE_ERR_MSG | TCL_GLOBAL_ONLY);
      break;
    case MOPS_VTTRIM:
      to = Tcl_NewStringObj("Trim",-1);
      Tcl_ObjSetVar2(interp,toa,ton,to,TCL_LEAVE_ERR_MSG | TCL_GLOBAL_ONLY);
      break;
    }

  Tcl_SetStringObj(ton,"DrawWire",-1);
  to = Tcl_NewIntObj(view->drawwire);
  Tcl_ObjSetVar2(interp,toa,ton,to,TCL_LEAVE_ERR_MSG | TCL_GLOBAL_ONLY);

  Tcl_SetStringObj(ton,"DrawSurf",-1);
  to = Tcl_NewIntObj(view->drawsurf);
  Tcl_ObjSetVar2(interp,toa,ton,to,TCL_LEAVE_ERR_MSG | TCL_GLOBAL_ONLY);

  Tcl_SetStringObj(ton,"DrawSel",-1);
  to = Tcl_NewIntObj(view->drawsel);
  Tcl_ObjSetVar2(interp,toa,ton,to,TCL_LEAVE_ERR_MSG | TCL_GLOBAL_ONLY);

  Tcl_SetStringObj(ton,"Redraw",-1);
  to = Tcl_NewIntObj(view->redraw);
  Tcl_ObjSetVar2(interp,toa,ton,to,TCL_LEAVE_ERR_MSG | TCL_GLOBAL_ONLY);

  Tcl_SetStringObj(ton,"Grid",-1);
  to = Tcl_NewDoubleObj(view->grid);
  Tcl_ObjSetVar2(interp,toa,ton,to,TCL_LEAVE_ERR_MSG | TCL_GLOBAL_ONLY);

  Tcl_SetStringObj(ton,"DrawGrid",-1);
  to = Tcl_NewIntObj(view->drawgrid);
  Tcl_ObjSetVar2(interp,toa,ton,to,TCL_LEAVE_ERR_MSG | TCL_GLOBAL_ONLY);

  Tcl_SetStringObj(ton,"Lighting",-1);
  to = Tcl_NewIntObj(view->lighting);
  Tcl_ObjSetVar2(interp,toa,ton,to,TCL_LEAVE_ERR_MSG | TCL_GLOBAL_ONLY);

  Tcl_SetStringObj(ton,"From_X",-1);
  to = Tcl_NewDoubleObj(view->from_x);
  Tcl_ObjSetVar2(interp,toa,ton,to,TCL_LEAVE_ERR_MSG | TCL_GLOBAL_ONLY);
  Tcl_SetStringObj(ton,"From_Y",-1);
  to = Tcl_NewDoubleObj(view->from_y);
  Tcl_ObjSetVar2(interp,toa,ton,to,TCL_LEAVE_ERR_MSG | TCL_GLOBAL_ONLY);
  Tcl_SetStringObj(ton,"From_Z",-1);
  to = Tcl_NewDoubleObj(view->from_z);
  Tcl_ObjSetVar2(interp,toa,ton,to,TCL_LEAVE_ERR_MSG | TCL_GLOBAL_ONLY);

  Tcl_SetStringObj(ton,"To_X",-1);
  to = Tcl_NewDoubleObj(view->to_x);
  Tcl_ObjSetVar2(interp,toa,ton,to,TCL_LEAVE_ERR_MSG | TCL_GLOBAL_ONLY);
  Tcl_SetStringObj(ton,"To_Y",-1);
  to = Tcl_NewDoubleObj(view->to_y);
  Tcl_ObjSetVar2(interp,toa,ton,to,TCL_LEAVE_ERR_MSG | TCL_GLOBAL_ONLY);
  Tcl_SetStringObj(ton,"To_Z",-1);
  to = Tcl_NewDoubleObj(view->to_z);
  Tcl_ObjSetVar2(interp,toa,ton,to,TCL_LEAVE_ERR_MSG | TCL_GLOBAL_ONLY);

  Tcl_SetStringObj(ton,"Roll",-1);
  to = Tcl_NewDoubleObj(view->roll);
  Tcl_ObjSetVar2(interp,toa,ton,to,TCL_LEAVE_ERR_MSG | TCL_GLOBAL_ONLY);

  Tcl_SetStringObj(ton,"Zoom",-1);
  to = Tcl_NewDoubleObj(view->zoom);
  Tcl_ObjSetVar2(interp,toa,ton,to,TCL_LEAVE_ERR_MSG | TCL_GLOBAL_ONLY);

  Tcl_SetStringObj(ton,"Width",-1);
  to = Tcl_NewIntObj(width);
  Tcl_ObjSetVar2(interp,toa,ton,to,TCL_LEAVE_ERR_MSG | TCL_GLOBAL_ONLY);

  Tcl_SetStringObj(ton,"Height",-1);
  to = Tcl_NewIntObj(height);
  Tcl_ObjSetVar2(interp,toa,ton,to,TCL_LEAVE_ERR_MSG | TCL_GLOBAL_ONLY);

  Tcl_IncrRefCount(toa); Tcl_DecrRefCount(toa);
  Tcl_IncrRefCount(ton); Tcl_DecrRefCount(ton);

  return TCL_OK;
}


/********************************************/
/* updatePosition_cb:
  update position fields of a view          */
/********************************************/
int
updatePosition_cb(struct Togl *togl, int argc, char *argv[])
{
 Tcl_Interp *interp = Togl_Interp (togl);
 mops_view_info *view = Togl_GetClientData(togl);
 char *fname="updatePosition";

  if(argc!=4)
    {
      mops_error(interp,MOPS_ARGS,fname,"posx posy");
      return TCL_OK;
    }
  
  Tcl_GetInt(interp,argv[2],&(view->pos_x));
  Tcl_GetInt(interp,argv[3],&(view->pos_y));

  return TCL_OK;
}


/********************************************/
/* calls object type specific get_point
   functions to let the object decide
   which point should be edited in single
   point editing mode or for which point
   the weight should be changed             */
/********************************************/
int
mops_get_point(mops_object *o, double objX, double objY, double objZ)
{
  int mops_status = MOPS_OK;

  switch(o->type)
    {
    case MOPS_OTNURBCURVE:
      mops_status = mops_get_nurbcurve_point(o,objX,objY,objZ);
      break;

    case MOPS_OTNURBPATCH:
      mops_status = mops_get_nurbpatch_point(o,objX,objY,objZ);
      break;

    case MOPS_OTCUSTOM:
      mops_status = mops_get_custom_point(o,objX,objY,objZ);
      break;
    case MOPS_OTLIGHT:
      mops_status = mops_get_light_point(o,objX,objY,objZ);
      break;

    }

 return mops_status;
}


/********************************************/
/*mops_window_to_object:
  transforms the window coordinates winX,
  winY, to object coordinates for the given
  object *o
  Assumes the standard transformations are
  in use, even for custom objects!
  If a custom object does not use these,
  it should not change the default
  transformation values as well!            */
/********************************************/
int
mops_window_to_object(struct Togl *togl, mops_object *o,
		      double winX, double winY,
		      double *objX, double *objY, double *objZ)
{
  int mops_status = MOPS_OK;
  /*  int width = Togl_Width (togl);*/
  int height = Togl_Height (togl);
  GLint viewport[4];
  GLdouble modelMatrix[16], projMatrix[16], winx, winy;
  GLfloat winz = 0.0;
  GLfloat m[4][4];

  if(!o)
    return mops_status;

  winx = winX;
  winy = height - winY;

  glReadPixels((GLint)winx,(GLint)winy,1,1,GL_DEPTH_COMPONENT,GL_FLOAT,&winz);

  /* fprintf(stderr,"WinZ:%g\n",winz); */

  glGetIntegerv(GL_VIEWPORT, viewport);

  mops_setup_projection(togl);

  glGetDoublev(GL_PROJECTION_MATRIX, projMatrix);


  glMatrixMode (GL_MODELVIEW);
  glPushMatrix();

  mops_get_all_transform(mops_current_level->next);

  glTranslated(o->movx, o->movy, o->movz);

  mops_build_rotmatrix(m, o->quat);
  glMultMatrixf(&m[0][0]);

  glScaled (o->scalx, o->scaly, o->scalz);

  glGetDoublev(GL_MODELVIEW_MATRIX, modelMatrix);

  gluUnProject(winx, winy, (GLdouble)winz, modelMatrix, projMatrix, viewport,
	       objX, objY, objZ);

  glPopMatrix ();

 return mops_status;
}

/********************************************/
/*mops_window_to_ncu:
  transforms the window coordinates winX,
  winY, to the corresponding parametric
  value u on the NURB curve o               */
/********************************************/
int
mops_window_to_ncu(struct Togl *togl, mops_object *o,
		   double *winX, double *winY, double *u)
{
 int mops_status = MOPS_OK;
 /*  int width = Togl_Width (togl);*/
 int height = Togl_Height (togl);
 GLint viewport[4];
 GLdouble modelMatrix[16], projMatrix[16], winx, winy;
 GLfloat winz = 0.0;
 GLfloat m[4][4];
/* XXXX let's hope nobody changes MOPS_CLSEL */
 GLfloat pixel1[3] = {0.9,0.9,0.9}, pixel[3] = {0};
 mops_nurbcurve_object *c = NULL;
 int dx[25] = {0,1,1,0,-1,-1,-1,0,1, 2,2,2,1,0,-1,-2,-2,-2,-2,-2,-1,0,1,2,2};
 int dy[25] = {0,0,-1,-1,-1,0,1,1,1, 0,-1,-2,-2,-2,-2,-2,-1,0,1,2,2,2,2,2,1};
 int found, i = 0, j = 0, k = 0, /*maxtry = 1000,*/ stride, samples = 10;
 int starti = 0, endi = 0;
 double point[3] = {0}/*, guess = 0.0, e1 = 0.05, e2 = 0.05*/;
 double distance = 0.0, min_distance = 0.0;
 double *cp = NULL, U[10/* XXXX samples! */] = {0}, startu, endu;

  if(!o)
    return MOPS_ERROR;

  if(!o->type == MOPS_OTNURBCURVE)
    return MOPS_ERROR;

  c = (mops_nurbcurve_object *)o->object;

  /*
  winx = winX;
  winy = height - winY;
  */

  /* first, find point on curve in window coordinates */
  found = MOPS_FALSE;
  while(!found)
    {
      winx = *winX+dx[i];
      winy = height-*winY-dy[i];
      i++;

      if(i > 24)
	{
	  return MOPS_ERROR;
	}

      glReadPixels((GLint)winx,(GLint)winy,1,1,
		   GL_RGB,GL_FLOAT,&pixel);
      if((pixel1[0] < pixel[0]) && (pixel1[1] < pixel[1]) &&
	 (pixel1[2] < pixel[2]))
	{
	  found = MOPS_TRUE;
	}
    }

  /* get winz */
  glReadPixels((GLint)winx,(GLint)winy,1,1,GL_DEPTH_COMPONENT,GL_FLOAT,&winz);

  /* get object coordinates */
  glGetIntegerv(GL_VIEWPORT, viewport);

  mops_setup_projection(togl);

  glGetDoublev(GL_PROJECTION_MATRIX, projMatrix);

  glMatrixMode (GL_MODELVIEW);
  glPushMatrix();
   mops_get_all_transform(mops_current_level->next);
   glTranslated(o->movx, o->movy, o->movz);
   mops_build_rotmatrix(m, o->quat);
   glMultMatrixf(&m[0][0]);
   glScaled (o->scalx, o->scaly, o->scalz);
   glGetDoublev(GL_MODELVIEW_MATRIX, modelMatrix);
   gluUnProject(winx, winy, (GLdouble)winz, modelMatrix, projMatrix, viewport,
	        &(point[0]), &(point[1]), &(point[2]));
  glPopMatrix ();

  /* get guess */
  stride = c->has_weights?4:3;
  if(!(cp = calloc(samples*stride, sizeof(double))))
    return TCL_OK;

  startu = c->knotv[c->order-1];
  endu = c->knotv[c->length];
  for(k = 0; k < 3; k++)
    {
      U[0] = startu;

      for(i = 1; i < samples; i++)
	{
	  U[i] = U[i-1]+(endu - startu)/samples;
	}

      for(i = 0; i < samples; i++)
       {
	if(c->has_weights)
	 mops_status = mops_nb_CurvePoint4D(c->length-1, c->order-1, c->knotv,
					  c->controlv, U[i], &(cp[i*stride]));
	else
	 mops_status = mops_nb_CurvePoint3D(c->length-1, c->order-1, c->knotv,
					  c->controlv, U[i], &(cp[i*stride]));
       }

      min_distance = DBL_MAX;
      j = 0;
      for(i = 0; i < samples; i++)
	{
	  distance = MOPS_VLEN((point[0] - cp[j]),
			       (point[1] - cp[j+1]),
			       (point[2] - cp[j+2]));

	  if(distance < min_distance)
	    {
	      *u = U[i];
	      starti = i-1;
	      endi = i+1;
	      min_distance = distance;
	    }

	  j += stride;
	}

      if(starti < 0)
	startu = U[0];
      else
	startu = U[starti];

      if(endi >= samples)
	endu = U[samples-1];
      else
	endu = U[endi];
      

    }

  /*
  if(guess <= c->knotv[c->order-1])
    guess = U[1];

  if(guess > c->knotv[c->length])
    guess = c->knotv[c->length];


  mops_nb_FinduFromPoint(c, point, guess, maxtry, e1, e2, u);


  if((*u <= c->knotv[c->order-1]) || (*u > c->knotv[c->length]))
    { free(cp); return MOPS_ERROR; }
  */

  *winX = winx;
  *winY = winy;

 free(cp);

 return mops_status;
}

/********************************************/
/*mops_window_to_world:
  transforms the window coordinates winX,
  winY, to world coordinates                */
/********************************************/
int
mops_window_to_world(struct Togl *togl,	double winX, double winY,
		      double *worldX, double *worldY, double *worldZ)
{
  int mops_status = MOPS_OK;
  /*  int width = Togl_Width (togl);*/
  int height = Togl_Height (togl);
  GLint viewport[4];
  GLdouble modelMatrix[16], projMatrix[16], winx, winy;
  GLdouble winz = 0.0;


  winx = winX;
  winy = height - winY;

  glGetIntegerv(GL_VIEWPORT, viewport);
  mops_setup_projection(togl);
  glGetDoublev(GL_PROJECTION_MATRIX, projMatrix);

  glMatrixMode (GL_MODELVIEW);
  glPushMatrix();
  glLoadIdentity();
  glGetDoublev(GL_MODELVIEW_MATRIX, modelMatrix);
  glPopMatrix ();

  gluUnProject(winx, winy, winz, modelMatrix, projMatrix, viewport,
	       worldX, worldY, worldZ);

 return mops_status;
}

/********************************************/
/* selectPoints_cb:
                                    */
/********************************************/
int
selectPoints_cb(struct Togl *togl, int argc, char *argv[])
{
 int mops_status = MOPS_OK;
 Tcl_Interp *interp = Togl_Interp (togl);
 /*  mops_view_info *view = Togl_GetClientData(togl);*/
 double height = Togl_Height(togl);
 int i, have_it, j,k;
 double winX = 0.0, winY = 0.0, winX2 = 0.0, winY2 = 0.0, dtemp = 0.0;
 double objX = 0.0, objY = 0.0, objZ = 0.0;
 /*  double pickepsilon = mops_prefs.pick_epsilon;*/
 mops_point_object *newp = NULL, *point = NULL, *last = NULL;
 int multiple = MOPS_FALSE;
 /* XXXX let's hope nobody changes MOPS_CLSEL */
 GLfloat pixel1[3] = {0.9,0.9,0.9}, pixel[3] = {0};
 mops_list_object *sel = mops_current_selection;

  if(mops_current_selection)
    {
      Tcl_GetDouble(interp, argv[2], &winX);
      Tcl_GetDouble(interp, argv[3], &winY);

      if(argc > 4)
	{
	  multiple = MOPS_TRUE;
	  Tcl_GetDouble(interp, argv[4], &winX2);
	  Tcl_GetDouble(interp, argv[5], &winY2);

	}
      else
	{
	  winX2 = winX;
	  winY2 = winY;
	}

      if(winX2 < winX)
	{
	  dtemp = winX2;
	  winX2 = winX;
	  winX = dtemp;
	}

      if(winY2 < winY)
	{
	  dtemp = winY2;
	  winY2 = winY;
	  winY = dtemp;
	}

      while(sel)
      {
      mops_point_edit_object = sel->object;

      for(j=0;j<=(winX2-winX);j++)
      for(k=0;k<=(winY2-winY);k++)
      {
	glReadPixels((GLint)winX+j,(GLint)height-(winY+k),1,1,
		     GL_RGB,GL_FLOAT,&pixel);

	/* omit background pixels, to speed things up a bit */
	if(!multiple || (multiple && ((pixel1[0] < pixel[0]) &&
				      (pixel1[1] < pixel[1]) &&
				      (pixel1[2] < pixel[2]))))
	{
	  mops_status = mops_window_to_object(togl, sel->object,
					  winX+j,winY+k,&objX,&objY,&objZ);
	  /*      if(view->zoom>10)
		  mops_prefs.pick_epsilon *= view->zoom;*/
	  mops_point_edit_coords = NULL;
	  mops_status = mops_get_point(sel->object, objX, objY, objZ);
	  /*      mops_prefs.pick_epsilon = pickepsilon;*/

	  if(mops_point_edit_coords)
	    {
	      have_it = 0;
	      for(i=0;i<mops_point_edit_coords_number;i++)
		{
		  /* do we have that point selected already? */
		  point = sel->object->selp;
		  last = NULL;
		  have_it = 0;
		  while(point)
		    {
		      if(point->point == mops_point_edit_coords[i])
			{ /* we have that point already, so we delete
			     it from the selection if we are not in
			     multiple selection mode */
			  if(!multiple)
			    {

			      if(last)
				last->next = point->next;
			      else
				sel->object->selp = point->next;

			      free(point);
			    }
			  have_it = 1;
			  point = NULL; /* break loop */
			}

		      last = point;
		      if(point)
			point = point->next;
		    }

		  if(!have_it)
		    {
		      /* create new point object */
		      newp = NULL;
		      if(!(newp = calloc(1,sizeof(mops_point_object))))
			{
			  mops_error(interp, MOPS_OUTOFMEM, "selectPoints",
				     NULL);
			  return TCL_OK;
			} 
		      newp->next = sel->object->selp;
		      sel->object->selp = newp;
		      newp->point = mops_point_edit_coords[i];
		      newp->homogenous = mops_point_edit_coords_homogenous;
		    }

		} /* for */
	    } /*if*/
	} /* if */
      } /* for */

      sel = sel->next;
      } /* while */

    }
  else
    mops_error(interp, MOPS_NOSEL, "selectPoints", NULL);

 return TCL_OK;
}

/********************************************/
/* deselectPoints_cb:
                                    */
/********************************************/
int
deselectPoints_cb(struct Togl *togl, int argc, char *argv[])
{
 mops_list_object *sel = mops_current_selection;

  while(sel)
    {
      
      mops_clear_point_selection(sel->object);

      sel = sel->next;
    } /* while */

 return TCL_OK;
}

/********************************************/
/* startPointEdit_cb:
   prepares everything for the point editing
   modes                                    */
/********************************************/
int
startPointEdit_cb(struct Togl *togl, int argc, char *argv[])
{
  int mops_status = MOPS_OK;
  Tcl_Interp *interp = Togl_Interp (togl);
  /*  mops_view_info *view = Togl_GetClientData(togl);*/
  double winX = 0.0, winY = 0.0, objX = 0.0, objY = 0.0, objZ = 0.0;
  /*  double pickepsilon = mops_prefs.pick_epsilon;*/

  if(mops_current_selection)
    {
      Tcl_GetDouble(interp, argv[2], &winX);
      Tcl_GetDouble(interp, argv[3], &winY);

      mops_status = mops_window_to_object(togl,mops_current_selection->object,
					  winX,winY,&objX,&objY,&objZ);

      /*
	if(view->zoom>10)
	mops_prefs.pick_epsilon *= view->zoom;
	*/
      mops_status = mops_get_point(mops_current_selection->object,
				   objX, objY, objZ);
      /*      mops_prefs.pick_epsilon = pickepsilon;*/
      mops_point_edit_object = mops_current_selection->object;
    }
  else
    mops_error(interp, MOPS_NOSEL, "StartPointEdit", NULL);

 return TCL_OK;
}


/********************************************/
/* pointEditD_cb:
 */
/********************************************/
int
pointEditD_cb(struct Togl *togl, int argc, char *argv[])
{
 int mops_status = MOPS_OK;
 Tcl_Interp *interp = Togl_Interp (togl);
 double winX = 0.0, winY = 0.0, objX = 0.0, objY = 0.0, objZ = 0.0;
 char *n1 = "editPointDarray";
 Tcl_Obj *to = NULL, *toa = NULL, *ton = NULL;
 int i;
 double *coords;
 char cmd[] = "editPointDp", cmd2[] = "redrawAllViews";

  if(mops_current_selection)
    {
      toa = Tcl_NewStringObj(n1,-1);
      ton = Tcl_NewStringObj(n1,-1);

      Tcl_GetDouble(interp, argv[2], &winX);
      Tcl_GetDouble(interp, argv[3], &winY);

      mops_status = mops_window_to_object(togl,mops_current_selection->object,
					  winX,winY,&objX,&objY,&objZ);

      mops_status = mops_get_point(mops_current_selection->object,
				   objX, objY, objZ);

      mops_point_edit_object = mops_current_selection->object;

      if(mops_point_edit_coords)
	{
	  coords = mops_point_edit_coords[0];

	  Tcl_SetStringObj(ton,"x",-1);
	  to = Tcl_NewDoubleObj(coords[0]);
	  Tcl_ObjSetVar2(interp,toa,ton,to,
			 TCL_LEAVE_ERR_MSG | TCL_GLOBAL_ONLY);

	  Tcl_SetStringObj(ton,"y",-1);
	  to = Tcl_NewDoubleObj(coords[1]);
	  Tcl_ObjSetVar2(interp,toa,ton,to,
			 TCL_LEAVE_ERR_MSG | TCL_GLOBAL_ONLY);

	  Tcl_SetStringObj(ton,"z",-1);
	  to = Tcl_NewDoubleObj(coords[2]);
	  Tcl_ObjSetVar2(interp,toa,ton,to,
			 TCL_LEAVE_ERR_MSG | TCL_GLOBAL_ONLY);

	  if(mops_point_edit_coords_homogenous)
	    {
	      Tcl_SetStringObj(ton,"w",-1);
	      to = Tcl_NewDoubleObj(coords[3]);
	      Tcl_ObjSetVar2(interp,toa,ton,to,
			     TCL_LEAVE_ERR_MSG | TCL_GLOBAL_ONLY);
	    }

	  Tcl_Eval(interp,cmd);

	  for(i=0;i<mops_point_edit_coords_number;i++)
	    {
	      coords = mops_point_edit_coords[i];
	      Tcl_SetStringObj(ton,"x",-1);
	      to = Tcl_ObjGetVar2(interp,toa,ton,
				  TCL_LEAVE_ERR_MSG | TCL_GLOBAL_ONLY);
	      Tcl_GetDoubleFromObj(interp,to, &coords[0]);
	      Tcl_SetStringObj(ton,"y",-1);
	      to = Tcl_ObjGetVar2(interp,toa,ton,
				  TCL_LEAVE_ERR_MSG | TCL_GLOBAL_ONLY);
	      Tcl_GetDoubleFromObj(interp,to, &coords[1]);
	      Tcl_SetStringObj(ton,"z",-1);
	      to = Tcl_ObjGetVar2(interp,toa,ton,
				  TCL_LEAVE_ERR_MSG | TCL_GLOBAL_ONLY);
	      Tcl_GetDoubleFromObj(interp,to, &coords[2]);

	      if(mops_point_edit_coords_homogenous)
		{
		  Tcl_SetStringObj(ton,"w",-1);
		  to = Tcl_ObjGetVar2(interp,toa,ton,
				      TCL_LEAVE_ERR_MSG | TCL_GLOBAL_ONLY);
		  Tcl_GetDoubleFromObj(interp,to, &coords[3]);
		}

	    } /* for */

	  Tcl_Eval(interp,cmd2);

	} /* if */

      Tcl_IncrRefCount(toa); Tcl_DecrRefCount(toa);
      Tcl_IncrRefCount(ton); Tcl_DecrRefCount(ton);

      mops_point_edit_object->modified = MOPS_TRUE;
	  
    }
  else
    mops_error(interp, MOPS_NOSEL, "pointEditD", NULL);

  mops_status = mops_notify_parent();

 return TCL_OK;
}


/********************************************/
/* insertPoint_cb:
 */
/********************************************/
int
insertPoint_cb(struct Togl *togl, int argc, char *argv[])
{
 int mops_status = MOPS_OK;
 Tcl_Interp *interp = Togl_Interp (togl);
 double winX = 0.0, winY = 0.0, objX = 0.0, objY = 0.0, objZ = 0.0;
 char fname[] = "insert_point";
 mops_nurbcurve_object *curve = NULL;
 double *cv = NULL, u = 0.0;
 int i=0, j=0, k=0, add=0, index = -1, s=0, nq = 0;
 double min_distance = mops_prefs.pick_epsilon, distance = 0.0;
 double *newcontrolv = NULL, *oldcontrolv = NULL, *newknotv = NULL;
 int inserted;

  if(mops_current_selection)
    {
      if(mops_current_selection->object->type != MOPS_OTNURBCURVE)
	{
	  mops_error(interp,MOPS_ERROR,fname,"object is not a NURB curve");
	  return TCL_OK;
	}

      mops_clear_point_selection(mops_current_selection->object);

      Tcl_GetDouble(interp, argv[2], &winX);
      Tcl_GetDouble(interp, argv[3], &winY);

      mops_status = mops_window_to_object(togl,mops_current_selection->object,
					  winX,winY,&objX,&objY,&objZ);

      curve = (mops_nurbcurve_object *)mops_current_selection->object->object;
      
      cv = curve->controlv;

      if(min_distance == 0.0)
	min_distance = DBL_MAX;
      add = ((curve->has_weights)?4:3);

      for(i = 0; i < curve->length; i++)
	{
	  distance = MOPS_VLEN((objX - curve->controlv[j]),
			       (objY -  curve->controlv[j+1]),
			       (objZ -  curve->controlv[j+2]));

	  if(distance < min_distance)
	    {
	      index = i;
	      min_distance = distance;
	    }

	  j += add;
	} /* for */

      /* no point picked? */
      if(index == -1)
	{
	  return TCL_OK;
	}

      if(curve->closed)
	{
	  /*
	   * if the curve is closed we simply insert the new point
	   * in the controlvector between the picked point and the
	   * next different
	   */
	  oldcontrolv = curve->controlv;

	  if(index == curve->length-1)
	    index--;

	  curve->length++;
	  if(!(newcontrolv = calloc(curve->length*add, sizeof(double))))
	    return MOPS_OUTOFMEM;

	  j = 0;
	  inserted = MOPS_FALSE;
	  for(i = 0; i < (curve->length-1); i++)
	    {
	      if(i>=index && !inserted)
		{
		  memcpy(&(newcontrolv[j*add]), &(oldcontrolv[i*add]),
			 add*sizeof(double));
		  
		  if(oldcontrolv[i*add] != oldcontrolv[(i+1)*add] ||
		     oldcontrolv[i*add+1] != oldcontrolv[(i+1)*add+1] ||
		     oldcontrolv[i*add+2] != oldcontrolv[(i+1)*add+2])
		    {
		      newcontrolv[(j+1)*add] = oldcontrolv[i*add] +
			((oldcontrolv[(i+1)*add] -
			  oldcontrolv[i*add])/2.0);

		      newcontrolv[(j+1)*add+1] = oldcontrolv[i*add+1] +
			((oldcontrolv[(i+1)*add+1] -
			  oldcontrolv[i*add+1])/2.0);

		      newcontrolv[(j+1)*add+2] = oldcontrolv[i*add+2] +
			((oldcontrolv[(i+1)*add+2] -
			  oldcontrolv[i*add+2])/2.0);

		      if(add == 4)
			newcontrolv[(j+1)*add+3] = oldcontrolv[i*add+3] +
			  ((oldcontrolv[(i+1)*add+3] -
			    oldcontrolv[i*add+3])/2.0);

		      j++;
		      inserted = MOPS_TRUE;
		    }

		}
	      else
		{
		  memcpy(&(newcontrolv[j*add]), &(oldcontrolv[i*add]),
			 add*sizeof(double));
		}
	      j++;
	    } /* for */

	  if(!inserted)
	    {
	      free(newcontrolv);
	      curve->length--;
	      mops_error(interp,MOPS_ERROR,fname,
			 "Cannot insert point here!");
	      return TCL_OK;
	    }

	  free(curve->controlv);
	  curve->controlv = newcontrolv;

	  if(curve->knot_type == MOPS_KTCUSTOM)
	    {
	      mops_error(interp,MOPS_WARNING,fname,
			 "Changed knot type to B-Spline!");
	      curve->knot_type = MOPS_KTBSPLINE;
	    }

	  mops_status = mops_create_nurbcurve_knots(curve);
	  mops_status = mops_nurbcurve_close(curve);
	}
      else
	{ /* curve is not closed */

	  mops_nurbcurve_findufromcontrol(index, curve->length+curve->order,
				     curve->order, curve->knotv, &u, &k,&s);

	  curve->length++;
	  if(!(newcontrolv = calloc(curve->length*
				    (curve->has_weights?4:3), sizeof(double))))
	    return MOPS_OUTOFMEM;

	  if(!(newknotv = calloc(curve->length+curve->order, sizeof(double))))
	    return MOPS_OUTOFMEM;

	  if(u <= curve->knotv[curve->order-2] ||
	     u >= curve->knotv[curve->length])
	    {
	      j = 0;k =0;
	      for(i=0;i<curve->length-1;i++)
		{
		  if(i == index)
		    {
		      /* insert point here */
		      newcontrolv[k] = curve->controlv[j];
		      newcontrolv[k+1] = curve->controlv[j+1];
		      newcontrolv[k+2] = curve->controlv[j+2];
		      if(curve->has_weights)
			newcontrolv[k+3] = curve->controlv[j+3];
		      k+=add;

		      if(i == curve->length-2)
			{ /* the new point is the new last point */
			  newcontrolv[k] = curve->controlv[j] +
			    ((curve->controlv[j]-curve->controlv[j-add])/2.0);
			  newcontrolv[k+1] = curve->controlv[j+1]+
			    ((curve->controlv[j+1]-curve->controlv[j+1-add])/2.0);
			  newcontrolv[k+2] = curve->controlv[j+2]+
			    ((curve->controlv[j+2]-curve->controlv[j+2-add])/2.0);
			  if(curve->has_weights)
			    newcontrolv[k+3] = curve->controlv[j+3]+
			      ((curve->controlv[j+3]-curve->controlv[j+3-add])/2.0);
			}
		      else
			{
			  newcontrolv[k] = curve->controlv[j] +
			    ((curve->controlv[j+add]-curve->controlv[j])/2.0);
			  newcontrolv[k+1] = curve->controlv[j+1]+
			    ((curve->controlv[j+1+add]-curve->controlv[j+1])/2.0);
			  newcontrolv[k+2] = curve->controlv[j+2]+
			    ((curve->controlv[j+2+add]-curve->controlv[j+2])/2.0);
			  if(curve->has_weights)
			    newcontrolv[k+3] = curve->controlv[j+3]+
			      ((curve->controlv[j+3+add]-curve->controlv[j+3])/2.0);
			}
		    }
		  else
		    {
		      newcontrolv[k] = curve->controlv[j];
		      newcontrolv[k+1] = curve->controlv[j+1];
		      newcontrolv[k+2] = curve->controlv[j+2];
		      if(curve->has_weights)
			newcontrolv[k+3] = curve->controlv[j+3];
		    }
		  
		  k+=add;
		  j+=add;
		} /* for */

	      if(curve->knot_type == MOPS_KTCUSTOM)
		{
		  mops_error(interp,MOPS_WARNING,fname,
			     "Changed knot type to NURB!");
		  curve->knot_type = MOPS_KTNURB;
		}

	    }
	  else
	    {
	      if(curve->has_weights)
		mops_status = mops_nb_CurveInsertKnot4D(curve->length-2,
			curve->order-1, curve->knotv, curve->controlv, u, k,
			0, 1, &nq, newknotv, newcontrolv);
	      else
		mops_status = mops_nb_CurveInsertKnot3D(curve->length-2,
			  curve->order-1, curve->knotv, curve->controlv, u, k,
			  0, 1, &nq, newknotv, newcontrolv);

	    }

	  free(curve->controlv);
	  curve->controlv = newcontrolv;

	  free(curve->knotv);
	  curve->knotv = newknotv;
	  
	  if(curve->knot_type == MOPS_KTBEZIER)
	    curve->order++;

	  if(curve->knot_type != MOPS_KTCUSTOM)
	    mops_status = mops_create_nurbcurve_knots(curve);

	}

       mops_nurbcurve_recreatemp(curve);

       mops_status = mops_notify_parent();

    }
  else
    mops_error(interp, MOPS_NOSEL, fname, NULL);

 return TCL_OK;
}

/********************************************/
/* splitCurve_cb:
 */
/********************************************/
int
splitCurve_cb(struct Togl *togl, int argc, char *argv[])
{
 int mops_status = MOPS_OK;
 mops_view_info *view = Togl_GetClientData(togl);
 Tcl_Interp *interp = Togl_Interp (togl);
 double winX = 0.0, winY = 0.0;
 char fname[] = "splitCurve_cb";
 double u = 0.0;

  if(argc > 2)
    {
      if(!strcmp(argv[2],"-start"))
	{
	  view->drawhandles = MOPS_FALSE;
	  display_cb(togl);
	  return TCL_OK;
	}
      if(!strcmp(argv[2],"-end"))
	{
	  view->drawhandles = MOPS_TRUE;
	  display_cb(togl);
	  return TCL_OK;
	}
    }
  else
    return TCL_OK;

  if(mops_current_selection)
    {
      if(mops_current_selection->object->type != MOPS_OTNURBCURVE)
	{
	  mops_error(interp,MOPS_ERROR,fname,"object is not a NURB curve");
	  return TCL_OK;
	}

      mops_clear_point_selection(mops_current_selection->object);

      Tcl_GetDouble(interp, argv[2], &winX);
      Tcl_GetDouble(interp, argv[3], &winY);

      mops_status = mops_window_to_ncu(togl,mops_current_selection->object,
				       &winX,&winY,&u);

      if(mops_status)
	return TCL_OK;


      mops_status = mops_split_ncurve(interp, mops_current_selection->object,
				      u);

      if(mops_status)
	{
	  mops_error(interp, mops_status, fname, NULL);
	  return TCL_OK;
	}

      mops_status = mops_notify_parent();

    }
  else
    mops_error(interp, MOPS_NOSEL, fname, NULL);

 return TCL_OK;
}


/********************************************/
/* findU_cb:
 */
/********************************************/
int
findU_cb(struct Togl *togl, int argc, char *argv[])
{
 int mops_status = MOPS_OK;
 mops_view_info *view = Togl_GetClientData(togl);
 Tcl_Interp *interp = Togl_Interp (togl);
 double winX = 0.0, winY = 0.0;
 static int fvalid = MOPS_FALSE;
 static double fX = 0.0, fY = 0.0; 
 char fname[] = "findU_cb";
 double u = 0.0;
 Tcl_Obj *to = NULL, *ton = NULL;
 char cmd[] = "puts $u";

  if(argc > 2)
    {
      if(!strcmp(argv[2],"-start"))
	{
	  view->drawhandles = MOPS_FALSE;
	  display_cb(togl);
	  fvalid = MOPS_FALSE;
	  return TCL_OK;
	}
      if(!strcmp(argv[2],"-end"))
	{
	  view->drawhandles = MOPS_TRUE;

	  glClearColor (view->bgr, view->bgg,
			view->bgb, view->bga);
	  glClear (GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

	  /* draw scene */
	  mops_draw_scene(togl);
 
	  /* draw cross */
	  if(fvalid)
	    {
	      glMatrixMode(GL_PROJECTION);

	      glPushMatrix();
	      glLoadIdentity();
	      glOrtho(0, Togl_Width(togl), 0, Togl_Height(togl),
		      -100.0, 100.0);
	      
	      glMatrixMode(GL_MODELVIEW);
	      glPushMatrix();
	      glLoadIdentity();
	  
	      glBegin(GL_LINES);
	      glVertex3d(fX-4,fY+4,0.0);
	      glVertex3d(fX+4,fY-4,0.0);
	      glVertex3d(fX-4,fY-4,0.0);
	      glVertex3d(fX+4,fY+4,0.0);
	      glEnd();

	      glPopMatrix();
	      glMatrixMode(GL_PROJECTION);
	      glPopMatrix();
	      glMatrixMode(GL_MODELVIEW);
	    }	      

	  glFlush();    

	  Togl_SwapBuffers (togl);
	  fvalid = MOPS_FALSE;
	  return TCL_OK;
	}
    }
  else
    return TCL_OK;

  if(mops_current_selection)
    {
      if(mops_current_selection->object->type != MOPS_OTNURBCURVE)
	{
	  mops_error(interp,MOPS_ERROR,fname,"object is not a NURB curve");
	  return TCL_OK;
	}

      mops_clear_point_selection(mops_current_selection->object);

      Tcl_GetDouble(interp, argv[2], &winX);
      Tcl_GetDouble(interp, argv[3], &winY);

      mops_status = mops_window_to_ncu(togl,mops_current_selection->object,
				       &winX,&winY,&u);

      if(mops_status)
	return TCL_OK;

      ton = Tcl_NewStringObj("u",-1);
      to = Tcl_NewDoubleObj(u);
      Tcl_ObjSetVar2(interp,ton,NULL,to,TCL_LEAVE_ERR_MSG | TCL_GLOBAL_ONLY);
      Tcl_Eval(interp, cmd);
      Tcl_IncrRefCount(ton);Tcl_DecrRefCount(ton);

      mops_status = mops_notify_parent();

      fvalid = MOPS_TRUE;
      fX = winX;
      fY = winY;
    }
  else
    mops_error(interp, MOPS_NOSEL, fname, NULL);

 return TCL_OK;
}


/********************************************/
/* deletePoint_cb:
 */
/********************************************/
int
deletePoint_cb(struct Togl *togl, int argc, char *argv[])
{
 int mops_status = MOPS_OK;
 Tcl_Interp *interp = Togl_Interp (togl);
 double winX = 0.0, winY = 0.0, objX = 0.0, objY = 0.0, objZ = 0.0;
 char fname[] = "delete_point";
 mops_nurbcurve_object *curve = NULL;
 double *cv = NULL;
 int i=0, j=0, k=0, add=0, index = -1;
 double min_distance = mops_prefs.pick_epsilon, distance = 0.0;
 double *newcontrolv = NULL, *newknotv = NULL;

  if(mops_current_selection)
    {
      if(mops_current_selection->object->type != MOPS_OTNURBCURVE)
	{
	  mops_error(interp,MOPS_ERROR,fname,"object is not a NURB curve");
	  return TCL_OK;
	}

      mops_clear_point_selection(mops_current_selection->object);

      Tcl_GetDouble(interp, argv[2], &winX);
      Tcl_GetDouble(interp, argv[3], &winY);

      mops_status = mops_window_to_object(togl,mops_current_selection->object,
					  winX,winY,&objX,&objY,&objZ);

      curve = (mops_nurbcurve_object *)mops_current_selection->object->object;
      
      cv = curve->controlv;

      if(min_distance == 0.0)
	min_distance = DBL_MAX;
      add = ((curve->has_weights)?4:3);
      index = -1;
      for(i = 0; i < curve->length; i++)
	{
	  distance = MOPS_VLEN((objX - cv[j]),
			       (objY -  cv[j+1]),
			       (objZ -  cv[j+2]));

	  if(distance < min_distance)
	    {
	      index = i;
	      min_distance = distance;
	    }

	  j += add;
	} /* for */

      if((index != -1) && (curve->length-1 < 2))
	{
	  mops_error(interp,MOPS_ERROR,fname,"need atleast two points");
	  return TCL_OK;
	}

      if((curve->closed) &&
	 (index > (curve->length-curve->order)))
	{
	  index = curve->order-(curve->length-index)-1;
	}

      /* create new curve */
      if(index != -1)
	{
	  curve->length--;
	  if(!(newcontrolv = calloc(curve->length*add, sizeof(double))))
	    return MOPS_OUTOFMEM;

	  /* copy controlv */
	  j = 0;
	  k = 0;
	  for(i=0; i<=curve->length; i++)
	    {
	      if(i != index)
		{
		  newcontrolv[k] = curve->controlv[j];
		  newcontrolv[k+1] = curve->controlv[j+1];
		  newcontrolv[k+2] = curve->controlv[j+2];
		  if(curve->has_weights)
		    newcontrolv[k+3] = curve->controlv[j+3];
		  k+=add;
		}
		
	      j+=add;
	    } /* for */

	  free(curve->controlv);
	  curve->controlv = newcontrolv;

	  if(curve->knot_type == MOPS_KTBEZIER)
	    curve->order--;

	  if(curve->length < curve->order)
	    curve->order = curve->length;

	  /* generate new knots? */
	  if(curve->knot_type == MOPS_KTCUSTOM)
	    {
	      if(!(newknotv = calloc(curve->order+curve->length,
				     sizeof(double))))
		{ return MOPS_OUTOFMEM; }

	      memcpy(newknotv, curve->knotv, (curve->length)*sizeof(double));
	      memcpy(&(newknotv[curve->length]),
		     &(curve->knotv[curve->length+1]),
		     curve->order*sizeof(double));
	      /*
	      for(i=0;i<curve->length+curve->order;i++)
		{
		  newknotv[i] = curve->knotv[i];
		}
	      */
	      free(curve->knotv);
	      curve->knotv = newknotv;
	    }
	  else
	    {
	      mops_status = mops_create_nurbcurve_knots(curve);
	    }

	} /* if */

      if(curve->closed)
	{
	  mops_status = mops_nurbcurve_close(curve);
	  if(mops_status) 
	    {
	     mops_error(interp, mops_status, fname, "cannot close this curve");
	     curve->closed = MOPS_FALSE;
	    }
	}
      
      mops_nurbcurve_recreatemp(curve);

      mops_status = mops_notify_parent();

    }
  else
    mops_error(interp, MOPS_NOSEL, fname, NULL);

 return TCL_OK;
}


/********************************************/
/* editPoint_cb:
   action callback for single point editing */
/********************************************/
int
editPoint_cb(struct Togl *togl, int argc, char *argv[])
{
 int mops_status = MOPS_OK;
 Tcl_Interp *interp = Togl_Interp (togl);
 mops_view_info *view = Togl_GetClientData(togl);
 static double oldwinx = 0.0, oldwiny = 0.0;
 double winx = 0.0, winy = 0.0;
 double movX, movY, movZ, dx=0.0, dy=0.0, dz=0.0, *coords;
 float euler[3] = {0};
 int i = 0;
 static GLdouble m[16] = {0};
 mops_object *o = mops_point_edit_object;

  if(!o) return TCL_OK;

  if(argc == 5)
    {
      if(!strcmp(argv[2],"-winxy"))
      {
	Tcl_GetDouble(interp, argv[3], &winx);
	Tcl_GetDouble(interp, argv[4], &winy);

	if(view->grid != 0.0)
	  {
	    mops_snap_togrid(togl,&winx,&winy);
	  }
      }
      else
	if(!strcmp(argv[2],"-start"))
	  {
	    Tcl_GetDouble(interp, argv[3], &winx);
	    Tcl_GetDouble(interp, argv[4], &winy);

	    if(view->grid != 0.0)
	      {
		mops_snap_togrid(togl,&winx,&winy);
	      }

	    oldwinx = winx;
	    oldwiny = winy;

	    glMatrixMode(GL_MODELVIEW);
	    glPushMatrix();
	    glScaled (1.0/o->scalx, 1.0/o->scaly, 1.0/o->scalz);

	    mops_quat_to_euler(o->quat,euler);
	    glRotatef(MOPS_R2D(euler[0]), 0.0, 0.0, 1.0);
	    glRotatef(MOPS_R2D(euler[1]), 0.0, 1.0, 0.0);
	    glRotatef(MOPS_R2D(euler[2]), 1.0, 0.0, 0.0);

	    glTranslated(-o->movx, -o->movy, -o->movz);
	    mops_get_all_itransform(mops_current_level->next);
	    glGetDoublev(GL_MODELVIEW_MATRIX, m);
	    glPopMatrix();

	    }
    }
  else
    {
      Tcl_SetResult (interp, "wrong # of args, should be \"movPointac [-start winx winy]|[-winxy winx winy]\"",
		     TCL_STATIC);
      return TCL_ERROR;
    } 

  dx = -(oldwinx - winx) * view->conv_x;
  dy = (oldwiny - winy) * view->conv_y;

  oldwinx = winx;
  oldwiny = winy;

  /* Side or Top view? */
  if(view->type == MOPS_VTSIDE)
    {
      dz = -dx;
      dx = 0.0;
    }

  if(view->type == MOPS_VTTOP)
    {
      dx = dx;
      dz = -dy;
      dy = 0.0;
    }

  movX = m[0]*dx+m[4]*dy+m[8]*dz;
  movY = m[1]*dx+m[5]*dy+m[9]*dz;
  movZ = m[2]*dx+m[6]*dy+m[10]*dz;

  if(mops_point_edit_coords)
    {
      for(i=0;i<mops_point_edit_coords_number;i++)
	{
	  coords = mops_point_edit_coords[i]; 

	  coords[0] += movX; 
	  coords[1] += movY;
	  coords[2] += movZ;
	  
	}
    }

  o->modified = MOPS_TRUE;

  if(!mops_prefs.lazynotify)
    mops_status = mops_notify_parent();

 return TCL_OK;
}


/********************************************/
/* editWeight_cb: 
   action callback for single point weight
   editing                                  */
/********************************************/
int
editWeight_cb(struct Togl *togl, int argc, char *argv[])
{
 int mops_status = MOPS_OK;
 Tcl_Interp *interp = Togl_Interp (togl);
 /*  mops_view_info *view = Togl_GetClientData(togl);*/
 double dx, winx = 0.0, new_weight, *coords;
 static double oldwinx = 0.0;
 static double olddx = 0.0;
 int i = 0;

  /* parse args */
  if(argc == 4)
    {
      if(!strcmp(argv[2],"-winx"))
	Tcl_GetDouble(interp, argv[3], &winx);
      else
	if(!strcmp(argv[2],"-start"))
	  {
	    Tcl_GetDouble(interp, argv[3], &winx);
	    oldwinx = winx;
	    olddx = 0.0;
	  }
    }
  else
    {
      Tcl_SetResult (interp,
"wrong # of args, should be \"editWac [-start winx]|[-winx winx]\"",
		     TCL_STATIC);
      return TCL_ERROR;
    }

  dx = oldwinx - winx;
  oldwinx = winx;

  if(mops_point_edit_coords)
    {
      for(i=0;i<mops_point_edit_coords_number;i++)
	{
	  coords = mops_point_edit_coords[i]; 
	  if(mops_point_edit_coords_homogenous)
	    {
	      new_weight = coords[3];
	      if(dx>0.0)
		new_weight *= 1.1;
	      else
		new_weight *= 0.9;
	      coords[3] = new_weight;
	    }
	}
    }

  mops_point_edit_object->modified = MOPS_TRUE;

  if(!mops_prefs.lazynotify)
    mops_status = mops_notify_parent();

 return TCL_OK;
}


/********************************************/
/* mops_zoomToOb:                           */
/********************************************/
int
mops_zoomToOb(struct Togl *togl, int argc, char *argv[])
{
 int mops_status = MOPS_OK;
 mops_view_info *view = Togl_GetClientData(togl);
 mops_object *o = NULL;
 mops_list_object *sel = mops_current_selection;
 double bb[24] = {0};
 double xmin = DBL_MAX, xmax = -DBL_MAX, ymin = DBL_MAX;
 double ymax = -DBL_MAX, zmin = DBL_MAX, zmax = -DBL_MAX;
 double cog[3] = {0}, dt[3] = {0}, dt2[3] = {0}, l, lx, ly, lz;
 int i, a;
 GLdouble m[16] = {0};

  if(sel)
    {
      while(sel)
	{
	  o = sel->object;

	  mops_status = mops_get_bb(o, bb);
	  
	  a = 0;
	  for(i = 0; i < 8; i++)
	    {
	      if(bb[a] < xmin)
		xmin = bb[a];
	      if(bb[a] > xmax)
		xmax = bb[a];
	      a += 3;
	    }

	  a = 1;
	  for(i = 0; i < 8; i++)
	    {
	      if(bb[a] < ymin)
		ymin = bb[a];
	      if(bb[a] > ymax)
		ymax = bb[a];
	      a += 3;
	    }

	  a = 2;
	  for(i = 0; i < 8; i++)
	    {
	      if(bb[a] < zmin)
		zmin = bb[a];
	      if(bb[a] > zmax)
		zmax = bb[a];
	      a += 3;
	    }
	  sel = sel->next;
	}

      /* calc center of gravity (middle of bounding box) */
      cog[0] = xmin+((fabs(xmax-xmin))/2.0);
      cog[1] = ymin+((fabs(ymax-ymin))/2.0);
      cog[2] = zmin+((fabs(zmax-zmin))/2.0);

      dt[0] = view->to_x - cog[0];
      dt[1] = view->to_y - cog[1];
      dt[2] = view->to_z - cog[2];

      /* let the view point to objects COG */
      view->to_x -= dt[0];
      view->to_y -= dt[1];
      view->to_z -= dt[2];
      view->from_x -= dt[0];
      view->from_y -= dt[1];
      view->from_z -= dt[2];

      reshape_cb(togl);

      /* calc new zoom factor */
      glMatrixMode (GL_MODELVIEW);
      glPushMatrix();
       glLoadIdentity();
       glRotated(view->rotx, 1.0, 0.0, 0.0);
       glRotated(view->roty, 0.0, 1.0, 0.0);
       glRotated(view->rotz, 0.0, 0.0, 1.0);
       glGetDoublev(GL_MODELVIEW_MATRIX, m);
      glPopMatrix();

      dt[0] = xmin;
      dt[1] = ymin;
      dt[2] = zmin;

      dt2[0] = xmax;
      dt2[1] = ymax;
      dt2[2] = zmax;

      /*
      mops_apply_transform3(dt, m);
      mops_apply_transform3(dt2, m);
      */

      lx = fabs(dt[0]-dt2[0]);
      ly = fabs(dt[1]-dt2[1]);
      lz = fabs(dt[2]-dt2[2]);

      
      if(ly>lx)
	l = ly;
      else
	l = lx;
      if(l<lz)
	l = lz;

      if(l < 1E-6)
	l = 3.0;

      view->zoom = l*0.9;

      if(view->type == MOPS_VTPERSP)
	{
	  view->zoom /= 12; 
	  /* normalize */
	  /*
	  dt[0] = view->from_x - view->to_x;
	  dt[1] = view->from_y - view->to_y;
	  dt[2] = view->from_z - view->to_z;
	  l = MOPS_V3LEN(dt);
	  */
	}

      if(view->zoom < 1E-6)
	view->zoom = 3.0;

      reshape_cb(togl);
      display_cb(togl);
    }

 return TCL_OK;
}


/********************************************/

/********************************************/
/*Tcl_AppInit || mops_InitStandalone:       */
/********************************************/
#ifdef MOPSSTANDALONE
int mops_InitStandalone(Tcl_Interp *interp);
int
mops_InitStandalone(Tcl_Interp *interp)
#else
int
Tcl_AppInit (Tcl_Interp *interp)
#endif
{
int mops_status = MOPS_OK;

#ifndef MOPSSTANDALONE
  if (Tcl_Init (interp) == TCL_ERROR)
    return TCL_ERROR;

  if (Tk_Init (interp) == TCL_ERROR)
    return TCL_ERROR;
#endif

  if (Togl_Init (interp) == TCL_ERROR)
    return TCL_ERROR;

  Togl_CreateFunc  (create_cb);
  Togl_DestroyFunc (destroy_cb);
  Togl_DisplayFunc (display_cb);
  Togl_ReshapeFunc (reshape_cb);
  Togl_CreateCommand ("setViewConfig", setViewConfig_cb);
  Togl_CreateCommand ("getViewConfig", getViewConfig_cb);
  Togl_CreateCommand ("updatePosition", updatePosition_cb);
  Togl_CreateCommand ("startPointEdit", startPointEdit_cb);
  Togl_CreateCommand ("selectPoints", selectPoints_cb);
  Togl_CreateCommand ("deselectPoints", deselectPoints_cb);
  Togl_CreateCommand ("editPac", editPoint_cb);
  Togl_CreateCommand ("editWac", editWeight_cb);
  Togl_CreateCommand ("pointEditD", pointEditD_cb);
  Togl_CreateCommand ("insertPoint", insertPoint_cb);
  Togl_CreateCommand ("deletePoint", deletePoint_cb);
  Togl_CreateCommand ("splitCurve", splitCurve_cb);
  Togl_CreateCommand ("findU", findU_cb);

  Togl_CreateCommand ("movObac", mops_movob_actcb);
  Togl_CreateCommand ("movVwac", mops_moview_actcb);
  Togl_CreateCommand ("zomVwac", mops_zoomview_actcb);
  Togl_CreateCommand ("rotObac", mops_rotob_actcb);
  Togl_CreateCommand ("rotObaac", mops_rotoba_actcb);
  Togl_CreateCommand ("sc1DXObac", mops_sc1DXob_actcb);
  Togl_CreateCommand ("sc1DYObac", mops_sc1DYob_actcb);
  Togl_CreateCommand ("sc1DZObac", mops_sc1DZob_actcb);
  Togl_CreateCommand ("sc1DXWorldac", mops_sc1DXworld_actcb);
  Togl_CreateCommand ("sc1DYWorldac", mops_sc1DYworld_actcb);
  Togl_CreateCommand ("sc1DZWorldac", mops_sc1DZworld_actcb);
  Togl_CreateCommand ("sc2DObac", mops_sc2Dob_actcb);
  Togl_CreateCommand ("sc3DObac", mops_sc3Dob_actcb);
  Togl_CreateCommand ("wrib", mops_wrib_cb);
  Togl_CreateCommand ("zoomToOb", mops_zoomToOb);
  Togl_CreateCommand ("movzVwac", mops_movezview_actcb);

  Tcl_CreateCommand (interp, "scanShader", mops_scan_shader_tcmd,
		     (ClientData) NULL, (Tcl_CmdDeleteProc *) NULL);

  Tcl_CreateCommand (interp, "crtNSphere2", mops_crtnsphere2_tcmd,
		     (ClientData) NULL, (Tcl_CmdDeleteProc *) NULL);

  Tcl_CreateCommand (interp, "splitNPatch", mops_split_to_curves_tcmd,
		     (ClientData) NULL, (Tcl_CmdDeleteProc *) NULL);

  Tcl_CreateCommand (interp, "buildNPatch", mops_build_from_curves_tcmd,
		     (ClientData) NULL, (Tcl_CmdDeleteProc *) NULL);
		     
  Tcl_CreateCommand (interp, "newScene", mops_new_scene_tcmd,
		     (ClientData) NULL, (Tcl_CmdDeleteProc *) NULL);
  Tcl_CreateCommand (interp, "replaceScene", mops_replace_scene_tcmd,
		     (ClientData) NULL, (Tcl_CmdDeleteProc *) NULL);
  Tcl_CreateCommand (interp, "insertScene", mops_insert_scene_tcmd,
		     (ClientData) NULL, (Tcl_CmdDeleteProc *) NULL);
  Tcl_CreateCommand (interp, "saveScene", mops_save_scene_tcmd,
		     (ClientData) NULL, (Tcl_CmdDeleteProc *) NULL);

  Tcl_CreateCommand (interp, "getCLevel", mops_get_clevel_tcmd,
		     (ClientData) NULL, (Tcl_CmdDeleteProc *) NULL);
  Tcl_CreateCommand (interp, "setSelect", mops_set_select_tcmd,
		     (ClientData) NULL, (Tcl_CmdDeleteProc *) NULL);

  Tcl_CreateCommand (interp, "goTop", mops_gotop_tcmd,
		     (ClientData) NULL, (Tcl_CmdDeleteProc *) NULL);

  Tcl_CreateCommand (interp, "goUp", mops_goup_tcmd,
		     (ClientData) NULL, (Tcl_CmdDeleteProc *) NULL);

  Tcl_CreateCommand (interp, "goDown", mops_godown_tcmd,
		     (ClientData) NULL, (Tcl_CmdDeleteProc *) NULL);

  Tcl_CreateCommand (interp, "getVersion", mops_get_version_tcmd,
		     (ClientData) NULL, (Tcl_CmdDeleteProc *) NULL);

  /* transformations */
  Tcl_CreateCommand (interp, "movOb", mops_movob_tcmd,
		     (ClientData) NULL, (Tcl_CmdDeleteProc *) NULL);
  Tcl_CreateCommand (interp, "scalOb", mops_scalob_tcmd,
		     (ClientData) NULL, (Tcl_CmdDeleteProc *) NULL);
  Tcl_CreateCommand (interp, "rotOb", mops_rotob_tcmd,
		     (ClientData) NULL, (Tcl_CmdDeleteProc *) NULL);
  Tcl_CreateCommand (interp, "movSel", mops_movsel_tcmd,
		     (ClientData) NULL, (Tcl_CmdDeleteProc *) NULL);
  Tcl_CreateCommand (interp, "scalSel", mops_scalsel_tcmd,
		     (ClientData) NULL, (Tcl_CmdDeleteProc *) NULL);
  Tcl_CreateCommand (interp, "rotSel", mops_rotsel_tcmd,
		     (ClientData) NULL, (Tcl_CmdDeleteProc *) NULL);

  Tcl_CreateCommand (interp, "delegateTran", mops_delegtran_tcmd,
		     (ClientData) NULL, (Tcl_CmdDeleteProc *) NULL);

  /* clipboard */
  Tcl_CreateCommand (interp, "delOb", mops_delob_tcmd,
		     (ClientData) NULL, (Tcl_CmdDeleteProc *) NULL);
  Tcl_CreateCommand (interp, "cutOb", mops_cutob_tcmd,
		     (ClientData) NULL, (Tcl_CmdDeleteProc *) NULL);
  Tcl_CreateCommand (interp, "pasOb", mops_pasob_tcmd,
		     (ClientData) NULL, (Tcl_CmdDeleteProc *) NULL);
  Tcl_CreateCommand (interp, "copOb", mops_copob_tcmd,
		     (ClientData) NULL, (Tcl_CmdDeleteProc *) NULL);

  /* create */
  Tcl_CreateCommand (interp, "crtOb", mops_crtob_tcmd,
		     (ClientData) NULL, (Tcl_CmdDeleteProc *) NULL);

  /* tools */
  Tcl_CreateCommand (interp, "revertNC", mops_revert_ncurve_tcmd,
		     (ClientData) NULL, (Tcl_CmdDeleteProc *) NULL);

  /* properties */
  Tcl_CreateCommand (interp, "getQAttr", mops_get_qattrprop_tcmd,
		     (ClientData) NULL, (Tcl_CmdDeleteProc *) NULL);
  Tcl_CreateCommand (interp, "setQAttr", mops_set_qattrprop_tcmd,
		     (ClientData) NULL, (Tcl_CmdDeleteProc *) NULL);
  Tcl_CreateCommand (interp, "getTransfProp", mops_get_transfprop_tcmd,
		     (ClientData) NULL, (Tcl_CmdDeleteProc *) NULL);
  Tcl_CreateCommand (interp, "setTransfProp", mops_set_transfprop_tcmd,
		     (ClientData) NULL, (Tcl_CmdDeleteProc *) NULL);
  Tcl_CreateCommand (interp, "getAttrProp", mops_get_attrprop_tcmd,
		     (ClientData) NULL, (Tcl_CmdDeleteProc *) NULL);
  Tcl_CreateCommand (interp, "setAttrProp", mops_set_attrprop_tcmd,
		     (ClientData) NULL, (Tcl_CmdDeleteProc *) NULL);
  Tcl_CreateCommand (interp, "getNCAttrProp", mops_get_ncattrprop_tcmd,
		     (ClientData) NULL, (Tcl_CmdDeleteProc *) NULL);
  Tcl_CreateCommand (interp, "setNCAttrProp", mops_set_ncattrprop_tcmd,
		     (ClientData) NULL, (Tcl_CmdDeleteProc *) NULL);
  Tcl_CreateCommand (interp, "getNPAttrProp", mops_get_npattrprop_tcmd,
		     (ClientData) NULL, (Tcl_CmdDeleteProc *) NULL);
  Tcl_CreateCommand (interp, "setNPAttrProp", mops_set_npattrprop_tcmd,
		     (ClientData) NULL, (Tcl_CmdDeleteProc *) NULL);
  Tcl_CreateCommand (interp, "getLAttrProp", mops_get_lattrprop_tcmd,
		     (ClientData) NULL, (Tcl_CmdDeleteProc *) NULL);
  Tcl_CreateCommand (interp, "setLAttrProp", mops_set_lattrprop_tcmd,
		     (ClientData) NULL, (Tcl_CmdDeleteProc *) NULL);
  Tcl_CreateCommand (interp, "getSProp", mops_get_shaderprop_tcmd,
		     (ClientData) NULL, (Tcl_CmdDeleteProc *) NULL);
  Tcl_CreateCommand (interp, "setSProp", mops_set_shaderprop_tcmd,
		     (ClientData) NULL, (Tcl_CmdDeleteProc *) NULL);
  Tcl_CreateCommand (interp, "delShader",mops_del_shader_tcmd,
		     (ClientData) NULL, (Tcl_CmdDeleteProc *) NULL);

  Tcl_CreateCommand (interp, "getPrefsProp", mops_get_prefsprop_tcmd,
		     (ClientData) NULL, (Tcl_CmdDeleteProc *) NULL);
  Tcl_CreateCommand (interp, "setPrefsProp", mops_set_prefsprop_tcmd,
		     (ClientData) NULL, (Tcl_CmdDeleteProc *) NULL);

  Tcl_CreateCommand (interp, "getRiOptProp", mops_get_rioptprop_tcmd,
		     (ClientData) NULL, (Tcl_CmdDeleteProc *) NULL);
  Tcl_CreateCommand (interp, "setRiOptProp", mops_set_rioptprop_tcmd,
		     (ClientData) NULL, (Tcl_CmdDeleteProc *) NULL);

  Tcl_CreateCommand (interp, "getBoxAttr", mops_get_boxattrprop_tcmd,
		     (ClientData) NULL, (Tcl_CmdDeleteProc *) NULL);
  Tcl_CreateCommand (interp, "setBoxAttr", mops_set_boxattrprop_tcmd,
		     (ClientData) NULL, (Tcl_CmdDeleteProc *) NULL);

  Tcl_CreateCommand (interp, "getCustomObjects", mops_get_custom_objects_tcmd,
		     (ClientData) NULL, (Tcl_CmdDeleteProc *) NULL);

  Tcl_CreateCommand (interp, "getNCPoint", mops_get_ncurve_point_tcmd,
		     (ClientData) NULL, (Tcl_CmdDeleteProc *) NULL);
  Tcl_CreateCommand (interp, "getNPPoint", mops_get_npatch_point_tcmd,
		     (ClientData) NULL, (Tcl_CmdDeleteProc *) NULL);
  Tcl_CreateCommand (interp, "setNCPoint", mops_set_ncurve_point_tcmd,
		     (ClientData) NULL, (Tcl_CmdDeleteProc *) NULL);
  Tcl_CreateCommand (interp, "setNPPoint", mops_set_npatch_point_tcmd,
		     (ClientData) NULL, (Tcl_CmdDeleteProc *) NULL);

  Tcl_CreateCommand (interp, "getTmpFile", mops_gettmp_tcmd,
		     (ClientData) NULL, (Tcl_CmdDeleteProc *) NULL);

  Tcl_CreateCommand (interp, "getType", mops_gettype_tcmd,
		     (ClientData) NULL, (Tcl_CmdDeleteProc *) NULL);

  Tcl_CreateCommand (interp, "hasChild", mops_haschild_tcmd,
		     (ClientData) NULL, (Tcl_CmdDeleteProc *) NULL);

  Tcl_CreateCommand (interp, "setON", mops_set_oname_tcmd,
		     (ClientData) NULL, (Tcl_CmdDeleteProc *) NULL);

  Tcl_CreateCommand (interp, "swapOb", mops_swapob_tcmd,
		     (ClientData) NULL, (Tcl_CmdDeleteProc *) NULL);

  Tcl_CreateCommand (interp, "insKn", mops_insertkn_ncurve_tcmd,
		     (ClientData) NULL, (Tcl_CmdDeleteProc *) NULL);

  Tcl_CreateCommand (interp, "splitNC", mops_split_ncurve_tcmd,
		     (ClientData) NULL, (Tcl_CmdDeleteProc *) NULL);

  Tcl_CreateCommand (interp, "undo", mops_undo_tcmd,
		     (ClientData) NULL, (Tcl_CmdDeleteProc *) NULL);

  Tcl_CreateCommand (interp, "saveUndo", mops_save_undo_tcmd,
		     (ClientData) NULL, (Tcl_CmdDeleteProc *) NULL);

  Tcl_CreateCommand (interp, "setTags", mops_set_tags_tcmd,
		     (ClientData) NULL, (Tcl_CmdDeleteProc *) NULL);

  Tcl_CreateCommand (interp, "getTags", mops_get_tags_tcmd,
		     (ClientData) NULL, (Tcl_CmdDeleteProc *) NULL);

  Tcl_CreateCommand (interp, "hide", mops_hide_tcmd,
		     (ClientData) NULL, (Tcl_CmdDeleteProc *) NULL);

  Tcl_CreateCommand (interp, "show", mops_show_tcmd,
		     (ClientData) NULL, (Tcl_CmdDeleteProc *) NULL);

  Tcl_CreateCommand (interp, "resolveIn", mops_resolve_instance_tcmd,
		     (ClientData) NULL, (Tcl_CmdDeleteProc *) NULL);

  Tcl_CreateCommand (interp, "convOb", mops_convert_object_tcmd,
		     (ClientData) NULL, (Tcl_CmdDeleteProc *) NULL);

  Tcl_CreateCommand (interp, "forceNot", mops_forcenot_tcmd,
		     (ClientData) NULL, (Tcl_CmdDeleteProc *) NULL);

  Tcl_CreateCommand (interp, "projectNC", mops_project_nc_tcmd,
		     (ClientData) NULL, (Tcl_CmdDeleteProc *) NULL);

  Tcl_CreateCommand (interp, "refineNC", mops_refine_curve_tcmd,
		     (ClientData) NULL, (Tcl_CmdDeleteProc *) NULL);

  Tcl_CreateCommand (interp, "elevateNC", mops_elevate_curve_tcmd,
		     (ClientData) NULL, (Tcl_CmdDeleteProc *) NULL);

  Tcl_CreateCommand (interp, "clampNC", mops_clamp_curve_tcmd,
		     (ClientData) NULL, (Tcl_CmdDeleteProc *) NULL);

  Tcl_CreateCommand (interp, "rpNC", mops_reparam_curve_tcmd,
		     (ClientData) NULL, (Tcl_CmdDeleteProc *) NULL);

  Tcl_CreateCommand (interp, "coll", mops_collapse_tcmd,
		     (ClientData) NULL, (Tcl_CmdDeleteProc *) NULL);

  Tcl_CreateCommand (interp, "expl", mops_explode_tcmd,
		     (ClientData) NULL, (Tcl_CmdDeleteProc *) NULL);


#ifndef MOPSSTANDALONE
Tcl_SetVar(interp,"MOPSSTANDALONE", "0", TCL_LEAVE_ERR_MSG |
	   TCL_GLOBAL_ONLY);
#else
Tcl_SetVar(interp, "MOPSSTANDALONE", "1", TCL_LEAVE_ERR_MSG |
	   TCL_GLOBAL_ONLY);
#endif

#ifndef MOPSWRAPPED
Tcl_SetVar(interp,"MOPSWRAPPED", "0", TCL_LEAVE_ERR_MSG |
	   TCL_GLOBAL_ONLY);
#else
Tcl_SetVar(interp, "MOPSWRAPPED", "1", TCL_LEAVE_ERR_MSG |
	   TCL_GLOBAL_ONLY);
#endif


#ifndef MOPSHAVEAFFINE
Tcl_SetVar(interp,"MOPSHAVEAFFINE", "0", TCL_LEAVE_ERR_MSG |
	   TCL_GLOBAL_ONLY);
#else
Tcl_SetVar(interp, "MOPSHAVEAFFINE", "1", TCL_LEAVE_ERR_MSG |
	   TCL_GLOBAL_ONLY);
#endif

#ifndef MOPSNOSLCARGS
Tcl_SetVar(interp,"MOPSNOSLCARGS", "0", TCL_LEAVE_ERR_MSG |
	   TCL_GLOBAL_ONLY);
#else
Tcl_SetVar(interp, "MOPSNOSLCARGS", "1", TCL_LEAVE_ERR_MSG |
	   TCL_GLOBAL_ONLY);
#endif

#ifndef MOPSIDR
Tcl_SetVar(interp,"MOPSIDR", "0", TCL_LEAVE_ERR_MSG |
	   TCL_GLOBAL_ONLY);
#else
Tcl_SetVar(interp, "MOPSIDR", "1", TCL_LEAVE_ERR_MSG |
	   TCL_GLOBAL_ONLY);
#endif

  if((mops_status = mops_init(interp)) != MOPS_OK)
    {
      Tcl_SetResult (interp,
                     "MOPS-Initialization failed!",
                     TCL_STATIC);
      return TCL_ERROR;
    }

 return TCL_OK;
} /* Tcl_AppInit */


#ifndef MOPSSTANDALONE
/********************************************/
/*main:                                     */
/********************************************/
int
main (int argc, char **argv)
{
  Tk_Main (argc, argv, Tcl_AppInit);
  return 0;
} /* main */
#endif
