#ifndef __mops_h__
#define __mops_h__

/*
 * The Mops, a free 3D Tcl/Tk/OpenGL/RenderMan modeler
 *
 * The Mops is copyrighted (C) 1998-1999 by Randolf Schultz
 * (rschultz@informatik.uni-rostock.de)
 *
 * See the file License for details.
 *
 */

#ifdef __cplusplus
extern "C" {
#endif

#include <math.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>

#ifdef WIN32
#define WIN32_LEAN_AND_MEAN
#include <windows.h>
#undef WIN32_LEAN_AND_MEAN
#endif

#include <GL/glu.h>

#include <togl.h>

#ifndef MOPSHAVEAFFINE
#include <ri.h>
#else
#include <ributil.h>
#endif

#ifndef MOPSNOSLCARGS
#include <slc.h>
#endif

/* The Object-Structure */

/* Mops Object */
typedef struct mops_object_s {
  struct mops_object_s *next;  /* next in same hierarchie-level */
  struct mops_object_s *down;

  char *name;
  unsigned int type;
  int selected;
  int modified;
  unsigned int refcount;

  /* generic attributes */
  double movx, movy, movz;
  double rotx, roty, rotz;
  float quat[4];
  double scalx, scaly, scalz;

  int colr, colg, colb, cola; 
  int opr, opg, opb, opa; 

  int avr, avg, avb, ava;
  int emr, emg, emb, ema;
  int spr, spg, spb, spa;
  
  double patch_size, elem_size, min_size;

  int zonal; /* dontset, none, zonal_receives, zonal_shoots, full_zonal */

  double shading_rate;
  int shading_interpolation; /* constant 0, smooth 1 */

  int cast_shadows; /* Os, none, opaque, surface */
  int true_displacement; /* 0 no, 1 yes */
  int d_bound; /* coordinate system */
  double d_bound_val; /* displacement bound */
  int has_caustics;

  /* visibility */
  int editor;
  int camera;
  int reflection;
  int shadow;

  /* shaders */
  struct mops_shader_s *shaders;
  
  void *object; /* type specific object */
  void *attrib; /* type specific attrib */

  /* selected points of this object */
  struct mops_point_object_s *selp;

  /* tags of this object */
  struct mops_tag_object_s *tags;
} mops_object;

/* lists and stacks of objects can be made of list objects: */
typedef struct mops_list_object_s
{
  struct mops_list_object_s *next;
  mops_object *object;
} mops_list_object;

typedef struct mops_nurbcurve_object_s
{
  int order; 
  int modified;
  int knot_type; /* bezier, bspline, nurb, custom */
  int length;
  int has_weights;
  int closed;
  double *controlv;
  double *knotv;

  GLUnurbsObj *no;
  double glu_sampling_tolerance;

  /* misc */
  int weight_handles;
  double handle_size;

  /* multiple points */
  struct mops_mpoint_object_s *mpoints;

  /* internal use only */
  int hide_curve;
} mops_nurbcurve_object;

typedef struct mops_nurbpatch_object_s
{
  int uorder; 
  int vorder; 
  int modified;
  int uknot_type; /* bezier, bspline, nurb, custom */
  int vknot_type; /* bezier, bspline, nurb, custom */
  int width, height;
  int has_weights;

  double *controlv;
  double *uknotv;
  double *vknotv;
  double *texv;

  GLUnurbsObj *no;
  double glu_sampling_tolerance;
  int glu_display_mode;
  int weight_handles;
  double handle_size;
} mops_nurbpatch_object;

typedef struct mops_light_object_s
{
  int type; /* custom, point, spot, distant */
  int on; /* 0 off, 1 on */
  int shadows; /* 0 off, 1 on */
  int samples; /* default 1 */
  int colr, colg, colb, cola;
  double intensity;
  double cone_angle;
  double cone_delta_angle;
  double beam_distribution;
} mops_light_object;

typedef struct mops_level_object_s
{
  char type;
} mops_level_object;

typedef struct mops_box_object_s
{
 double width, length, height;
} mops_box_object;

typedef struct mops_sphere_object_s
{
 char closed;
 double radius;
 double zmin, zmax;
 double thetamax;
} mops_sphere_object;

typedef struct mops_cone_object_s
{
 char closed;
 double radius;
 double height;
 double thetamax;
} mops_cone_object;

typedef struct mops_disk_object_s
{
 double radius;
 double height;
 double thetamax;
} mops_disk_object;

typedef struct mops_cylinder_object_s
{
 char closed;
 double radius;
 double zmin,zmax;
 double thetamax;
} mops_cylinder_object;

typedef struct mops_hyperboloid_s
{
 char closed;
 double p1x,p1y,p1z;
 double p2x,p2y,p2z;
 double thetamax;
} mops_hyperboloid_object;

typedef struct mops_paraboloid_object_s
{
 char closed;
 double rmax;
 double zmin,zmax;
 double thetamax;
} mops_paraboloid_object;

typedef struct mops_torus_object_s
{
 char closed;
 double majorrad, minorrad;
 double phimin, phimax;
 double thetamax;
} mops_torus_object;

typedef struct mops_custom_object_s
{
  char *type;
  unsigned int provides;
  void *object;
} mops_custom_object;


/* Functions */
typedef int (mops_createcb) (Tcl_Interp *interp, int argc, char *argv[],
			     mops_object *o);

typedef int (mops_create_custom_cb) (Tcl_Interp *interp,
				     int argc, char *argv[], void **c);

typedef int (mops_draw_custom_cb) (struct Togl *togl, mops_object *o);

typedef int (mops_delete_custom_cb) (void *c);

typedef int (mops_copy_custom_cb) (void *source, void **dest);

typedef int (mops_getprop_custom_cb) (Tcl_Interp *interp);

typedef int (mops_getpoint_custom_cb) (mops_object *o,
				       double objX, double objY, double objZ);

typedef int (mops_wrib_custom_cb) (mops_object *o);

typedef int (mops_read_custom_cb) (FILE *fileptr,
				   mops_custom_object *custom);

typedef int (mops_write_custom_cb) (FILE *fileptr,
				    mops_custom_object *custom);

typedef int (mops_notify_custom_cb) (mops_object *o);

typedef int (mops_converter_cb) (Tcl_Interp *interp, mops_object *o);

typedef int (mops_provider_cb) (mops_object *o, void **);

typedef int (mops_bbc_cb) (mops_object *o, double *bb);

/* Core object types */
#define MOPS_OTCUSTOM        0
#define MOPS_OTNURBPATCH     1
#define MOPS_OTNURBCURVE     2
#define MOPS_OTLEVEL         3
#define MOPS_OTLIGHT         4
#define MOPS_OTROOT          5
#define MOPS_OTSPHERE        6
#define MOPS_OTDISK          7
#define MOPS_OTCONE          8
#define MOPS_OTCYLINDER      9
#define MOPS_OTPARABOLOID   10
#define MOPS_OTHYPERBOLOID  11
#define MOPS_OTTORUS        12
#define MOPS_OTBOX          13
#define MOPS_OTINSTANCE     14
#define MOPS_OTATTRIB 20
#define MOPS_OTTRAFO 21
#define MOPS_OTSHADER 30
#define MOPS_OTVIEW   40
#define MOPS_OTTAG    50


/* Mops Level Types */
#define MOPS_LTUNION        0
#define MOPS_LTINTERSECTION 1
#define MOPS_LTDIFFERENCE   2
#define MOPS_LTPRIMITIVE    3
#define MOPS_LTLEVEL        4
#define MOPS_LTTRIMLOOP     5
#define MOPS_LTDOWN         6
#define MOPS_LTENDLEVEL    10

/* Mops Light Types */
#define MOPS_LITCUSTOM  0
#define MOPS_LITPOINT   1
#define MOPS_LITSPOT    2
#define MOPS_LITDIST    3

/* NURB Knot Types */
#define MOPS_KTBEZIER     0
#define MOPS_KTBSPLINE    1
#define MOPS_KTNURB       2
#define MOPS_KTCUSTOM     3

/* View Types */
#define MOPS_VTFRONT  0
#define MOPS_VTSIDE   1
#define MOPS_VTTOP    2
#define MOPS_VTPERSP  3
#define MOPS_VTTRIM   4

/* ViewInfo */
typedef struct mops_view_info_s
{
  struct mops_view_info_s *next;
  struct Togl *togl;
  int type; /* Persp, Front, Side, Top, Trim */
  double grid; /* gridsize, 0.0 == no grid */
  double bgr, bgg, bgb, bga; /* background color */
  int drawwire; /* draw wireframes */
  int drawsurf; /* draw surfaces/curves */
  int drawsel; /* draw selected objects only */
  int redraw; /* automatic redraw */
  int drawgrid; /* draw grid */
  int lighting;

  /* camera */
  double from_x, from_y, from_z;
  double to_x, to_y, to_z;
  double up_x, up_y, up_z;
  double roll, zoom;

  /* temporarily in use for rotation with cursor keys */
  double rotx, roty, rotz;

  /* how to get from win to world coordinates */
  double conv_x, conv_y;

  /* clipping rectangle, currently used to draw
     a rectangle while drag selection of points */
  double clip_xmin, clip_xmax, clip_ymin, clip_ymax;
  int draw_clip;

  /* position */
  int pos_x, pos_y;

  /* internal */
  int drawhandles;

}mops_view_info;

/* shaders */
typedef struct mops_shader_s
{
struct mops_shader_s *next;
int type; /* light volume surface displacement transformation */
int subtype; /* interior exterior atmosphere */
char *name;                   
char *loc; /* unused */
struct mops_shader_arg_s *arg;
}mops_shader;

/* shader arguments */
typedef struct mops_shader_arg_s
{
struct mops_shader_arg_s *next;
char *name;
int arraylen;
int type; /* color point vector scalar string */
 union {
  float color[4];
  float point[4];
  float scalar;
  char *string;
 } val;
} mops_shader_arg;

typedef struct mops_prefs_object_s
{
  double pick_epsilon;
  char onerror; /* 0 stop, 1 continue */
  char *ribfile;
  char *image;
  int saveviews; /* 0 do not save views */
  int readriopt; /* 0 do not save RiOptions */
  int resolveinstances;
  int checklights;
  int ristandard;
  int lazynotify;
  /* internal */
} mops_prefs_object;

/* Picking */
#define MOPS_DEFAULT_PICK_EPSILON 0.2;

typedef struct mops_riopt_object_s
{
  double Variance;
  int Samples_X;
  int Samples_Y;
  char FilterFunc;
  int FilterWidth;
  int FilterHeight;
  double ExpGain;
  double ExpGamma;
  double RGBA_ONE;
  double RGBA_MIN;
  double RGBA_MAX;
  double RGBA_Dither;

  int MinSamples;
  int MaxSamples;
  int MaxRayLevel;
  double ShadowBias;
  char PRManSpec;
  int RadSteps;
  int PatchSamples;

  char *textures;
  char *includes;
  char *shaders;

  int texturemem;
  int geommem;

  int width;
  int height;

} mops_riopt_object;


typedef struct mops_point_object_s
{
  struct mops_point_object_s *next;
  int homogenous;
  double *point;
} mops_point_object;

typedef struct mops_mpoint_object_s
{
  struct mops_mpoint_object_s *next;
 int multiplicity;
 double **points;
} mops_mpoint_object;

typedef struct mops_tag_object_s
{
 struct mops_tag_object_s *next;
 char *name;
 char *type;
 char *val;
} mops_tag_object;


/* Shader Types */
#define MOPS_STSURFACE 0
#define MOPS_STDISPLACEMENT 1
#define MOPS_STIMAGER 2
#define MOPS_STLIGHT 3
#define MOPS_STVOLUME 4
#define MOPS_STTRANSFORM 5
/* the following types are used in the subtype field
   of a shader only */
#define MOPS_STATMOSPHERE 6
#define MOPS_STEXTERIOR 7
#define MOPS_STINTERIOR 8


/* Shader Argument Types */
#define MOPS_SACOLOR 0
#define MOPS_SAPOINT 1
#define MOPS_SASCALAR 2
#define MOPS_SASTRING 3
#define MOPS_SAVECTOR 4

/* The Truth */
#define MOPS_TRUE      1
#define MOPS_FALSE     0

/* Error, Error, Error */
#define MOPS_OK           0
#define MOPS_ERROR        1
#define MOPS_WARNING      2
#define MOPS_ARGS         3
#define MOPS_OUTOFMEM     5
#define MOPS_OPENFILE    10
#define MOPS_CLOSEFILE   11
#define MOPS_NOTAMOPS    12
#define MOPS_EOF         13
#define MOPS_UEOF        14
#define MOPS_READOBJ     15
#define MOPS_SAVEOBJ     16
#define MOPS_RESOLV      20
#define MOPS_REFCOUNT    21
#define MOPS_HIER        30
#define MOPS_TYPEEXISTS  40
#define MOPS_RIBOUTERR   50
#define MOPS_NOSEL       60
#define MOPS_HASH        70

#define MOPS_OUTOFMEMSTR "Out of Memory!"
#define MOPS_ARGSSTR "Insufficient arguments, need: "
#define MOPS_OPENFILESTR "Unable to open file: "
#define MOPS_CLOSEFILESTR "Unable to close file: "
#define MOPS_NOTAMOPSSTR "Not a mops file!"
#define MOPS_NOSELSTR "No object(s) selected!"
#define MOPS_REFCOUNTSTR "Reference counter is not zero!"
#define MOPS_HASHSTR "Could not find entry in hashtable!"
#define MOPS_HIERSTR "Operation not permitted by hierarchy!"
#define MOPS_TYPEEXISTSSTR "Custom object already registered?"

#define MOPS_CRERROR if(mops_status){mops_error(interp, mops_status, fname, NULL); return TCL_OK;}

/* prototypes */

/* action.c */
int mops_movob_tcmd(ClientData clientData, Tcl_Interp * interp,
		    int argc, char *argv[]);

int mops_movsel_tcmd(ClientData clientData, Tcl_Interp * interp,
		     int argc, char *argv[]);

int mops_scalob_tcmd(ClientData clientData, Tcl_Interp * interp,
		     int argc, char *argv[]);

int mops_scalsel_tcmd(ClientData clientData, Tcl_Interp * interp,
		      int argc, char *argv[]);

int mops_rotob_tcmd(ClientData clientData, Tcl_Interp * interp,
		    int argc, char *argv[]);

int mops_rotsel_tcmd(ClientData clientData, Tcl_Interp * interp,
		     int argc, char *argv[]);

int mops_rotob_actcb(struct Togl *togl, int argc, char *argv[]);

int mops_rotoba_actcb(struct Togl *togl, int argc, char *argv[]);

int mops_movob_actcb(struct Togl *togl, int argc, char *argv[]);

int mops_sc1DXob_actcb(struct Togl *togl, int argc, char *argv[]);

int mops_sc1DYob_actcb(struct Togl *togl, int argc, char *argv[]);

int mops_sc1DZob_actcb(struct Togl *togl, int argc, char *argv[]);

int mops_sc1DXworld_actcb(struct Togl *togl, int argc, char *argv[]);

int mops_sc1DYworld_actcb(struct Togl *togl, int argc, char *argv[]);

int mops_sc1DZworld_actcb(struct Togl *togl, int argc, char *argv[]);

int mops_sc2Dob_actcb(struct Togl *togl, int argc, char *argv[]);

int mops_sc3Dob_actcb(struct Togl *togl, int argc, char *argv[]);

int mops_moview_actcb(struct Togl *togl, int argc, char *argv[]);

int mops_zoomview_actcb(struct Togl *togl, int argc, char *argv[]);

int mops_movezview_actcb(struct Togl *togl, int argc, char *argv[]);


/* bbc.c */
int mops_get_bb(mops_object *o, double *bbox);
int mops_get_sub_bb(mops_object *o, double *bbox);


/* box.c */
int mops_crtbox_tcmd(Tcl_Interp * interp,
		     int argc, char *argv[], mops_object *o);

int mops_free_box(mops_box_object *box);

int mops_copy_box(mops_box_object *source,
		  mops_box_object **dest);

int mops_draw_box_c(struct Togl *togl, mops_object *o);

int mops_draw_box(struct Togl *togl, mops_object *o);

int mops_wrib_box(mops_object *o);

int mops_get_box_properties(Tcl_Interp *interp);

int mops_set_boxattrprop_tcmd(ClientData clientData, Tcl_Interp *interp,
			      int argc, char *argv[]);

int mops_get_boxattrprop_tcmd(ClientData clientData, Tcl_Interp *interp,
			      int argc, char *argv[]);

/* clipsel.c */
int mops_set_select_tcmd(ClientData clientData, Tcl_Interp * interp,
			 int argc, char *argv[]);

int mops_free_selection(void);

int mops_add_selection(mops_object *o);

int mops_add_level_object(mops_object *new_object);

int mops_del_level_object(void);

int mops_del_all_level_objects(void);

int mops_delob_tcmd(ClientData clientData, Tcl_Interp * interp,
		    int argc, char *argv[]);

int mops_cutob_tcmd(ClientData clientData, Tcl_Interp * interp,
		    int argc, char *argv[]);

int mops_pasob_tcmd(ClientData clientData, Tcl_Interp * interp,
		    int argc, char *argv[]);

int mops_copob_tcmd(ClientData clientData, Tcl_Interp * interp,
		    int argc, char *argv[]);

/* custom.c */
int mops_init_custom_ht(void);

int mops_register_custom_objecttype(char *type_name,
				mops_create_custom_cb *customcreate_cb,
				mops_draw_custom_cb *customdraws_cb,
				mops_draw_custom_cb *customdrawh_cb,
				mops_draw_custom_cb *customdraww_cb,
				mops_delete_custom_cb *customdelete_cb,
				mops_copy_custom_cb *customcopy_cb,
				mops_getprop_custom_cb *customgetprop_cb,
				mops_getpoint_custom_cb *customgetpoint_cb,
				mops_wrib_custom_cb *customwrib_cb,
				mops_read_custom_cb *customread_cb,
				mops_write_custom_cb *customwrite_cb,
				mops_notify_custom_cb *customnotify_cb
			        );

int mops_custom_create_tcmd(Tcl_Interp * interp,
			    int argc, char *argv[], mops_object *o);

int mops_get_custom_objects_tcmd(ClientData clientData, Tcl_Interp * interp,
				 int argc, char *argv[]);

int mops_get_custom_objectname(mops_object *o, char **result);

int mops_draw_custom_w(struct Togl *togl, mops_object *o);

int mops_draw_custom_s(struct Togl *togl, mops_object *o);

int mops_draw_custom_h(struct Togl *togl, mops_object *o);

int mops_get_custom_point(mops_object *o,
			  double objX, double objY, double objZ);

int mops_get_custom_properties(Tcl_Interp *interp, mops_object *o);

int mops_free_custom(mops_custom_object *c);

int mops_copy_custom(mops_custom_object *source,
		     mops_custom_object **dest);

int mops_wrib_custom(mops_object *o);

int mops_read_custom(FILE *fileptr, mops_object *o);

int mops_write_custom(FILE *fileptr, mops_object *o);

int mops_notify_custom(mops_object *o);

int mops_register_provider(Tcl_Interp *interp, char *type,
			   unsigned int target_type,
			   mops_provider_cb *cb);

int mops_provide_object(mops_object *o, void **target);

int mops_register_bbccb(Tcl_Interp *interp, char *type, mops_bbc_cb *cb);

int mops_bb_custom(mops_object *o, double *bb);


/* error.c */
void mops_error(Tcl_Interp *interp, int code, char *where, char *what);

/* extrude.c */
int Extrude_Init(Tcl_Interp *interp);

/* hier.c */
int mops_link_new_object(mops_object *new_object);

int mops_unlink_object(mops_object *object);

int mops_clear_scene(void);

int mops_delete_object(mops_object *del);

int mops_delete_multi_objects(mops_object *del);

int mops_copy_object(mops_object *source, mops_object **dest);

int mops_set_oname_tcmd(ClientData clientData, Tcl_Interp * interp,
			int argc, char *argv[]);

int mops_get_clevel_tcmd(ClientData clientData, Tcl_Interp * interp,
			 int argc, char *argv[]);

int mops_get_objectname(mops_object *o, int typeonly, char **result);

int mops_godown_tcmd(ClientData clientData, Tcl_Interp * interp,
		     int argc, char *argv[]);

int mops_goup_tcmd(ClientData clientData, Tcl_Interp * interp,
		     int argc, char *argv[]);

int mops_gotop_tcmd(ClientData clientData, Tcl_Interp * interp,
		     int argc, char *argv[]);

int mops_swapob_tcmd(ClientData clientData, Tcl_Interp * interp,
		     int argc, char *argv[]);

/* icurve.c */
int ICurve_Init(Tcl_Interp *interp);

/* instance.c */
int mops_create_oid_ht(mops_object *o);

int mops_create_readoid_ht(mops_object *o);

int mops_connect_instances(mops_object *o);

int mops_create_oid(char **dest);

int mops_create_orig_ids(mops_object *o);

int mops_create_instance_ids(mops_object *o);

int mops_check_refcount(mops_object *o);

int mops_crtinstance_tcmd(Tcl_Interp * interp,
			  int argc, char *argv[], mops_object *o);

int mops_get_instance_properties(Tcl_Interp *interp);

int mops_wrib_instance(mops_object *o, char *file);

int mops_resolve_instance(Tcl_Interp *interp, mops_object *i);

int mops_resolve_instance_tcmd(ClientData clientData, Tcl_Interp * interp,
			       int argc, char *argv[]);


/* level.c */
int mops_crtlevel_tcmd(Tcl_Interp * interp,
		       int argc, char *argv[], mops_object *o);

int mops_free_level(mops_level_object *level);

int mops_copy_level(mops_level_object *source,
		    mops_level_object **dest);

/* lights.c */
int mops_crtlight_tcmd(Tcl_Interp * interp,
		       int argc, char *argv[], mops_object *o);

int mops_create_light(mops_light_object **lightptr);

int mops_free_light(mops_light_object *light);

int mops_copy_light(mops_light_object *source,
		    mops_light_object **dest);

int mops_get_light_point(mops_object *o, double objX, double objY,
			 double objZ);

int mops_draw_light_h(struct Togl *togl, mops_object *o);

int mops_draw_light(struct Togl *togl, mops_object *o);

int mops_get_light_properties(Tcl_Interp *interp, mops_object *o);

int mops_set_lattrprop_tcmd(ClientData clientData, Tcl_Interp * interp,
			    int argc, char *argv[]);

int mops_get_lattrprop_tcmd(ClientData clientData, Tcl_Interp * interp,
			    int argc, char *argv[]);
int mops_check_lights(mops_object *o);

int mops_wrib_lights(mops_object *o, char *file, int wrib_archive);


/* misc.c */
int mops_add_surfshader_property(Tcl_Interp *interp);

int mops_add_dispshader_property(Tcl_Interp *interp);

int mops_add_intshader_property(Tcl_Interp *interp);

int mops_add_extshader_property(Tcl_Interp *interp);

int mops_default_attributes(mops_object *o);

int mops_copy_attributes(mops_object *src, mops_object *dst);

int mops_copy_transform(mops_object *src, mops_object *dst);

void mops_get_all_transform(mops_list_object *o);

void mops_get_all_itransform(mops_list_object *o);

void mops_get_all_isrtransform(mops_list_object *o);
void mops_get_all_srtransform(mops_list_object *o);

void mops_get_all_irtransform(mops_list_object *lo);

void mops_apply_transform3(double *c, double *m);

void mops_apply_transform4(double *c, double *m);

int mops_get_version_tcmd(ClientData clientData, Tcl_Interp * interp,
			  int argc, char *argv[]);

int mops_crtob_tcmd(ClientData clientData, Tcl_Interp * interp,
		    int argc, char *argv[]);

int mops_register_crt(Tcl_Interp *interp, char *name, mops_createcb *func);

int mops_gettmp_tcmd(ClientData clientData, Tcl_Interp * interp,
		     int argc, char *argv[]);

int mops_gettype_tcmd(ClientData clientData, Tcl_Interp * interp,
		     int argc, char *argv[]);

int mops_haschild_tcmd(ClientData clientData, Tcl_Interp * interp,
		       int argc, char *argv[]);

int mops_notify_parent(void);

int mops_delegtran_tcmd(ClientData clientData, Tcl_Interp * interp,
			int argc, char *argv[]);

void mops_clear_point_selection(mops_object *o);


int mops_hide_tcmd(ClientData clientData, Tcl_Interp * interp,
		   int argc, char *argv[]);

int mops_show_tcmd(ClientData clientData, Tcl_Interp * interp,
		   int argc, char *argv[]);

int mops_convert_object_tcmd(ClientData clientData, Tcl_Interp * interp,
			     int argc, char *argv[]);

int mops_register_converter(Tcl_Interp *interp, char *type,
			    mops_converter_cb *cb);

int mops_snap_togrid(struct Togl *togl, double *winx, double *winy);

int mops_get_object(mops_object *o, unsigned int type, mops_object **dst);

int mops_force_notification(mops_object *o);

int mops_forcenot_tcmd(ClientData clientData, Tcl_Interp * interp,
		       int argc, char *argv[]);

/* mops.c */
void display_cb (struct Togl *togl);
void reshape_cb (struct Togl *togl);

void mops_render_attributes(mops_object *o);
void mops_draw_instance(struct Togl *togl, mops_object *i);
void mops_draw_object(struct Togl *togl, mops_object *o);

/* nb.c */
int mops_nb_LUDecompose(int n, double *A, int *pivot);

int mops_nb_LUInvert(int n, double *inv, int *pivot);

int mops_nb_GlobalInterpolation4D(int n, double *Q, double *ub,
				  double *Uc, int d);

int mops_nb_GlobalInterpolation4DD(int n, double *Q, double *ub,
				   double *Uc, int d, double *D1, double *D2);

int mops_nb_RemoveKnot(int stride, int n, int p, double *U, double *Pw,
		       int r, int s, int num, double *Ubar, double *Qw);

int mops_nb_DegreeElevateCurve(int stride, int n, int p, double *U,
			       double *Pw, int t, int *nh, double *Uh,
			       double *Qw);

int mops_nb_SolveTridiagonal(int n, double *Q, double *U, double *P);

int mops_nb_CurveInsertKnot4D(int np, int p, double *UP, double *Pw, double u,
			      int k, int s, int r, int *nq, double *UQ,
			      double *Qw);

int mops_nb_CurveInsertKnot3D(int np, int p, double *UP, double *P, double u,
			      int k, int s, int r, int *nq, double *UQ,
			      double *Q);

void mops_nb_BasisFuns(int i, double u, int p, double *U, double *N);

int mops_nb_FindSpan(int n, int p, double u, double *U);

int mops_nb_FindSpanMult(int n, int p, double u, double *U, int *s);

int mops_nb_CurvePoint4D(int n, int p, double *U, double *Pw, double u,
			 double *C);

int mops_nb_CurvePoint3D(int n, int p, double *U, double *Pw, double u,
			 double *C);

int mops_nb_SurfacePoint4D(int n, int m, int p, int q, double *U, double *V,
			   double *Pw, double u, double v, double *C);

int mops_nb_SurfacePoint3D(int n, int m, int p, int q, double *U, double *V,
			   double *P, double u, double v, double *C);

void mops_nb_ComputeFirstDer3D(int n, int p, double *U, double *P, double u,
			       double *CK);

void mops_nb_ComputeFirstDer4D(int n, int p, double *U, double *Pw, double u,
			       double *CK);

void mops_nb_CompFirstDerSurf3D(int n, int m, int p, int q, double *U,
				double *V, double *P, double u, double v,
				double *C);

void mops_nb_CompFirstDerSurf4D(int n, int m, int p, int q, double *U,
				double *V, double *Pw, double u, double v,
				double *C);

int mops_intersect_lines3D(double *p1, double *t1,
			   double *p2, double *t2,
			   double *p);

int mops_nb_CreateNurbsCircle(double r, double ths, double the,
			    int *length, double **knotv, double **controlv);

void mops_nb_FinduFromPoint(mops_nurbcurve_object *curve, double *point,
			    double guess, int max_try, double e1, double e2,
			    double *u);

void mops_nb_RefineKnotVectCurve(int stride, int n, int p, double *U,
				 double *Pw, double *X, int r, double *Ubar,
				 double *Qw);

/* nurbs.c */
int mops_wrib_trimcurves(mops_object *o);

int mops_wrib_npatch(mops_object *o);

int mops_create_nurbpatch_knots(mops_nurbpatch_object *patch);

int mops_create_nurbcurve_knots(mops_nurbcurve_object *curve);

int mops_check_knots(int length, int order, int knot_count, double *knotv);

int mops_draw_nurbpatch(struct Togl *togl, mops_object *o);

int mops_draw_nurbpatch_c(struct Togl *togl, mops_object *o);

int mops_draw_nurbpatch_h(struct Togl *togl, mops_object *o);

int mops_draw_nurbcurve(struct Togl *togl, mops_object *o);

int mops_draw_nurbcurve_c(struct Togl *togl, mops_object *o);

int mops_draw_nurbcurve_h(struct Togl *togl, mops_object *o);

int mops_create_nurbcurve(int order, int length, int knot_type,
			  int has_weights, double *controlv, double *knotv,
			  mops_nurbcurve_object **curveptr);

int mops_create_nurbpatch(int uorder, int vorder, int width, int height,
		      int uknot_type, int vknot_type, int has_weights,
		      double *controlv, double *uknotv, double *vknotv,
		      mops_nurbpatch_object **patchptr);

int mops_free_nurbcurve(mops_nurbcurve_object *curve);

int mops_free_nurbpatch(mops_nurbpatch_object *patch);

int mops_copy_nurbcurve(mops_nurbcurve_object *source,
			mops_nurbcurve_object **dest);

int mops_copy_nurbpatch(mops_nurbpatch_object *source,
			mops_nurbpatch_object **dest);

int mops_get_nurbcurve_point(mops_object *o, double objX, double objY,
			     double objZ);
int mops_get_nurbpatch_point(mops_object *o, double objX, double objY,
			     double objZ);

int mops_nurbcurve_close(mops_nurbcurve_object *curve);

int mops_nurbcurve_convert_weight(mops_nurbcurve_object *curve,
				  int new_weight);

int mops_nurbpatch_convert_weight(mops_nurbpatch_object *patch,
				  int new_weight);

int mops_nurbpatch_convert_width(mops_nurbpatch_object *patch,
				  int new_width);

int mops_nurbpatch_convert_height(mops_nurbpatch_object *patch,
				  int new_height);

int mops_nurbcurve_convert_length(mops_nurbcurve_object *curve,
				  int new_length);

int mops_set_ncattrprop_tcmd(ClientData clientData, Tcl_Interp * interp,
			     int argc, char *argv[]);

int mops_get_ncattrprop_tcmd(ClientData clientData, Tcl_Interp * interp,
			     int argc, char *argv[]);

int mops_set_npattrprop_tcmd(ClientData clientData, Tcl_Interp * interp,
			     int argc, char *argv[]);

int mops_get_npattrprop_tcmd(ClientData clientData, Tcl_Interp * interp,
			     int argc, char *argv[]);

int mops_get_nurbpatch_properties(Tcl_Interp *interp);

int mops_get_nurbcurve_properties(Tcl_Interp *interp);

void mops_nurbcurve_clearmp(mops_nurbcurve_object *c);

int mops_nurbcurve_collapseselp(mops_object *o);

int mops_nurbcurve_explodemp(mops_object *o);

int mops_nurbcurve_recreatemp(mops_nurbcurve_object *c);

/* nurbutil.c */

int mops_getuv_frompatch(mops_nurbpatch_object *p, double objX, double objY,
			 double objZ, double *u, double *v, double *err);

int mops_interpolate_npatchu(mops_nurbpatch_object *patch);

int mops_reparam_curve(mops_nurbcurve_object *curve, double *knotv);

int mops_clamp_curve(mops_nurbcurve_object *curve);

int mops_rescale_knotv(int n, double *knotv);

int mops_skin_nurbcurves(mops_object *curves, int order, int knot_type,
			 int interpolate, mops_nurbpatch_object **skin);

void mops_nurbcurve_findufromcontrol(int index, int nknots, int order,
				double *knots, double *u, int *k, int *s);

int mops_crtncurve_tcmd(Tcl_Interp * interp,
			int argc, char *argv[], mops_object *o);

int mops_crtnpatch_tcmd(Tcl_Interp * interp,
			int argc, char *argv[], mops_object *o);

int mops_crtnsphere2_tcmd(ClientData clientData, Tcl_Interp *interp,
		      int argc, char *argv[]);

int mops_crtncircle_tcmd(Tcl_Interp * interp,
			 int argc, char *argv[], mops_object *o);

int mops_crtnsphere_tcmd(Tcl_Interp * interp,
			 int argc, char *argv[], mops_object *o);

int mops_revolve_tcmd(Tcl_Interp * interp,
		      int argc, char *argv[], mops_object *o);

int mops_extrude_tcmd(Tcl_Interp * interp,
		      int argc, char *argv[], mops_object *o);

int mops_sweep_tcmd(Tcl_Interp * interp,
		    int argc, char *argv[], mops_object *o);

int mops_concat_nc_tcmd(Tcl_Interp * interp,
			int argc, char *argv[], mops_object *o);

int mops_create_nurbcircle(mops_nurbcurve_object **curve);

int mops_create_nurbhcircle(mops_nurbcurve_object **curve);

int mops_sweep_nurb(mops_object *o, mops_object *o2, int sections,
		    int rotate, mops_nurbpatch_object **patch);

int mops_revolve_nurb(mops_object *o, double arc,
		      mops_nurbpatch_object **patch);

int mops_extrude_nurb(double height, mops_object *o,
		      mops_nurbpatch_object **patch);

int mops_bevel_nurb(int type, double radius, mops_object *o,
		    mops_nurbpatch_object **patch);

int mops_create_cap(double z, mops_nurbcurve_object *curve,
		    double *ominx, double *omaxx,
		    double *ominy, double *omaxy, double *oangle,
		    mops_nurbpatch_object **cap);

void mops_get_nc_orientation(mops_nurbcurve_object *curve, double *oangle);

int mops_split_to_curves_tcmd(ClientData clientData, Tcl_Interp *interp,
			      int argc, char *argv[]);

int mops_build_from_curves_tcmd(ClientData clientData, Tcl_Interp *interp,
				int argc, char *argv[]);

int mops_revert_ncurve(mops_nurbcurve_object *curve);

int mops_revert_ncurve_tcmd(ClientData clientData, Tcl_Interp *interp,
			int argc, char *argv[]);

int mops_crt_trimrect_tcmd(Tcl_Interp * interp,
			   int argc, char *argv[], mops_object *o);

int mops_crt_closedbsp_tcmd(Tcl_Interp * interp,
			    int argc, char *argv[], mops_object *o);

int mops_get_ncurve_point_tcmd(ClientData clientData, Tcl_Interp *interp,
			       int argc, char *argv[]);

int mops_get_npatch_point_tcmd(ClientData clientData, Tcl_Interp *interp,
			       int argc, char *argv[]);

int mops_set_ncurve_point_tcmd(ClientData clientData, Tcl_Interp *interp,
			       int argc, char *argv[]);

int mops_set_npatch_point_tcmd(ClientData clientData, Tcl_Interp *interp,
			       int argc, char *argv[]);

int mops_insertkn_ncurve_tcmd(ClientData clientData, Tcl_Interp *interp,
			      int argc, char *argv[]);

int mops_split_ncurve(Tcl_Interp *interp, mops_object *src, double u);

int mops_split_ncurve_tcmd(ClientData clientData, Tcl_Interp *interp,
			   int argc, char *argv[]);

int mops_project_nc_tcmd(ClientData clientData, Tcl_Interp *interp,
			 int argc, char *argv[]);

int mops_refine_curve_tcmd(ClientData clientData, Tcl_Interp *interp,
			   int argc, char *argv[]);

int mops_elevate_curve_tcmd(ClientData clientData, Tcl_Interp *interp,
			  int argc, char *argv[]);

int mops_clamp_curve_tcmd(ClientData clientData, Tcl_Interp *interp,
			  int argc, char *argv[]);

int mops_reparam_curve_tcmd(ClientData clientData, Tcl_Interp *interp,
			    int argc, char *argv[]);

int mops_collapse_tcmd(ClientData clientData, Tcl_Interp *interp,
		       int argc, char *argv[]);

int mops_explode_tcmd(ClientData clientData, Tcl_Interp *interp,
		       int argc, char *argv[]);

/* prop.c */
int mops_set_prefsprop_tcmd(ClientData clientData, Tcl_Interp * interp,
			    int argc, char *argv[]);

int mops_get_prefsprop_tcmd(ClientData clientData, Tcl_Interp * interp,
			    int argc, char *argv[]);

int mops_set_transfprop_tcmd(ClientData clientData, Tcl_Interp * interp,
			     int argc, char *argv[]);

int mops_get_transfprop_tcmd(ClientData clientData, Tcl_Interp * interp,
			     int argc, char *argv[]);

int mops_set_attrprop_tcmd(ClientData clientData, Tcl_Interp * interp,
			     int argc, char *argv[]);

int mops_get_attrprop_tcmd(ClientData clientData, Tcl_Interp * interp,
			     int argc, char *argv[]);

int mops_set_rioptprop_tcmd(ClientData clientData, Tcl_Interp * interp,
			     int argc, char *argv[]);

int mops_get_rioptprop_tcmd(ClientData clientData, Tcl_Interp * interp,
			     int argc, char *argv[]);

int mops_get_root_properties(Tcl_Interp *interp);

int mops_get_level_properties(Tcl_Interp *interp);

int mops_get_view_properties(Tcl_Interp *interp, mops_object *o);


/* quadric.c */
int mops_crtquad_tcmd(Tcl_Interp * interp,
		      int argc, char *argv[], mops_object *o);

int mops_free_quadric(void *quadric, int type);

int mops_copy_quadric(void *source,
		      void **dest, int type);

int mops_draw_sphere_c(struct Togl *togl, mops_object *o);

int mops_draw_sphere(struct Togl *togl, mops_object *o);

int mops_draw_disk_c(struct Togl *togl, mops_object *o);

int mops_draw_disk(struct Togl *togl, mops_object *o);

int mops_draw_cone_c(struct Togl *togl, mops_object *o);

int mops_draw_cone(struct Togl *togl, mops_object *o);

int mops_draw_torus_c(struct Togl *togl, mops_object *o);

int mops_draw_torus(struct Togl *togl, mops_object *o);

int mops_draw_cylinder_c(struct Togl *togl, mops_object *o);

int mops_draw_cylinder(struct Togl *togl, mops_object *o);

int mops_draw_hyperboloid_c(struct Togl *togl, mops_object *o);

int mops_draw_hyperboloid(struct Togl *togl, mops_object *o);

int mops_draw_paraboloid_c(struct Togl *togl, mops_object *o);

int mops_draw_paraboloid(struct Togl *togl, mops_object *o);

int mops_wrib_sphere(mops_object *o);

int mops_wrib_disk(mops_object *o);

int mops_wrib_cone(mops_object *o);

int mops_wrib_cylinder(mops_object *o);

int mops_wrib_torus(mops_object *o);

int mops_wrib_hyperboloid(mops_object *o);

int mops_wrib_paraboloid(mops_object *o);

int mops_get_quadric_properties(Tcl_Interp *interp, int type);

int mops_set_qattrprop_tcmd(ClientData clientData, Tcl_Interp *interp,
			    int argc, char *argv[]);

int mops_get_qattrprop_tcmd(ClientData clientData, Tcl_Interp *interp,
			    int argc, char *argv[]);

/* quat.c */
void mops_axis_to_quat(float a[3], float phi, float q[4]);

void mops_add_quat(float q1[4], float q2[4], float dest[4]);

void mops_norm_quat(float q[4]);

void mops_build_rotmatrix(float m[4][4], float q[4]);

void mops_quat_to_euler(float q[4], float euler[3]);

/* read.c */
int mops_read_string(FILE *fileptr, char **result);

int mops_read_light(FILE *fileptr, mops_object *o);

int mops_read_box(FILE *fileptr, mops_object *o);

int mops_read_sphere(FILE *fileptr, mops_object *o);

int mops_read_disk(FILE *fileptr, mops_object *o);

int mops_read_cone(FILE *fileptr, mops_object *o);

int mops_read_cylinder(FILE *fileptr, mops_object *o);

int mops_read_hyperboloid(FILE *fileptr, mops_object *o);

int mops_read_paraboloid(FILE *fileptr, mops_object *o);

int mops_read_torus(FILE *fileptr, mops_object *o);

int mops_read_nurbpatch(FILE *fileptr, mops_object *o);

int mops_read_nurbcurve(FILE *fileptr, mops_object *o);

int mops_insert_scene_tcmd(ClientData clientData, Tcl_Interp * interp,
			   int argc, char *argv[]);

int mops_replace_scene_tcmd(ClientData clientData, Tcl_Interp * interp,
			    int argc, char *argv[]);

int mops_new_scene_tcmd(ClientData clientData, Tcl_Interp * interp,
			int argc, char *argv[]);


/* revolve.c */
int Revolve_Init(Tcl_Interp *interp);

/* shader.c */
int mops_scan_shader_tcmd(ClientData clientData, Tcl_Interp * interp,
			  int argc, char *argv[]);

int mops_copy_shader_arg(mops_shader_arg *source,
			 mops_shader_arg **dest);

int mops_copy_shader(mops_shader *source,
		     mops_shader **dest);

int mops_add_shader(mops_object *o,
		    mops_shader *new);

int mops_free_shader(mops_shader *shader);

int mops_wrib_shaders(mops_object *o);

int mops_set_shaderprop_tcmd(ClientData clientData, Tcl_Interp * interp,
			     int argc, char *argv[]);

int mops_get_shaderprop_tcmd(ClientData clientData, Tcl_Interp * interp,
			     int argc, char *argv[]);

int mops_del_shader_tcmd(ClientData clientData, Tcl_Interp * interp,
			 int argc, char *argv[]);

/* skin.c */
int Skin_Init(Tcl_Interp *interp);

/* sweep.c */
int Sweep_Init(Tcl_Interp *interp);

/* tags.c */
int mops_register_tagtype(Tcl_Interp *interp, char *name, char **result);

int mops_set_tags_tcmd(ClientData clientData, Tcl_Interp * interp,
		       int argc, char *argv[]);

int mops_get_tags_tcmd(ClientData clientData, Tcl_Interp * interp,
		       int argc, char *argv[]);

int mops_copy_tag(mops_tag_object *source,
		  mops_tag_object **dest);

int mops_free_tag(mops_tag_object *tag);

/* undo.c */
int mops_clear_undo_info(void);

int mops_save_undo_tcmd(ClientData clientData, Tcl_Interp * interp,
		    int argc, char *argv[]);

int mops_undo_tcmd(ClientData clientData, Tcl_Interp * interp,
		   int argc, char *argv[]);

/* wrib.c */
int mops_wrib_cb(struct Togl *togl, int argc, char *argv[]);

int mops_wrib_scene(char *file, char *image,
		    double fromx, double fromy, double fromz,
		    double tox, double toy, double toz,
		    double roll, double zoom,
		    int width, int height, int type);

int mops_wrib_attributes(mops_object *o);

int mops_wrib_trafos(mops_object *o);

int mops_wrib_object(mops_object *o, char *file);

/* write.c */
int mops_write_light(FILE *fileptr, mops_object *o);

int mops_write_box(FILE *fileptr, mops_object *o);

int mops_write_sphere(FILE *fileptr, mops_object *o);

int mops_write_disk(FILE *fileptr, mops_object *o);

int mops_write_cone(FILE *fileptr, mops_object *o);

int mops_write_cylinder(FILE *fileptr, mops_object *o);

int mops_write_hyperboloid(FILE *fileptr, mops_object *o);

int mops_write_paraboloid(FILE *fileptr, mops_object *o);

int mops_write_torus(FILE *fileptr, mops_object *o);

int mops_write_nurbpatch(FILE *fileptr, mops_object *o);

int mops_write_nurbcurve(FILE *fileptr, mops_object *o);

int mops_write_object(FILE *fileptr, mops_object *o);

int mops_save_scene_tcmd(ClientData clientData, Tcl_Interp * interp,
			 int argc, char *argv[]);


/* global variables/pointers */
extern Tcl_Interp *mops_interp;
extern mops_prefs_object mops_prefs;
extern mops_riopt_object mops_riopt;
extern mops_object *mops_root;
extern mops_object **mops_next_object;

extern mops_list_object *mops_current_selection;
extern mops_list_object *mops_last_selection;

extern mops_list_object *mops_current_level;
extern mops_list_object *mops_view_list;

extern mops_object *mops_object_clipboard;

extern mops_object *mops_undo_objects;
extern mops_list_object *mops_undo_list;

extern int mops_point_edit_coords_homogenous;
extern int mops_point_edit_coords_number;
extern double **mops_point_edit_coords;

extern Tcl_HashTable mops_createcb_ht;

extern Tcl_HashTable mops_tagtypes_ht;

extern Tcl_HashTable mops_oidptr_ht;

extern Tcl_HashTable mops_converter_ht;
extern Tcl_HashTable mops_provider_ht;

extern char *mops_riattr_tagtype;
extern char *mops_oi_tagtype;
extern Tcl_HashTable mops_custom_id_ht;

extern int mops_errno;

extern mops_object *mops_last_read_object;

extern int mops_read_version;

extern RtLightHandle mops_light_handle;

extern char mops_version_ma[];
extern char mops_version_mi[];

extern RtVoid Subdivide(RtPointer data, RtFloat detail);
extern RtVoid Free(RtPointer data);
/* Basic vmath */
#define MOPS_VLEN(x, y, z) sqrt((x*x)+(y*y)+(z*z))

#define MOPS_V3LEN(v) sqrt((v[0]*v[0])+(v[1]*v[1])+(v[2]*v[2]))

#define MOPS_V2LEN(v) sqrt((v[0]*v[0])+(v[1]*v[1]))

#define MOPS_V3ZERO(v) {v[0]=0.0; v[1] = 0.0; v[2]=0.0}

#define MOPS_V3SUB(r,v1,v2) {r[0]=v1[0]-v2[0];r[1]=v1[1]-v2[1];r[2]=v1[2]-v2[2];}
#define MOPS_V3ADD(r,v1,v2) {r[0]=v1[0]+v2[0];r[1]=v1[1]+v2[1];r[2]=v1[2]+v2[2];}
#define MOPS_V3MUL(r,v1,v2) {r[0]=v1[0]*v2[0];r[1]=v1[1]*v2[1];r[2]=v1[2]*v2[2];}
#define MOPS_V3CROSS(r,v1,v2) {r[0]=(v1[1] * v2[2]) - (v1[2] * v2[1]);r[1]=(v1[2] * v2[0]) - (v1[0] * v2[2]);r[2]=(v1[0] * v2[1]) - (v1[1] * v2[0]);}

#define MOPS_V3SCAL(v,f) {(v[0])*=(f);(v[1])*=(f);(v[2])*=(f);}

#define MOPS_V3NORM(v) {MOPS_V3SCAL((v),(1.0/(MOPS_V3LEN(v))));}

#define MOPS_V3DOT(v1,v2) (v1[0]*v2[0] + v1[1]*v2[1] + v1[2]*v2[2])

#define MOPS_V2DOT(v1,v2) (v1[0]*v2[0] + v1[1]*v2[1])

#define MOPS_PI 3.1415926535897932384626433

#define MOPS_D2R(x) ((x)*MOPS_PI/180.0)

#define MOPS_R2D(x) ((x)*180.0/MOPS_PI)

/* Warning: v1 and v2 must be different locations in memory! */
#define MOPS_APTRAN3(v1,v2,m) {v1[0]=v2[0]*m[0]+v2[1]*m[4]+v2[2]*m[8]+1.0*m[12];v1[1]=v2[0]*m[1]+v2[1]*m[5]+v2[2]*m[9]+1.0*m[13];v1[2]=v2[0]*m[2]+v2[1]*m[6]+v2[2]*m[10]+1.0*m[14];}

/* Warning: v1 and v2 must be different locations in memory! */
#define MOPS_APTRAN4(v1,v2,m) {v1[0]=v2[0]*m[0]+v2[1]*m[4]+v2[2]*m[8]+v2[3]*m[12];v1[1]=v2[0]*m[1]+v2[1]*m[5]+v2[2]*m[9]+v2[3]*m[13];v1[2]=v2[0]*m[2]+v2[1]*m[6]+v2[2]*m[10]+v2[3]*m[14];v1[3]=v2[0]*m[3]+v2[1]*m[7]+v2[2]*m[11]+v2[3]*m[15];}

#define MOPS_POINTER 8.0

#define MOPS_QRENORMCOUNT 11


/* standard colors */
#define MOPS_CLGRID (0.6,0.6,0.6)
#define MOPS_CLSEL (1.0,1.0,1.0)
#define MOPS_CLUSEL (0.0,0.1,0.45)
#define MOPS_CLSELP (0.5,0.0,0.0)

#define MOPS_VERSIONSTR "0.42d"
#define MOPS_VERSIONSTRMI "0"

#ifdef __cplusplus
}
#endif


#endif                  /* __mops_h__ */

