/*
 * The Mops, a free 3D Tcl/Tk/OpenGL/RenderMan modeler
 *
 * The Mops is copyrighted (C) 1998-1999 by Randolf Schultz
 * (rschultz@informatik.uni-rostock.de)
 *
 * See the file License for details.
 *
 */

#include <limits.h>
#include <float.h>
#include "mops.h"

/* NURBS */

/* functions local to this module */
int mops_draw_trimcurve(struct Togl *togl, mops_object *co, GLUnurbsObj *no);

void mops_print_gluerr(GLenum err);

/********************************************/
/* mops_print_gluerr:
   GLU error callback, write error to stderr
   or console, if available                 */
/********************************************/
void
mops_print_gluerr(GLenum err)
{
 char fname[] = "draw_nurb";

  if(mops_interp)
    mops_error(mops_interp, MOPS_ERROR, fname, (char *)gluErrorString(err));
  else
    fprintf(stderr, "%s\n", gluErrorString(err));

 return;
}


/********************************************/
/* mops_create_nurbpatch_knots:
   create knot vectors according to
   the knot_type fields for patch *patch    */
/********************************************/
int
mops_create_nurbpatch_knots(mops_nurbpatch_object *patch)
{
 int uorder = 0, vorder = 0, width = 0, height = 0;
 int uknot_count = 0, vknot_count = 0;
 int i = 0, j = 0, kts = 0;
 int start = 0;
 double *newuknotv = NULL, *newvknotv = NULL;

  uorder = patch->uorder;
  vorder = patch->vorder;
  width = patch->width;
  height = patch->height;
  uknot_count = width + uorder;
  vknot_count = height + vorder;

  /* calloc new knot-arrays */
  if(!(newuknotv = calloc(uknot_count, sizeof(double))))
    return MOPS_OUTOFMEM;
  if(!(newvknotv = calloc(vknot_count, sizeof(double))))
    { free(newuknotv); return MOPS_OUTOFMEM; }

  /* free old knot-arrays */
  if(patch->uknotv != NULL) free(patch->uknotv);
  patch->uknotv = newuknotv;
  if(patch->vknotv != NULL) free(patch->vknotv);
  patch->vknotv = newvknotv;

  /* fill knot-arrays */
  switch(patch->uknot_type)
    {
    case MOPS_KTBEZIER:
      for(i=0; i<uknot_count/2; i++)
	(patch->uknotv)[i] = 0.0;
      for(i=uknot_count/2; i<uknot_count; i++)
	(patch->uknotv)[i] = 1.0;
      break;

    case MOPS_KTBSPLINE:
      j=0;
      start = floor(uknot_count/2);
      for(i=-start; i<start; i++)
	{
	  (patch->uknotv)[j] = i/fabs((double)(uknot_count-(uorder-1)));
	  j++;
	}
      if(fmod(uknot_count,2.0) > 0.0)
	(patch->uknotv)[j] = i/fabs((double)(uknot_count-(uorder-1)));
      break;

    case MOPS_KTNURB:
      for(i=0; i<uorder; i++)
	(patch->uknotv)[i] = 0.0;
      j=1;
      kts = 1 + uknot_count - (uorder*2);
      for(i=uorder; i<=uknot_count-uorder; i++)
	(patch->uknotv)[i] = j++/(double)kts;
      for(i=uknot_count-uorder; i<uknot_count; i++)
	(patch->uknotv)[i] = 1.0;
      break;

    case MOPS_KTCUSTOM:
      /* user specified own knot vertices */
      break;
    }

  switch(patch->vknot_type)
    {
    case MOPS_KTBEZIER:
      for(i=0; i<vknot_count/2; i++)
	(patch->vknotv)[i] = 0.0;
      for(i=vknot_count/2; i<vknot_count; i++)
	(patch->vknotv)[i] = 1.0;
      break;

    case MOPS_KTBSPLINE:
      j=0;
      start = floor(vknot_count/2);
      for(i=-start; i<start; i++)
	{
	  (patch->vknotv)[j] = i/fabs((double)(vknot_count-(vorder-1)));
	  j++;
	}
      if(fmod(vknot_count,2.0) > 0.0)
	(patch->vknotv)[j] = i/fabs((double)(vknot_count-(vorder-1)));
      break;

    case MOPS_KTNURB:
      for(i=0; i<vorder; i++)
	(patch->vknotv)[i] = 0.0;
      j=1;
      kts = 1 + vknot_count - (vorder*2);
      for(i=vorder; i<=vknot_count-vorder; i++)
	(patch->vknotv)[i] = j++/((double)kts);
      for(i=vknot_count-vorder; i<vknot_count; i++)
	(patch->vknotv)[i] = 1.0;
      break;

    case MOPS_KTCUSTOM:
      /* user specified own knot vertices */
      break;
    }

 return MOPS_OK;
}


/********************************************/
/* mops_create_nurbcurve_knots:
                                            */
/********************************************/
int
mops_create_nurbcurve_knots(mops_nurbcurve_object *curve)
{
 int order = 0, length = 0, knot_count = 0;
 int i = 0, j = 0, kts = 0;
 double *newknotv = NULL;

  order = curve->order;
  length = curve->length;

  knot_count = length + order;

  /* calloc new knot-arrays */
  if((newknotv = calloc(knot_count, sizeof(double))) == NULL)
    return MOPS_OUTOFMEM;

  /* free old knot-array */
  if(curve->knotv) free(curve->knotv);
  curve->knotv = newknotv;


  /* fill knot-arrays */
  switch(curve->knot_type)
    {
    case MOPS_KTBEZIER:
      for(i=0; i<knot_count/2; i++)
	(curve->knotv)[i] = 0.0;
      for(i=knot_count/2; i<knot_count; i++)
	(curve->knotv)[i] = 1.0;
      break;

    case MOPS_KTBSPLINE:
      j=0;
      /*
      start = floor(knot_count/2);
      for(i=-start; i<start; i++)
	{
	  (curve->knotv)[j] = i/fabs((double)(knot_count-(order-1)));
	  j++;
	}
      if(fmod(knot_count,2.0) > 0.0)
	(curve->knotv)[j] = i/fabs((double)(knot_count-(order-1)));
      */
      for(i = 0; i < knot_count; i++)
	{
	  (curve->knotv)[i] = (double)i/knot_count;

	}
      break;

    case MOPS_KTNURB:
      for(i=0; i<order; i++)
	(curve->knotv)[i] = 0.0;
      j=1;
      kts = 1 + knot_count - (order*2);
      for(i=order; i<=knot_count-order; i++)
	(curve->knotv)[i] = j++/((double)kts);
      for(i=knot_count-order; i<knot_count; i++)
	(curve->knotv)[i] = 1.0;
      break;

    case MOPS_KTCUSTOM:
      /* user specified own knot vertices */
      break;
    }

 return MOPS_OK;
}


/********************************************/
/* mops_check_knots:
   returns:
   o 0 - the knots are ok
   o 1 - too few knots
   o 2 - too much knots
   o 3 - knot multiplicity too high
   o 4 - decreasing knots                   */
/********************************************/
int
mops_check_knots(int length, int order, int knot_count, double *knotv)
{
 int i, mult_count = 1;

  if(knot_count < (length+order)) 
    return 1; 

  if(knot_count > (length+order)) 
    return 2; 

  for(i = 0; i < (knot_count-1); i++)
    {
      if(knotv[i] == knotv[i+1])
	{
	  mult_count++;
	  if(mult_count > order)
	    return 3;
	}
      else
	{
	  mult_count = 1;

	  if(knotv[i] > knotv[i+1])
	    return 4;
	}
    }

 return 0;
}


/********************************************/
/* mops_draw_trimcurve:
                                            */
/********************************************/
int
mops_draw_trimcurve(struct Togl *togl, mops_object *o, GLUnurbsObj *no)
{
 int order = 0, length = 0, knot_count = 0, i = 0, a = 0, b = 0;
 mops_nurbcurve_object *curve = (mops_nurbcurve_object *) o->object;
 static GLfloat *knots = NULL, *controls = NULL;
 GLdouble m[16] = {0}, x = 0.0, y = 0.0, w = 1.0;
 GLfloat mr[4][4];

  /* get curves transformation-matrix */
  glMatrixMode (GL_MODELVIEW);
  glPushMatrix();
  glLoadIdentity ();
  glTranslated(o->movx, o->movy, o->movz);

  mops_build_rotmatrix(mr, o->quat);
  glMultMatrixf(&mr[0][0]);

  glScaled (o->scalx, o->scaly, o->scalz);
  glGetDoublev(GL_MODELVIEW_MATRIX, m);
  glPopMatrix();

  if(controls)
    {
      free(controls);
      controls = NULL;
    }
  if(knots)
    {
      free(knots);
      knots = NULL;
    }

  order = curve->order;
  length = curve->length;

  knot_count = length + order;

  if((knots = calloc(knot_count, sizeof(GLfloat))) == NULL)
    return MOPS_OUTOFMEM;
  if((controls = calloc(length*3, sizeof(GLfloat))) == NULL)
    { free(knots); knots = NULL; return MOPS_OUTOFMEM; }

  a=0;
  for(i=0;i<knot_count;i++)
    {
      knots[a] = (GLfloat)curve->knotv[a];
      a++;
    }
  a=0; b=0;
  for(i=0;i<length;i++)
    {
      x = (GLdouble)curve->controlv[b];b++;
      y = (GLdouble)curve->controlv[b];b++;

      b++;
      if(curve->has_weights)
	{w = (GLfloat)curve->controlv[b];b++;}
      else
	w = 1.0;

      controls[a] = m[0]*x + m[4]*y + m[12]*w;
      controls[a+1] = m[1]*x + m[5]*y + m[13]*w;
      controls[a+2] = m[3]*x + m[7]*y +	m[15]*w;
      a+=3;
    }

  gluNurbsCurve(no, (GLint)knot_count, knots,
		  (GLint)3, controls,
		  (GLint)curve->order, GLU_MAP1_TRIM_3);

 return MOPS_OK;
}


/********************************************/
/* mops_draw_nurbpatch:
                                            */
/********************************************/
int
mops_draw_nurbpatch(struct Togl *togl, mops_object *o)
{
 int mops_status = MOPS_OK;
 int uorder = 0, vorder = 0, width = 0, height = 0;
 int uknot_count = 0, vknot_count = 0;
 mops_nurbpatch_object *patch = (mops_nurbpatch_object *) o->object;
 mops_object *trim = NULL, *loop = NULL;
 GLdouble sampling_tolerance = 30.0;
 static GLfloat *uknots = NULL, *vknots = NULL, *controls = NULL;
 int i=0,j=0,a=0;


  if(controls)
    {
      free(controls);
      controls = NULL;
    }
  if(uknots)
    {
      free(uknots);
      uknots = NULL;
    }
  if(vknots)
    {
      free(vknots);
      vknots = NULL;
    }

  uorder = patch->uorder;
  vorder = patch->vorder;
  width = patch->width;
  height = patch->height;

  sampling_tolerance = patch->glu_sampling_tolerance;

  uknot_count = width + uorder;
  vknot_count = height + vorder;

  if((uknots = calloc(uknot_count, sizeof(GLfloat))) == NULL)
    return MOPS_OUTOFMEM; 
  if((vknots = calloc(vknot_count, sizeof(GLfloat))) == NULL)
    { free(uknots); uknots = NULL; return MOPS_OUTOFMEM; }
  if(patch->has_weights)
   {
     if((controls = calloc(width*height*4, sizeof(GLfloat))) == NULL)
       {
	 free(uknots); uknots = NULL;
	 free(vknots); vknots = NULL;
	 return MOPS_OUTOFMEM;
       }
   }
  else
    {
      if((controls = calloc(width*height*3, sizeof(GLfloat))) == NULL)
       {
	 free(uknots); uknots = NULL;
	 free(vknots); vknots = NULL;
	 return MOPS_OUTOFMEM;
       }
    }

  a=0;
  for(i=0;i<uknot_count;i++)
    {
      uknots[a] = (GLfloat)patch->uknotv[a];
      a++;
    }
  a=0;
  for(i=0;i<vknot_count;i++)
    {
      vknots[a] = (GLfloat)patch->vknotv[a];
      a++;
    }
  a=0;
  for(i=0;i<width;i++)
    {
  for(j=0;j<height;j++)
    {
      controls[a] = (GLfloat)patch->controlv[a]; 
      a++;
      controls[a] = (GLfloat)patch->controlv[a]; 
      a++;
      controls[a] = (GLfloat)patch->controlv[a]; 
      a++;
      if(patch->has_weights)
	{
	  controls[a] = (GLfloat)patch->controlv[a]; 
	  a++;
	}
    }
    }
  if(!patch->no)
    {
      patch->no = gluNewNurbsRenderer();
      if(patch->no == NULL)
       {
	 free(uknots); uknots = NULL;
	 free(vknots); vknots = NULL;
	 free(controls); controls = NULL;
	 return MOPS_OUTOFMEM;
       }
    }

#ifdef WIN32
  gluNurbsCallback(patch->no, GLU_ERROR, (GLUnurbsErrorProc)mops_print_gluerr);
#else
  gluNurbsCallback(patch->no, GLU_ERROR, mops_print_gluerr);
#endif

  gluBeginSurface(patch->no);

  gluNurbsProperty(patch->no, GLU_SAMPLING_TOLERANCE, sampling_tolerance);

  if(!patch->glu_display_mode)
    gluNurbsProperty(patch->no, GLU_DISPLAY_MODE, GLU_FILL);
  else
    if(patch->glu_display_mode == 1)
      gluNurbsProperty(patch->no, GLU_DISPLAY_MODE, GLU_OUTLINE_POLYGON);
    else
      gluNurbsProperty(patch->no, GLU_DISPLAY_MODE, GLU_OUTLINE_PATCH);

  if(patch->has_weights)
    gluNurbsSurface(patch->no, (GLint)uknot_count, uknots,
		    (GLint)vknot_count, vknots,
		    (GLint)height*4, (GLint)4, controls,
		    (GLint)(patch->uorder), (GLint)(patch->vorder),
		    GL_MAP2_VERTEX_4);
  else
    gluNurbsSurface(patch->no, (GLint)uknot_count, uknots,
		    (GLint)vknot_count, vknots,
		    (GLint)height*3, (GLint)3, controls,
		    (GLint)patch->uorder, (GLint)patch->vorder,
		    GL_MAP2_VERTEX_3);


  if(o->down)
    {
      trim = o->down;

      while(trim)
	{
	  if(trim->type == MOPS_OTNURBCURVE)
	    {
	      gluBeginTrim(patch->no);
	      mops_status = mops_draw_trimcurve(togl, trim, patch->no);
	      gluEndTrim(patch->no);
	    }

	  if(trim->type == MOPS_OTLEVEL)
	    { /* XXXX check, whether level is of type trimloop? */
	      loop = trim->down;
	      if(loop)
		{
		  gluBeginTrim(patch->no);
		  while(loop)
		    {
		      if(loop->type == MOPS_OTNURBCURVE)
			mops_status = mops_draw_trimcurve(togl,
							  loop, patch->no);
		      loop = loop->next;
		    }
		  gluEndTrim(patch->no);
		}
	    }
	 
	  trim = trim->next;
	}
      

      }
      
  gluEndSurface(patch->no);

 return MOPS_OK;
}


/********************************************/
/* mops_draw_nurbpatch_c:
   draw only the control-polygon of a NURB
   patch as wireframe                       */
/********************************************/
int
mops_draw_nurbpatch_c(struct Togl *togl, mops_object *o)
{
 int width = 0, height = 0;
 int i = 0, j = 0, a = 0;
 mops_nurbpatch_object *patch = (mops_nurbpatch_object *) o->object;
 GLdouble *ver;

  width = patch->width;
  height = patch->height;

  ver = patch->controlv;
  a = 0;
  for(i=0; i<width; i++)
    {
      glBegin(GL_LINE_STRIP);
	for(j=0; j<height; j++)
	  {
	    glVertex3dv((GLdouble *)&ver[a]);
	    if(patch->has_weights)
	      a+=4;
	    else
	      a+=3;
	    
	  }
      glEnd();
    }
  
  a = 0.0;
  for(j=0; j<height; j++)
    {
      if(patch->has_weights)
	a = j * 4;
      else
	a = j * 3;
      
      glBegin(GL_LINE_STRIP);
	for(i=0; i<width; i++)
	  {
	    glVertex3dv((GLdouble *)&ver[a]);
	    if(patch->has_weights)
	      a+=4 * height;
	    else
	      a+=3 * height;

	  }
      glEnd();
    }

 return MOPS_OK;
}


/********************************************/
/* mops_draw_nurbpatch_h:
   draws only handles (for the vertices of
   the control-polygon) of a NURB patch     */
/********************************************/
int
mops_draw_nurbpatch_h(struct Togl *togl, mops_object *o)
{
 int width = 0, height = 0;
 int i = 0, j = 0, a = 0;
 double w = 0.0;
 mops_nurbpatch_object *patch = (mops_nurbpatch_object *) o->object;
 GLdouble *ver, point_size = 6.0;
 GLUquadricObj *qobj = NULL;
 GLdouble mvm[16], pm[16], winx, winy, winz;
 GLint vp[4];

  width = patch->width;
  height = patch->height;
  point_size = patch->handle_size;
  ver = patch->controlv;

  if(!patch->weight_handles)
    {
      a = 0;
      glPointSize(point_size);
      glBegin(GL_POINTS);
      for(i=0; i<width; i++)
	{
	  for(j=0; j<height; j++)
	    {
	      glVertex3dv((GLdouble *)&ver[a]);
	      if(patch->has_weights)
		a+=4;
	      else
		a+=3;
	    }
	}
      glEnd();  
    }
  else
    { /* draw with weights */
      if(patch->has_weights)
	{
	  glGetDoublev(GL_MODELVIEW_MATRIX, mvm);
	  glGetDoublev(GL_PROJECTION_MATRIX, pm);
	  glGetIntegerv(GL_VIEWPORT, vp);

	  glMatrixMode(GL_PROJECTION);
	  /*	  glPushMatrix();*/
	  glLoadIdentity();
	  glOrtho(0, Togl_Width(togl), 0, Togl_Height(togl), -100.0, 100.0);

	  glMatrixMode(GL_MODELVIEW);
	  /* glPushMatrix();*/
	  glLoadIdentity();

  	  if(!(qobj = gluNewQuadric()))
	    return MOPS_OUTOFMEM;

	  for(i=0; i<width; i++)
	    {
	      for(j=0; j<height; j++)
		{
		  
		  w = ver[a+3];
		  gluProject((GLdouble)ver[a], (GLdouble)ver[a+1],
			     (GLdouble)ver[a+2], mvm, pm, vp,
			     &winx, &winy, &winz);

		  /* XXXX picking? */
		  glTranslated(winx,winy,winz);
		
		  point_size = patch->handle_size /** 0.033*/;
		  if(w<1.0)
		    {
		      point_size /= w;
		      gluDisk(qobj,point_size-(point_size/10),point_size,8,1);
		    }
		  else
		    {
		      point_size *= w;
		      gluDisk(qobj,0.0,point_size,8,1);
		    }
		  glTranslated(-winx,-winy,-winz);
		
		  a+=4;
		} /* for */
	    } /* for */

	  gluDeleteQuadric(qobj);

	  /*	  glPopMatrix();*/
	  glLoadMatrixd(mvm);

	  glMatrixMode(GL_PROJECTION);
	  glLoadMatrixd(pm);
	  glMatrixMode(GL_MODELVIEW);

	}
    }

 return MOPS_OK;
}


/********************************************/
/* mops_draw_nurbcurve:
                                            */
/********************************************/
int
mops_draw_nurbcurve(struct Togl *togl, mops_object *o)
{
 int order = 0, length = 0, knot_count = 0, i = 0, a = 0;
 mops_nurbcurve_object *curve = (mops_nurbcurve_object *) o->object;
 GLdouble sampling_tolerance = 30.0;
 static GLfloat *knots = NULL, *controls = NULL;

  if(curve->hide_curve)
    return MOPS_OK;

  if(controls)
    {
      free(controls);
      controls = NULL;
    }
  if(knots)
    {
      free(knots);
      knots = NULL;
    }
  order = curve->order;
  length = curve->length;

  sampling_tolerance = curve->glu_sampling_tolerance;

  knot_count = length + order;

  if((knots = calloc(knot_count, sizeof(GLfloat))) == NULL)
    return MOPS_OUTOFMEM;
  if((controls = calloc(length*4, sizeof(GLfloat))) == NULL)
    { free(knots); knots = NULL; return MOPS_OUTOFMEM; }

  a=0;
  for(i=0;i<knot_count;i++)
    {
      knots[a] = (GLfloat)curve->knotv[a];
      a++;
    }
  a=0;
  for(i=0;i<length;i++)
    {
      controls[a] = (GLfloat)curve->controlv[a]; 
      a++;
      controls[a] = (GLfloat)curve->controlv[a]; 
      a++;
      controls[a] = (GLfloat)curve->controlv[a]; 
      a++;
      if(curve->has_weights)
	{
	  controls[a] = (GLfloat)curve->controlv[a]; 
	  a++;
	}
    }

  if(!curve->no)
    {
      curve->no = gluNewNurbsRenderer();
      if(curve->no == NULL)
	{ 
	  free(knots); knots = NULL; 
	  free(controls); controls = NULL; 
	  return MOPS_OUTOFMEM;
	}
    }

#ifdef WIN32
  gluNurbsCallback(curve->no, GLU_ERROR, (GLUnurbsErrorProc)mops_print_gluerr);
#else
  gluNurbsCallback(curve->no, GLU_ERROR, mops_print_gluerr);
#endif

  gluBeginCurve(curve->no);
  gluNurbsProperty(curve->no, GLU_SAMPLING_TOLERANCE, sampling_tolerance);

  if(curve->has_weights)
    gluNurbsCurve(curve->no, (GLint)knot_count, knots,
		    (GLint)4, controls,
		    (GLint)curve->order, GL_MAP1_VERTEX_4);
  else
    gluNurbsCurve(curve->no, (GLint)knot_count, knots,
		    (GLint)3, controls,
		    (GLint)curve->order, GL_MAP1_VERTEX_3);
  
  gluEndCurve(curve->no);

 return MOPS_OK;
}


/********************************************/
/* mops_draw_nurbcurve_c:
   draws only the control-polygon of a NURB
   curve as wireframe                       */
/********************************************/
int
mops_draw_nurbcurve_c(struct Togl *togl, mops_object *o)
{
 int length = 0, i = 0, a = 0;
 mops_nurbcurve_object *curve = (mops_nurbcurve_object *) o->object;
 GLdouble *ver = NULL;
 GLdouble mvm[16], pm[16], win1x, win1y, win1z, win2x, win2y, win2z;
 GLdouble p1x, p1y, p2x, p2y, alpha;
 GLint vp[4];

  length = curve->length;

  ver = curve->controlv;

  glBegin(GL_LINE_STRIP);
  for(i=0; i<length;i++)
    {
      glVertex3dv((GLdouble *)&ver[a]);
      if(curve->has_weights)
	a += 4;
      else
	a += 3;

    }
  glEnd();

  /* draw direction pointer */
  glGetDoublev(GL_MODELVIEW_MATRIX, mvm);
  glGetDoublev(GL_PROJECTION_MATRIX, pm);
  glGetIntegerv(GL_VIEWPORT, vp);
  length = curve->length - 1;
  length *= (curve->has_weights?4:3);
  gluProject((GLdouble)ver[length], (GLdouble)ver[length+1],
	     (GLdouble)ver[length+2], mvm, pm, vp,
	     &win1x, &win1y, &win1z);

  length = curve->length - 2;
  length *= (curve->has_weights?4:3);
  gluProject((GLdouble)ver[length], (GLdouble)ver[length+1],
	     (GLdouble)ver[length+2], mvm, pm, vp,
	     &win2x, &win2y, &win2z);

  p1x = win1x+MOPS_POINTER;
  p1y = win1y+MOPS_POINTER;

  p2x = win1x+MOPS_POINTER;
  p2y = win1y-MOPS_POINTER;

  if(win2x-win1x != 0.0)
    {
      alpha = MOPS_R2D(atan((win2y-win1y)/(win2x-win1x)));

      if(win2x<win1x)
	alpha = 180+alpha;
    }
  else
    if(win2y>win1y)
      alpha = 90;
    else
      alpha = -90;

  glMatrixMode(GL_PROJECTION);
  glPushMatrix();
  glLoadIdentity();
  glOrtho(0, Togl_Width(togl), 0, Togl_Height(togl), -100.0, 100.0);

  glMatrixMode(GL_MODELVIEW);
  glPushMatrix();
  glLoadIdentity();
  glTranslated(win1x,win1y,0.0);
  glRotated(alpha,0.0,0.0,1.0);
  glTranslated(-win1x,-win1y,0.0);
  glBegin(GL_LINE_STRIP);
  glVertex3d(p1x,p1y,win1z);
  glVertex3d(win1x,win1y,win1z);
  glVertex3d(p2x,p2y,win1z);
  glEnd();

  glPopMatrix();
  glMatrixMode(GL_PROJECTION);
  glPopMatrix();
  glMatrixMode(GL_MODELVIEW);

 return MOPS_OK;
}


/********************************************/
/* mops_draw_nurbcurve_h:
   draws only handles (for the vertices of
   the control-polygon) of a NURB curve     */
/********************************************/
int
mops_draw_nurbcurve_h(struct Togl *togl, mops_object *o)
{
 int length = 0, i = 0, a = 0;
 double w = 0.0;
 mops_nurbcurve_object *curve = (mops_nurbcurve_object *) o->object;
 mops_mpoint_object *p = NULL;
 GLdouble *ver = NULL, point_size = 6.0;
 GLUquadricObj *qobj = NULL;
 GLdouble mvm[16], pm[16], winx, winy, winz;
 GLint vp[4];

  point_size = curve->handle_size;
  length = curve->length;

  ver = curve->controlv;
  if(!curve->weight_handles)
    {
      glPointSize(point_size);

      glBegin(GL_POINTS);
      for(i=0; i<length; i++)
	{
	  glVertex3dv((GLdouble *)&ver[a]);
	  if(curve->has_weights)
	    a += 4;
	  else
	    a += 3;
	}
      glEnd();

      if(curve->mpoints)
	{
	  glPointSize(point_size*1.25);
	  glBegin(GL_POINTS);
	   p = curve->mpoints;
	   while(p)
	     {
	       glVertex3dv((GLdouble *)(p->points[0]));
	       p = p->next;
	     }
	  glEnd();
	  glPointSize(point_size);
	}

    }
  else
    { /* Weights -> handlesize */
      if(curve->has_weights)
	{
	  glGetDoublev(GL_MODELVIEW_MATRIX, mvm);
	  glGetDoublev(GL_PROJECTION_MATRIX, pm);
	  glGetIntegerv(GL_VIEWPORT, vp);

	  glMatrixMode(GL_PROJECTION);
	  /*	  glPushMatrix();*/
	  glLoadIdentity();
	  glOrtho(0, Togl_Width(togl), 0, Togl_Height(togl), -100.0, 100.0);
	  
	  glMatrixMode(GL_MODELVIEW);
	  /* glPushMatrix();*/
	  glLoadIdentity();

	  if(!(qobj = gluNewQuadric()))
	    return MOPS_OUTOFMEM;

	  for(i=0; i<length; i++)
	  {
	    gluProject((GLdouble)ver[a], (GLdouble)ver[a+1],
		       (GLdouble)ver[a+2], mvm, pm, vp,
		       &winx, &winy, &winz);

	    /* XXXX picking? */
	    glTranslated(winx,winy,winz);


	    w = ver[a+3];
	    point_size = curve->handle_size /** 0.033*/;
	    if(w<1.0)
	      {
		point_size /= w;
		gluDisk(qobj,point_size-(point_size/10),point_size,8,1);
	      }
	    else
	      {
		point_size *= w;
		gluDisk(qobj,0.0,point_size,8,1);
	      }
	    glTranslated(-winx,-winy,-winz);

	    a += 4;
	  }

	  gluDeleteQuadric(qobj);
	  /*	  glPopMatrix();*/
	  glLoadMatrixd(mvm);
	  glMatrixMode(GL_PROJECTION);
	  glLoadMatrixd(pm);
	  glMatrixMode(GL_MODELVIEW);
	  
	}
    }
 return MOPS_OK;
}


/********************************************/
/* mops_create_nurbcurve:
   create a NURB curve                      */
/********************************************/
int
mops_create_nurbcurve(int order, int length, int knot_type, int has_weights,
		      double *controlv, double *knotv,
		      mops_nurbcurve_object **curveptr)
{
 int mops_status = MOPS_OK;
 double *newcontrolv = NULL;
 mops_nurbcurve_object *curve = NULL;
  
  if(!(curve = calloc(1, sizeof(mops_nurbcurve_object))))
    return MOPS_OUTOFMEM;
  
  curve->order = order;
  curve->length = length;
  curve->knot_type = knot_type;
  curve->has_weights = has_weights;
  
  curve->controlv = controlv;
  if(!controlv)
    {
      if(has_weights)
	{
	  if(!(newcontrolv = calloc(4*length, sizeof(double))))
	    {
	      free(curve);
	      return MOPS_OUTOFMEM;
	    }
	}
      else
	{
	  if(!(newcontrolv = calloc(3*length, sizeof(double))))
	    {
	      free(curve);
	      return MOPS_OUTOFMEM; 	
	    }
	}

      curve->controlv = newcontrolv;

    }

  if(knot_type != MOPS_KTCUSTOM && knotv == NULL)
    { /* we need to create knots */
      mops_status = mops_create_nurbcurve_knots(curve);
      if(mops_status)
	{
	  if(newcontrolv) free(newcontrolv);
	  free(curve);
	  return mops_status;
	}
    }
  else /* user specified own knots */
    curve->knotv = knotv;

  /* Misc. Defaults */
  curve->glu_sampling_tolerance = 30.0;
  curve->handle_size = 6.0;

  *curveptr = curve;

 return MOPS_OK;
}


/********************************************/
/* mops_create_nurbpatch:
   create a NURB patch                      */
/********************************************/
int
mops_create_nurbpatch(int uorder, int vorder, int width, int height,
		      int uknot_type, int vknot_type, int has_weights,
		      double *controlv, double *uknotv, double *vknotv,
		      mops_nurbpatch_object **patchptr)
{
  int mops_status = MOPS_OK;
  double *newcontrolv = NULL;
  mops_nurbpatch_object *patch = NULL;

  if(!(patch = calloc(1, sizeof(mops_nurbpatch_object))))
    return MOPS_OUTOFMEM;

  patch->uorder = uorder;
  patch->vorder = vorder;
  patch->width = width;
  patch->height = height;
  patch->uknot_type = uknot_type;
  patch->vknot_type = vknot_type;
  patch->has_weights = has_weights;

  patch->controlv = controlv;

  if(!controlv)
    {
      if(has_weights)
	{
	  if(!(newcontrolv = calloc(4*width*height, sizeof(double))))
	    {
	      free(patch);
	      return MOPS_OUTOFMEM;
	    }
	}
      else
	{
	  if(!(newcontrolv = calloc(3*width*height, sizeof(double))))
	    {
	      free(patch);
	      return MOPS_OUTOFMEM;
	    }
	}

      patch->controlv = newcontrolv;

    }

  mops_status = mops_create_nurbpatch_knots(patch);
  if(mops_status)
    {
      if(newcontrolv) free(newcontrolv);
      free(patch);
      return mops_status;
    }

  if((uknot_type == MOPS_KTCUSTOM) && uknotv)
    {
      free(patch->uknotv);
      patch->uknotv = uknotv;
    }
  if((vknot_type == MOPS_KTCUSTOM) && vknotv)
    {
      free(patch->vknotv);
      patch->vknotv = vknotv;
    }

  patch->glu_sampling_tolerance = 30.0;
  patch->handle_size = 6.0;

  *patchptr = patch;

  return MOPS_OK;
}


/********************************************/
/* mops_free_nurbcurve:
   free a NURB curve                        */
/********************************************/
int
mops_free_nurbcurve(mops_nurbcurve_object *curve)
{

  if(!curve)
    return MOPS_OK;

  if(curve->controlv)
    free(curve->controlv);

  if(curve->knotv)
    free(curve->knotv);

  /* free GLU NURBS renderer */
  if(curve->no)
    gluDeleteNurbsRenderer(curve->no);

  free(curve);

  return MOPS_OK;
}

/********************************************/
/* mops_free_nurbpatch:
   free a NURB patch                        */
/********************************************/
int
mops_free_nurbpatch(mops_nurbpatch_object *patch)
{
 int mops_status = MOPS_OK;

  if(!patch)
    return MOPS_OK;

  if(patch->controlv)
    free(patch->controlv); 

  if(patch->uknotv)
    free(patch->uknotv); 
  if(patch->vknotv)
    free(patch->vknotv);  

  /* free GLU NURBS renderer */
  if(patch->no)
    gluDeleteNurbsRenderer(patch->no);

  free(patch);

 return mops_status;
}

/********************************************/
/* mops_copy_nurbpatch:
   copy a NURB patch                        */
/********************************************/
int
mops_copy_nurbpatch(mops_nurbpatch_object *source,
		    mops_nurbpatch_object **dest)
{
 int mops_status = MOPS_OK;
 mops_nurbpatch_object *newo = NULL;

  if(!(newo = calloc(1,sizeof(mops_nurbpatch_object))))
    return MOPS_OUTOFMEM; 

  memcpy(newo, source, sizeof(mops_nurbpatch_object)); 
  /* danger! links point to original hierachy */

  if(source->has_weights)
    {
      if(!(newo->controlv = calloc(4*source->width*source->height,
				   sizeof(double))))
	{ free(newo); return MOPS_OUTOFMEM; }

      memcpy(newo->controlv, source->controlv,
	     (size_t)(4*source->width*source->height * sizeof(double)));
    }
  else
    {
      if(!(newo->controlv = calloc(3*source->width*source->height,
				   sizeof(double))))
	{ free(newo); return MOPS_OUTOFMEM; }
      memcpy(newo->controlv, source->controlv,
	     (size_t)(3*source->width*source->height * sizeof(double)));
    }

  if(!(newo->uknotv = calloc(source->width+source->uorder,sizeof(double))))
    { mops_free_nurbpatch(newo); return MOPS_OUTOFMEM; }
  memcpy(newo->uknotv, source->uknotv,
	 (size_t)((source->width+source->uorder)*sizeof(double)));

  if(!(newo->vknotv = calloc(source->height+source->vorder,sizeof(double))))
    { mops_free_nurbpatch(newo); return MOPS_OUTOFMEM; }
  memcpy(newo->vknotv, source->vknotv,
	 (size_t)((source->height+source->vorder)*sizeof(double)));

  /* no GLUnurbsObj yet */
  newo->no = NULL;

  *dest = newo;

 return mops_status;
}


/********************************************/
/* mops_copy_nurbcurve:
   copy a NURB curve                        */
/********************************************/
int
mops_copy_nurbcurve(mops_nurbcurve_object *source,
		    mops_nurbcurve_object **dest)
{
  int mops_status = MOPS_OK;
  mops_nurbcurve_object *newo = NULL;

  if(!(newo = calloc(1,sizeof(mops_nurbcurve_object))))
    return MOPS_OUTOFMEM; 

  memcpy(newo, source, sizeof(mops_nurbcurve_object)); 
  /* danger! links point to original hierachy */

  newo->mpoints = NULL;

  if(source->has_weights)
    {
      if(!(newo->controlv = calloc(4*source->length, sizeof(double))))
	{ mops_free_nurbcurve(newo); return MOPS_OUTOFMEM; }
      memcpy(newo->controlv, source->controlv,
	     (size_t)(4*source->length * sizeof(double)));
    }
  else
    {
      if(!(newo->controlv = calloc(3*source->length, sizeof(double))))
	{ mops_free_nurbcurve(newo); return MOPS_OUTOFMEM; }
      memcpy(newo->controlv, source->controlv,
	     (size_t)(3*source->length * sizeof(double)));
    }


  if(!(newo->knotv = calloc(source->length+source->order,sizeof(double))))
    { mops_free_nurbcurve(newo); return MOPS_OUTOFMEM; }
  memcpy(newo->knotv, source->knotv,
	 (size_t)((source->length+source->order)*sizeof(double)));

  /* no GLUnurbsObj yet */
  newo->no = NULL;

  mops_nurbcurve_recreatemp(newo);

  *dest = newo;

  return mops_status;
}


/********************************************/
/*mops_get_nurbcurve_point:
  calculates nearest controlpoint of curve
  to point obj, puts the address of that 
  point into global point_edit_stack
  for single point editing                  */
/********************************************/
int
mops_get_nurbcurve_point(mops_object *o, double objX, double objY,
			 double objZ)
{
 int mops_status = MOPS_OK;
 mops_nurbcurve_object *curve = o->object;
 mops_mpoint_object *p = NULL;
 double *control = NULL, min_distance = mops_prefs.pick_epsilon,
   distance = 0.0;
 int i = 0, j = 0, add = 0, found = MOPS_FALSE;
 double *pecoords = NULL;


  if(min_distance == 0.0)
    min_distance = DBL_MAX;

  if(mops_point_edit_coords)
    free(mops_point_edit_coords);

  mops_point_edit_coords = NULL;

  control = curve->controlv;
  add = ((curve->has_weights)?4:3);

  for(i = 0; i < curve->length; i++)
    {
      distance = MOPS_VLEN((objX - control[j]),
			   (objY - control[j+1]),
			   (objZ - control[j+2]));

      if(distance < min_distance)
	{
	  pecoords = &(control[j]);
	  min_distance = distance;
	}

      j += add;
    }

  if(!pecoords)
    return MOPS_OK; /* XXXX should this return a 'MOPS_FAILEDPICK' ? */

  mops_point_edit_coords_homogenous = curve->has_weights;

  if(curve->mpoints)
    {
      p = curve->mpoints;
      while(p && !found)
	{
	  found = MOPS_FALSE;
	  for(i = 0; i < p->multiplicity; i++)
	    {
	      if(p->points[i] == pecoords)
		{
		  found = MOPS_TRUE;
		  mops_point_edit_coords_number = p->multiplicity;
		  if(!(mops_point_edit_coords = calloc(p->multiplicity,
						       sizeof(double*))))
		    return MOPS_OUTOFMEM;
		  memcpy(mops_point_edit_coords, p->points, p->multiplicity*
			 sizeof(double *));
		}
	    }

	  p = p->next;
	}

    }

  if(!found)
    {
      if(!(mops_point_edit_coords = calloc(1,sizeof(double*))))
	return MOPS_OUTOFMEM;
      mops_point_edit_coords_number = 1;
      mops_point_edit_coords[0] = pecoords;
    }

 return mops_status;
}


/********************************************/
/*mops_get_nurbpatch_point:
  calculates nearest controlpoint of patch
  to point obj, puts the address of that 
  point into global point_edit_stack
  for single point editing                  */
/********************************************/
int
mops_get_nurbpatch_point(mops_object *o, double objX, double objY,
			 double objZ)
{
 int mops_status = MOPS_OK;
 mops_nurbpatch_object *patch = o->object;
 double *control = NULL, min_distance = mops_prefs.pick_epsilon,
   distance = 0.0;
 int h = 0, i = 0, j = 0, add = 0;
 double *pecoords = NULL;

  if(min_distance == 0.0)
    min_distance = DBL_MAX;

  if(mops_point_edit_coords)
    free(mops_point_edit_coords);

  mops_point_edit_coords = NULL;

  control = patch->controlv;
  add = ((patch->has_weights)?4:3);

  for(h = 0; h < patch->width; h++)
    for(i = 0; i < patch->height; i++)
      {
	distance = MOPS_VLEN((objX - control[j]),
			     (objY - control[j+1]),
			     (objZ - control[j+2]));

	if(distance < min_distance)
	  {
	    pecoords = &(control[j]);
	    min_distance = distance;
	  }

	j += add;
      }
  
  /* XXXX should this return a 'MOPS_FAILEDPICK' ? */  
  if(!pecoords)
    return MOPS_OK;

  if(patch->has_weights)
    mops_point_edit_coords_homogenous = MOPS_TRUE;
  else
    mops_point_edit_coords_homogenous = MOPS_FALSE;

  if(!(mops_point_edit_coords = calloc(1,sizeof(double*))))
    return MOPS_OUTOFMEM;	  ;
  mops_point_edit_coords_number = 1;
  mops_point_edit_coords[0] = pecoords;

 return mops_status;
}

/********************************************/
/*mops_nurbcurve_convert_weight:
                                            */
/********************************************/
int
mops_nurbcurve_convert_weight(mops_nurbcurve_object *curve, int new_weight)
{
 int mops_status = MOPS_OK;
 int i = 0, j = 0, k = 0;
 double *newcontrolv;

  if( (!new_weight) && curve->has_weights)
    { /* delete weights */
     
      if(!(newcontrolv = calloc(curve->length*3,sizeof(double))))
	return MOPS_OUTOFMEM;

      for(i = 0; i<curve->length; i++)
	{
	  newcontrolv[j] = curve->controlv[k]; 
	  newcontrolv[j+1] = curve->controlv[k+1]; 
	  newcontrolv[j+2] = curve->controlv[k+2]; 

	  j += 3;
	  k += 4;
	}

      free(curve->controlv);
      curve->controlv = newcontrolv;
      curve->has_weights = MOPS_FALSE;
    }
  else
  if( new_weight && (!curve->has_weights))
    { /* add weights */

      if(!(newcontrolv = calloc(curve->length*4,sizeof(double))))
	return MOPS_OUTOFMEM;

      for(i = 0; i<curve->length; i++)
	{
	  newcontrolv[j] = curve->controlv[k]; 
	  newcontrolv[j+1] = curve->controlv[k+1]; 
	  newcontrolv[j+2] = curve->controlv[k+2]; 
	  newcontrolv[j+3] = 1.0;
	  j += 4;
	  k += 3;
	}

      free(curve->controlv);
      curve->controlv = newcontrolv;
      curve->has_weights = MOPS_TRUE;
    }

 return mops_status;
}

/********************************************/
/*mops_nurbcurve_close:
                                            */
/********************************************/
int
mops_nurbcurve_close(mops_nurbcurve_object *curve)
{
 int mops_status = MOPS_OK;
 double *controlv = NULL, *end = NULL;
 int stride = 0, i;

 /* close curve */
 stride = (curve->has_weights?4:3);
 if(curve->length >= ((curve->order-1)*2))
   {
     controlv = curve->controlv;
     end = &(controlv[(curve->length-(curve->order-1))*stride]);

     for(i = 0; i < (curve->order-1); i++)
       {
	 memcpy(end, controlv, stride*sizeof(double));
	 controlv += stride;
	 end += stride;
       }

     curve->closed = MOPS_TRUE;
   }
 else
   {
     curve->closed = MOPS_FALSE;
     return MOPS_ERROR;
   }

 return mops_status;
}


/********************************************/
/*mops_nurbpatch_convert_weight:
                                            */
/********************************************/
int
mops_nurbpatch_convert_weight(mops_nurbpatch_object *patch, int new_weight)
{
 int mops_status = MOPS_OK;
 int i = 0, j = 0, k = 0, l = 0;
 double *newcontrolv;

  if( (!new_weight) && patch->has_weights)
    { /* delete weights */
     
      if(!(newcontrolv = calloc(patch->width*patch->height*3,sizeof(double))))
	return MOPS_OUTOFMEM;

      for(i = 0; i<patch->width; i++)
	for(l = 0; l<patch->height; l++)
	  {
	    newcontrolv[j] = patch->controlv[k]; 
	    newcontrolv[j+1] = patch->controlv[k+1]; 
	    newcontrolv[j+2] = patch->controlv[k+2]; 

	    j += 3;
	    k += 4;
	  }

      free(patch->controlv);
      patch->controlv = newcontrolv;
      patch->has_weights = MOPS_FALSE;
    }
  else
  if( new_weight && (!patch->has_weights))
    { /* add weights */

      if(!(newcontrolv = calloc(patch->width*patch->height*4,sizeof(double))))
	return MOPS_OUTOFMEM;

      for(i = 0; i<patch->width; i++)
	for(l = 0; l<patch->height; l++)
	  {
	    newcontrolv[j] = patch->controlv[k]; 
	    newcontrolv[j+1] = patch->controlv[k+1]; 
	    newcontrolv[j+2] = patch->controlv[k+2]; 
	    newcontrolv[j+3] = 1.0;
	    j += 4;
	    k += 3;
	  }

      free(patch->controlv);
      patch->controlv = newcontrolv;
      patch->has_weights = MOPS_TRUE;
    }

 return mops_status;
}


/********************************************/
/*mops_nurbcurve_convert_length:            */
/********************************************/
int
mops_nurbcurve_convert_length(mops_nurbcurve_object *curve, int new_length)
{
 int mops_status = MOPS_OK;
 int a, b, i, j, stride;
 int *newpersec = NULL, new = 0;
 double *ncontrolv = NULL, v[3] = {0}, t = 0.0, *cv = NULL;

  if(new_length == curve->length)
    return mops_status;

  stride = curve->has_weights?4:3;

  if(!(ncontrolv = calloc(stride*new_length, sizeof(double))))
    return MOPS_OUTOFMEM;

  if(new_length < curve->length)
    {
      a = 0;
      for(i=0; i<new_length; i++)
	{
	  memcpy(&ncontrolv[a], &(curve->controlv[a]), stride*sizeof(double));
	  a+=stride;
	}
    }
  else
    {
      /* distribute new points */
      new = new_length - curve->length;

      if(!(newpersec = calloc((curve->length-1), sizeof(int))))

	return MOPS_OUTOFMEM;
      cv = curve->controlv;

      while(new)
	for(i = 0; i < (curve->length-1); i++)
	  {
	    if((cv[i*stride] != cv[(i+1)*stride]) ||
	       (cv[i*stride+1] != cv[(i+1)*stride+1]) ||
	       (cv[i*stride+2] != cv[(i+1)*stride+2]))
	      {
		if(!(curve->closed && (i>(curve->length-curve->order))))
		  if(new)
		    {
		      (newpersec[i])++;
		      new--;
		    }
	      } 
	  }

      a = 0;
      b = 0;
      for(i = 0; i < (curve->length-1); i++)
	{
	  memcpy(&ncontrolv[b], &(curve->controlv[a]), stride*sizeof(double));
	  b+=stride;

	  if((cv[i*stride] != cv[(i+1)*stride]) ||
	     (cv[i*stride+1] != cv[(i+1)*stride+1]) ||
	     (cv[i*stride+2] != cv[(i+1)*stride+2]))
	    {
	      if(!(curve->closed && (i>(curve->length-curve->order))))
	      for(j = 1; j <= newpersec[i]; j++)
		{

		  v[0] = curve->controlv[a+stride] - curve->controlv[a];
		  v[1] = curve->controlv[a+stride+1] - curve->controlv[a+1];
		  v[2] = curve->controlv[a+stride+2] - curve->controlv[a+2];

		  t = j/(newpersec[i]+1.0);
	      
		  MOPS_V3SCAL(v,t);

		  ncontrolv[b] = curve->controlv[a]+v[0];
		  ncontrolv[b+1] = curve->controlv[a+1]+v[1];
		  ncontrolv[b+2] = curve->controlv[a+2]+v[2];
		  if(curve->has_weights)
		    ncontrolv[b+3] = 1.0;
		  b+=stride;
		}
	    }

	  a+=stride;	       
	}
      memcpy(&ncontrolv[(new_length-1)*stride],
	     &(curve->controlv[(curve->length-1)*stride]),
	     stride*sizeof(double));

      free(newpersec);

    }

  free(curve->controlv);
  curve->controlv = ncontrolv;

  curve->length = new_length;

  if(curve->closed)
    {
      mops_status = mops_nurbcurve_close(curve);
    }

 return mops_status;
}


/********************************************/
/*mops_nurbpatch_convert_width:
  change width of a NURBPatch               */
/********************************************/
int
mops_nurbpatch_convert_width(mops_nurbpatch_object *patch,
			     int new_width)
{
 int mops_status = MOPS_OK;
 int i, j, k, a, b, stride;
 int *newpersec = NULL, new = 0;
 double *ncontrolv = NULL, v[3] = {0}, t = 0.0;

  if(new_width == patch->width)
    return mops_status;

  stride = patch->has_weights?4:3;

  if(!(ncontrolv = calloc(patch->height*new_width*stride,sizeof(double))))
    return MOPS_OUTOFMEM;

  if(new_width < patch->width)
    {
      a = 0; b = 0;
      for(i = 0; i < new_width; i++)
	{
	  memcpy(&(ncontrolv[b]),&(patch->controlv[a]),
		 patch->height*stride*sizeof(double));

	  a += (patch->height*stride);
	  b += (patch->height*stride);
	}

    }
  else
    {
      /* distribute new points */
      new = new_width-patch->width;

      if(!(newpersec = calloc((patch->width-1), sizeof(int))))
	return MOPS_OUTOFMEM;

      while(new)
	for(i = 0; i < (patch->width-1); i++)
	  {
	    if(new)
	      {
		(newpersec[i])++;
		new--;
	      }
	  }

      a = 0;
      b = 0;
      for(k = 0; k < (patch->width-1); k++)
	{

	  memcpy(&(ncontrolv[b]),&(patch->controlv[a]),
		 patch->height*stride*sizeof(double));

	  b += (patch->height*stride);

	  for(j = 1; j <= newpersec[k]; j++)
	    {
	      t = j/(newpersec[k]+1.0);
	      a = k*patch->height*stride;
	      for(i = 0; i < patch->height; i++)
		{
		  v[0] = patch->controlv[a+(stride*patch->height)] -
		    patch->controlv[a];
		  v[1] = patch->controlv[a+(stride*patch->height)+1] -
		    patch->controlv[a+1];
		  v[2] = patch->controlv[a+(stride*patch->height)+2] -
		    patch->controlv[a+2];

		  MOPS_V3SCAL(v,t);

		  ncontrolv[b] = patch->controlv[a]+v[0];
		  ncontrolv[b+1] = patch->controlv[a+1]+v[1];
		  ncontrolv[b+2] = patch->controlv[a+2]+v[2];
		  if(patch->has_weights)
		    ncontrolv[b+3] = 1.0;

		  a+=stride;
		  b+=stride;

		} /* for i */

	    } /* for j */

	  if(newpersec[k] == 0)
	    a += (patch->height*stride);

	} /* for k */

      memcpy(&ncontrolv[(new_width-1)*patch->height*stride],
	     &(patch->controlv[(patch->width-1)*patch->height*stride]),
	     patch->height*stride*sizeof(double));

      free(newpersec);

    } /* if */

  free(patch->controlv);
  patch->controlv = ncontrolv;
  patch->width = new_width;

  return mops_status;
}

/********************************************/
/*mops_nurbpatch_convert_height:
  change height of a NURBPatch              */
/********************************************/
int
mops_nurbpatch_convert_height(mops_nurbpatch_object *patch,
			      int new_height)
{
 int mops_status = MOPS_OK;
 int i, j, k, a, b, stride;
 int *newpersec = NULL, new = 0;
 double *ncontrolv = NULL, v[3] = {0}, t = 0.0;

  if(new_height == patch->height)
    return mops_status;

  stride = patch->has_weights?4:3;

  if(!(ncontrolv = calloc(patch->width*new_height*stride,sizeof(double))))
    return MOPS_OUTOFMEM;

  if(new_height < patch->height)
    {
      a = 0; b = 0;
      for(i = 0; i < patch->width; i++)
	{
	  memcpy(&(ncontrolv[b]),&(patch->controlv[a]),
		 new_height*stride*sizeof(double));

	  a += (patch->height*stride);
	  b += (new_height*stride);
	}

    }
  else
    {
      /* distribute new points */
      new = new_height-patch->height;

      if(!(newpersec = calloc((patch->height-1), sizeof(int))))
	return MOPS_OUTOFMEM;

      while(new)
	for(i = 0; i < (patch->height-1); i++)
	  {
	    if(new)
	      {
		(newpersec[i])++;
		new--;
	      }
	  }

      a = 0;
      b = 0;
      for(k = 0; k < patch->width; k++)
	{
	  for(i = 0; i < (patch->height-1); i++)
	    {
	      memcpy(&ncontrolv[b], &(patch->controlv[a]),
		     stride*sizeof(double));
	      b+=stride;

	      for(j = 1; j <= newpersec[i]; j++)
		{
		  v[0] = patch->controlv[a+stride] - patch->controlv[a];
		  v[1] = patch->controlv[a+stride+1] - patch->controlv[a+1];
		  v[2] = patch->controlv[a+stride+2] - patch->controlv[a+2];


		  t = j/(newpersec[i]+1.0);
	      
		  MOPS_V3SCAL(v,t);

		  ncontrolv[b] = patch->controlv[a]+v[0];
		  ncontrolv[b+1] = patch->controlv[a+1]+v[1];
		  ncontrolv[b+2] = patch->controlv[a+2]+v[2];
		  if(patch->has_weights)
		    ncontrolv[b+3] = 1.0;
		  b+=stride;
		} /* for */

	      a+=stride;

	    } /* for */

	  memcpy(&ncontrolv[b/*+(new_height-1)*stride*/],
		 &(patch->controlv[a/*+(patch->height-1)*stride*/]),
		 stride*sizeof(double));

	  a += stride;
	  b += stride;
	} /* for */

      free(newpersec);

    } /* if */

  
  free(patch->controlv);
  patch->controlv = ncontrolv;
  patch->height = new_height;

  return mops_status;
}


int
mops_wrib_trimcurves(mops_object *o)
{
 int mops_status = MOPS_OK;
 int curvecount = 0, i, j, k, a, b, c, totalcurves, totalcontrol, totalknots;
 RtInt nloops = 0, *ncurves, *order, *n;
 RtFloat *knot, *min, *max, *u, *v, *w;
 mops_object *trim = NULL, *loop = NULL;
 mops_nurbcurve_object *curve = NULL;
 GLdouble m[16] = {0};
 GLfloat mr[4][4];
 double x,y,z,w2;

  if(!o)
    return MOPS_OK;

  if(!(o->type == MOPS_OTNURBPATCH))
    return MOPS_OK;

  /* parse trimcurves */
  /* count loops */
  trim = o->down;
  
  if(trim)
    {
      curve = (mops_nurbcurve_object *)(trim->object);
    }
  else
    return MOPS_OK;

  while(trim)
    {
      if(trim->type == MOPS_OTNURBCURVE)
	nloops++;

      if(trim->type == MOPS_OTLEVEL)
	if(trim->down)
	  nloops++;

      trim = trim->next;
    } /* while */

  if(nloops == 0)
    return MOPS_OK;

  /* count curves per loop */
  if(!(ncurves = calloc(nloops,sizeof(RtInt))))
    return MOPS_OUTOFMEM;

  trim = o->down;
  
  if(trim)
    {
      curve = (mops_nurbcurve_object *)(trim->object);
    }
  curvecount = 0;
  totalcurves = 0;
  totalcontrol = 0;
  totalknots = 0;
  nloops = 0;
  while(trim)
    {
      if(trim->type == MOPS_OTNURBCURVE)
	{
	  totalcurves++;
	  curvecount++;
	  curve = (mops_nurbcurve_object *)(trim->object);
	  totalcontrol += curve->length;
	  
	  totalknots += curve->length;
	  totalknots += curve->order;

	  ncurves[nloops] = 1;
	  curvecount = 0;
	  nloops++;
	  
	  /* set new first */
	  if(trim->next)
	    {
	      curve = (mops_nurbcurve_object *)(trim->next->object);
	    }
	} /* if */

      if(trim->type == MOPS_OTLEVEL)
	{
	  loop = trim->down;
	  while(loop)
	    {
	      if(loop->type == MOPS_OTNURBCURVE)
		{
		  totalcurves++;
		  curvecount++;
		  curve = (mops_nurbcurve_object *)(loop->object);
		  totalcontrol += curve->length;
	  
		  totalknots += curve->length;
		  totalknots += curve->order;

		  ncurves[nloops] = 1;
		  curvecount = 0;
		  nloops++;
		  
		} /* if */

	      loop = loop->next;
	    } /* while */

	} /* if */
      
      trim = trim->next;
    } /* while */

  if(!(order = calloc(totalcurves,sizeof(RtInt))))
    return MOPS_OUTOFMEM;
  if(!(n = calloc(totalcurves,sizeof(RtInt))))
    { free(order); return MOPS_OUTOFMEM; }
  if(!(knot = calloc(totalknots,sizeof(RtFloat))))
    { free(order); free(n); return MOPS_OUTOFMEM; }
  if(!(min = calloc(totalcurves,sizeof(RtFloat))))
    { free(order); free(n); free(knot); return MOPS_OUTOFMEM; }
  if(!(max = calloc(totalcurves,sizeof(RtFloat))))
    { free(order); free(n); free(knot); free(min); return MOPS_OUTOFMEM; }

  if(!(u = calloc(totalcontrol,sizeof(RtFloat))))
    { free(order); free(n); free(knot); free(min); free(max);
    return MOPS_OUTOFMEM; }
  if(!(v = calloc(totalcontrol,sizeof(RtFloat))))
    { free(order); free(n); free(knot); free(min); free(max); free(u);
    return MOPS_OUTOFMEM; }
  if(!(w = calloc(totalcontrol,sizeof(RtFloat))))
    { free(order); free(n); free(knot); free(min); free(max); free(u);
    free(v); return MOPS_OUTOFMEM; }

  trim = o->down;
  a = 0;
  b = 0;
  c = 0;
  /* compile arguments */
  for(i = 0;i < nloops; i++)
    for(j = 0;j < ncurves[i]; j++)

      while(trim)
      {
	if(trim->type == MOPS_OTNURBCURVE)
	  {
	    curve = (mops_nurbcurve_object *)(trim->object);

	    /* fill order[], n[], min[], max[] */
	    order[a] = (RtInt)curve->order;
	    n[a] = (RtInt)curve->length;
	    min[a] = (RtFloat)((curve->knotv)[curve->order - 1]);
	    max[a] = (RtFloat)((curve->knotv)[curve->length]);
	    a++;

	    /* get curves transformation-matrix */
	    glMatrixMode (GL_MODELVIEW);
	    glPushMatrix();
	    glLoadIdentity ();
	    glTranslated(trim->movx, trim->movy, trim->movz);

	    mops_build_rotmatrix(mr, trim->quat);
	    glMultMatrixf(&mr[0][0]);

	    glScaled (trim->scalx, trim->scaly, trim->scalz);
	    glGetDoublev(GL_MODELVIEW_MATRIX, m);
	    glPopMatrix();

	    /* copy & revert control (fill u[] v[] w[]) */
	    for(k = curve->length-1; k >= 0 ; k--)
	      {
		x = (RtFloat)((curve->controlv)[k*(curve->has_weights?4:3)]);
		
		y = (RtFloat)((curve->controlv)
			      [(k*(curve->has_weights?4:3))+1]);
		z = (RtFloat)((curve->controlv)
			      [(k*(curve->has_weights?4:3))+2]);

		if(curve->has_weights)
		  w2 = (RtFloat)((curve->controlv)
				 [(k*(curve->has_weights?4:3))+3]);
		else
		  w2 = (RtFloat)1.0;

		/* apply transformation */
		u[b] = m[0]*x + m[4]*y + m[8]*z + m[12]*w2;
		v[b] = m[1]*x + m[5]*y + m[9]*z + m[13]*w2;
		w[b] = w2;

		b++;
	      }

	    /* copy knots */
	    for(k = 0; k < curve->length + curve->order; k++)
	      {
		knot[c] = curve->knotv[k];
		c++;
	      }
	  } /* if */

	if(trim->type == MOPS_OTLEVEL)
	  if(trim->down)
	    {
	      loop = trim->down;
	      
		while(loop)
		  {
		    if(loop->type == MOPS_OTNURBCURVE)
		      {
			curve = (mops_nurbcurve_object *)(loop->object);

			/* fill order[], n[], min[], max[] */
			order[a] = (RtInt)curve->order;
			n[a] = (RtInt)curve->length;
			min[a] = (RtFloat)((curve->knotv)[curve->order - 1]);
			max[a] = (RtFloat)((curve->knotv)[curve->length]);
			a++;
			
			/* get curves transformation-matrix */
			glMatrixMode (GL_MODELVIEW);
			glPushMatrix();
			glLoadIdentity ();
			glTranslated(loop->movx, loop->movy, loop->movz);

			mops_build_rotmatrix(mr, loop->quat);
			glMultMatrixf(&mr[0][0]);
			
			glScaled (loop->scalx, loop->scaly, loop->scalz);
			glGetDoublev(GL_MODELVIEW_MATRIX, m);
			glPopMatrix();

			/* copy & revert control (fill u[] v[] w[]) */
			for(k = curve->length-1; k >= 0 ; k--)
			  {
			    x = (RtFloat)((curve->controlv)
					  [k*(curve->has_weights?4:3)]);
			    y = (RtFloat)((curve->controlv)
					  [(k*(curve->has_weights?4:3))+1]);
			    z = (RtFloat)((curve->controlv)
					  [(k*(curve->has_weights?4:3))+2]);

		            if(curve->has_weights)
			      w2 = (RtFloat)((curve->controlv)
					     [(k*(curve->has_weights?4:3))+3]);
			    else
			      w2 = (RtFloat)1.0;

			    /* apply transformation */
			    u[b] = m[0]*x + m[4]*y + m[8]*z + m[12]*w2;
			    v[b] = m[1]*x + m[5]*y + m[9]*z + m[13]*w2;
			    w[b] = w2;

			    b++;
			  }

			/* copy knots */
			for(k = 0; k < curve->length + curve->order; k++)
			  {
			    knot[c] = curve->knotv[k];
			    c++;
			  }
		      } /* if */

		  loop = loop->next;
		  } /* while */

	    }

	trim = trim->next;
      }

  if(nloops)
    {
      /* Finally, write the TrimCurves */
      RiTrimCurve(nloops, ncurves, order, knot, min, max, n, u, v, w);

      /* clean up the mess */
      free(ncurves);
      free(order);
      free(knot);
      free(min);
      free(max);
      free(n);
      free(u);
      free(v);
      free(w);
    }

  return mops_status;
}


int
mops_wrib_npatch(mops_object *o)
{
  int mops_status = MOPS_OK;
  mops_nurbpatch_object *patch = NULL;
  RtInt nu,nv,uorder,vorder;
  RtFloat *uknots = NULL, *vknots = NULL;
  RtFloat *controls = NULL;
  int i=0,j=0,a=0,b=0;

  if(!o)
    return MOPS_OK;

  patch = (mops_nurbpatch_object*)(o->object);

  nu = (RtInt)patch->width;
  uorder = (RtInt)patch->uorder;
  nv = (RtInt)patch->height;
  vorder = (RtInt)patch->vorder;

  if((uknots = calloc(nu+uorder, sizeof(RtFloat))) == NULL)
    return MOPS_OUTOFMEM; 
  if((vknots = calloc(nv+vorder, sizeof(RtFloat))) == NULL)
    return MOPS_OUTOFMEM;  

  if(patch->has_weights)
    {
      if((controls = calloc(nu*nv*4, sizeof(RtFloat))) == NULL)
	return MOPS_OUTOFMEM;
    }
  else
    {
      if((controls = calloc(nu*nv*3, sizeof(RtFloat))) == NULL)
	return MOPS_OUTOFMEM;
    }
  a=0;
  for(i=0;i<nu+uorder;i++)
    {
      uknots[a] = (RtFloat)patch->uknotv[a];
      a++;
    }
  a=0;
  for(i=0;i<nv+vorder;i++)
    {
      vknots[a] = (RtFloat)patch->vknotv[a];
      a++;
    }
  a=0;
  /* RenderMan expects u-major order! */
  for(i=0;i<nv;i++)
    {
      b = i*(patch->has_weights?4:3);
      for(j=0;j<nu;j++)
	{
	  controls[a] = (RtFloat)patch->controlv[b]; 
	  a++;
	  controls[a] = (RtFloat)patch->controlv[b+1]; 
	  a++;
	  controls[a] = (RtFloat)patch->controlv[b+2]; 
	  a++;
	  if(patch->has_weights)
	    {
	      controls[a] = (RtFloat)patch->controlv[b+3]; 
	      a++;
	      b+=(nv*4);
	    }
	  else
	    b+=(nv*3);
	}
    }

  mops_status = mops_wrib_shaders(o);

  RiNuPatch(nu, uorder, uknots, uknots[uorder-1], uknots[nu],
	    nv, vorder, vknots, vknots[vorder-1], vknots[nv],
	    patch->has_weights?"Pw":"P", controls, NULL);


  free(uknots);
  free(vknots);
  free(controls);

  return mops_status;
}



/* C -> Tcl! */
int
mops_get_ncattrprop_tcmd(ClientData clientData, Tcl_Interp * interp,
			 int argc, char *argv[])
{
 mops_object *o = NULL;
 mops_list_object *sel = mops_current_selection;
 char *n1="ncattrPropData";
 mops_nurbcurve_object *curve = NULL;
 int i;
 Tcl_Obj *to = NULL, *toa = NULL, *ton = NULL;

  if(!sel) /* oops? */
    return TCL_OK;

  o = sel->object; 
  curve = (mops_nurbcurve_object *) o->object;

  toa = Tcl_NewStringObj(n1,-1);
  ton = Tcl_NewStringObj("Length",-1);
  to = Tcl_NewIntObj(curve->length);
  Tcl_ObjSetVar2(interp,toa,ton,to,TCL_LEAVE_ERR_MSG |
		 TCL_GLOBAL_ONLY);

  Tcl_SetStringObj(ton,"Order",-1);
  to = Tcl_NewIntObj(curve->order);
  Tcl_ObjSetVar2(interp,toa,ton,to,TCL_LEAVE_ERR_MSG |
		 TCL_GLOBAL_ONLY);

  Tcl_SetStringObj(ton,"Has-Weights",-1);
  to = Tcl_NewIntObj(curve->has_weights);
  Tcl_ObjSetVar2(interp,toa,ton,to,TCL_LEAVE_ERR_MSG |
		 TCL_GLOBAL_ONLY);

  Tcl_SetStringObj(ton,"Closed",-1);
  to = Tcl_NewIntObj(curve->closed);
  Tcl_ObjSetVar2(interp,toa,ton,to,TCL_LEAVE_ERR_MSG |
		 TCL_GLOBAL_ONLY);

  Tcl_SetStringObj(ton,"Knot-Type",-1);
  to = Tcl_NewIntObj(curve->knot_type);
  Tcl_ObjSetVar2(interp,toa,ton,to,TCL_LEAVE_ERR_MSG |
		 TCL_GLOBAL_ONLY);

  Tcl_SetVar2(interp,n1,"Knots","", TCL_LEAVE_ERR_MSG |
	      TCL_GLOBAL_ONLY); 
  Tcl_SetStringObj(ton,"Knots",-1);
  for(i=0; i<curve->length+curve->order; i++)
    {
      
      to = Tcl_NewDoubleObj((curve->knotv)[i]);

      Tcl_ObjSetVar2(interp,toa,ton,to,TCL_APPEND_VALUE | 
		     TCL_LIST_ELEMENT | TCL_LEAVE_ERR_MSG |
		     TCL_GLOBAL_ONLY);      
    }

  Tcl_SetStringObj(ton,"Handlesize",-1);
  to = Tcl_NewDoubleObj(curve->handle_size);
  Tcl_ObjSetVar2(interp,toa,ton,to,TCL_LEAVE_ERR_MSG |
		 TCL_GLOBAL_ONLY);

  Tcl_SetStringObj(ton,"W-Handles",-1);
  to = Tcl_NewIntObj(curve->weight_handles);
  Tcl_ObjSetVar2(interp,toa,ton,to,TCL_LEAVE_ERR_MSG |
		 TCL_GLOBAL_ONLY);

  Tcl_SetStringObj(ton,"Tolerance",-1);
  to = Tcl_NewDoubleObj(curve->glu_sampling_tolerance);
  Tcl_ObjSetVar2(interp,toa,ton,to,TCL_LEAVE_ERR_MSG |
		 TCL_GLOBAL_ONLY);

  Tcl_IncrRefCount(toa);Tcl_DecrRefCount(toa);
  Tcl_IncrRefCount(ton);Tcl_DecrRefCount(ton);
  
 return TCL_OK;
}


/* Tcl -> C! */
int
mops_set_ncattrprop_tcmd(ClientData clientData, Tcl_Interp * interp,
			 int argc, char *argv[])
{
 int mops_status = MOPS_OK;
 mops_object *o = NULL;
 mops_list_object *sel = mops_current_selection;
 mops_nurbcurve_object *curve = NULL;
 char *n1="ncattrPropData";
 int new_order, new_length, new_has_weights, new_knot_type, new_closed;
 int old_length;
 double *nknotv = NULL;
 int updateKnots = 0;
 int knotc, i;
 char **knotv;
 Tcl_Obj *to = NULL, *toa = NULL, *ton = NULL;
 char fname[] = "setNCAttr";

  if(!sel) /* oops? */
    return TCL_OK;

  o = sel->object; 
  curve = (mops_nurbcurve_object *) o->object;

  mops_clear_point_selection(o);

  old_length = curve->length;

  toa = Tcl_NewStringObj(n1,-1);
  ton = Tcl_NewStringObj("Length",-1);
  to = Tcl_ObjGetVar2(interp,toa,ton,TCL_LEAVE_ERR_MSG | TCL_GLOBAL_ONLY);
  Tcl_GetIntFromObj(interp,to, &new_length);

  Tcl_SetStringObj(ton,"Order",-1);
  to = Tcl_ObjGetVar2(interp,toa,ton,TCL_LEAVE_ERR_MSG | TCL_GLOBAL_ONLY);
  Tcl_GetIntFromObj(interp,to, &new_order);

  Tcl_SetStringObj(ton,"Has-Weights",-1);
  to = Tcl_ObjGetVar2(interp,toa,ton,TCL_LEAVE_ERR_MSG | TCL_GLOBAL_ONLY);
  Tcl_GetIntFromObj(interp,to, &new_has_weights);

  Tcl_SetStringObj(ton,"Knot-Type",-1);
  to = Tcl_ObjGetVar2(interp,toa,ton,TCL_LEAVE_ERR_MSG | TCL_GLOBAL_ONLY);
  Tcl_GetIntFromObj(interp,to, &new_knot_type);

  Tcl_SetStringObj(ton,"Closed",-1);
  to = Tcl_ObjGetVar2(interp,toa,ton,TCL_LEAVE_ERR_MSG | TCL_GLOBAL_ONLY);
  Tcl_GetIntFromObj(interp,to, &new_closed);

  Tcl_SetStringObj(ton,"Handlesize",-1);
  to = Tcl_ObjGetVar2(interp,toa,ton,TCL_LEAVE_ERR_MSG | TCL_GLOBAL_ONLY);
  Tcl_GetDoubleFromObj(interp,to, &(curve->handle_size));

  Tcl_SetStringObj(ton,"W-Handles",-1);
  to = Tcl_ObjGetVar2(interp,toa,ton,TCL_LEAVE_ERR_MSG | TCL_GLOBAL_ONLY);
  Tcl_GetIntFromObj(interp,to, &(curve->weight_handles));

  Tcl_SetStringObj(ton,"Tolerance",-1);
  to = Tcl_ObjGetVar2(interp,toa,ton,TCL_LEAVE_ERR_MSG | TCL_GLOBAL_ONLY);
  Tcl_GetDoubleFromObj(interp,to, &(curve->glu_sampling_tolerance));

  Tcl_IncrRefCount(toa);Tcl_DecrRefCount(toa);
  Tcl_IncrRefCount(ton);Tcl_DecrRefCount(ton);


  /* apply changed values to curve */

  if(new_length != curve->length && (new_length > 1))
    {
      mops_status = mops_nurbcurve_convert_length(curve, new_length);

      if(mops_status)
       mops_error(interp,MOPS_ERROR,fname,"could not convert length of curve");

      updateKnots = 1;
    }

  if((curve->order != new_order) && (new_order > 1))
    {
      curve->order = new_order;
      updateKnots = 1;
    }

  if(curve->has_weights != new_has_weights)
    {
      mops_status = mops_nurbcurve_convert_weight(curve, new_has_weights);
      if(mops_status)
	mops_error(interp,MOPS_ERROR,fname,"could not convert weights");

    }

  if((new_knot_type != curve->knot_type) || (updateKnots))
    {
      curve->knot_type = new_knot_type;
      updateKnots = 1;

    }

  /*
    for all Pink's that do not know that a BezierCurve always has
    order = length
    */
  if(curve->knot_type == MOPS_KTBEZIER)
    if(curve->order != curve->length)
      {
	mops_error(interp,MOPS_WARNING,fname,
		   "changing order to match length!");
	curve->order = curve->length;
      }

  if(curve->length < curve->order)
    {
      mops_error(interp,MOPS_WARNING,fname,
		 "lowering order to match length!");
      curve->order = curve->length;
    }

  if((updateKnots) && (curve->knot_type != MOPS_KTCUSTOM))
    mops_status = mops_create_nurbcurve_knots(curve);


  /* decompose knot-list (create custom knot sequence) */
  if(curve->knot_type == MOPS_KTCUSTOM)
    {
      Tcl_SplitList(interp,Tcl_GetVar2(interp, n1, "Knots",
				       TCL_LEAVE_ERR_MSG | TCL_GLOBAL_ONLY),
		    &knotc, &knotv);

      if(!(nknotv = calloc(knotc, sizeof(double))))
	{
	  mops_error(interp,MOPS_OUTOFMEM,fname,NULL);
	  Tcl_Free((char *) knotv);
	  return TCL_OK;
	}

      for(i = 0; i < knotc; i++)
	{
	  Tcl_GetDouble(interp,knotv[i],&nknotv[i]);
	} /* for */

      if(!(mops_status = mops_check_knots(new_length,new_order,knotc,nknotv)))
	{/* the knots are ok */
	  free(curve->knotv);
	  curve->knotv = nknotv;
	  /*
	  if(curve->knotv[0] != 0.0 || curve->knotv[
	      curve->length+curve->order-1] != 1.0)
	    {
	      mops_status = mops_rescale_knotv(curve->length+curve->order,
					       curve->knotv);
	    }
	  */

	}
      else
	{/* the knots are wrong */
	  /* first, tell the user what went wrong */
	  switch (mops_status)
	    {
	    case 1:
	      mops_error(interp,MOPS_ERROR,fname,
			 "knot sequence has too few knots");
	      break;
	    case 2:
	      mops_error(interp,MOPS_ERROR,fname,
			 "knot sequence has too much knots");
	      break;
	    case 3:
	      mops_error(interp,MOPS_ERROR,fname,
			 "knot multiplicity higher than order");
	      break;
	    case 4:
	      mops_error(interp,MOPS_ERROR,fname,
			 "knot sequence has decreasing knots");
	      break;

	    }

	  free(nknotv);

	  /* create new knots */
	  mops_error(interp,MOPS_WARNING,fname,
			 "falling back to knot type NURB!");
	  curve->knot_type = MOPS_KTNURB;
	  mops_status = mops_create_nurbcurve_knots(curve);

	  if(mops_status)
	    mops_error(interp, MOPS_ERROR, fname, "error creating new knots");


	}

      Tcl_Free((char *) knotv);
    }

  if(new_closed && !curve->closed)
    {
      mops_status = mops_nurbcurve_close(curve);
      if(mops_status)
	{
	  mops_error(interp,MOPS_ERROR,fname,"could not close curve");
	}
      else
	{
	  if((curve->knot_type == MOPS_KTNURB) ||
	     (curve->knot_type == MOPS_KTBEZIER))
	    {
	      mops_error(interp, MOPS_WARNING, fname,
			 "changing knot type to B-Spline!");
	      curve->knot_type = MOPS_KTBSPLINE;
	      mops_status = mops_create_nurbcurve_knots(curve);

	      if(mops_status)
		mops_error(interp, MOPS_ERROR, fname,
			   "error creating new knots");
	    }
	}
    }
  else
    {
      curve->closed = new_closed;
    }

  /*  if(curve->mpoints)*/
  mops_status = mops_nurbcurve_recreatemp(curve);

  MOPS_CRERROR

  mops_status = mops_notify_parent();

 return TCL_OK;
}


/* C -> Tcl! */
int
mops_get_npattrprop_tcmd(ClientData clientData, Tcl_Interp * interp,
			 int argc, char *argv[])
{
 mops_object *o = NULL;
 mops_list_object *sel = mops_current_selection;
 char *n1="npattrPropData";
 mops_nurbpatch_object *patch = NULL;
 int i;
 Tcl_Obj *to = NULL, *toa = NULL, *ton = NULL;

  if(!sel) /* oops? */
    return TCL_OK;

  o = sel->object; 
  patch = (mops_nurbpatch_object *) o->object;

  toa = Tcl_NewStringObj(n1,-1);
  ton = Tcl_NewStringObj("Width",-1);
  to = Tcl_NewIntObj(patch->width);
  Tcl_ObjSetVar2(interp,toa,ton,to,TCL_LEAVE_ERR_MSG | TCL_GLOBAL_ONLY);

  Tcl_SetStringObj(ton,"Height",-1);
  to = Tcl_NewIntObj(patch->height);
  Tcl_ObjSetVar2(interp,toa,ton,to,TCL_LEAVE_ERR_MSG | TCL_GLOBAL_ONLY);

  Tcl_SetStringObj(ton,"Order_U",-1);
  to = Tcl_NewIntObj(patch->uorder);
  Tcl_ObjSetVar2(interp,toa,ton,to,TCL_LEAVE_ERR_MSG | TCL_GLOBAL_ONLY);

  Tcl_SetStringObj(ton,"Order_V",-1);
  to = Tcl_NewIntObj(patch->vorder);
  Tcl_ObjSetVar2(interp,toa,ton,to,TCL_LEAVE_ERR_MSG | TCL_GLOBAL_ONLY);

  Tcl_SetStringObj(ton,"Has-Weights",-1);
  to = Tcl_NewIntObj(patch->has_weights);
  Tcl_ObjSetVar2(interp,toa,ton,to,TCL_LEAVE_ERR_MSG | TCL_GLOBAL_ONLY);

  Tcl_SetStringObj(ton,"Knot-Type_U",-1);
  to = Tcl_NewIntObj(patch->uknot_type);
  Tcl_ObjSetVar2(interp,toa,ton,to,TCL_LEAVE_ERR_MSG | TCL_GLOBAL_ONLY);

  Tcl_SetStringObj(ton,"Knot-Type_V",-1);
  to = Tcl_NewIntObj(patch->vknot_type);
  Tcl_ObjSetVar2(interp,toa,ton,to,TCL_LEAVE_ERR_MSG | TCL_GLOBAL_ONLY);

  Tcl_SetVar2(interp,n1,"Knots_U","", TCL_LEAVE_ERR_MSG |
	      TCL_GLOBAL_ONLY); 
  Tcl_SetStringObj(ton,"Knots_U",-1);
  for(i=0; i<patch->width+patch->uorder; i++)
    {
      to = Tcl_NewDoubleObj((patch->uknotv)[i]);
      Tcl_ObjSetVar2(interp,toa,ton,to,TCL_APPEND_VALUE | 
		     TCL_LIST_ELEMENT | TCL_LEAVE_ERR_MSG |
		     TCL_GLOBAL_ONLY);
    }

  Tcl_SetVar2(interp,n1,"Knots_V","", TCL_LEAVE_ERR_MSG |
	      TCL_GLOBAL_ONLY); 
  Tcl_SetStringObj(ton,"Knots_V",-1);
  for(i=0; i<patch->height+patch->vorder; i++)
    {
      to = Tcl_NewDoubleObj((patch->vknotv)[i]);
      Tcl_ObjSetVar2(interp,toa,ton,to,TCL_APPEND_VALUE | 
		     TCL_LIST_ELEMENT | TCL_LEAVE_ERR_MSG |
		     TCL_GLOBAL_ONLY);      
    }

  Tcl_SetStringObj(ton,"Handlesize",-1);
  to = Tcl_NewDoubleObj(patch->handle_size);
  Tcl_ObjSetVar2(interp,toa,ton,to,TCL_LEAVE_ERR_MSG | TCL_GLOBAL_ONLY);

  Tcl_SetStringObj(ton,"W-Handles",-1);
  to = Tcl_NewIntObj(patch->weight_handles);
  Tcl_ObjSetVar2(interp,toa,ton,to,TCL_LEAVE_ERR_MSG | TCL_GLOBAL_ONLY);

  Tcl_SetStringObj(ton,"Tolerance",-1);
  to = Tcl_NewDoubleObj(patch->glu_sampling_tolerance);
  Tcl_ObjSetVar2(interp,toa,ton,to,TCL_LEAVE_ERR_MSG | TCL_GLOBAL_ONLY);

  Tcl_SetStringObj(ton,"Display",-1);
  to = Tcl_NewIntObj(patch->glu_display_mode);
  Tcl_ObjSetVar2(interp,toa,ton,to,TCL_LEAVE_ERR_MSG | TCL_GLOBAL_ONLY);

  Tcl_IncrRefCount(toa);Tcl_DecrRefCount(toa);
  Tcl_IncrRefCount(ton);Tcl_DecrRefCount(ton);
  
 return TCL_OK;
}


/* Tcl -> C! */
int
mops_set_npattrprop_tcmd(ClientData clientData, Tcl_Interp * interp,
			 int argc, char *argv[])
{
 int mops_status = MOPS_OK;
 mops_object *o = NULL;
 mops_list_object *sel = mops_current_selection;
 mops_nurbpatch_object *patch = NULL;
 char *n1="npattrPropData";
 int new_width, new_height, new_uorder, new_vorder;
 int new_uknot_type, new_vknot_type, new_has_weights; 
 double *old_uknots = NULL, *old_vknots = NULL;
 int updateKnots = 0;
 int knotc, i;
 double *nknotv;
 char **knotv;
 Tcl_Obj *to = NULL, *toa = NULL, *ton = NULL;
 char fname[] = "setNPAttr";

  if(!sel) /* oops? */
    return TCL_OK;

  o = sel->object; 
  patch = (mops_nurbpatch_object *) o->object;

  mops_clear_point_selection(o);

  toa = Tcl_NewStringObj(n1,-1);

  ton = Tcl_NewStringObj("Width",-1);
  to = Tcl_ObjGetVar2(interp,toa,ton,TCL_LEAVE_ERR_MSG | TCL_GLOBAL_ONLY);
  Tcl_GetIntFromObj(interp,to, &new_width);

  Tcl_SetStringObj(ton,"Height",-1);
  to = Tcl_ObjGetVar2(interp,toa,ton,TCL_LEAVE_ERR_MSG | TCL_GLOBAL_ONLY);
  Tcl_GetIntFromObj(interp,to, &new_height);

  Tcl_SetStringObj(ton,"Order_U",-1);
  to = Tcl_ObjGetVar2(interp,toa,ton,TCL_LEAVE_ERR_MSG | TCL_GLOBAL_ONLY);
  Tcl_GetIntFromObj(interp,to, &new_uorder);

  Tcl_SetStringObj(ton,"Order_V",-1);
  to = Tcl_ObjGetVar2(interp,toa,ton,TCL_LEAVE_ERR_MSG | TCL_GLOBAL_ONLY);
  Tcl_GetIntFromObj(interp,to, &new_vorder);

  Tcl_SetStringObj(ton,"Has-Weights",-1);
  to = Tcl_ObjGetVar2(interp,toa,ton,TCL_LEAVE_ERR_MSG | TCL_GLOBAL_ONLY);
  Tcl_GetIntFromObj(interp,to, &new_has_weights);

  Tcl_SetStringObj(ton,"Knot-Type_U",-1);
  to = Tcl_ObjGetVar2(interp,toa,ton,TCL_LEAVE_ERR_MSG | TCL_GLOBAL_ONLY);
  Tcl_GetIntFromObj(interp,to, &new_uknot_type);

  Tcl_SetStringObj(ton,"Knot-Type_V",-1);
  to = Tcl_ObjGetVar2(interp,toa,ton,TCL_LEAVE_ERR_MSG | TCL_GLOBAL_ONLY);
  Tcl_GetIntFromObj(interp,to, &new_vknot_type);


  Tcl_SetStringObj(ton,"Handlesize",-1);
  to = Tcl_ObjGetVar2(interp,toa,ton,TCL_LEAVE_ERR_MSG | TCL_GLOBAL_ONLY);
  Tcl_GetDoubleFromObj(interp,to, &(patch->handle_size));

  Tcl_SetStringObj(ton,"W-Handles",-1);
  to = Tcl_ObjGetVar2(interp,toa,ton,TCL_LEAVE_ERR_MSG | TCL_GLOBAL_ONLY);
  Tcl_GetIntFromObj(interp,to, &(patch->weight_handles));

  Tcl_SetStringObj(ton,"Tolerance",-1);
  to = Tcl_ObjGetVar2(interp,toa,ton,TCL_LEAVE_ERR_MSG | TCL_GLOBAL_ONLY);
  Tcl_GetDoubleFromObj(interp,to, &(patch->glu_sampling_tolerance));

  Tcl_SetStringObj(ton,"Display",-1);
  to = Tcl_ObjGetVar2(interp,toa,ton,TCL_LEAVE_ERR_MSG | TCL_GLOBAL_ONLY);
  Tcl_GetIntFromObj(interp,to, &(patch->glu_display_mode));

  Tcl_IncrRefCount(toa);Tcl_DecrRefCount(toa);
  Tcl_IncrRefCount(ton);Tcl_DecrRefCount(ton);

  if((new_width != patch->width) && (new_width > 1))
    {
      mops_status = mops_nurbpatch_convert_width(patch, new_width);
      if(mops_status)
       mops_error(interp,MOPS_ERROR,fname,"could not convert width of patch");

      updateKnots = 1;
    }

  if((new_height != patch->height) && (new_height > 1))
    {
      mops_status = mops_nurbpatch_convert_height(patch, new_height);
      if(mops_status)
       mops_error(interp,MOPS_ERROR,fname,"could not convert height of patch");

      updateKnots = 1;
    }

  if((patch->uorder != new_uorder) && (new_uorder > 1))
    {
      patch->uorder = new_uorder;
      updateKnots = 1;
    }

  if((patch->vorder != new_vorder) && (new_vorder > 1))
    {
      patch->vorder = new_vorder;
      updateKnots = 1;
    }

  if(patch->has_weights != new_has_weights)
    mops_status = mops_nurbpatch_convert_weight(patch, new_has_weights);

  if(new_uknot_type != patch->uknot_type)
    {
      patch->uknot_type = new_uknot_type;
      updateKnots = 1;
    }

  if(new_vknot_type != patch->vknot_type)
    {
      patch->vknot_type = new_vknot_type;
      updateKnots = 1;
    }

  if(patch->uknot_type == MOPS_KTBEZIER)
    if(patch->uorder != patch->width)
      {
	mops_error(interp,MOPS_WARNING,fname,
		   "changing uorder to match width!");
	patch->uorder = patch->width;
      }

  if(patch->width < patch->uorder)
    {
      mops_error(interp,MOPS_WARNING,fname,
		 "lowering uorder to match width!");
      patch->uorder = patch->width;
    }

  if(patch->vknot_type == MOPS_KTBEZIER)
    if(patch->vorder != patch->height)
      {
	mops_error(interp,MOPS_WARNING,fname,
		   "changing vorder to match height!");
	patch->vorder = patch->height;
      }

  if(patch->height < patch->vorder)
    {
      mops_error(interp,MOPS_WARNING,fname,
		 "lowering vorder to match height!");
      patch->vorder = patch->height;
    }

  if(updateKnots)
    mops_status = mops_create_nurbpatch_knots(patch);

  updateKnots = MOPS_FALSE;

  /* decompose uknot-list */
  if(patch->uknot_type == MOPS_KTCUSTOM)
    {
      Tcl_SplitList(interp,Tcl_GetVar2(interp, n1, "Knots_U",
				       TCL_LEAVE_ERR_MSG | TCL_GLOBAL_ONLY),
		    &knotc, &knotv);

      if(!(nknotv = calloc(knotc, sizeof(double))))
	{
	  mops_error(interp,MOPS_OUTOFMEM,fname,NULL);
	  Tcl_Free((char *) knotv);
	  return TCL_OK;
	}

      for(i = 0; i < knotc; i++)
	{
	  Tcl_GetDouble(interp,knotv[i],&nknotv[i]);
	} /* for */

      if(!(mops_status = mops_check_knots(new_width,new_uorder,knotc,nknotv)))
	{/* the knots are ok */
	  free(patch->uknotv);
	  patch->uknotv = nknotv;
	}
      else
	{/* the knots are wrong */
	  /* first, tell the user what went wrong */
	  switch (mops_status)
	    {
	    case 1:
	      mops_error(interp,MOPS_ERROR,fname,
			 "uknot sequence has too few knots");
	      break;
	    case 2:
	      mops_error(interp,MOPS_ERROR,fname,
			 "uknot sequence has too much knots");
	      break;
	    case 3:
	      mops_error(interp,MOPS_ERROR,fname,
			 "uknot multiplicity higher than order");
	      break;
	    case 4:
	      mops_error(interp,MOPS_ERROR,fname,
			 "uknot sequence has decreasing knots");
	      break;

	    }

	  free(nknotv);

	  /* create new knots */
	  mops_error(interp,MOPS_WARNING,fname,
			 "falling back to uknot type NURB!");
	  patch->uknot_type = MOPS_KTNURB;
	  updateKnots = MOPS_TRUE;

	}

      Tcl_Free((char *) knotv);
    }

  /* decompose vknot-list */
  if(patch->vknot_type == MOPS_KTCUSTOM)
    {
      Tcl_SplitList(interp,Tcl_GetVar2(interp, n1, "Knots_V",
				       TCL_LEAVE_ERR_MSG | TCL_GLOBAL_ONLY),
		    &knotc, &knotv);
      if(!(nknotv = calloc(knotc, sizeof(double))))
	{
	  mops_error(interp,MOPS_OUTOFMEM,fname,NULL);
	  Tcl_Free((char *) knotv);
	  return TCL_OK;
	}

      for(i = 0; i < knotc; i++)
	{
	  Tcl_GetDouble(interp,knotv[i],&nknotv[i]);
	} /* for */

      if(!(mops_status = mops_check_knots(new_height,new_vorder,knotc,nknotv)))
	{/* the knots are ok */
	  free(patch->vknotv);
	  patch->vknotv = nknotv;
	}
      else
	{/* the knots are wrong */
	  /* first, tell the user what went wrong */
	  switch (mops_status)
	    {
	    case 1:
	      mops_error(interp,MOPS_ERROR,fname,
			 "vknot sequence has too few knots");
	      break;
	    case 2:
	      mops_error(interp,MOPS_ERROR,fname,
			 "vknot sequence has too much knots");
	      break;
	    case 3:
	      mops_error(interp,MOPS_ERROR,fname,
			 "vknot multiplicity higher than order");
	      break;
	    case 4:
	      mops_error(interp,MOPS_ERROR,fname,
			 "vknot sequence has decreasing knots");
	      break;

	    }

	  free(nknotv);

	  /* create new knots */
	  mops_error(interp,MOPS_WARNING,fname,
			 "falling back to vknot type NURB!");
	  patch->vknot_type = MOPS_KTNURB;
	  updateKnots = MOPS_TRUE;

	}

      Tcl_Free((char *) knotv);
    }

  if(updateKnots)
    {
      if(patch->uknot_type == MOPS_KTCUSTOM)
	old_uknots = patch->uknotv;

      if(patch->vknot_type == MOPS_KTCUSTOM)
	old_vknots = patch->vknotv;

      mops_status = mops_create_nurbpatch_knots(patch);

      if(patch->uknot_type == MOPS_KTCUSTOM)
	{
	  free(patch->uknotv);
	  patch->uknotv = old_uknots;
	}
      if(patch->vknot_type == MOPS_KTCUSTOM)
	{
	  free(patch->vknotv);
	  patch->vknotv = old_vknots;
	}

    }

  mops_status = mops_notify_parent();

 return TCL_OK;
}


int
mops_get_nurbpatch_properties(Tcl_Interp *interp)
{
 int mops_status = MOPS_OK;

  /* Surface-Shader */
  mops_status = mops_add_surfshader_property(interp);

  /* Displacement-Shader */
  mops_status = mops_add_dispshader_property(interp);

  /* Interior-Shader */
  mops_status = mops_add_intshader_property(interp);

  /* Exterior-Shader */
  mops_status = mops_add_extshader_property(interp);

  /* NURB-patch-attributes */
  Tcl_SetVar(interp,"propertyList","NPatchAttrib", TCL_APPEND_VALUE |\
	     TCL_LIST_ELEMENT | TCL_LEAVE_ERR_MSG | TCL_GLOBAL_ONLY);

  Tcl_SetVar(interp,"propertyProcList","buildNPAttrGUI", TCL_APPEND_VALUE |\
	     TCL_LIST_ELEMENT | TCL_LEAVE_ERR_MSG | TCL_GLOBAL_ONLY);

  Tcl_SetVar(interp,"propertyGetProcList","getNPAttrProp",
	     TCL_APPEND_VALUE | TCL_LIST_ELEMENT |\
	     TCL_LEAVE_ERR_MSG | TCL_GLOBAL_ONLY);

  Tcl_SetVar(interp,"propertySetProcList","setNPAttrProp",
	     TCL_APPEND_VALUE | TCL_LIST_ELEMENT |\
	     TCL_LEAVE_ERR_MSG | TCL_GLOBAL_ONLY);

  Tcl_SetVar(interp,"propertyDataList","npattrPropData",
	     TCL_APPEND_VALUE | TCL_LIST_ELEMENT |\
	     TCL_LEAVE_ERR_MSG | TCL_GLOBAL_ONLY);

 return mops_status;
}


int
mops_get_nurbcurve_properties(Tcl_Interp *interp)
{
 int mops_status = MOPS_OK;

  Tcl_SetVar(interp,"propertyList","NCurveAttrib", TCL_APPEND_VALUE |\
	     TCL_LIST_ELEMENT | TCL_LEAVE_ERR_MSG | TCL_GLOBAL_ONLY);

  Tcl_SetVar(interp,"propertyProcList","buildNCAttrGUI", TCL_APPEND_VALUE |\
	     TCL_LIST_ELEMENT | TCL_LEAVE_ERR_MSG | TCL_GLOBAL_ONLY);

  Tcl_SetVar(interp,"propertyGetProcList","getNCAttrProp",
	     TCL_APPEND_VALUE | TCL_LIST_ELEMENT |\
	     TCL_LEAVE_ERR_MSG | TCL_GLOBAL_ONLY);

  Tcl_SetVar(interp,"propertySetProcList","setNCAttrProp",
	     TCL_APPEND_VALUE | TCL_LIST_ELEMENT |\
	     TCL_LEAVE_ERR_MSG | TCL_GLOBAL_ONLY);

  Tcl_SetVar(interp,"propertyDataList","ncattrPropData",
	     TCL_APPEND_VALUE | TCL_LIST_ELEMENT |\
	     TCL_LEAVE_ERR_MSG | TCL_GLOBAL_ONLY);

 return mops_status;
}

/* delete all mpoints from curve c */
void
mops_nurbcurve_clearmp(mops_nurbcurve_object *c)
{
  mops_mpoint_object *next = NULL, *p = NULL;

  if(!c)
    return;

  p = c->mpoints;

  while(p)
    {
      next = p->next;
      if(p->points)
	free(p->points);
      free(p);
      p = next;
    }

  c->mpoints = NULL;

 return;
}


/* collapse selected points */
int
mops_nurbcurve_collapseselp(mops_object *o)
{
 int mops_status = MOPS_OK;
 mops_nurbcurve_object *c = NULL;
 mops_mpoint_object *new = NULL, *p = NULL, *t = NULL, **last = NULL;
 mops_point_object *selp = NULL;
 double *first = NULL;
 int count = 0, i, found = MOPS_FALSE;

  if(!o)
    return MOPS_OK;

  if(!o->type == MOPS_OTNURBCURVE)
    return MOPS_ERROR;

  c = (mops_nurbcurve_object *)o->object;

  selp = o->selp;

  if(!selp)
    return MOPS_OK;

  /* count points to collapse */
  while(selp)
    {
      count++;
      selp = selp->next;
    }

  if(!(new = calloc(1,sizeof(mops_mpoint_object))))
    return MOPS_OUTOFMEM;
  if(!(new->points = calloc(count, sizeof(double *))))
    return MOPS_OUTOFMEM;

  /* fill mpoint */
  selp = o->selp;
  i = 0;
  first = selp->point;
  while(selp)
    {
      new->points[i] = selp->point;
      i++;
      if(selp->homogenous)
	memcpy(selp->point, first, 4*sizeof(double));
      else
	memcpy(selp->point, first, 3*sizeof(double));

      selp = selp->next;
    }
  new->multiplicity = count;

  /* find and delete all mpoints eventually
     containing a selected point */
  selp = o->selp;
  while(selp)
    {
      p = c->mpoints;
      last = &(c->mpoints);
      while(p)
	{
	  found = MOPS_FALSE;
	  for(i = 0; i < p->multiplicity; i++)
	    {
	      if(p->points[i] == selp->point)
		found = MOPS_TRUE;
	    }

	  if(found)
	    {
	      *last = p->next;
	      t = p->next;
	      free(p->points);
	      free(p);
	      p = t;
	    }
	  else
	    {
	      last = &(p->next);
	      p = p->next;
	    }
	}

      selp = selp->next;
    }

  /* link new mpoint */
  new->next = c->mpoints;
  c->mpoints = new;

 return mops_status;
}

/* explode selected mpoints */
int
mops_nurbcurve_explodemp(mops_object *o)
{
 int mops_status = MOPS_OK;
 mops_nurbcurve_object *c = NULL;
 mops_mpoint_object *p = NULL, *t = NULL, **last = NULL;
 mops_point_object *selp = NULL;
 int found = MOPS_FALSE, i;

  if(!o)
    return MOPS_OK;

  if(!o->type == MOPS_OTNURBCURVE)
    return MOPS_ERROR;

  c = (mops_nurbcurve_object *)o->object;

  selp = o->selp;

  if(!selp)
    return MOPS_OK;

  /* we simply delete all mpoints, that contain a selected point */
  while(selp)
    {      
      p = c->mpoints;
      last = &(c->mpoints);
      while(p)
	{
	  found = MOPS_FALSE;
	  for(i = 0; i < p->multiplicity; i++)
	    {
	      if(p->points[i] == selp->point)
		found = MOPS_TRUE;
	    }

	  if(found)
	    {
	      *last = p->next;
	      t = p->next;
	      free(p->points);
	      free(p);
	      p = t;
	    }
	  else
	    {
	      last = &(p->next);
	      p = p->next;
	    }
	}

      selp = selp->next;
    }

 return mops_status;
}


/* recreate mpoints from identical controlpoints */
int
mops_nurbcurve_recreatemp(mops_nurbcurve_object *c)
{
 int mops_status = MOPS_OK;

 mops_mpoint_object *p = NULL, *new = NULL;
 double *ta, **tmp = NULL;
 int found = MOPS_FALSE, a, b, i, j, stride, count;

  if(!c)
    return MOPS_OK;

  mops_nurbcurve_clearmp(c);

  if(!(tmp = calloc(c->length, sizeof(double *))))
    return MOPS_OUTOFMEM;

  stride = c->has_weights?4:3;
  a = 0;
  for(j = 0; j < c->length; j++)
    {
      ta = &(c->controlv[a]);

      /* count identical points */
      count = 0;
      b = 0;
      for(i = 0; i < c->length; i++)
	{
	  if(!memcmp(ta, &(c->controlv[b]), stride*sizeof(double)))
	    {
	      tmp[count] = &(c->controlv[b]);
	      count++;
	    }

	  b += stride;
	}

      /* create new mp, if it is not already there */
      if(count > 1)
	{
	  p = c->mpoints;
	  found = MOPS_FALSE;
	  while(p && !found)
	    {
	      if(!memcmp(ta, p->points[0], stride*sizeof(double)))
		found = MOPS_TRUE;

	      p = p->next;
	    }

	  if(!found)
	    {
	      if(!(new = calloc(1, sizeof(mops_mpoint_object))))
		return MOPS_OUTOFMEM;
	      if(!(new->points = calloc(count, sizeof(double *))))
		return MOPS_OUTOFMEM;
	      new->multiplicity = count;
	      memcpy(new->points, tmp, count*sizeof(double *));

	      new->next = c->mpoints;
	      c->mpoints = new;
	    }

	} /* if */

      a += stride;
    } /* for */


  free(tmp);

 return mops_status;
}


