/*
 * The Mops, a free 3D Tcl/Tk/OpenGL/RenderMan modeler
 *
 * The Mops is copyrighted (C) 1998-1999 by Randolf Schultz
 * (rschultz@informatik.uni-rostock.de)
 *
 * See the file License for details.
 *
 */

#include <limits.h>
#include <float.h>
#include "mops.h"

/* NURB Utils */

int
mops_getuv_frompatch(mops_nurbpatch_object *p, double objX, double objY,
		     double objZ, double *u, double *v, double *err)
{
 int mops_status = MOPS_OK;
 int samples = 10, i = 0, j = 0, k = 0, stride = 0;
 double su = 0.0, sv = 0.0, du = 0.0, dv = 0.0, P[4] = {0}, min_P[4] = {0};
 double distance = 0.0, min_distance = DBL_MAX;
 double startu = 0.0, endu = 0.0, startv = 0.0, endv = 0.0;

  stride = p->has_weights?4:3;

  startu = p->uknotv[p->uorder-1];
  endu = p->uknotv[p->width];
  startv = p->vknotv[p->vorder-1];
  endv = p->vknotv[p->height];

  for(i = 0; i < 3; i++)
    {
      du = (endu - startu)/samples;
      dv = (endv - startv)/samples;

      su = startu;
      /* calc samples */
      for(j = 0; j < samples; j++)
	{
	  sv = startv;
	  for(k = 0; k < samples; k++)
	    {
	      if(p->has_weights)
		{
		  mops_status = mops_nb_SurfacePoint4D(
			p->width-1, p->height-1, p->uorder-1, p->vorder-1,
			p->uknotv, p->vknotv, p->controlv, su, sv, P);

		  P[0] /= P[3];
		  P[1] /= P[3];
		  P[2] /= P[3];
		}
	      else
		mops_status = mops_nb_SurfacePoint3D(
		 p->width-1, p->height-1, p->uorder-1, p->vorder-1,
		 p->uknotv, p->vknotv, p->controlv, su, sv, P);

	      distance = MOPS_VLEN((P[0] - objX),
				   (P[1] - objY),
				   (P[2] - objZ));

	      if(distance < min_distance)
		{
		  min_distance = distance;
		  memcpy(min_P, P, 3*sizeof(double));
		  *u = su;
		  *v = sv;
		  *err = distance;
		}

	      sv += dv;
	    } /* for */

	  su += du;
	} /* for */

      startu = *u-du;
      if(startu < p->uknotv[p->uorder-1])
	startu = p->uknotv[p->uorder-1];
      endu = *u+du;
      if(endu > p->uknotv[p->width])
	endu = p->uknotv[p->width];
      startv = *v-dv;
      if(startv < p->vknotv[p->vorder-1])
	startv = p->vknotv[p->vorder-1];
      endv = *v+dv;
      if(endv > p->vknotv[p->height])
	endv = p->vknotv[p->height];

    } /* for */


/*
  fprintf(stderr,"uv_P: %g %g %g\n",
	  min_P[0], min_P[1], min_P[2]);
*/

 return mops_status;
}


int
mops_interpolate_npatchu(mops_nurbpatch_object *patch)
{
 int mops_status = MOPS_OK;
 int i, k, N, K, stride, ind, ind2, pu;
 double *vk = NULL, *d = NULL, *Pw = NULL, v[3] = {0};
 double *V = NULL, *Q = NULL;

  K = patch->width;
  N = patch->height;
  stride = patch->has_weights?4:3;
  Pw = patch->controlv;
  pu = patch->uorder-1;

  if(!(vk = calloc(K>N?K:N, sizeof(double))))
    return MOPS_OUTOFMEM;

  if(!(d = calloc(N, sizeof(double))))
    {
      free(vk); return MOPS_OUTOFMEM;
    }

  if(!(V = calloc(K+patch->uorder, sizeof(double))))
    {
      free(vk); free(d); return MOPS_OUTOFMEM;
    }

  if(!(Q = calloc(K*4, sizeof(double))))
    {
      free(vk); free(d); free(V); return MOPS_OUTOFMEM;
    }

  /* find average chord length */
  for(i = 0; i < N; i++)
    {
      d[i] = 0.0;
      ind = i*stride;
      ind2 = i*stride;
      for(k = 1; k < K; k++)
	{
	  ind += N*stride;
	  v[0] = Pw[ind]   - Pw[ind2];
	  v[1] = Pw[ind+1] - Pw[ind2+1];
	  v[2] = Pw[ind+2] - Pw[ind2+2];
	  d[i] += MOPS_V3LEN(v);
	  ind2 += N*stride;
	}
    }

  /* create knotv */
  ind = N*stride;
  ind2 = 0;
  vk[0] = 0.0;
  for(k = 1; k < K; k++)
    {
      vk[k] = 0.0;
      for(i = 0; i < N; i++)
	{
	  v[0] = Pw[ind]   - Pw[ind2];
	  v[1] = Pw[ind+1] - Pw[ind2+1];
	  v[2] = Pw[ind+2] - Pw[ind2+2];

	  vk[k] += (MOPS_V3LEN(v)/d[i]);
	  ind += stride;
	  ind2 += stride;
	}

      vk[k] /= N;
      vk[k] += vk[k-1];
    }
  vk[(K>N?K:N)-1] = 1.0;


  for(i = 1; i < (K-pu); i++)
    {
      V[i+pu] = 0.0;

      for(k = i; k < (i+pu); k++)
	{
	  V[i+pu] += vk[k];
	}

      V[i+pu] /= pu;
    }
  for(i = 0; i <= pu; i++)
    V[i] = 0.0;
  for(i = (K/*-pu-1*/); i < (K+patch->uorder); i++)
    V[i] = 1.0;

  /* interpolate */
  for(i = 0; i < N; i++)
    {
      ind = i*stride;
      for(k = 0; k < K; k++)
	{
	  memcpy(&(Q[k*4]), &(Pw[ind]), stride*sizeof(double));
	  if(stride != 4)
	    Q[k*4+3] = 1.0;
	  ind += N*stride;
	}

      mops_status = mops_nb_GlobalInterpolation4D(K-1, Q, vk, V, pu);

      if(mops_status)
	{ free(d); free(vk); free(V); free(Q); return mops_status; }

      ind = i*stride;
      for(k = 0; k < K; k++)
	{
	  memcpy(&(Pw[ind]), &(Q[k*4]), stride*sizeof(double));
	  ind += N*stride;
	}
    }

  free(patch->uknotv);
  patch->uknotv = V;
  patch->uknot_type = MOPS_KTCUSTOM;
  
  free(vk);
  free(d);
  free(Q);
  
 return MOPS_OK;
}


int
mops_reparam_curve(mops_nurbcurve_object *curve, double *knotv)
{
 int mops_status = MOPS_OK;
 int i, j, p, stride, klen, r, s, in,rm;
 double *controlv = NULL, *cknotv = NULL;
 double *X = NULL, *Ufoo = NULL, *Qw = NULL;
 int *Y = NULL;

  stride = curve->has_weights?4:3;
  p = curve->order-1;
  controlv = curve->controlv;
  cknotv = curve->knotv;

  klen = curve->length+curve->order;

  if(!(X = calloc(klen, sizeof(double))))
    return MOPS_OUTOFMEM;
  if(!(Y = calloc(klen, sizeof(int))))
    { free(X); return MOPS_OUTOFMEM; }

  /* find knots to insert */
  r = 0;
  for(i = p; i < curve->length; i++)
    {
      in = MOPS_TRUE;
      for(j = p; j < curve->length; j++)
	{
	  if(cknotv[j] == knotv[i])
	    {
	      in = MOPS_FALSE;
	    }
	}

      if(in)
	{
	  fprintf(stderr,"inserting: %g\n",knotv[i]);
	  X[r] = knotv[i];
	  r++;
	}
    }

  if(!r)
    {
      free(X); free(Y); return MOPS_OK;
    }

  /* find knots to remove after insertion */
  s = 0;
  for(i = p; i < curve->length; i++)
    {
      rm = MOPS_TRUE;
      for(j = p; j < curve->length; j++)
	{
	  if(cknotv[i] == knotv[j])
	    {
	      rm = MOPS_FALSE;
	    }
	}

      if(rm)
	{
	  fprintf(stderr,"removing: %g\n",cknotv[i]);
	  Y[s] = i;
	  s++;
	}
    }

  /* refine curve with new knots */
  if(!(Ufoo = calloc(klen + r, sizeof(double))))
    {
      free(X); free(Y); return MOPS_OUTOFMEM;
    }

  if(!(Qw = calloc((curve->length + r + 2)*stride, sizeof(double))))
    {
      free(X); free(Y); free(Ufoo); return MOPS_OUTOFMEM;
    }


  mops_nb_RefineKnotVectCurve(stride, curve->length-1, p,
			      cknotv, controlv, X, r-1, Ufoo, Qw);

  free(curve->knotv);
  curve->knotv = Ufoo;
  
  free(curve->controlv);
  curve->controlv = Qw;

  curve->length += r;

  /* remove old knots */
  for(i = 0; i < s; i++)
    {
      if(!(Ufoo = calloc(curve->length+curve->order-1, sizeof(double))))
	{
	  free(X); free(Y); return MOPS_OUTOFMEM;
	}

      if(!(Qw = calloc((curve->length - 1)*stride, sizeof(double))))
	{
	  free(X); free(Y); free(Ufoo); return MOPS_OUTOFMEM;
	}

      mops_status = mops_nb_RemoveKnot(stride, curve->length-1,
		     curve->order-1, curve->knotv, curve->controlv,
		     Y[i], 1, 1, Ufoo, Qw);

      if(mops_status)
	{
	  free(X); free(Y); free(Ufoo); free(Qw); return mops_status;
	}

      free(curve->knotv);
      curve->knotv = Ufoo;

      free(curve->controlv);
      curve->controlv = Qw;

      curve->length--;
    } /* for */

  free(X);
  free(Y);

 return MOPS_OK;
}


int
mops_clamp_curve(mops_nurbcurve_object *curve)
{
 int mops_status = MOPS_OK;
 double *newcontrolv = NULL, *newknotv = NULL;
 double u;
 int stride, r, k, s, nq = 0;

  stride = curve->has_weights?4:3;  

  /* clamp start */
  k = curve->order-1;
  u = curve->knotv[k];
  s = 0;
  r = curve->order - 1;
  curve->length += r;

  newcontrolv = NULL;
  if(!(newcontrolv = calloc(curve->length*stride, sizeof(double))))
    return MOPS_OUTOFMEM;

  newknotv = NULL;	  
  if(!(newknotv = calloc(curve->length+curve->order, sizeof(double))))
    return MOPS_OUTOFMEM;

  if(curve->has_weights)
    mops_status = mops_nb_CurveInsertKnot4D(curve->length-r-1,
		       curve->order-1, curve->knotv, curve->controlv, u, k,
		       s, r, &nq, newknotv, newcontrolv);
  else
    mops_status = mops_nb_CurveInsertKnot3D(curve->length-r-1,
			curve->order-1, curve->knotv, curve->controlv, u, k,
			s, r, &nq, newknotv, newcontrolv);

  free(curve->controlv);
  curve->controlv = newcontrolv;

  free(curve->knotv);
  curve->knotv = newknotv;

  /* clamp end */
  k = curve->length;
  u = curve->knotv[k];
  k--;
  s = 0;
  r = curve->order - 1;
  curve->length += r;

  newcontrolv = NULL;
  if(!(newcontrolv = calloc(curve->length*stride, sizeof(double))))
    return MOPS_OUTOFMEM;

  newknotv = NULL;	  
  if(!(newknotv = calloc(curve->length+curve->order, sizeof(double))))
    return MOPS_OUTOFMEM;

  if(curve->has_weights)
    mops_status = mops_nb_CurveInsertKnot4D(curve->length-r-1,
		       curve->order-1, curve->knotv, curve->controlv, u, k,
		       s, r, &nq, newknotv, newcontrolv);
  else
    mops_status = mops_nb_CurveInsertKnot3D(curve->length-r-1,
			curve->order-1, curve->knotv, curve->controlv, u, k,
			s, r, &nq, newknotv, newcontrolv);

  free(curve->controlv);
  curve->controlv = newcontrolv;

  free(curve->knotv);
  curve->knotv = newknotv;

  /* create new controlv, knotv, discarding the first p and last p knots */
  curve->length -= (curve->order-1)*2;
  newcontrolv = NULL;
  if(!(newcontrolv = calloc(curve->length*stride, sizeof(double))))
    return MOPS_OUTOFMEM;
  
  newknotv = NULL;
  if(!(newknotv = calloc(curve->length+curve->order, sizeof(double))))
    return MOPS_OUTOFMEM;

  memcpy(newcontrolv, &(curve->controlv[(curve->order-1)*stride]),
	 curve->length*stride*sizeof(double));

  memcpy(newknotv, &(curve->knotv[curve->order-1]),
	 (curve->length+curve->order)*sizeof(double));

  free(curve->controlv);
  curve->controlv = newcontrolv;

  free(curve->knotv);
  curve->knotv = newknotv;

 return MOPS_OK;
}


int
mops_rescale_knotv(int n, double *knotv)
{
 double *tmpknv, min, max, len;
 int i;

  max = knotv[0];
  min = knotv[0];

  for(i = 0; i < n; i++)
    {
      if(knotv[i] > max)
	max = knotv[i];
     
      if(knotv[i] < min)
	min = knotv[i];
    }

  if(min < 0.0 && max > 0.0)
    len = max+fabs(min);
  else
    len = max - min;

  if(!(tmpknv = calloc(n, sizeof(double))))
    return MOPS_OUTOFMEM;

  memcpy(tmpknv, knotv, n*sizeof(double));

  knotv[0] = 0.0;
  for(i = 1; i < n-1; i++)
    {
      knotv[i] = (tmpknv[i] - tmpknv[0])/len;
    }
  knotv[n-1] = 1.0;

  free(tmpknv);

 return MOPS_OK;
}


int
mops_unify_knots(double *Ua, int Ualen, double *Ub, int Ublen,
		 double **Ubar, int *Ubarlen)
{
 int i = 0, ia = 0, ib = 0, done = MOPS_FALSE;
 double t = 0.0, *U = NULL;

  if(!(U = calloc(Ualen+Ublen, sizeof(double))))
   {
     return MOPS_OUTOFMEM;
   }

  while(!done)
    {
      if(Ua[ia] == Ub[ib])
	{
	  t = Ua[ia];
	  ia++;
	  ib++;
	}
      else
	{
	  if(Ua[ia] < Ub[ib])
	    {
	      t = Ua[ia];
	      ia++;
	    }
	  else
	    {
	      t = Ub[ib];
	      ib++;
	    }
	}
      U[i] = t;
      i++;
      if((ia >= Ualen || ib >= Ublen))
	done = MOPS_TRUE;
    }


  if(!(*Ubar = realloc(U, i*sizeof(double))))
    {
      return MOPS_OUTOFMEM;
    }

  /*
  if(!(*Ubar = calloc(i, sizeof(double))))
    {
      return MOPS_OUTOFMEM;
    }

  memcpy(*Ubar, U, i*sizeof(double));

  free(U);
  */

  *Ubarlen = i;

 return MOPS_OK;
}


int
mops_merge_knots(mops_nurbcurve_object *curve, double *Ubar, int Ubarlen)
{
 double *X = NULL, *Ufoo = NULL, *U = NULL, *Qw = NULL;
 int stride, r, ia, ib, done;

  stride = curve->has_weights?4:3;

  if(!(X = calloc(Ubarlen, sizeof(double))))
   {
     return MOPS_OUTOFMEM;
   }
  
  U = curve->knotv;

  /* find knots to insert */
  r = 0;
  ia = 0;
  ib = 0;
  done = MOPS_FALSE;
  while(!done)
    {
      if(Ubar[ib] == U[ia])
	{
	  ib++;
	  ia++;
	}
      else
	{
	  X[r] = Ubar[ib];
	  r++;
	  ib++;
	}
      done = ((ia >= (curve->length+curve->order)) || (ib >= Ubarlen));
    }

  
  if(r == 0)
    {
      free(X); return MOPS_OK;
    }

  if(!(Ufoo = calloc((curve->length + curve->order + r), 
		     sizeof(double))))
    {
      free(X); return MOPS_OUTOFMEM;
    }

  if(!(Qw = calloc((curve->length + r+2)*(curve->has_weights?4:3), 
		   sizeof(double))))
    {
      free(X); free(Ufoo); return MOPS_OUTOFMEM;
    }

  mops_nb_RefineKnotVectCurve(stride, curve->length-1,
			      curve->order-1, curve->knotv, curve->controlv,
			      X, r-1, Ufoo, Qw);

  free(curve->knotv);
  curve->knotv = Ufoo;
  
  free(curve->controlv);
  curve->controlv = Qw;

  free(X);
	  
  curve->length += r;

 return MOPS_OK;
}


int
mops_skin_nurbcurves(mops_object *curves, int order, int knot_type,
		     int interpolate, mops_nurbpatch_object **skin)
{
 int mops_status = MOPS_OK;
 mops_object *o = NULL;
 mops_nurbcurve_object *curve = NULL;
 int numcurves = 0, max_order = 0;
 int stride, nh = 0, numknots = 0, t = 0, i, j, a, b;
 int Ualen, Ublen, Ubarlen, clamp_me;
 double *Uh = NULL, *Qw = NULL, *realUh = NULL, *realQw = NULL;
 double *Ubar = NULL, *Ua = NULL, *Ub = NULL;
 double *skc = NULL, u = 0.0;
 GLdouble m[16] = {0};
 GLfloat mr[4][4];

  /* clamp curves */
  o = curves;
  while(o)
    {
      curve = (mops_nurbcurve_object *) o->object;
      clamp_me = MOPS_FALSE;
      if(curve->knot_type == MOPS_KTBSPLINE)
	{
	  clamp_me = MOPS_TRUE;
	}
      else
	{
	  if(curve->knot_type == MOPS_KTCUSTOM)
	    {
	      a = 1;
	      u = curve->knotv[0];
	      for(i = 1; i < curve->order; i++)
		if(u == curve->knotv[i])
		  a++;

	      j = curve->length+curve->order-1;
	      b = 1;
	      u = curve->knotv[j];
	      for(i = j-1; i >= curve->length; i--)
		if(u == curve->knotv[i])
		  b++;

	      if((a < curve->order) || (b < curve->order))
		{
		  clamp_me = MOPS_TRUE;
		}
	    }
	}

      if(clamp_me)
	mops_status = mops_clamp_curve(curve);
      
      o = o->next;
    }

  /* rescale knots to range 0.0 - 1.0 */
  o = curves;
  while(o)
    {
      curve = (mops_nurbcurve_object *) o->object;
      if(curve->knotv[0] != 0.0 || curve->knotv[
	  curve->length+curve->order-1] != 1.0)
	{
	  mops_status = mops_rescale_knotv(curve->length+curve->order,
					   curve->knotv);
	}
      o = o->next;
    }

  /* find max order */
  o = curves;
  while(o)
    {
      curve = (mops_nurbcurve_object *) o->object;
      if(curve->order > max_order)
	max_order = curve->order;

      numcurves++;
      o = o->next;
    }

  /* degree elevate */
  o = curves;
  while(o)
    {
      curve = (mops_nurbcurve_object *) o->object;
      if(curve->order < max_order)
	{
	  stride = curve->has_weights?4:3;
	  t = max_order - curve->order;

	  /* alloc new knotv & new controlv */
	  if(!(Uh = calloc((curve->length + curve->length*t +
			    curve->order + t), 
			   sizeof(double))))
	    {
	      return MOPS_OUTOFMEM;
	    }
	  if(!(Qw = calloc((curve->length + curve->length*t)*
			   (curve->has_weights?4:3), sizeof(double))))
	    {
	      free(Uh);
	      return MOPS_OUTOFMEM;
	    }

	  mops_status = mops_nb_DegreeElevateCurve(stride,
	    curve->length-1, curve->order-1, curve->knotv, curve->controlv, 
	    t, &nh, Uh, Qw);

	  if(mops_status)
	    {
	      free(Uh); free(Qw); return mops_status;
	    }

	  if(!(realQw = realloc(Qw, nh*(curve->has_weights?4:3)*
				sizeof(double))))
	    {
	      return MOPS_OUTOFMEM;
	    }

	  if(!(realUh = realloc(Uh, (nh+curve->order+t)*sizeof(double))))
	    {
	      return MOPS_OUTOFMEM;
	    }

	  free(curve->knotv);
	  curve->knotv = realUh;

	  free(curve->controlv);
	  curve->controlv = realQw;

	  curve->knot_type = MOPS_KTCUSTOM;
	  
	  curve->order += t;
	  
	  curve->length = nh;

	  numknots += (curve->order + curve->length);

	  Qw = NULL;
	  Uh = NULL;
	  realQw = NULL;
	  realUh = NULL;
	} /* if */
      o = o->next;
    } /* while */

  /* unify knots */
  o = curves;
  curve = (mops_nurbcurve_object *) o->object;
  Ua = curve->knotv;
  Ualen = curve->length+curve->order;

  o = o->next;
  while(o)
    {
      curve = (mops_nurbcurve_object *)o->object;
      Ub = curve->knotv;
      Ublen = curve->length+curve->order;

      mops_status = mops_unify_knots(Ua, Ualen, Ub, Ublen, &Ubar, &Ubarlen);

      if(mops_status)
	{
	  fprintf(stderr,"Mops: Memory may have leaked!");
	  return mops_status;
	}

      Ua = Ubar;
      Ualen = Ubarlen;

      o = o->next;
    }

  /* merge knots */
  o = curves;
  while(o)
    {
      curve = (mops_nurbcurve_object *) o->object;

      mops_status = mops_merge_knots(curve, Ubar, Ubarlen);
      if(mops_status)
	{
	  free(Ubar); return mops_status;
	}
      o = o->next;
    }

  /* construct patch */
  o = curves;
  curve = (mops_nurbcurve_object *) o->object;

  if(!(skc = calloc((curve->length * numcurves * 4), sizeof(double))))
    {
      free(Ubar); return MOPS_OUTOFMEM;
    }

  b = 0;
  i = 0;

  while(o)
    {
      curve = (mops_nurbcurve_object *) o->object;
      /* get curves transformation-matrix */
      glMatrixMode (GL_MODELVIEW);
      glPushMatrix();
       glLoadIdentity ();
      
       glTranslated(o->movx, o->movy, o->movz);
	
       mops_build_rotmatrix(mr, o->quat);
       glMultMatrixf(&mr[0][0]);
  
       glScaled (o->scalx, o->scaly, o->scalz);
       glGetDoublev(GL_MODELVIEW_MATRIX, m);
      glPopMatrix();

      stride = curve->has_weights?4:3;

      for(i = 0; i < curve->length; i++)
	{
	  a = i*stride;

	  if(stride == 3)
	    mops_apply_transform3(&(curve->controlv[a]), m);
	  else
	    mops_apply_transform4(&(curve->controlv[a]), m);

	  memcpy(&(skc[b]), &(curve->controlv[a]), stride*sizeof(double));

	  if(!curve->has_weights)
	    skc[b+3] = 1.0;

	  b += 4;
	}

      o = o->next;
    }

  if(order > numcurves)
    order = numcurves;
  if(order < 2)
    order = 2;

  if(knot_type == MOPS_KTBEZIER)
    if(order < numcurves)
      order = numcurves;

  mops_status = mops_create_nurbpatch(order, curve->order,
				      numcurves, curve->length,
				      knot_type, MOPS_KTCUSTOM, MOPS_TRUE,
				      skc, NULL, Ubar, skin);

  if(mops_status)
    {
      free(Ubar); return mops_status;
    }

  if(interpolate)
    mops_status = mops_interpolate_npatchu(*skin);

 return mops_status;
}


void
mops_nurbcurve_findufromcontrol(int index, int nknots, int order,
				double *knots,
				double *u, int *k, int *s)
{
 int j=0, i=0;


 while(knots[j]==knots[j+1])
   j++;

 /*  j = order-2;*/

  while(j<(nknots-1))
    {
      *s = 0;
      if(i == index)
	{
	  if(knots[j] == knots[j+1])
	    {
	      *k = j;
	      *u = knots[j];
	    }
	  else
	    {
	      *k = j;
	      *u = knots[j]+(knots[j+1]-knots[j])/2.0;
	    }
	  return;
	}

      i++;
      j++;
    }

 return;
}


int
mops_create_nurbcircle(mops_nurbcurve_object **curve)
{
 int mops_status = MOPS_OK;
 mops_nurbcurve_object *new = NULL;
 double *controlv = NULL, *knotv = NULL;
 int i = 0;
 double controls[36] = {
   0.0,1.0,0.0,1.0,
   1.0,1.0,0.0,1.0,
   1.0,0.0,0.0,1.0,
   1.0,-1.0,0.0,1.0,
   0.0,-1.0,0.0,1.0, 
   -1.0,-1.0,0.0,1.0, 
   -1.0,0.0,0.0,1.0, 
   -1.0,1.0,0.0,1.0,
   0.0,1.0,0.0,1.0
 };
 double knots[12] = {
   0.0, 0.0, 0.0, 0.25, 0.25, 0.5, 0.5, 0.75, 0.75, 1.0, 1.0, 1.0
 };

  i = 7;
  controls[i] = sqrt(2.0)/2.0;
  controls[i-2] *= sqrt(2.0)/2.0;
  controls[i-3] *= sqrt(2.0)/2.0;
  i+=8;
  controls[i] = sqrt(2.0)/2.0;
  controls[i-2] *= sqrt(2.0)/2.0;
  controls[i-3] *= sqrt(2.0)/2.0;
  i+=8;
  controls[i] = sqrt(2.0)/2.0;
  controls[i-2] *= sqrt(2.0)/2.0;
  controls[i-3] *= sqrt(2.0)/2.0;
  i+=8;
  controls[i] = sqrt(2.0)/2.0;
  controls[i-2] *= sqrt(2.0)/2.0;
  controls[i-3] *= sqrt(2.0)/2.0;

  if(!(new = calloc(1, sizeof(mops_nurbcurve_object))))
    return MOPS_OUTOFMEM;
  if(!(controlv = calloc(9*4, sizeof(double))))
    { free(new); return MOPS_OUTOFMEM; }
  if(!(knotv = calloc(12, sizeof(double))))
    { free(new); free(controlv); return MOPS_OUTOFMEM; }
  
  new->order = 3;
  new->length = 9;
  new->knot_type = MOPS_KTCUSTOM;
  new->has_weights = MOPS_TRUE;
 
  memcpy(controlv,controls,9*4*sizeof(double)); 
  new->controlv = controlv;
  
  memcpy(knotv,knots,12*sizeof(double)); 
  new->knotv = knotv;
 
  new->glu_sampling_tolerance = 30.0;
  new->handle_size = 6.0;

  *curve = new;

 return mops_status;
}

int
mops_create_nurbcirclearc(double arc, mops_nurbcurve_object **curve)
{
 int mops_status = MOPS_OK;
 mops_nurbcurve_object *new = NULL;

  if(!(new = calloc(1, sizeof(mops_nurbcurve_object))))
    return MOPS_OUTOFMEM;

  if(arc<0.0)
    {
      mops_status = mops_nb_CreateNurbsCircle(1.0, arc, 0.0, &(new->length),
					      &new->knotv, &new->controlv);
    }
  else
    {
      mops_status = mops_nb_CreateNurbsCircle(1.0, 0.0, arc, &(new->length),
					      &new->knotv, &new->controlv);
    }

  if(mops_status)
    { free(new); return mops_status; }

  new->order = 3;
  new->knot_type = MOPS_KTCUSTOM;
  new->has_weights = MOPS_TRUE;
  new->glu_sampling_tolerance = 30.0;
  new->handle_size = 6.0;

  *curve = new;

 return mops_status;
}

/* 
   controls for Cobb-NURBSphere taken from:
   "NURB Curves and Surfaces, from Projective Geometry to Practical Use",
   by G. Farin
 */

int
mops_create_nurbcobbsphere(mops_nurbpatch_object **patch)
{
 int mops_status = MOPS_OK;
 mops_nurbpatch_object *new = NULL;
 double *controls = NULL;
 double t = 0.0, d = 0.0;

  if(!(controls=calloc(100, sizeof(double))))
    return MOPS_OUTOFMEM;

  t = sqrt(3.0);
  d = sqrt(2.0);

  controls[0] = 4.0*(1.0-t);
  controls[1] = 4.0*(1.0-t);
  controls[2] = 4.0*(1.0-t);
  controls[3] = 4.0*(3.0-t);

  controls[4] = -d;
  controls[5] = d * (t - 4.0);
  controls[6] = d * (t - 4.0);
  controls[7] = d * (3.0*t - 2.0);
  
  controls[8] = 0.0;
  controls[9] = 4.0*(1.0-2.0*t)/3.0;
  controls[10] = 4.0*(1.0-2.0*t)/3.0;
  controls[11] = 4.0*(5.0-t)/3.0;

  controls[12] = d;
  controls[13] = d * (t - 4.0);
  controls[14] = d * (t - 4.0);
  controls[15] = d * (3*t - 2.0);

  controls[16] = 4.0*(t - 1.0);
  controls[17] = 4.0*(1.0-t);
  controls[18] = 4.0*(1.0-t);
  controls[19] = 4.0*(3.0-t);

  controls[20] = d*(t - 4.0);
  controls[21] = -d;
  controls[22] = d*(t-4.0);
  controls[23] = d*(3.0*t-2.0);

  controls[24] = (2.0-3.0*t)/2.0;
  controls[25] = (2.0-3.0*t)/2.0;
  controls[26] = -(t+6.0)/2.0;
  controls[27] = (t+6.0)/2.0;

  controls[28] = 0.0;
  controls[29] = d*(2.0*t-7.0)/3.0;
  controls[30] = -5.0*sqrt(6.0)/3.0;
  controls[31] = d*(t+6.0)/3.0;

  controls[32] = (3.0*t-2.0)/2.0;
  controls[33] = (2.0-3.0*t)/2.0;
  controls[34] = -(t+6.0)/2.0;
  controls[35] = (t+6.0)/2.0;

  controls[36] = d*(4.0-t);
  controls[37] = -d;
  controls[38] = d*(t-4.0);
  controls[39] = d*(3.0*t-2.0);

  controls[40] = 4.0*(1.0-2*t)/3.0;
  controls[41] = 0.0;
  controls[42] = 4.0*(1.0-2*t)/3.0;
  controls[43] = 4.0*(5.0-t)/3.0;

  controls[44] = d*(2.0*t-7.0)/3.0;
  controls[45] = 0.0;
  controls[46] = -5.0*sqrt(6.0)/3.0;
  controls[47] = d*(t+6.0)/3.0;

  controls[48] = 0.0;
  controls[49] = 0.0;
  controls[50] = 4.0*(t-5.0)/3.0;
  controls[51] = 4.0*(5.0*t-1.0)/9.0;

  controls[52] = -d*(2*t-7.0)/3.0;
  controls[53] = 0.0;
  controls[54] = -5.0*sqrt(6.0)/3.0;
  controls[55] = d*(t+6.0)/3.0;

  controls[56] = -4.0*(1.0-2.0*t)/3.0;
  controls[57] = 0.0;
  controls[58] = 4.0*(1.0-2.0*t)/3.0;
  controls[59] = 4.0*(5.0-t)/3.0;

  controls[60] = d*(t-4.0);
  controls[61] = d;
  controls[62] = d*(t-4.0);
  controls[63] = d*(3.0*t-2);

  controls[64] = (2.0-3.0*t)/2.0;
  controls[65] = -(2.0-3.0*t)/2.0;
  controls[66] = -(t+6.0)/2.0;
  controls[67] = (t+6.0)/2.0;

  controls[68] = 0.0;
  controls[69] = -d*(2.0*t-7.0)/3.0;
  controls[70] = -5.0*sqrt(6.0)/3.0;
  controls[71] = d*(t+6.0)/3.0;

  controls[72] = (3.0*t-2.0)/2.0;
  controls[73] = -(2.0-3.0*t)/2.0;
  controls[74] = -(t+6.0)/2.0;
  controls[75] = (t+6.0)/2.0;

  controls[76] = d*(4.0-t);
  controls[77] = d;
  controls[78] = d*(t-4.0);
  controls[79] = d*(3.0*t-2.0);

  controls[80] = 4.0*(1.0-t);
  controls[81] = -4.0*(1.0-t);
  controls[82] = 4.0*(1.0-t);
  controls[83] = 4.0*(3.0-t);

  controls[84] = -d;
  controls[85] = -d*(t-4.0);
  controls[86] = d*(t-4.0);
  controls[87] = d*(3.0*t-2.0);

  controls[88] = 0.0;
  controls[89] = -4.0*(1.0-2.0*t)/3.0;
  controls[90] = 4.0*(1.0-2.0*t)/3.0;
  controls[91] = 4.0*(5.0-t)/3.0;

  controls[92] = d;
  controls[93] = -d*(t-4.0);
  controls[94] = d*(t-4.0);
  controls[95] = d*(3.0*t-2.0);

  controls[96] = 4.0*(t-1.0);
  controls[97] = -4.0*(1.0-t);
  controls[98] = 4.0*(1.0-t);
  controls[99] = 4.0*(3.0-t);

  mops_status = mops_create_nurbpatch(5, 5, 5, 5,
				      MOPS_KTNURB,MOPS_KTNURB,
				      MOPS_TRUE,controls,NULL,NULL,&new);

  if(mops_status)
    {
      free(controls);
      return mops_status;
    }

 new->glu_sampling_tolerance = 30.0;
 new->handle_size = 6.0;

 *patch = new;

 return mops_status;
}


int
mops_create_nurbhcircle(mops_nurbcurve_object **curve)
{
 int mops_status = MOPS_OK;
 mops_nurbcurve_object *new = NULL;
 double *controlv = NULL, *knotv = NULL;
 int i = 0;
 double controls[20] = {
   0.0,1.0,0.0,1.0,
   1.0,1.0,0.0,1.0,
   1.0,0.0,0.0,1.0,
   1.0,-1.0,0.0,1.0,
   0.0,-1.0,0.0,1.0, 
 };
 double knots[8] = {
   0.0, 0.0, 0.0, 0.5, 0.5, 1.0, 1.0, 1.0
 };

  i = 4;
  controls[i] *= sqrt(2.0)/2.0;
  controls[i+1] *= sqrt(2.0)/2.0;
  controls[i+3] = sqrt(2.0)/2.0;
  i+=8;
  controls[i] *= sqrt(2.0)/2.0;
  controls[i+1] *= sqrt(2.0)/2.0;
  controls[i+3] = sqrt(2.0)/2.0;


  if(!(new = calloc(1, sizeof(mops_nurbcurve_object))))
    return MOPS_OUTOFMEM;
  if(!(controlv = calloc(5*4, sizeof(double))))
    { free(new); return MOPS_OUTOFMEM; }
  if(!(knotv = calloc(12, sizeof(double))))
    { free(new); free(controlv); return MOPS_OUTOFMEM; }
  
  new->order = 3;
  new->length = 5;
  new->knot_type = MOPS_KTCUSTOM;
  new->has_weights = MOPS_TRUE;

  memcpy(controlv,controls,5*4*sizeof(double)); 
  new->controlv = controlv;

  memcpy(knotv,knots,8*sizeof(double)); 
  new->knotv = knotv;

  new->glu_sampling_tolerance = 30.0;
  new->handle_size = 6.0;

  *curve = new;

 return mops_status;
}


int
mops_sweep_nurb(mops_object *o, mops_object *o2, int sections,
		int rotate, mops_nurbpatch_object **patch)
{
 int mops_status = MOPS_OK;
 mops_nurbpatch_object *new = NULL;
 mops_nurbcurve_object *tr,*cs;
 double *controlv = NULL;
 int i=0, j=0, stride;
 double u, p1[4], p2[4], v1[3];
 double xangle = 0.0, yangle = 0.0, zangle = 0.0, der[3] = {0};
 double plen = 0.0;
 GLdouble m[16] = {0}, mcs[16] = {0}, mtr[16] = {0};
 GLfloat mr[4][4];

  if(!o || !o2 || !patch)
    return MOPS_OK;

  if((o->type != MOPS_OTNURBCURVE) || (o2->type != MOPS_OTNURBCURVE))
    return MOPS_OK;

  cs = (mops_nurbcurve_object *)(o->object);
  tr = (mops_nurbcurve_object *)(o2->object);

  stride = (cs->has_weights?4:3);

  /* get curves transformation-matrix */
  glMatrixMode (GL_MODELVIEW);
  glPushMatrix();
  glLoadIdentity ();
  /*
  glTranslated(o->movx, o->movy, o->movz);
  */
  mops_build_rotmatrix(mr, o->quat);
  glMultMatrixf(&mr[0][0]);
  
  glScaled (o->scalx, o->scaly, o->scalz);
  glGetDoublev(GL_MODELVIEW_MATRIX, mcs);
  glPopMatrix();

  
  glPushMatrix();
  glLoadIdentity ();
  glTranslated(o2->movx, o2->movy, o2->movz);
  mops_build_rotmatrix(mr, o2->quat);
  glMultMatrixf(&mr[0][0]);
  glScaled (o2->scalx, o2->scaly, o2->scalz);
  glGetDoublev(GL_MODELVIEW_MATRIX, mtr);
  glPopMatrix();
  

  /* calloc the new patch */
  if(!(new = calloc(1, sizeof(mops_nurbpatch_object))))
    return MOPS_OUTOFMEM;
  if(!(controlv = calloc(cs->length*(sections+1)*stride, sizeof(double))))
    { free(new); return MOPS_OUTOFMEM; }
  if(!(new->vknotv = calloc(cs->length+cs->order, sizeof(double))))
    { free(new); free(controlv); return MOPS_OUTOFMEM; }
  if(!(new->uknotv = calloc(sections+4, sizeof(double))))
    { free(new->vknotv); free(new); free(controlv); return MOPS_OUTOFMEM; }

  new->vorder = cs->order;
  new->uorder = 4;
  new->controlv = controlv;
  
  new->vknot_type = cs->knot_type;
  new->uknot_type = MOPS_KTNURB;
  new->width = sections+1;
  new->height = cs->length;
  new->has_weights = cs->has_weights;

  new->glu_sampling_tolerance = cs->glu_sampling_tolerance;
  new->handle_size = cs->handle_size;

  mops_status = mops_create_nurbpatch_knots(new);
  if(mops_status)
    { free(new); free(controlv); return mops_status; }

  if(cs->knot_type == MOPS_KTCUSTOM)
    {
      memcpy(new->vknotv,cs->knotv,(cs->length+cs->order)*sizeof(double));
    }

  if(tr->has_weights)
    mops_nb_CurvePoint4D(tr->length-1,tr->order-1,tr->knotv,tr->controlv,
			 tr->knotv[tr->order-1], p1);
  else
    mops_nb_CurvePoint3D(tr->length-1,tr->order-1,tr->knotv,tr->controlv,
			 tr->knotv[tr->order-1], p1);

  plen = fabs(tr->knotv[tr->length] - tr->knotv[tr->order-1]);

  /* copy cross sections controlv section+1 times */
  for(i=0;i<=sections;i++)
    {
      memcpy(&controlv[i*stride*cs->length],cs->controlv,
	     cs->length*stride*sizeof(double));

      u = tr->knotv[tr->order-1]+(((double)i/sections)*plen);

      if(tr->has_weights)
	mops_nb_CurvePoint4D(tr->length-1,tr->order-1,tr->knotv,tr->controlv,
			     u, p2);
      else
	mops_nb_CurvePoint3D(tr->length-1,tr->order-1,tr->knotv,tr->controlv,
			     u, p2);

      if(rotate)
	{
	  xangle = 0.0;
	  yangle = 0.0;
	  zangle = 0.0;

	  if(tr->has_weights)
	    mops_nb_ComputeFirstDer4D(tr->length-1,tr->order-1,tr->knotv,
				      tr->controlv, u, der);
	  else
	    mops_nb_ComputeFirstDer3D(tr->length-1,tr->order-1,tr->knotv,
				      tr->controlv, u, der);

	  if(der[0]==0.0&&der[1]==0.0)
	    zangle = 0.0;
	  else
	    zangle = MOPS_R2D(acos(der[0]/MOPS_V2LEN(der)));
	  if(der[1] < 0.0)
	    zangle = -zangle;

	  v1[0] = der[0]; v1[1] = der[2];
	  if(v1[0]==0.0&&v1[1]==0.0)
	    yangle = 0.0;
	  else
	    yangle = MOPS_R2D(acos(v1[0]/MOPS_V2LEN(v1)));
	  if(v1[0] < 0.0)
	    yangle = 180-yangle;
	  if(v1[1] > 0.0)
	    yangle = -yangle;

          v1[0] = der[1]; v1[1] = der[2];
	  if(v1[0]==0.0&&v1[1]==0.0)
	    xangle = 0.0;
	  else
	    xangle = MOPS_R2D(acos(v1[0]/MOPS_V2LEN(v1)));
	  if(v1[0] < 0.0)
	    xangle = 180-xangle;
	  if(v1[1] > 0.0)
	    xangle = -xangle;

	}
			     
      glPushMatrix();
      glLoadIdentity();

      glMultMatrixd(mtr);

      MOPS_V3SUB(p2,p2,p1)
      glTranslated(p2[0], p2[1], p2[2]);
      glTranslated(p1[0], p1[1], p1[2]);
      
      if(rotate)
	{
	  glRotated(zangle,0.0,0.0,1.0);
	  glRotated(yangle,0.0,1.0,0.0);
	  glRotated(xangle,1.0,0.0,0.0);
	}
	
      glMultMatrixd(mcs);
      glGetDoublev(GL_MODELVIEW_MATRIX, m);
      glPopMatrix();
      
      /* sweep profile */
      for(j=0;j<cs->length;j++)
	{
	  if(stride == 3)
	      mops_apply_transform3(&controlv[i*cs->length*stride+j*stride],m);
	    else
	      mops_apply_transform4(&controlv[i*cs->length*stride+j*stride],m);
	}

    }

  *patch = new;

 return mops_status;
}


int
mops_revolve_nurb(mops_object *o, double arc, mops_nurbpatch_object **patch)
{
 int mops_status = MOPS_OK;
 mops_nurbpatch_object *new = NULL;
 mops_nurbcurve_object *curve;
 double *uknotv = NULL, *tcontrolv = NULL;
 double radius = 0.0, w = 0.0, ww = 0.0,x,y,z;
 int i=0,j=0,a=0,b=0, c = 0;
 GLdouble m[16] = {0}, point[4] = {0};
 GLfloat mr[4][4];

  ww = sqrt(2.0)/2.0;

  if(!o)
    return MOPS_OK;

  if(o->type != MOPS_OTNURBCURVE)
    return MOPS_OK;

  curve = (mops_nurbcurve_object *)(o->object);

  /* get curves transformation-matrix */
  glMatrixMode (GL_MODELVIEW);
  glPushMatrix();
  glLoadIdentity ();
  glTranslated(o->movx, o->movy, o->movz);

  mops_build_rotmatrix(mr, o->quat);
  glMultMatrixf(&mr[0][0]);

  glScaled (o->scalx, o->scaly, o->scalz);
  glGetDoublev(GL_MODELVIEW_MATRIX, m);
  glPopMatrix();

  if(arc >= 360.0 || arc < -360.0 || arc == 0.0)
    {
      arc = 360.0;
    }

  /* calloc the new patch */
  if(!(new = calloc(1, sizeof(mops_nurbpatch_object))))
    return MOPS_OUTOFMEM;

  if(!(uknotv = calloc(curve->length+curve->order,sizeof(double))))
    {
      free(new); return MOPS_OUTOFMEM;
    }

  memcpy(uknotv,curve->knotv,(curve->length+curve->order)*sizeof(double));
  new->uknotv = uknotv;
  new->uorder = curve->order;
  new->vorder = 3;
  new->uknot_type = curve->knot_type;
  new->vknot_type = MOPS_KTCUSTOM;
  new->width = curve->length;
  new->has_weights = MOPS_TRUE;
  
  new->glu_sampling_tolerance = curve->glu_sampling_tolerance;
  new->handle_size = curve->handle_size;

  /* fill controlv */
  a = 0;
  for(j = 0; j < curve->length; j++)
    {
      /* transform point */
      x = curve->controlv[a];
      y = curve->controlv[a+1];
      z = curve->controlv[a+2];
      if(curve->has_weights)
	w = curve->controlv[a+3];
      else
	w = 1.0;
	  
      point[0] = m[0]*x + m[4]*y + m[8]*point[2] +
	m[12]*w;
      point[1] = m[1]*x + m[5]*y + m[9]*point[2] +
	m[13]*w;
      point[3] = m[3]*x + m[7]*y + m[11]*point[2] +
	m[15]*w;
	  
      /* project point onto XY-Plane! */
      point[2] = 0.0; /* XXXX loss of data! */

      radius = point[0];

      if(tcontrolv)
	free(tcontrolv);
      tcontrolv = NULL;
      if(new->vknotv)
	free(new->vknotv);
      new->vknotv = NULL;

      if(arc>0.0)
	{
	 mops_status = mops_nb_CreateNurbsCircle(radius, 0.0, arc,
						 &(new->height),&new->vknotv,
						 &tcontrolv);
	}
      else
	{
	 mops_status = mops_nb_CreateNurbsCircle(radius, arc, 0.0,
						 &(new->height), &new->vknotv,
						 &tcontrolv);
	}

      if(!new->controlv)
	{
	  if(!(new->controlv = calloc(new->height*new->width*4,
				      sizeof(double))))
	    return MOPS_OUTOFMEM;
	}

      /* copy to real controlv */
      b = 0;
      c = j*new->height*4;
      for(i=0;i<new->height;i++)
	{
	  new->controlv[c] = tcontrolv[b];
	  new->controlv[c+1] = point[1]*tcontrolv[b+3];
	  new->controlv[c+2] = tcontrolv[b+1];
	  new->controlv[c+3] = tcontrolv[b+3]*w;
	  b+=4;
	  c+=4;
	} /* for */

      if(curve->has_weights)
	a += 4;
      else
	a += 3;
    } /* for */
  
 
  *patch = new;

 return mops_status;
}


int
mops_extrude_nurb(double height, mops_object *o,
		  mops_nurbpatch_object **patch)
{
 int mops_status = MOPS_OK;
 mops_nurbpatch_object *new = NULL;
 mops_nurbcurve_object *curve;
 double vknots[4] = {0.0, 0.0, 1.0, 1.0}; /* uknots are taken from curve! */
 double *uknotv = NULL, *vknotv = NULL, *controlv = NULL;
 double x,y,z,w;
 int j=0,a=0,b=0;
 GLdouble m[16] = {0}, point[4] = {0};
 GLfloat mr[4][4];

  if(!o)
    return MOPS_OK;

  if(o->type != MOPS_OTNURBCURVE)
    return MOPS_OK;

  curve = (mops_nurbcurve_object *)(o->object);

  /* get curves transformation-matrix */
  glMatrixMode (GL_MODELVIEW);
  glPushMatrix();
  glLoadIdentity ();
  glTranslated(o->movx, o->movy, o->movz);

  mops_build_rotmatrix(mr, o->quat);
  glMultMatrixf(&mr[0][0]);

  glScaled (o->scalx, o->scaly, o->scalz);
  glGetDoublev(GL_MODELVIEW_MATRIX, m);
  glPopMatrix();

  /* calloc the new patch */
  if(!(new = calloc(1, sizeof(mops_nurbpatch_object))))
    return MOPS_OUTOFMEM;
  if(!(controlv = calloc(4*4*curve->length, sizeof(double))))
    { free(new); return MOPS_OUTOFMEM; }
  if(!(vknotv = calloc(4, sizeof(double))))
    { free(new); free(controlv); return MOPS_OUTOFMEM; }
  if(!(uknotv = calloc(curve->length+curve->order,sizeof(double))))
    { free(new); free(controlv); free(vknotv); return MOPS_OUTOFMEM; }
  
  memcpy(uknotv,curve->knotv,(curve->length+curve->order)*sizeof(double));
  new->uknotv = uknotv;
  memcpy(vknotv,vknots,4*sizeof(double));
  new->vknotv = vknotv;
  new->uorder = curve->order;
  new->vorder = 2; /* linear! */
  new->uknot_type = curve->knot_type;
  new->vknot_type = MOPS_KTCUSTOM;
  new->width = curve->length;
  new->height = 2;
  new->has_weights = MOPS_TRUE;
  
  new->glu_sampling_tolerance = curve->glu_sampling_tolerance;
  new->handle_size = curve->handle_size;
  
  /* fill controlv */ 
  a = 0;
  b = 0;
  for(j = 0; j < curve->length; j++)
    {

      /*      memcpy(point, &(curve->controlv[a]), 4*sizeof(GLdouble));*/

     /* transform point */
      x = curve->controlv[a];
      y = curve->controlv[a+1];
      z = curve->controlv[a+2];
      if(curve->has_weights)
	w = curve->controlv[a+3];
      else
	w = 1.0;

      point[0] = m[0]*x + m[4]*y + m[8]*z + m[12]*w;
      point[1] = m[1]*x + m[5]*y + m[9]*z + m[13]*w;
      point[2] = 0.0;
      point[3] = m[3]*x + m[7]*y + m[11]*z + m[15]*w;

      /* build a profile */
      controlv[b] = point[0];
      controlv[b+1] = point[1];
      controlv[b+2] = point[2];
      controlv[b+3] = w;

      b+=4;

      controlv[b] = point[0];
      controlv[b+1] = point[1];
      controlv[b+2] = point[2]+(height*w);
      controlv[b+3] = w;

      b += 4;
      if(curve->has_weights)
	a += 4;
      else
	a += 3;
    }

  new->controlv = controlv;

  *patch = new;

 return mops_status;
}



int
mops_get_tangentfromcontrol(int closed, int n, int p, int stride, double *P, int a, double *t)
{
 int mops_status = MOPS_OK;
 int found = MOPS_FALSE, wrapped = MOPS_FALSE;
 int b, i1, i2;
 int before, after;

 if(closed)
   {
     if(a == 0)
       a = (n-p);
     if(a == n)
       a = p;
     if(a > (n-p))
       a -= (n-p);
   }

 /* find a good point after P[a] */
 b = a+1;
 while(!found)
   {
     if(b >= n)
       {
	 if(wrapped)
	   return MOPS_ERROR;
	 wrapped = MOPS_TRUE;
	 b = 0;
       }

     i1 = a*stride;
     i2 = b*stride;
     if((P[i1] != P[i2]) || (P[i1+1] != P[i2+1]))
       {
	 found = MOPS_TRUE;
       }
     else
       {
	 b++;
       }

   }

 after = b;

 /* find a good point before P[a] */
 found = MOPS_FALSE;
 wrapped = MOPS_FALSE;
 b = a-1;
 while(!found)
   {
     if(b < 0)
       {
	 if(wrapped)
	   return MOPS_ERROR;
	 wrapped = MOPS_TRUE;
	 b = (n-1);
       }

     i1 = a*stride;
     i2 = b*stride;
     if((P[i1] != P[i2]) || (P[i1+1] != P[i2+1]))
       {
	 found = MOPS_TRUE;
       }
     else
       {
	 b--;
       }

   }

 before = b;

 /* now calc the tangent */
 t[0] = P[after*stride] - P[before*stride];
 t[1] = P[(after*stride)+1] - P[(before*stride)+1];


 return mops_status;
}


int
mops_bevel_nurb(int type, double radius, mops_object *o,
		mops_nurbpatch_object **patch)
{
 int mops_status = MOPS_OK;
 mops_nurbpatch_object *new = NULL;
 mops_nurbcurve_object *curve = NULL;
 double uknots_round[6] = {0.0, 0.0, 0.0, 1.0, 1.0, 1.0};
 double uknots_linear[4] = {0.0, 0.0, 1.0, 1.0};
 double uknots_ridge[8] = {0.0, 0.0, 0.0, 0.25, 0.75, 1.0, 1.0, 1.0};
 /* vknots are taken from curve! */
 double *uknotv = NULL, *vknotv = NULL, *controlv = NULL;
 double x,y,z,w;
 double tangent[3]={0}, normal[3]={0}, zaxis[3]={0.0,0.0,-1.0};
 double ww = sqrt(2.0)/2.0, displacex, displacey;
 int i=0, j=0, a=0, b=0, k = 0;
 GLdouble m[16] = {0}, point[4] = {0};
 GLfloat mr[4][4];


 curve = (mops_nurbcurve_object *)o->object;

  /* calloc the new patch */
  if(!(new = calloc(1, sizeof(mops_nurbpatch_object))))
    return MOPS_OUTOFMEM;
  if(type == 0)
    {
      if(!(controlv = calloc(4*3*curve->length, sizeof(double))))
	{ free(new); return MOPS_OUTOFMEM; }
      if(!(uknotv = calloc(6, sizeof(double))))
	{ free(new); free(controlv); return MOPS_OUTOFMEM; }
      memcpy(uknotv,uknots_round,6*sizeof(double));
      new->uorder = 3;
      new->width = 3;
    }
  if(type == 1)
    {
      if(!(controlv = calloc(4*2*curve->length, sizeof(double))))
	{ free(new); return MOPS_OUTOFMEM; }
      if(!(uknotv = calloc(4, sizeof(double))))
	{ free(new); free(controlv); return MOPS_OUTOFMEM; }
      memcpy(uknotv,uknots_linear,4*sizeof(double));
      new->uorder = 2;
      new->width = 2;
    }
  if(type == 2)
    {
      if(!(controlv = calloc(4*5*curve->length, sizeof(double))))
	{ free(new); return MOPS_OUTOFMEM; }
      if(!(uknotv = calloc(8, sizeof(double))))
	{ free(new); free(controlv); return MOPS_OUTOFMEM; }
      memcpy(uknotv,uknots_ridge,8*sizeof(double));
      new->uorder = 3;
      new->width = 5;
    }

  if(!(vknotv = calloc(curve->length+curve->order,sizeof(double))))
    { free(new); free(controlv); free(vknotv); return MOPS_OUTOFMEM; }
  memcpy(vknotv,curve->knotv,(curve->length+curve->order)*sizeof(double));
  new->vknotv = vknotv;
  new->uknotv = uknotv;
  new->vorder = curve->order;
  new->uknot_type = curve->knot_type;
  new->vknot_type = MOPS_KTCUSTOM;
  new->height = curve->length;
  new->has_weights = MOPS_TRUE;
  new->glu_sampling_tolerance = curve->glu_sampling_tolerance;
  new->handle_size = curve->handle_size;

  glMatrixMode (GL_MODELVIEW);
  glPushMatrix();
  glLoadIdentity();
  glTranslated(o->movx, o->movy, o->movz);

  mops_build_rotmatrix(mr, o->quat);
  glMultMatrixf(&mr[0][0]);

  glScaled (o->scalx, o->scaly, o->scalz);
  glGetDoublev(GL_MODELVIEW_MATRIX, m);
  glPopMatrix();

  
  /* fill controlv */
  /* first loop */
  b = 0;
  for(i = 0; i < new->width; i++)
    {
      a = 0;
      for(j = 0; j < curve->length; j++)
	{
	  if(curve->has_weights)
	    {
	      memcpy(&controlv[b],&curve->controlv[a],4*sizeof(double));
	      mops_apply_transform4(&controlv[b],m);
	      a+=4;
	    }
	  else
	    {
	      memcpy(&controlv[b],&curve->controlv[a],3*sizeof(double));
	      controlv[b+3] = 1.0;
	      mops_apply_transform3(&controlv[b],m);
	      a+=3;
	    }


	  controlv[b+2]=0.0;

	  b+=4;
	}
    }
  b = curve->length*4;
  /* transform second loop */
  if(type == 0)
    {
      /*
      if(curve->knot_type == MOPS_KTBSPLINE)
	{
	  x = controlv[(curve->length-curve->order)*4];
	  y = controlv[(curve->length-curve->order)*4+1];
	  z = controlv[(curve->length-curve->order)*4+2];
	  w = controlv[(curve->length-curve->order)*4+3];
	}
      else
	{
	  x = controlv[(curve->length-2)*4];
	  y = controlv[(curve->length-2)*4+1];
	  z = controlv[(curve->length-2)*4+2];
	  w = controlv[(curve->length-2)*4+3];
	}
      */
      for(j = 0; j < curve->length; j++)
	{
	  /* get displacement direction */
	  /*
	  if(j == (curve->length-1))
	    {
	      if(curve->knot_type == MOPS_KTBSPLINE)
		{
		  tangent[0] = x-controlv[(curve->order-1)*4];
		  tangent[1] = y-controlv[(curve->order-1)*4+1];
		}
	      else
		{
		  tangent[0] = x-controlv[4];
		  tangent[1] = y-controlv[5];
		}
	    }
	  else
	    {
	      tangent[0] = x-controlv[b+4];
	      tangent[1] = y-controlv[b+5];
	    }
	  */

	  mops_get_tangentfromcontrol(curve->closed, curve->length,
				      curve->order-1, 4, controlv, j, tangent);

	  w = controlv[b+3];
	  x = controlv[b]/w;
	  y = controlv[b+1]/w;
	  z = controlv[b+2]/w;

	  
	  MOPS_V3CROSS(normal, tangent, zaxis)
	  MOPS_V3SCAL(normal, radius-(radius*w))

	  /* create transformation matrix */
	  glMatrixMode (GL_MODELVIEW);
	  glPushMatrix();
	   glLoadIdentity();
	   glTranslated(normal[0], normal[1], radius*w);
	   glGetDoublev(GL_MODELVIEW_MATRIX, m);
	  glPopMatrix();

	  /* transform point */
	  point[0] = m[0]*x + m[4]*y + m[8]*z + m[12]*1.0;
	  point[1] = m[1]*x + m[5]*y + m[9]*z + m[13]*1.0;
	  point[2] = m[2]*x + m[6]*y + m[10]*z + m[14]*1.0;

	  controlv[b] = point[0]*(w*ww);
	  controlv[b+1] = point[1]*(w*ww);
	  controlv[b+2] = point[2]*(w*ww);

	  controlv[b+3] = ww*w;
	  
	  b+=4;
	}
    }

  if(type == 2)
    {
      /* transform the middle 3 loops */
      for(k=0;k<3;k++)
	{
	  if((k == 0) || (k == 2))
	    {
	      ww = 0.8535;
	      if(k == 0)
		{
		  displacex = 0.8535/ww;
		  displacey = 0.3535/ww;
		}
	      else
		{
		  displacex = 0.3535/ww;
		  displacey = 0.8535/ww;
		}
	    }
	  else
	    {
	      ww = 1.1;
	      displacex = 0.5/ww;
	      displacey = 0.5/ww;
	    }
	  /*
	  if(curve->knot_type == MOPS_KTBSPLINE)
	    {
	      x = controlv[(curve->length-curve->order)*4];
	      y = controlv[(curve->length-curve->order)*4+1];
	      z = controlv[(curve->length-curve->order)*4+2];
	      w = controlv[(curve->length-curve->order)*4+3];
	    }
	  else
	    {
	      x = controlv[(curve->length-2)*4];
	      y = controlv[(curve->length-2)*4+1];
	      z = controlv[(curve->length-2)*4+2];
	      w = controlv[(curve->length-2)*4+3];
	    }
	  */
	  for(j = 0; j < curve->length; j++)
	    {
	      /* get displacement direction */
	      /*
	      if(j == (curve->length-1))
		{
		  if(curve->knot_type == MOPS_KTBSPLINE)
		    {
		      tangent[0] = x-controlv[(curve->order-1)*4];
		      tangent[1] = y-controlv[(curve->order-1)*4+1];
		    }
		  else
		    {
		      tangent[0] = x-controlv[4];
		      tangent[1] = y-controlv[5];
		    }
		}
	      else
		{
		  tangent[0] = x-controlv[b+4];
		  tangent[1] = y-controlv[b+5];
		}
	      */
	      mops_get_tangentfromcontrol(curve->closed, curve->length,
					  curve->order-1, 4, controlv, j,
					  tangent);
	      w = controlv[b+3];
	      x = controlv[b];
	      y = controlv[b+1];
	      z = controlv[b+2];

	      MOPS_V3CROSS(normal, tangent, zaxis)
	      MOPS_V3SCAL(normal,(radius*(1.0-displacex)))

	      /* create transformation matrix */
	      glMatrixMode (GL_MODELVIEW);
	      glPushMatrix();
	       glLoadIdentity();
	       glTranslated(normal[0], normal[1], displacey*radius*w);
	       glGetDoublev(GL_MODELVIEW_MATRIX, m);
	      glPopMatrix();

	      /* transform point */
	      point[0] = m[0]*x + m[4]*y + m[8]*z + m[12]*1.0;
	      point[1] = m[1]*x + m[5]*y + m[9]*z + m[13]*1.0;
	      point[2] = m[2]*x + m[6]*y + m[10]*z + m[14]*1.0;

	      controlv[b] = point[0]*ww;
	      controlv[b+1] = point[1]*ww;
	      controlv[b+2] = point[2]*ww;

	      controlv[b+3] = w*ww;
	  
	      b+=4;
	    } /* for */
	} /* for */
    } /* if */

  /* transform last loop */
  /*
  if(curve->knot_type == MOPS_KTBSPLINE)
    {
      x = controlv[(curve->length-curve->order)*4];
      y = controlv[(curve->length-curve->order)*4+1];
      z = controlv[(curve->length-curve->order)*4+2];
      w = controlv[(curve->length-curve->order)*4+3];
    }
  else
    {
      x = controlv[(curve->length-2)*4];
      y = controlv[(curve->length-2)*4+1];
      z = controlv[(curve->length-2)*4+2];
      w = controlv[(curve->length-2)*4+3];
    }
  */
  for(j = 0; j < curve->length; j++)
    {
      /* get displacement direction */
      /*
      if(j == (curve->length-1))
	{
	  if(curve->knot_type == MOPS_KTBSPLINE)
	    {
	      tangent[0] = x-controlv[(curve->order-1)*4];
	      tangent[1] = y-controlv[(curve->order-1)*4+1];
	    }
	  else
	    {
	      tangent[0] = x-controlv[4];
	      tangent[1] = y-controlv[5];
	    }
	}
      else
	{
	  tangent[0] = x-controlv[b+4];
	  tangent[1] = y-controlv[b+5];
	}
      */

      mops_get_tangentfromcontrol(curve->closed, curve->length,
				  curve->order-1, 4, controlv, j, tangent);

      x = controlv[b];
      y = controlv[b+1];
      z = controlv[b+2];
      w = controlv[b+3];

      MOPS_V3CROSS(normal,tangent,zaxis)
      MOPS_V3SCAL(normal, radius)

      /* create transformation matrix */
      glMatrixMode (GL_MODELVIEW);
      glPushMatrix();
       glLoadIdentity();
       glTranslated(normal[0], normal[1], radius*w);
       glGetDoublev(GL_MODELVIEW_MATRIX, m);
      glPopMatrix();

      /* transform point */
      point[0] = m[0]*x + m[4]*y + m[8]*z + m[12]*1.0;
      point[1] = m[1]*x + m[5]*y + m[9]*z + m[13]*1.0;
      point[2] = m[2]*x + m[6]*y + m[10]*z + m[14]*1.0;
      /*point[3] = m[3]*x + m[7]*y + m[11]*z + m[15]*1.0;*/

      controlv[b] = point[0];
      controlv[b+1] = point[1];
      controlv[b+2] = point[2];
      controlv[b+3] = w;

      b+=4;
    }

  new->controlv = controlv;

  *patch = new;

 return mops_status;
}


int
mops_create_cap(double z, mops_nurbcurve_object *curve,
		double *ominx, double *omaxx,
		double *ominy, double *omaxy, double *oangle,
		mops_nurbpatch_object **cap)
{
 int mops_status = MOPS_OK;
 mops_nurbpatch_object *new = NULL;
 double knotv[4] = {0.0,0.0,1.0,1.0};
 double minx, miny, maxx, maxy, angle;
 int i, stride;

  /* calloc the new patch */
  if(!(new = calloc(1, sizeof(mops_nurbpatch_object))))
    return MOPS_OUTOFMEM;
  if(!(new->vknotv = calloc(4, sizeof(double))))
    { free(new); return MOPS_OUTOFMEM; }
  if(!(new->uknotv = calloc(4, sizeof(double))))
    { free(new); free(new->vknotv); return MOPS_OUTOFMEM; }
  if(!(new->controlv = calloc(4*3, sizeof(double))))
    { free(new); free(new->vknotv); free(new->uknotv); return MOPS_OUTOFMEM; }

  new->width = 2;
  new->height = 2;
  new->uorder = 2;
  new->vorder = 2;
  memcpy(new->uknotv,knotv,4*sizeof(double));
  memcpy(new->vknotv,knotv,4*sizeof(double));

  i = 0;
  minx = curve->controlv[0]; maxx = minx;
  miny = curve->controlv[1]; maxy = miny;
  angle = 0.0;
  stride = (curve->has_weights?4:3);
  while(i<curve->length*stride)
    {
      if(curve->controlv[i] > maxx)
	maxx = curve->controlv[i];
      if(curve->controlv[i] < minx)
	minx = curve->controlv[i];
      if(curve->controlv[i+1] > maxy)
	maxy = curve->controlv[i+1];
      if(curve->controlv[i+1] < miny)
	miny = curve->controlv[i+1];
		
      /* compute direction */
      if((i<(curve->length-1)*stride)&&(i>stride))
	{
	  angle +=
	    ((curve->controlv[i+stride] - curve->controlv[i-stride])*
	     (curve->controlv[i+1] - curve->controlv[i+1-stride])) -
	    ((curve->controlv[i] - curve->controlv[i-stride])*
	     (curve->controlv[i+stride+1] - curve->controlv[i+1-stride]));
	}

		   
      i+=stride;
    }
	    
  new->controlv[0] = minx;
  new->controlv[1] = miny;

  new->controlv[3] = minx;
  new->controlv[4] = maxy;

  new->controlv[6] = maxx;
  new->controlv[7] = miny;

  new->controlv[9] = maxx;
  new->controlv[10] = maxy;

  for(i=2;i<=11;i+=3)
    new->controlv[i] = z;

  *ominx = minx;
  *omaxx = maxx;
  *ominy = miny;
  *omaxy = maxy;
  *oangle = angle;

  *cap = new;
  
 return mops_status;
}


void
mops_get_nc_orientation(mops_nurbcurve_object *curve, double *oangle)
{
 double angle = 0.0, maxx, minx, maxy, miny;
 int i, i1 = 0, i2 = 0, i3 = 0, a, stride;
 int pi[4], itmp;
 double *p[4], *dtmp;
 double *cntrlv = NULL;


  cntrlv = curve->controlv;
  maxx = cntrlv[0];
  minx = cntrlv[0];
  maxy = cntrlv[1];
  miny = cntrlv[1];
  p[0] = &(cntrlv[0]);
  p[1] = &(cntrlv[0]);
  p[2] = &(cntrlv[0]);
  p[3] = &(cntrlv[0]);

  stride = curve->has_weights?4:3;
  a = stride;
  for(i = 1; i < curve->length; i++)
    {
      if(cntrlv[a] < minx)
	{
	  minx = cntrlv[a];
	  p[0] = &(cntrlv[a]);
	  pi[0] = i;
	}
      if(cntrlv[a] > maxx)
	{
	  maxx = cntrlv[a];
	  p[1] = &(cntrlv[a]);
	  pi[1] = i;
	}
      if(cntrlv[a+1] < miny)
	{
	  miny = cntrlv[a+1];
	  p[2] = &(cntrlv[a]);
	  pi[2] = i;
	}
      if(cntrlv[a+1] > maxy)
	{
	  maxy = cntrlv[a+1];
	  p[3] = &(cntrlv[a]);
	  pi[3] = i;
	}

      a += stride;
    }

  /* bubbles everywhere! */
  for(a = 0; a < 3; a++)
    {
      for(i = 0; i < (3-a); i++)
	{
	  if(pi[i] < pi[i+1])
	    {
	      itmp = pi[i+1];
	      pi[i+1] = pi[i];
	      pi[i] = itmp;
	      dtmp = p[i+1];
	      p[i+1] = p[i];
	      p[i] = dtmp;
	    }
	}
    }

  for(i = 0; i < 3; i++)
    {
      switch (i)
	{
	case 0:
	  i1 = 3; i2 = 0; i3 = 1;
	  break;
	case 1:
	  i1 = 0; i2 = 1; i3 = 2;
	  break;
	case 2:
	  i1 = 1; i2 = 2; i3 = 3;
	  break;
	case 3:
	  i1 = 2; i2 = 3; i3 = 0;
	}

      angle += (((p[i3])[0]-(p[i1])[0]) *
		((p[i2])[1]-(p[i1])[1])) -
	       (((p[i2])[0]-(p[i1])[0]) *
		((p[i3])[1]-(p[i1])[1]));
    }

  *oangle = angle;
}


/*
int
mops_concat_nurbcurve(mops_object *o1, mops_object *o2,
		      mops_nurbcurve_object **newcurve)
{
 int mops_status = MOPS_OK;
 mops_nurbcurve_object *new = NULL;
 mops_nurbcurve_object *curve1, *curve2;
 double *knotv = NULL, *controlv = NULL;
 double x,y,w;
 int i=0,j=0,a=0,b=0;
 GLdouble m1[16] = {0}, m2[16] = {0}, point[4] = {0};

  if((!o1) || (!o2) || (!newcurve))
    return MOPS_OK;

  if((o1->type != MOPS_OTNURBCURVE) || (o2->type != MOPS_OTNURBCURVE))
    return MOPS_OK;

  curve1 = (mops_nurbcurve_object *)(o1->object);
  curve2 = (mops_nurbcurve_object *)(o2->object);


  glMatrixMode (GL_MODELVIEW);
  glPushMatrix();
  glLoadIdentity();
  glTranslated(o1->movx, o1->movy, o1->movz);
  glRotated(o1->rotx, 1.0, 0.0, 0.0);
  glRotated(o1->roty, 0.0, 1.0, 0.0);
  glRotated(o1->rotz, 0.0, 0.0, 1.0);
  glScaled (o1->scalx, o1->scaly, o1->scalz);
  glGetDoublev(GL_MODELVIEW_MATRIX, m1);
  glPopMatrix();


  glPushMatrix();
  glLoadIdentity();
  glTranslated(o2->movx, o2->movy, o2->movz);
  glRotated(o2->rotx, 1.0, 0.0, 0.0);
  glRotated(o2->roty, 0.0, 1.0, 0.0);
  glRotated(o2->rotz, 0.0, 0.0, 1.0);
  glScaled (o2->scalx, o2->scaly, o2->scalz);
  glGetDoublev(GL_MODELVIEW_MATRIX, m2);
  glPopMatrix();


  if((new = calloc(1, sizeof(mops_nurbcurve_object))) == NULL)
    return MOPS_OUTOFMEM;
  if((controlv = calloc(4*curve1->length, sizeof(double))) == NULL)
    return MOPS_OUTOFMEM;
  if((knotv = calloc(4, sizeof(double))) == NULL)
    return MOPS_OUTOFMEM;

  new->knotv = knotv;
  new->order = curve1->order;
  new->knot_type = MOPS_KTCUSTOM;

  new->has_weights = MOPS_TRUE;

  new->glu_sampling_tolerance = 20;
 

  a = 0;

  *newcurve = new;

 return mops_status;
}
*/
/**********************************************/
/**********************************************/
/**********************************************/

int
mops_crtncurve_tcmd(Tcl_Interp * interp,
		   int argc, char *argv[], mops_object *o)
{
 int mops_status;
 int i = 1, order = 4, length = 4, knot_type = MOPS_KTNURB;
 int j = 0, has_weights = MOPS_TRUE;
 double *controlv;
 char fname[] = "create_ncurve";

  /* parse args */
 while(i+1 < argc)
   {
     if(!strcmp(argv[i],"-length"))
       {
	 Tcl_GetInt(interp, argv[i+1], &length);
	 if(length <= 0) length = 4;
	 i+=2;
       }
     else
       i++;
   }

 if(!(controlv = (calloc(length*4, sizeof(double)))))
   {
     mops_error(interp, MOPS_OUTOFMEM, fname, NULL);
     return TCL_ERROR;
   }

 /* default cordinates */
 j = 0;
 for(i=0;i<length;i++)
   {
     controlv[j] = i;
     controlv[j+1] = i;
     controlv[j+2] = 0.0;
     controlv[j+3] = 1.0;
     j+=4;
   }

 o->type = MOPS_OTNURBCURVE;

 if(order>length) order = length;

  /* we create the nurbcurve-object */
 mops_status = mops_create_nurbcurve(order, length,
				     knot_type, has_weights,
				     controlv, NULL,
				     (mops_nurbcurve_object **)&(o->object));

 if(mops_status)
   {
     free(controlv);
     mops_error(interp, mops_status, fname, NULL);
     return TCL_ERROR;
   }

 /* change knot_type immediately to custom */
 /* those curves are simply more flexible  */
 ((mops_nurbcurve_object *)o->object)->knot_type = MOPS_KTCUSTOM;

 return TCL_OK;
}


int
mops_crtnpatch_tcmd(Tcl_Interp * interp, int argc, char *argv[],
		    mops_object *o)
{
 int mops_status;
 int i = 1, uorder = 4, vorder = 4, width = 4, height = 4,
   knot_type = MOPS_KTNURB;
 int j = 0, k = 0, has_weights = MOPS_TRUE;
 double *controlv;
 char fname[] = "create_npatch";

  /* parse args */
  while(i+1 < argc)
    {
      if(!strcmp(argv[i],"-width"))
	{
	  Tcl_GetInt(interp, argv[i+1], &width);
	  if(width <= 0) width = 4;
	  i+=2;
	} else
      if(!strcmp(argv[i],"-height"))
	{
	  Tcl_GetInt(interp, argv[i+1], &height);
	  if(height <= 0) height = 4;
	  i+=2;
	}
      else
	i++;
    }

  if(!(controlv = (calloc(width*height*4, sizeof(double)))))
    {
      mops_error(interp, MOPS_OUTOFMEM, fname, NULL);
      return TCL_ERROR;
    }

  /* default cordinates */
  k = 0;
  for(i=0;i<width;i++)
    for(j=0;j<height;j++)
      {
	controlv[k] = i;
	controlv[k+1] = j;
	controlv[k+2] = 0.0;
	controlv[k+3] = 1.0;
	k+=4;
      }

  o->type = MOPS_OTNURBPATCH;

  if(uorder>width)
    uorder = width;

  if(vorder>height)
    vorder = height;

  /* we create the nurbpatch-object */
  mops_status = mops_create_nurbpatch(uorder, vorder, width, height,
				      knot_type, knot_type, has_weights,
				      controlv, NULL, NULL,
				      (mops_nurbpatch_object **)&(o->object));

  if(mops_status)
    {
      free(controlv);
      mops_error(interp, mops_status, fname, NULL);
      return TCL_ERROR;
    }

 return TCL_OK;
}


int
mops_crtncircle_tcmd(Tcl_Interp * interp,
		     int argc, char *argv[], mops_object *o)
{
 int mops_status;
 char fname[] = "create_ncircle";
 double arc = 0;

  o->type = MOPS_OTNURBCURVE;

  /* we create the nurbcurve-object */
  if(argc < 3)
    {
      mops_status = mops_create_nurbcircle(
			(mops_nurbcurve_object **)&(o->object));
    }
  else
    {
      Tcl_GetDouble(interp, argv[2], &arc);
      if(arc >= 360.0 || arc <= -360.0 || arc == 0.0)
	{
	  mops_status = mops_create_nurbcircle(
			(mops_nurbcurve_object **)&(o->object));
	}
      else
	{
	  mops_status = mops_create_nurbcirclearc(arc,
			(mops_nurbcurve_object **)&(o->object));
	}

    }

  if(mops_status)
    {
      mops_error(interp,mops_status,fname,NULL);
      return TCL_ERROR;
    }

  mops_nurbcurve_recreatemp((mops_nurbcurve_object *)o->object);

 return TCL_OK;
}


int
mops_crtnsphere_tcmd(Tcl_Interp * interp,
		     int argc, char *argv[], mops_object *o)
{
 int mops_status;
 mops_object *newc = NULL;
 char fname[] = "create_nsphere";


  if(!(newc = calloc(1,sizeof(mops_object))))
    {
      mops_error(interp, MOPS_OUTOFMEM, fname, NULL);
      return TCL_ERROR;
    }

  newc->type = MOPS_OTNURBCURVE;
  o->type = MOPS_OTNURBPATCH;

  mops_status = mops_default_attributes(newc);

  /* first, we create a half circle nurbcurve-object */
  mops_status = mops_create_nurbhcircle(
		 (mops_nurbcurve_object **)&(newc->object));
  if(mops_status)
    { 
      free (newc);
      mops_error(interp, mops_status, fname, NULL);
      return TCL_ERROR;
    }

  mops_status = mops_revolve_nurb(newc, 360.0,
                 (mops_nurbpatch_object **)&(o->object));

  if(mops_status)
    {
      mops_error(interp,mops_status,fname,NULL);
      return TCL_ERROR;
    }

  /* we do not need the half-circle any longer */
  mops_status = mops_delete_object(newc);


 return TCL_OK;
}


int
mops_crtnsphere2_tcmd(ClientData clientData, Tcl_Interp *interp,
		     int argc, char *argv[])
{
 int mops_status;
 int i;
 double rot[18] = {
   0.0, 0.0, 0.0,
   0.0, 90.0, 0.0,
   0.0, 180.0, 0.0,
   0.0, -90.0, 0.0,
   90.0, 0.0, 0.0,
   -90.0, 0.0, 0.0
 };
 float xaxis[3]={1.0,0.0,0.0};
 float yaxis[3]={0.0,1.0,0.0};
 float zaxis[3]={0.0,0.0,1.0};
 float quat[4];
 mops_object *new = NULL;
 char fname[] = "create_cobbsphere";

  for(i=0;i<6;i++)
    {
      if(!(new = calloc(1,sizeof(mops_object))))
	{
	  mops_error(interp, MOPS_OUTOFMEM, fname, NULL);
	  return TCL_ERROR;
	}

      new->type = MOPS_OTNURBPATCH;

      mops_status = mops_default_attributes(new);

      mops_status = mops_create_nurbcobbsphere(
			(mops_nurbpatch_object **)&(new->object));
      if(mops_status)
	{
	  mops_delete_object(new);
	  return TCL_ERROR;
	}

      new->rotx = rot[i*3];
      if(new->rotx != 0.0)
	{      
	  mops_axis_to_quat(xaxis,MOPS_D2R(rot[i*3]),quat);
	  mops_add_quat(quat,new->quat,new->quat);
	}
      new->roty = rot[(i*3)+1];
      if(new->roty != 0.0)
	{      
	  mops_axis_to_quat(yaxis,MOPS_D2R(rot[i*3+1]),quat);
	  mops_add_quat(quat,new->quat,new->quat);
	}
      new->rotz = rot[(i*3)+2];
      if(new->rotz != 0.0)
	{
	  mops_axis_to_quat(zaxis,MOPS_D2R(rot[i*3+2]),quat);
	  mops_add_quat(quat,new->quat,new->quat);
	}
      mops_status = mops_link_new_object(new);
      if(mops_status)
	{
	  mops_delete_object(new);
	  return TCL_ERROR;
	}

    }

 return TCL_OK;
}


int
mops_revolve_tcmd(Tcl_Interp * interp,
		  int argc, char *argv[], mops_object *o)
{
 int mops_status;
 mops_list_object *sel = mops_current_selection;
 mops_object *selo = NULL, *curve = NULL;
 char fname[] = "revolve"; 

  if(!sel)
    {
      mops_error(interp, MOPS_NOSEL, fname, NULL);
      return TCL_ERROR;
    }

  selo = sel->object;
  curve = selo;

  if(curve->type != MOPS_OTNURBCURVE)
    {
      mops_error(interp, MOPS_ERROR, fname, "object is not a NURBCurve");
      return TCL_ERROR;
    }

  o->type = MOPS_OTNURBPATCH;

 /* do the revolve */
  mops_status = mops_revolve_nurb(curve, 360.0,
				  (mops_nurbpatch_object **)&(o->object));

  if(mops_status)
    {
      mops_error(interp,mops_status,fname,NULL);
      return TCL_ERROR;
    }

 return TCL_OK;
}


int
mops_extrude_tcmd(Tcl_Interp * interp,
		  int argc, char *argv[], mops_object *o)
{
 int mops_status;
 double height = 1.0;
 mops_list_object *sel = mops_current_selection;
 mops_object *selo = NULL, *curve = NULL;
 char fname[] = "extrude";

  if(!sel)
    {
      mops_error(interp, MOPS_NOSEL, fname, NULL);
      return TCL_ERROR;
    }

  selo = sel->object;
  curve = selo;

  if(curve->type != MOPS_OTNURBCURVE)
    {
      mops_error(interp, MOPS_ERROR, fname, "object is not a NURBCurve");
      return TCL_ERROR;
    }

  if(argc>2)
    Tcl_GetDouble(interp, argv[2], &(height));

  o->type = MOPS_OTNURBPATCH;
 
  mops_status = mops_copy_attributes(curve,o);
  
  /* do the extrusion */
  mops_status = mops_extrude_nurb(height, curve,
				  (mops_nurbpatch_object **)&(o->object));

  if(mops_status)
    {
      mops_error(interp,mops_status,fname,NULL);
      return TCL_ERROR;
    }

 return TCL_OK;
}


int
mops_sweep_tcmd(Tcl_Interp * interp,
		  int argc, char *argv[], mops_object *o)
{
 int mops_status;
 mops_list_object *sel = mops_current_selection;
 char fname[] = "sweep";
 int sections = 4, rotate = 1;

  if(!sel)
    {
      mops_error(interp, MOPS_NOSEL, fname, NULL);
      return TCL_ERROR;
    }
  if(!sel->next)
    {
      mops_error(interp,MOPS_ERROR,fname,"select two NURB curves!");
      return TCL_ERROR;
    }
  if((sel->object->type != MOPS_OTNURBCURVE) ||
     (sel->next->object->type != MOPS_OTNURBCURVE))
    {
      mops_error(interp, MOPS_ERROR, fname, "object is not a NURBCurve");
      return TCL_ERROR;
    }
  if(argc<4)
    {
      mops_error(interp, MOPS_ARGS, fname, "numsections rotate");
      return TCL_ERROR;
    }
  Tcl_GetInt(interp,argv[2], &(sections));
  Tcl_GetInt(interp,argv[3], &(rotate));

  o->type = MOPS_OTNURBPATCH;
   
  /* do the sweep */
  mops_status = mops_sweep_nurb(sel->object,sel->next->object,sections,
			   rotate,(mops_nurbpatch_object **)&(o->object));

  if(mops_status)
    {
      mops_error(interp,mops_status,fname,NULL);
      return TCL_ERROR;
    }

 return TCL_OK;
}


int
mops_concat_nc_tcmd(Tcl_Interp * interp,
		    int argc, char *argv[], mops_object *o)
{
 int mops_status;
 double *newcontrolv = NULL;
 double *controlv1, *controlv2;
 int nc1w = MOPS_FALSE, nc2w = MOPS_FALSE;
 int has_weights = MOPS_FALSE, ktype = MOPS_KTNURB;
 int i = 0, a = 0, b = 0;
 GLdouble m1[16] = {0}, m2[16] = {0};
 mops_list_object *sel = mops_current_selection;
 mops_nurbcurve_object *nc1 = NULL, *nc2 = NULL;
 char fname[] = "concat_nurbcurves";

  if(!sel)
    {
      mops_error(interp, MOPS_NOSEL, fname, NULL);
      return TCL_ERROR;
    }
  if(!sel->next)
    {
      mops_error(interp,MOPS_ERROR,fname,"select two NURB curves!");
      return TCL_ERROR;
    }
  if((sel->object->type != MOPS_OTNURBCURVE) ||
     (sel->next->object->type != MOPS_OTNURBCURVE))
    {
      mops_error(interp, MOPS_ERROR, fname, "object is not a NURBCurve");
      return TCL_ERROR;
    }

  o->type = MOPS_OTNURBCURVE;

  glMatrixMode (GL_MODELVIEW);
  glPushMatrix();
  glLoadIdentity();
  glTranslated(sel->object->movx, sel->object->movy, sel->object->movz);
  glRotated(sel->object->rotx, 1.0, 0.0, 0.0);
  glRotated(sel->object->roty, 0.0, 1.0, 0.0);
  glRotated(sel->object->rotz, 0.0, 0.0, 1.0);
  glScaled (sel->object->scalx, sel->object->scaly, sel->object->scalz);
  glGetDoublev(GL_MODELVIEW_MATRIX, m1);
  glPopMatrix();

  glPushMatrix();
  glLoadIdentity();
  glTranslated(sel->next->object->movx, sel->next->object->movy,
	       sel->next->object->movz);
  glRotated(sel->next->object->rotx, 1.0, 0.0, 0.0);
  glRotated(sel->next->object->roty, 0.0, 1.0, 0.0);
  glRotated(sel->next->object->rotz, 0.0, 0.0, 1.0);
  glScaled (sel->next->object->scalx, sel->next->object->scaly,
	    sel->next->object->scalz);
  glGetDoublev(GL_MODELVIEW_MATRIX, m2);
  glPopMatrix();

  nc1 = (mops_nurbcurve_object *)sel->object->object;
  nc2 = (mops_nurbcurve_object *)sel->next->object->object;

  nc1w = nc1->has_weights;
  nc2w = nc2->has_weights;

  if((nc1w) || (nc2w))
    has_weights = MOPS_TRUE;

  if(nc1->knot_type != MOPS_KTCUSTOM)
    ktype = nc1->knot_type;

  controlv1 = nc1->controlv;
  controlv2 = nc2->controlv;

  if(!(newcontrolv = calloc(nc1->length*(has_weights?4:3)+
			    nc2->length*(has_weights?4:3),
			    sizeof(double))))
    {
      mops_error(interp,MOPS_OUTOFMEM,fname,NULL);
      return TCL_ERROR;
    }

  a = 0; b = 0;
  for(i=0;i<(nc1->length);i++)
    {
      newcontrolv[b] = controlv1[a];
      newcontrolv[b+1] = controlv1[a+1];
      newcontrolv[b+2] = controlv1[a+2];

      if(nc1w)
	newcontrolv[b+3] = controlv1[a+3];
      else
	if(has_weights)
	  newcontrolv[b+3] = 1.0;

      if(has_weights)
	mops_apply_transform4(&(newcontrolv[b]), m1);
      else
	mops_apply_transform3(&(newcontrolv[b]), m1);

      a += (nc1w?4:3);
      b += (has_weights?4:3);
    }
  a = 0;
  for(i=0;i<(nc2->length);i++)
    {
      newcontrolv[b] = controlv2[a];
      newcontrolv[b+1] = controlv2[a+1];
      newcontrolv[b+2] = controlv2[a+2];

      if(nc2w)
	newcontrolv[b+3] = controlv2[a+3];
      else
	if(has_weights)
	  newcontrolv[b+3] = 1.0;

      if(has_weights)
	mops_apply_transform4(&(newcontrolv[b]), m2);
      else
	mops_apply_transform3(&(newcontrolv[b]), m2);

      a += (nc2w?4:3);
      b += (has_weights?4:3);
    }

  mops_status = mops_create_nurbcurve(nc1->order, nc1->length+nc2->length,
				      ktype, has_weights, newcontrolv, NULL,
				      (mops_nurbcurve_object **)&(o->object));

  if(mops_status)
    {
      mops_error(interp,mops_status,fname,NULL);
      return TCL_ERROR;
    }

  mops_nurbcurve_recreatemp((mops_nurbcurve_object *)o->object);

 return TCL_OK;
}


int
mops_split_to_curves_tcmd(ClientData clientData, Tcl_Interp *interp,
			  int argc, char *argv[])
{
 int mops_status = MOPS_OK;
 mops_list_object *sel = mops_current_selection;
 mops_object *src = NULL, *new = NULL;
 mops_nurbpatch_object *patch = NULL;
 double *knotv = NULL, *controlv = NULL;
 int knots = 0, stride = 0,  dstlen = 0;
 int u = 0, i = 0, j = 0;
 GLdouble m[16] = {0};
 GLfloat mr[4][4];
 char fname[] = "split_to_curves";

  if(!sel)
    {
      mops_error(interp, MOPS_NOSEL, fname, NULL);
      return TCL_OK;
    }

  /* parse args */
  if(argc != 2)
    {
      mops_error(interp,MOPS_ARGS,fname,"u|v");
      return TCL_OK;
    }

  if(!strcmp(argv[1],"u"))
    u = 1;

  src = sel->object;
  if(src->type != MOPS_OTNURBPATCH)
    {
      mops_error(interp, MOPS_ERROR, fname, "object is not a NURBPatch");
      return TCL_OK;
    }

  patch = src->object;

  glMatrixMode (GL_MODELVIEW);
  glPushMatrix();
  glTranslated(src->movx, src->movy, src->movz);

  mops_build_rotmatrix(mr, src->quat);
  glMultMatrixf(&mr[0][0]);

  glScaled (src->scalx, src->scaly, src->scalz);
  glGetDoublev(GL_MODELVIEW_MATRIX, m);
  glPopMatrix();

  if(u)
    {
      dstlen = patch->width;
      knots = dstlen+patch->uorder;
      stride = (patch->has_weights?4:3);

      for(i=0; i<patch->height; i++)
	{
	  new = NULL;
	  if(!(new = calloc(1, sizeof(mops_object))))
	    {
	      mops_error(interp, MOPS_OUTOFMEM, fname, NULL);
	      return TCL_OK;
	    }	   
 
	  new->type = MOPS_OTNURBCURVE;
	  
	  if(!(knotv = calloc(knots, sizeof(double))))
	    {
	      free(new);
	      mops_error(interp, MOPS_OUTOFMEM, fname, NULL);
	      return TCL_OK;
	    }

	  memcpy(knotv, patch->uknotv, (size_t)(knots*sizeof(double)));

	  if(!(controlv = calloc(stride*dstlen, sizeof(double))))
	    {
	      free(new);
	      free(knotv);
	      mops_error(interp, MOPS_OUTOFMEM, fname, NULL);
	      return TCL_OK;
	    }

	  for(j=0;j<dstlen;j++)
	    {
	      memcpy(&(controlv[j*stride]),
		     &(patch->controlv[(i+(j*patch->height))*stride]),
		     (size_t)(stride*sizeof(double)));

	      if(patch->has_weights)
		mops_apply_transform4(&(controlv[j*stride]),m);
	      else
		mops_apply_transform3(&(controlv[j*stride]),m);

	    }

	  mops_status = mops_create_nurbcurve(patch->uorder, dstlen,
					      patch->uknot_type, patch->has_weights,
					      controlv, knotv,
					      (mops_nurbcurve_object **)&(new->object));

	  if(mops_status)
	    {
	      free(new); free(knotv); free(controlv);
	      mops_error(interp, mops_status, fname, NULL);
	      return TCL_OK;
	    }

	  mops_status = mops_default_attributes(new);
	  mops_status = mops_link_new_object(new);
	  if(mops_status)
	    {
	      free(new); free(knotv); free(controlv);
	      mops_error(interp, mops_status, fname, NULL);
	      return TCL_OK;
	    }

	  mops_nurbcurve_recreatemp((mops_nurbcurve_object *)new->object);

	} /* for */
 
    }
  else
    {
      dstlen = patch->height;
      knots = dstlen + patch->vorder;
      stride = (patch->has_weights?4:3) * dstlen;

      for(i=0; i<patch->width; i++)
	{
	  new = NULL;
	  if(!(new = calloc(1, sizeof(mops_object))))
	    {
	      mops_error(interp, MOPS_OUTOFMEM, fname, NULL);
	      return TCL_OK;
	    }

	  new->type = MOPS_OTNURBCURVE;
	  
	  knotv = NULL;
	  if(!(knotv = calloc(knots, sizeof(double))))
	    {
	      free(new);
	      mops_error(interp, MOPS_OUTOFMEM, fname, NULL);
	      return TCL_OK;
	    }
	  memcpy(knotv, patch->vknotv, (size_t)(knots*sizeof(double)));

	  controlv = NULL;
	  if(!(controlv = calloc(stride, sizeof(double))))
	    {
	      free(new); free(knotv);
	      mops_error(interp, MOPS_OUTOFMEM, fname, NULL);
	      return TCL_OK;
	    }

	  memcpy(controlv, &(patch->controlv[i*stride]),
		 (size_t)(stride*sizeof(double)));

	  for(j=0;j<dstlen;j++)
	    {
	      if(patch->has_weights)
		mops_apply_transform4(&(controlv[j*4]),m);
	      else
		mops_apply_transform3(&(controlv[j*3]),m);
	    }

	  mops_status = mops_create_nurbcurve(patch->vorder, dstlen,
					      patch->vknot_type, patch->has_weights,
					      controlv, knotv,
					      (mops_nurbcurve_object **)&(new->object));
	  if(mops_status)
	    {
	      free(new); free(knotv); free(controlv);
	      mops_error(interp, mops_status, fname, NULL);
	      return TCL_OK;
	    }
	  mops_status = mops_default_attributes(new);
	  mops_status = mops_link_new_object(new);
	  if(mops_status)
	    {
	      free(new); free(knotv); free(controlv);
	      mops_error(interp, mops_status, fname, NULL);
	      return TCL_OK;
	    }	

	} /* for */
    }

 return TCL_OK;
}


int
mops_build_from_curves_tcmd(ClientData clientData, Tcl_Interp *interp,
			  int argc, char *argv[])
{
 int mops_status = MOPS_OK;
 mops_list_object *sel = mops_current_selection;
 mops_object *src = NULL, *new = NULL;
 mops_nurbcurve_object *curve = NULL;
 double *newvknotv = NULL, *newcontrolv = NULL;
 int newwidth = 0, newheight = 0;
 int newvorder = 0, newvknots = 0, newuorder = 0;
 int newuknot_type = MOPS_KTNURB, newvknot_type = 0;
 int newhas_weights = 0;
 int i = 0, j = 0, a = 0;
 GLdouble m[16] = {0};
 GLfloat mr[4][4];
 char fname[] = "build_from_curves";

  if(!sel)
    {
      mops_error(interp, MOPS_NOSEL, fname, NULL);
      return TCL_OK;
    }

  src = sel->object;
  if(src->type != MOPS_OTNURBCURVE)
    {
      mops_error(interp, MOPS_ERROR, fname, "first object not a NURBCurve");
      return TCL_OK;
    }

  curve = src->object;
  newheight = curve->length;
  newhas_weights = curve->has_weights;
  newvorder = curve->order;
  newvknots = newheight + newvorder;
  newvknot_type = curve->knot_type;
  if(!(newvknotv = calloc(newvknots, sizeof(double))))
    {
      mops_error(interp, MOPS_OUTOFMEM, fname, NULL);
      return TCL_OK;
    }

  memcpy(newvknotv, curve->knotv, (size_t)(newvknots*sizeof(double)));

  /* parse selection */
  while(sel)
    {
      src = sel->object;
      curve = src->object;
      if(curve->length >= newheight)
	newwidth++;
      
      sel = sel->next;
    }

  /* enough curves to form a patch ? */
  if(newwidth < 2)
    {
      mops_error(interp, MOPS_ERROR, fname,
		 "not enough suitable curves selected");
      return TCL_OK;
    }
  /* create new patch */
  if(!(new = calloc(1, sizeof(mops_object))))
    {
      mops_error(interp, MOPS_OUTOFMEM, fname, NULL);
      free(newvknotv);
      return TCL_OK;
    }
  new->type = MOPS_OTNURBPATCH;

  if(!(newcontrolv = calloc(newwidth*newheight*(newhas_weights?4:3),
			    sizeof(double))))
    {
      mops_error(interp, MOPS_OUTOFMEM, fname, NULL);
      free(newvknotv); free(new);
      return TCL_OK;
    }
  glMatrixMode(GL_MODELVIEW);

  /* fill newcontrolv */
  sel = mops_current_selection;
  while(sel)
    {
      src = sel->object;
      curve = src->object;
      if(curve->length >= newheight)
	{
	  glPushMatrix();
	  glTranslated(src->movx, src->movy, src->movz);

	  mops_build_rotmatrix(mr, src->quat);
	  glMultMatrixf(&mr[0][0]);


	  glScaled (src->scalx, src->scaly, src->scalz);
	  glGetDoublev(GL_MODELVIEW_MATRIX, m);
	  glPopMatrix();

	  a = 0;
	  for(j=0;j<newheight;j++)
	    {
	      newcontrolv[i++] = curve->controlv[a++];
	      newcontrolv[i++] = curve->controlv[a++];
	      newcontrolv[i++] = curve->controlv[a++];
	      if(newhas_weights)
		{
		  if(curve->has_weights)
		    newcontrolv[i++] = curve->controlv[a++];
		  else
		    newcontrolv[i++] = 1.0;

		  mops_apply_transform4(&(newcontrolv[i-4]), m);
		}
	      else
		{
		  if(curve->has_weights)
		    a++;

		  mops_apply_transform3(&(newcontrolv[i-3]), m);
		}
	    } /* for */
	} /* if */

      sel = sel->next;
    } /* while */

  if(newwidth < 4)
    newuorder = newwidth;
  else
    newuorder = 4;

  mops_status = mops_create_nurbpatch(newuorder, newvorder,
				      newwidth, newheight,
				      newuknot_type, newvknot_type,
				      newhas_weights, newcontrolv,
				      NULL, newvknotv,
				     (mops_nurbpatch_object **)&(new->object));

  if(mops_status)
    {
      mops_error(interp, mops_status, fname, NULL);
      free(new); free(newvknotv); free(newcontrolv);
      return TCL_OK;
    }

  mops_status = mops_default_attributes(new);
  mops_status = mops_link_new_object(new);

  if(mops_status)
    {
      mops_error(interp, mops_status, fname, NULL);
      mops_delete_object(new);
      return TCL_OK;
    }

 return TCL_OK;
}


int
mops_revert_ncurve(mops_nurbcurve_object *curve)
{
 int stride, i, j, kncount;
 double dtemp;
 double *newknotv = NULL;

  if(!curve)
    return MOPS_OK;

  /* revert control */
  stride = (curve->has_weights?4:3);
  j = (curve->length - 1)*(curve->has_weights?4:3);
  i = 0;
  while(i < j)
    {
      dtemp = curve->controlv[j];
      curve->controlv[j] = curve->controlv[i];
      curve->controlv[i] = dtemp;

      dtemp = curve->controlv[j+1];
      curve->controlv[j+1] = curve->controlv[i+1];
      curve->controlv[i+1] = dtemp;

      dtemp = curve->controlv[j+2];
      curve->controlv[j+2] = curve->controlv[i+2];
      curve->controlv[i+2] = dtemp;
      
      if(curve->has_weights)
	{
	  dtemp = curve->controlv[j+3];
	  curve->controlv[j+3] = curve->controlv[i+3];
	  curve->controlv[i+3] = dtemp;
	}
      
      i+=stride;
      j-=stride;
    } /* while */

  /* revert knots */
  if(curve->knot_type == MOPS_KTCUSTOM)
    {
      kncount = curve->length+curve->order;

      if(!(newknotv = calloc(kncount, sizeof(double))))
	return MOPS_OUTOFMEM;

      newknotv[0] = curve->knotv[0];
      newknotv[kncount-1] = curve->knotv[kncount-1];
      j = kncount-2;
      for(i=1; i < (kncount-1); i++)
	{
	  newknotv[i] = newknotv[0]+(newknotv[kncount-1]-curve->knotv[j]);

	  j--;
	}
      free(curve->knotv);
      curve->knotv = newknotv;

    }

 return MOPS_OK;
}


int
mops_revert_ncurve_tcmd(ClientData clientData, Tcl_Interp *interp,
			int argc, char *argv[])
{
 int mops_status;
 mops_list_object *sel = mops_current_selection;
 mops_object *src = NULL;
 mops_nurbcurve_object *curve = NULL;
 char fname[] = "revert_ncurve";

  if(!sel)
    {
      mops_error(interp, MOPS_NOSEL, fname, NULL);
      return TCL_OK;
    }

  while(sel)
    {
      src = sel->object;
      if(src->type != MOPS_OTNURBCURVE)
	{
	  mops_error(interp, MOPS_ERROR, fname, "object is not a NURBCurve");
	}
      else
	{
	  curve = (mops_nurbcurve_object*)src->object;

	  mops_status = mops_revert_ncurve(curve);
	  if(mops_status)
	    {
	      mops_error(interp, mops_status, fname, NULL);
	      return TCL_OK;
	    }

	  mops_nurbcurve_recreatemp(curve);

	} /* if */

      sel = sel->next;
    } /* while */

  mops_notify_parent();

 return TCL_OK;
}


int
mops_crt_trimrect_tcmd(Tcl_Interp * interp,
		       int argc, char *argv[], mops_object *o)
{
 int mops_status;
 mops_list_object *lev = mops_current_level;
 mops_object *clevel = NULL, *parent = NULL;
 mops_nurbpatch_object *patch = NULL;
 mops_nurbcurve_object *curve = NULL;
 char fname[] = "create_trimrect"; 
 double def_controls[15] = { 1.0, 1.0, 0.0,
			     1.0,-1.0, 0.0,
			    -1.0,-1.0, 0.0,
			    -1.0, 1.0, 0.0,
			     1.0, 1.0, 0.0};
 double knots[7] = {0.0, 0.0, 0.25, 0.5, 0.75, 1.0, 1.0};
 int i = 0, create_trim = MOPS_TRUE;

  clevel = lev->object;
  mops_status = mops_del_level_object();
  parent = mops_current_level->object;
  mops_status = mops_add_level_object(clevel);

  if(parent->type != MOPS_OTNURBPATCH)
   {
     create_trim = MOPS_FALSE;
     /*
    mops_error(interp, MOPS_ERROR, fname, "parent object is not a NURBPatch");
    return TCL_ERROR;
     */
   }

  patch = (mops_nurbpatch_object *)parent->object;

  if(!(curve = calloc(1,sizeof(mops_nurbcurve_object))))
    {
      mops_error(interp, MOPS_OUTOFMEM, fname, NULL);
      return TCL_OK;
    }

  if(!(curve->controlv = calloc(15, sizeof(double))))
    {
      free(curve);
      mops_error(interp, MOPS_OUTOFMEM, fname, NULL);
      return TCL_OK;
    }

  if(!(curve->knotv = calloc(7, sizeof(double))))
    {
      free(curve->controlv); free(curve);
      mops_error(interp, MOPS_OUTOFMEM, fname, NULL);
      return TCL_OK;
    }

  curve->length = 5;
  curve->order = 2;
  curve->knot_type = MOPS_KTCUSTOM;
  curve->has_weights = MOPS_FALSE;
  curve->glu_sampling_tolerance = 30.0;
  curve->handle_size = 6.0;

  /* fill knotv */
  for(i=0;i<7;i++)
    curve->knotv[i] = knots[i];

  /* fill controlv */
  if(create_trim)
    {
      curve->controlv[0] = patch->uknotv[0];
      curve->controlv[1] = patch->vknotv[0];

      curve->controlv[3] = patch->uknotv[patch->width+patch->uorder-1];
      curve->controlv[4] = patch->vknotv[0];

      curve->controlv[6] = patch->uknotv[patch->width+patch->uorder-1];
      curve->controlv[7] = patch->vknotv[patch->height+patch->vorder-1];

      curve->controlv[9] = patch->uknotv[0];
      curve->controlv[10] = patch->vknotv[patch->height+patch->vorder-1];

      curve->controlv[12] = patch->uknotv[0];
      curve->controlv[13] = patch->vknotv[0];
    }
  else
    {
      memcpy(curve->controlv, def_controls, 15*sizeof(double));
    }

  o->type = MOPS_OTNURBCURVE;
  o->object = curve;

  mops_nurbcurve_recreatemp(curve);

 return TCL_OK;
}


int
mops_crt_closedbsp_tcmd(Tcl_Interp * interp,
			int argc, char *argv[], mops_object *o)
{
 int mops_status;
 mops_nurbcurve_object *curve = NULL;
 char fname[] = "create_closedbsp"; 
 double *controlv;
 int i=0,num=0,a = 0;
 double angle = 0.0;
 GLdouble m[16] = {0};

  if(argc < 3)
    {
      mops_error(interp,MOPS_ARGS,fname,"numpoints");
      return TCL_ERROR;
    }
  
  Tcl_GetInt(interp,argv[2],&num);
 
  if(num<=2)
    {
      mops_error(interp, MOPS_ERROR, fname, "numpoints must be >= 3");
      return TCL_ERROR;
    }

  if(!(controlv = calloc((num+3)*3, sizeof(double))))
    {
      free(curve);
      mops_error(interp, MOPS_OUTOFMEM, fname, NULL);
      return TCL_ERROR;
    }

  angle = 360.0/num;
  
  glMatrixMode(GL_MODELVIEW);
  glPushMatrix();
  glLoadIdentity ();
  for(i=0;i<num+3;i++)
    {
      a = i*3;
      controlv[a] = 1.0;
      controlv[a+1] = 0.0;
      controlv[a+2] = 0.0;
      glRotated(angle,0.0,0.0,1.0);
      glGetDoublev(GL_MODELVIEW_MATRIX, m);
      mops_apply_transform3(&(controlv[a]),m);
    }
  glPopMatrix();

  mops_status = mops_create_nurbcurve(4, num+3, MOPS_KTBSPLINE, MOPS_FALSE,
				      controlv, NULL, &curve);

  curve->glu_sampling_tolerance = 30.0;
  curve->handle_size = 6.0;
  curve->closed = MOPS_TRUE;
  mops_nurbcurve_close(curve);
  o->object = curve;
  o->type = MOPS_OTNURBCURVE;
  
  mops_nurbcurve_recreatemp(curve);

 return TCL_OK;
}


int
mops_get_ncurve_point_tcmd(ClientData clientData, Tcl_Interp *interp,
			   int argc, char *argv[])
{
 mops_list_object *sel = mops_current_selection;
 mops_object *src = NULL;
 mops_nurbcurve_object *curve = NULL;
 int stride = 3, index = 0;
 char fname[] = "get_ncurve_point";
 Tcl_Obj *to = NULL, *ton = NULL;

  if(argc!=6)
    {
      mops_error(interp, MOPS_ARGS, fname, "index varx vary varz varw");
      return TCL_OK;
    }

  if(!sel)
    {
      mops_error(interp, MOPS_NOSEL, fname, NULL);
      return TCL_OK;
    }

  while(sel)
    {
      src = sel->object;
      if(src->type != MOPS_OTNURBCURVE)
	{
	  mops_error(interp, MOPS_ERROR, fname, "object is not a NURBCurve");
	  return TCL_OK;
	}
      else
	{
	  curve = (mops_nurbcurve_object*)src->object;
	  stride = (curve->has_weights?4:3);

	  Tcl_GetInt(interp, argv[1], &index);
	  index *= stride;

	  if(index >= curve->length*stride || index < 0)
	    {
	      mops_error(interp, MOPS_ERROR, fname, "index out of range");
	      return TCL_OK;
	    }
	  ton = Tcl_NewStringObj(argv[2],-1);
	  to = Tcl_NewDoubleObj(curve->controlv[index]);
	  Tcl_ObjSetVar2(interp,ton,NULL,to,
			 TCL_LEAVE_ERR_MSG | TCL_GLOBAL_ONLY);

	  Tcl_SetStringObj(ton,argv[3],-1);
	  to = Tcl_NewDoubleObj(curve->controlv[index+1]);
	  Tcl_ObjSetVar2(interp,ton,NULL,to,
			 TCL_LEAVE_ERR_MSG | TCL_GLOBAL_ONLY);

	  Tcl_SetStringObj(ton,argv[4],-1);
	  to = Tcl_NewDoubleObj(curve->controlv[index+2]);
	  Tcl_ObjSetVar2(interp,ton,NULL,to,
			 TCL_LEAVE_ERR_MSG | TCL_GLOBAL_ONLY);
	  if(stride == 4)
	  {
	    Tcl_SetStringObj(ton,argv[5],-1);
	    to = Tcl_NewDoubleObj(curve->controlv[index+3]);
	    Tcl_ObjSetVar2(interp,ton,NULL,to,
			   TCL_LEAVE_ERR_MSG | TCL_GLOBAL_ONLY);
	  }

	  Tcl_IncrRefCount(ton);Tcl_DecrRefCount(ton);

	} /* if */

      sel = sel->next;
    } /* while */

 return TCL_OK;
}


int
mops_set_ncurve_point_tcmd(ClientData clientData, Tcl_Interp *interp,
			   int argc, char *argv[])
{
 mops_list_object *sel = mops_current_selection;
 mops_object *src = NULL;
 mops_nurbcurve_object *curve = NULL;
 double dtemp = 0.0;
 int stride = 3, index = 0;
 char fname[] = "set_ncurve_point";

  if(argc<6)
    {
      mops_error(interp, MOPS_ARGS, fname, "index x y z w");
      return TCL_OK;
    }

  if(!sel)
    {
      mops_error(interp, MOPS_NOSEL, fname, NULL);
      return TCL_OK;
    }

  while(sel)
    {
      src = sel->object;
      if(src->type != MOPS_OTNURBCURVE)
	{
	  mops_error(interp, MOPS_ERROR, fname, "object is not a NURBCurve");
	  return TCL_OK;
	}
      else
	{
	  curve = (mops_nurbcurve_object*)src->object;
	  stride = (curve->has_weights?4:3);

	  Tcl_GetInt(interp, argv[1], &index);
	  index *= stride;

	  if(index >= curve->length*stride || index < 0)
	    {
	      mops_error(interp, MOPS_ERROR, fname, "index out of range");
	      return TCL_OK;
	    }

	  Tcl_GetDouble(interp, argv[2], &dtemp);
	  curve->controlv[index] = dtemp;

	  Tcl_GetDouble(interp, argv[3], &dtemp);
	  curve->controlv[index+1] = dtemp;

	  Tcl_GetDouble(interp, argv[4], &dtemp);
	  curve->controlv[index+2] = dtemp;

	  if(stride == 4)
	  {
	    Tcl_GetDouble(interp, argv[5], &dtemp);
	    curve->controlv[index+3] = dtemp;
	  }
	  
	  mops_nurbcurve_recreatemp(curve);

	} /* if */

      sel = sel->next;
    } /* while */

  mops_notify_parent();

 return TCL_OK;
}


int
mops_get_npatch_point_tcmd(ClientData clientData, Tcl_Interp *interp,
			   int argc, char *argv[])
{
 mops_list_object *sel = mops_current_selection;
 mops_object *src = NULL;
 mops_nurbpatch_object *patch = NULL;
 int stride = 3, indexu = 0, indexv = 0;
 char fname[] = "get_npatch_point";
 Tcl_Obj *to = NULL, *ton = NULL;

  if(argc < 7)
    {
      mops_error(interp, MOPS_ARGS, fname,
		 "indexu indexv varx vary varz varw");
      return TCL_OK;
    }

  if(!sel)
    {
      mops_error(interp, MOPS_NOSEL, fname, NULL);
      return TCL_OK;
    }

  while(sel)
    {
      src = sel->object;
      if(src->type != MOPS_OTNURBPATCH)
	{
	  mops_error(interp, MOPS_ERROR, fname, "object is not a NURBPatch");
	  return TCL_OK;
	}
      else
	{
	  patch = (mops_nurbpatch_object*)src->object;
	  stride = (patch->has_weights?4:3);

	  Tcl_GetInt(interp, argv[1], &indexu);
	  if(indexu >= patch->width*stride || indexu < 0)
	    {
	      mops_error(interp, MOPS_ERROR, fname, "index u out of range");
	      return TCL_OK;
	    }
	  indexu *= patch->height;
	  indexu *= stride;

	  Tcl_GetInt(interp, argv[2], &indexv);
	  if(indexv >= patch->height*stride || indexv < 0)
	    {
	      mops_error(interp, MOPS_ERROR, fname, "index v out of range");
	      return TCL_OK;
	    }
	  indexu += (indexv*stride);

	  ton = Tcl_NewStringObj(argv[3],-1);
	  to = Tcl_NewDoubleObj(patch->controlv[indexu]);
	  Tcl_ObjSetVar2(interp,ton,NULL,to,
			 TCL_LEAVE_ERR_MSG | TCL_GLOBAL_ONLY);

	  Tcl_SetStringObj(ton,argv[4],-1);
	  to = Tcl_NewDoubleObj(patch->controlv[indexu+1]);
	  Tcl_ObjSetVar2(interp,ton,NULL,to,
			 TCL_LEAVE_ERR_MSG | TCL_GLOBAL_ONLY);

	  Tcl_SetStringObj(ton,argv[5],-1);
	  to = Tcl_NewDoubleObj(patch->controlv[indexu+2]);
	  Tcl_ObjSetVar2(interp,ton,NULL,to,
			 TCL_LEAVE_ERR_MSG | TCL_GLOBAL_ONLY);
	  if(stride == 4)
	  {
	    Tcl_SetStringObj(ton,argv[6],-1);
	    to = Tcl_NewDoubleObj(patch->controlv[indexu+3]);
	    Tcl_ObjSetVar2(interp,ton,NULL,to,
			   TCL_LEAVE_ERR_MSG | TCL_GLOBAL_ONLY);
	  }

	  Tcl_IncrRefCount(ton);Tcl_DecrRefCount(ton);

	} /* if */

      sel = sel->next;
    } /* while */

 return TCL_OK;
}


int
mops_set_npatch_point_tcmd(ClientData clientData, Tcl_Interp *interp,
			   int argc, char *argv[])
{
 mops_list_object *sel = mops_current_selection;
 mops_object *src = NULL;
 mops_nurbpatch_object *patch = NULL;
 double dtemp = 0.0;
 int stride = 3, indexu = 0, indexv = 0;
 char fname[] = "set_npatch_point";

  if(argc < 7)
    {
      mops_error(interp, MOPS_ARGS, fname, "indexu indexv x y z w");
      return TCL_OK;
    }

  if(!sel)
    {
      mops_error(interp, MOPS_NOSEL, fname, NULL);
      return TCL_OK;
    }

  while(sel)
    {
      src = sel->object;
      if(src->type != MOPS_OTNURBPATCH)
	{
	  mops_error(interp, MOPS_ERROR, fname, "object is not a NURBPatch");
	  return TCL_OK;
	}
      else
	{
	  patch = (mops_nurbpatch_object*)src->object;
	  stride = (patch->has_weights?4:3);

	  Tcl_GetInt(interp, argv[1], &indexu);
	  if(indexu >= patch->width || indexu < 0)
	    {
	      mops_error(interp, MOPS_ERROR, fname, "index u out of range");
	      return TCL_OK;
	    }
	  indexu *= stride;

	  Tcl_GetInt(interp, argv[2], &indexv);
	  if(indexv >= patch->height || indexv < 0)
	    {
	      mops_error(interp, MOPS_ERROR, fname, "index v out of range");
	      return TCL_OK;
	    }

	  indexu += (indexv*patch->width*stride);

	  Tcl_GetDouble(interp, argv[3], &dtemp);
	  patch->controlv[indexu] = dtemp;

	  Tcl_GetDouble(interp, argv[4], &dtemp);
	  patch->controlv[indexu+1] = dtemp;

	  Tcl_GetDouble(interp, argv[5], &dtemp);
	  patch->controlv[indexu+2] = dtemp;

	  if(stride == 4)
	  {
	    Tcl_GetDouble(interp, argv[6], &dtemp);
	    patch->controlv[indexu+3] = dtemp;	    
	  }

	} /* if */

      sel = sel->next;
    } /* while */

  mops_notify_parent();

 return TCL_OK;
}


int
mops_insertkn_ncurve_tcmd(ClientData clientData, Tcl_Interp *interp,
			  int argc, char *argv[])
{
 int mops_status = MOPS_OK;
 mops_list_object *sel = mops_current_selection;
 mops_object *src = NULL;
 mops_nurbcurve_object *curve = NULL;
 double u, *knots = NULL, *newcontrolv = NULL, *newknotv = NULL;
 int stride = 3, i, k = 0, s = 0, r = 0, nq = 0;
 char fname[] = "insertkn_ncurve";

  if(argc<3)
    {
      mops_error(interp, MOPS_ARGS, fname, "u r");
      return TCL_OK;
    }

  if(!sel)
    {
      mops_error(interp, MOPS_NOSEL, fname, NULL);
      return TCL_OK;
    }

  while(sel)
    {
      src = sel->object;
      if(src->type != MOPS_OTNURBCURVE)
	{
	  mops_error(interp, MOPS_ERROR, fname, "object is not a NURBCurve");
	  return TCL_OK;
	}
      else
	{
	  curve = (mops_nurbcurve_object*)src->object;
	  stride = (curve->has_weights?4:3);
	  knots = curve->knotv;

	  Tcl_GetDouble(interp, argv[1], &u);

	  if(u<knots[curve->order-1] || u>knots[curve->length])
	    {
	      mops_error(interp, MOPS_ERROR, fname,
			 "parameter u out of range");
	      return TCL_OK;
	    }

	  i = 0; k = 0;

	  k = mops_nb_FindSpanMult(curve->length-1,
				   curve->order-1, u,
				   knots, &s);

	  Tcl_GetInt(interp, argv[2], &r);

	  if(curve->order < r+s)
	    {
	      mops_error(interp, MOPS_ERROR, fname,
			 "knot insertion leads to illegal knot sequence");
	      return TCL_OK;
	    }

	  curve->length += r;

	  if(!(newcontrolv = calloc(curve->length*stride, sizeof(double))))
	    return MOPS_OUTOFMEM;
	  
	  if(!(newknotv = calloc(curve->length+curve->order, sizeof(double))))
	    return MOPS_OUTOFMEM;


	  if(curve->has_weights)
	    mops_status = mops_nb_CurveInsertKnot4D(curve->length-r-1,
			  curve->order-1, curve->knotv, curve->controlv, u, k,
			  s, r, &nq, newknotv, newcontrolv);
	  else
	    mops_status = mops_nb_CurveInsertKnot3D(curve->length-r-1,
			  curve->order-1, curve->knotv, curve->controlv, u, k,
			  s, r, &nq, newknotv, newcontrolv);

	  free(curve->controlv);
	  curve->controlv = newcontrolv;

	  free(curve->knotv);
	  curve->knotv = newknotv;
	  curve->knot_type = MOPS_KTCUSTOM;

	  mops_nurbcurve_recreatemp(curve);

	} /* if */

      sel = sel->next;
    } /* while */

  mops_notify_parent();

 return TCL_OK;
}


int
mops_split_ncurve(Tcl_Interp *interp, mops_object *src, double u)
{
 int mops_status = MOPS_OK;
 mops_object *new = NULL;
 mops_nurbcurve_object *curve = NULL;
 mops_nurbcurve_object *nc1 = NULL, *nc2 = NULL;
 double *knots = NULL, *newcontrolv = NULL, *newknotv = NULL;
 int stride = 3, k = 0, r = 0, s = 0, nq = 0, nc1len = 0;
 char fname[] = "split_ncurve";


 if(!src)
   return MOPS_OK;


 if(src->type != MOPS_OTNURBCURVE)
   {
     mops_error(interp, MOPS_ERROR, fname, "object is not a NURBCurve");
     return MOPS_ERROR;
   }
 else
   {
     curve = (mops_nurbcurve_object*)src->object;
     stride = (curve->has_weights?4:3);
     knots = curve->knotv;

     if((u <= knots[0/*curve->order-2*/]) || (u >= knots[curve->length]))
       {
	 mops_error(interp, MOPS_ERROR, fname,
		    "parameter u out of range");
	 return MOPS_ERROR;
       }

     k = 0;

     k = mops_nb_FindSpanMult(curve->length-1, curve->order-1, u,
			      knots, &s);

     r = curve->order-1-s;
	  

     curve->length += r;

     if(r != 0)
       {
	 newcontrolv = NULL;
	 if(!(newcontrolv = calloc(curve->length*stride, sizeof(double))))
	   return MOPS_OUTOFMEM;
	 newknotv = NULL;	  
	 if(!(newknotv = calloc(curve->length+curve->order,
				sizeof(double))))
	   return MOPS_OUTOFMEM;

	 if(curve->has_weights)
	   mops_status = mops_nb_CurveInsertKnot4D(curve->length-r-1,
			curve->order-1, curve->knotv, curve->controlv, u, k,
			s, r, &nq, newknotv, newcontrolv);
	 else
	   mops_status = mops_nb_CurveInsertKnot3D(curve->length-r-1,
			curve->order-1, curve->knotv, curve->controlv, u, k,
			s, r, &nq, newknotv, newcontrolv);

	 free(curve->controlv);
	 curve->controlv = newcontrolv;

	 free(curve->knotv);
	 curve->knotv = newknotv;
       }

     curve->knot_type = MOPS_KTCUSTOM;
     /* create two new curves */
     nc1 = curve;
     nc1->closed = MOPS_FALSE;
     mops_status = mops_copy_object(src, &new);
     MOPS_CRERROR
     mops_status = mops_link_new_object(new);  
     MOPS_CRERROR

       if(r != 0)
	 nc1len = k - (nc1->order-1) + 1 + (curve->order-1-s+r-1)/2 + 1;
       else
	 nc1len = k - (nc1->order-1) + 1;

     nc2 = (mops_nurbcurve_object*)new->object;
     nc2->length = (nc1->length+1) - nc1len;
     nc1->length = nc1len;

     newcontrolv = NULL;
     if(!(newcontrolv = calloc(nc1->length*stride, sizeof(double))))
       return MOPS_OUTOFMEM;
     newknotv = NULL;	  
     if(!(newknotv = calloc(nc1->length+nc1->order, sizeof(double))))
       return MOPS_OUTOFMEM;

     memcpy(newcontrolv,nc1->controlv,nc1->length*stride*sizeof(double));

     memcpy(newknotv,nc1->knotv,(nc1->length+nc1->order)*sizeof(double));

     free(nc2->controlv);
     nc2->controlv = NULL;
     free(nc2->knotv);
     nc2->knotv = NULL;

     if(!(nc2->controlv = calloc(nc2->length*stride, sizeof(double))))
       return MOPS_OUTOFMEM;

     if(!(nc2->knotv = calloc(nc2->length+nc2->order, sizeof(double))))
       return MOPS_OUTOFMEM;

     memcpy(nc2->controlv,&(nc1->controlv[(nc1->length-1)*stride]),
	    nc2->length*stride*sizeof(double));

     memcpy(nc2->knotv,&(nc1->knotv[nc1->length-1]),
	    (nc2->length+nc2->order)*sizeof(double));

     free(nc1->controlv);
     nc1->controlv = newcontrolv;
     free(nc1->knotv);
     nc1->knotv = newknotv;

     mops_nurbcurve_recreatemp(nc1);
     mops_nurbcurve_recreatemp(nc2);

     new->selected = MOPS_FALSE;
     new->modified = MOPS_TRUE;
     src->modified = MOPS_TRUE;

   } /* if */

 return MOPS_OK;
}


int
mops_split_ncurve_tcmd(ClientData clientData, Tcl_Interp *interp,
		       int argc, char *argv[])
{
 int mops_status = MOPS_OK;
 mops_list_object *sel = mops_current_selection;
 double u = 0.0;
 char fname[] = "split_ncurve_tcmd";

  if(argc<2)
    {
      mops_error(interp, MOPS_ARGS, fname, "u");
      return TCL_OK;
    }

  if(!sel)
    {
      mops_error(interp, MOPS_NOSEL, fname, NULL);
      return TCL_OK;
    }


  Tcl_GetDouble(interp, argv[1], &u);

  while(sel)
    {
      if(sel->object)
	{
	  if(sel->object->selp)
	    mops_clear_point_selection(sel->object);

	  if(sel->object->type == MOPS_OTNURBCURVE)
	    {
	      mops_status = mops_split_ncurve(interp, sel->object, u);

	      if(mops_status)
		{
		  mops_error(interp, mops_status, fname, NULL);
		  return TCL_OK;
		}
	    }
	  else
	    {
	      mops_error(interp, MOPS_ERROR, fname,
			 "object is not a NURBCurve");
	      return TCL_OK;
	    }
	}
      sel = sel->next;
    }

  mops_notify_parent();

 return TCL_OK;
}


int
mops_break_cbsp_tcmd(ClientData clientData, Tcl_Interp *interp,
		     int argc, char *argv[])
{
 mops_list_object *sel = mops_current_selection;
 mops_nurbcurve_object *curve = NULL;
 int stride = 3, i = 0;
 char fname[] = "break_cbsp_tcmd";
 double *newcontrolv = NULL;

  if(argc<2)
    {
      mops_error(interp, MOPS_ARGS, fname, "i");
      return TCL_OK;
    }

  if(!sel)
    {
      mops_error(interp, MOPS_NOSEL, fname, NULL);
      return TCL_OK;
    }


  Tcl_GetInt(interp, argv[1], &i);

  while(sel)
    {
      if(sel->object)
	{
	  if(sel->object->type == MOPS_OTNURBCURVE)
	    {
	      curve = (mops_nurbcurve_object *)sel->object->object;
	      stride = (curve->has_weights?4:3);

	      if(curve->knot_type != MOPS_KTBSPLINE)
		{
		  mops_error(interp, MOPS_ERROR, fname,
			     "knot type is not B-Spline");
		  return TCL_OK;
		}

	      newcontrolv = NULL;
	      if(!(newcontrolv = calloc(curve->length*stride, sizeof(double))))
		{
		  mops_error(interp, MOPS_OUTOFMEM, fname, NULL);
		  return TCL_OK;
		}

	      memcpy(newcontrolv, &(curve->controlv[curve->length-i]),
		     (curve->length-i)*stride*sizeof(double));

	      /*
		for(a = 0; a < i*stride; a++)
		{}
	      */
	    }
	  else
	    {
	      mops_error(interp, MOPS_ERROR, fname,
			 "object is not a NURBCurve");
	      return TCL_OK;
	    }
	}
      sel = sel->next;
    }

 return TCL_OK;
}


int
mops_project_nc_tcmd(ClientData clientData, Tcl_Interp *interp,
		       int argc, char *argv[])
{
 mops_list_object *sel = mops_current_selection;
 double u = 0.0, point[3], guess = 0.0, e1 = 0.0, e2 = 0.0;
 int maxtry;
 char fname[] = "project_point_tcmd";
 Tcl_Obj *to = NULL, *ton = NULL;

  if(argc<9)
    {
      mops_error(interp, MOPS_ARGS, fname, "x y z guess maxtry e1 e2 varname");
      return TCL_OK;
    }

  if(!sel)
    {
      mops_error(interp, MOPS_NOSEL, fname, NULL);
      return TCL_OK;
    }

  /* get args */
  Tcl_GetDouble(interp, argv[1], &(point[0]));
  Tcl_GetDouble(interp, argv[2], &(point[1]));
  Tcl_GetDouble(interp, argv[3], &(point[2]));

  Tcl_GetDouble(interp, argv[4], &guess);
  Tcl_GetInt(interp, argv[5], &maxtry);
  Tcl_GetDouble(interp, argv[6], &e1);
  Tcl_GetDouble(interp, argv[7], &e2);


  if(sel->object)
    {
      if(sel->object->type == MOPS_OTNURBCURVE)
	{
	  
	  mops_nb_FinduFromPoint(
			(mops_nurbcurve_object*)sel->object->object,
			point, guess, maxtry, e1, e2, &u);

	  ton = Tcl_NewStringObj(argv[8], -1);
	  to = Tcl_NewDoubleObj(u);
	  Tcl_ObjSetVar2(interp,ton,NULL,to,
			 TCL_LEAVE_ERR_MSG | TCL_GLOBAL_ONLY);

	  Tcl_IncrRefCount(ton);Tcl_DecrRefCount(ton);

	}
      else
	{
	  mops_error(interp, MOPS_ERROR, fname,
		     "object is not a NURBCurve");
	  return TCL_OK;
	}
    }


 return TCL_OK;
}


int
mops_refine_curve_tcmd(ClientData clientData, Tcl_Interp *interp,
		       int argc, char *argv[])
{
 mops_list_object *sel = mops_current_selection;
 mops_nurbcurve_object *curve = NULL;
 double *X = NULL, *Ubar = NULL, *Qw = NULL, *knotv, *Q = NULL;
 int aknotc = 0, count = 0, i, j, stride;
 char **aknotv;
 char fname[] = "refine_curve_tcmd";
 

  if(!sel)
    {
      mops_error(interp, MOPS_NOSEL, fname, NULL);
      return TCL_OK;
    }

  if(argc > 1)
    {
      Tcl_SplitList(interp, argv[1], &aknotc, &aknotv);
      if(!(X = calloc(aknotc, sizeof(double))))
	{
	  mops_error(interp,MOPS_OUTOFMEM,fname,NULL);
	  Tcl_Free((char *) aknotv);
	  return TCL_OK;
	}

      for(i = 0; i < aknotc; i++)
	{
	  Tcl_GetDouble(interp,aknotv[i],&X[i]);
	} /* for */

      count = aknotc;

      Tcl_Free((char *) aknotv);
    }


  if(sel->object)
    {
      if(sel->object->type == MOPS_OTNURBCURVE)
	{
	  if(sel->object->selp)
	    mops_clear_point_selection(sel->object);

	  curve = (mops_nurbcurve_object *)sel->object->object;
	  knotv = curve->knotv;
	  stride = curve->has_weights?4:3;

	  if(curve->closed)
	    {
	      /* special case: curves marked closed;
	       * we keep the p multiple points at the ends
	       * and add new points into the other sections
	       * taking care for other multiple points...
	       */
	      /* first, count new points */
	      count = 0;
	      Q = curve->controlv;
	      for(i = curve->order-2; i < curve->length-(curve->order-1); i++)
		{
		  if((Q[i*stride]   != Q[(i+1)*stride]) ||
		     (Q[i*stride+1] != Q[(i+1)*stride+1]) ||
		     (Q[i*stride+2] != Q[(i+1)*stride+2]))
		    {
		      count++;
		    }
		}
	      
	      if(count)
		{
		  /* alloc new control vector */
		  if(!(Qw = calloc((curve->length + count)*stride, 
				   sizeof(double))))
		    {
		      mops_error(interp, MOPS_OUTOFMEM, fname, NULL);
		      return TCL_OK;
		    }

		  /* copy first p points */
		  memcpy(Qw, Q, (curve->order-1)*stride*sizeof(double));

		  /* copy old & create new points */
		  j = curve->order-2;
		  for(i=curve->order-2; i<curve->length-(curve->order-1); i++)
		    {
		      memcpy(&(Qw[j*stride]), &(Q[i*stride]),
			     stride*sizeof(double));

		      if((Q[i*stride] != Q[(i+1)*stride]) ||
			 (Q[i*stride+1] != Q[(i+1)*stride+1]) ||
			 (Q[i*stride+2] != Q[(i+1)*stride+2]))
			{
			  Qw[(j+1)*stride] = Q[i*stride] +
			    ((Q[(i+1)*stride] - Q[i*stride])/2.0);

			  Qw[(j+1)*stride+1] = Q[i*stride+1] +
			    ((Q[(i+1)*stride+1] - Q[i*stride+1])/2.0);

			  Qw[(j+1)*stride+2] = Q[i*stride+2] +
			    ((Q[(i+1)*stride+2] - Q[i*stride+2])/2.0);

			  if(stride == 4)
			    {
			      Qw[(j+1)*stride+3] = Q[i*stride+3] +
				((Q[(i+1)*stride+3] - Q[i*stride+3])/2.0);
			    }

			  j++;
			}

		      j++;
		    }

		  /* copy last p points */
		  memcpy(&(Qw[(curve->length+count-(curve->order-1))*stride]),
			 &(Q[(curve->length-(curve->order-1))*stride]),
			 (curve->order-1)*stride*sizeof(double));

		  curve->length += count;
		  free(curve->controlv);
		  curve->controlv = Qw;
		  mops_nurbcurve_recreatemp(curve);
		  curve->knot_type = MOPS_KTBSPLINE;
		  mops_create_nurbcurve_knots(curve);
		} /* if count */
	    }
	  else
	    {
	      /* alloc X, new knotv & new controlv */
	      if(!X)
		{
		  count = 0;
		  for(i = curve->order-1; i < curve->length; i++)
		    {
		      if(knotv[i] != knotv[i+1])
			count++;
		    }

		  if(!(X = calloc(count, sizeof(double))))
		    {
		      mops_error(interp, MOPS_OUTOFMEM, fname, NULL);
		      return TCL_OK;
		    }
		}

	      if(!(Ubar = calloc((curve->length + curve->order + count), 
				 sizeof(double))))
		{
		  free(X);
		  mops_error(interp, MOPS_OUTOFMEM, fname, NULL);
		  return TCL_OK;
		}
	      if(!(Qw = calloc((curve->length + count+2)*
			       (curve->has_weights?4:3), 
			       sizeof(double))))
		{
		  free(X); free(Ubar);
		  mops_error(interp, MOPS_OUTOFMEM, fname, NULL);
		  return TCL_OK;
		}

	      if( argc < 2)
		{
		  /* fill X (contains just the new u values) */
		  count = 0;

		  for(i = curve->order-1; i < curve->length; i++)
		    {
		      if(knotv[i] != knotv[i+1])
			{
			  X[count] = knotv[i]+((knotv[i+1]-knotv[i])/2.0);
			  count++;
			}
		    }
		}

	      /* fill Ubar & Qw */
	      mops_nb_RefineKnotVectCurve(curve->has_weights?4:3,
					  curve->length-1, curve->order-1,
					  curve->knotv, curve->controlv,
					  X, count-1, Ubar, Qw);

	      free(curve->knotv);
	      curve->knotv = Ubar;

	      free(curve->controlv);
	      curve->controlv = Qw;

	      free(X);
	      
	      curve->length += count;
	      
	      mops_nurbcurve_recreatemp(curve);

	    } /* if curve closed */
	}
      else
	{
	  mops_error(interp, MOPS_ERROR, fname,
		     "object is not a NURBCurve");
	  
	  if(X)
	    free(X);

	  return TCL_OK;
	}
      
    }

  mops_notify_parent();

 return TCL_OK;
}


int
mops_elevate_curve_tcmd(ClientData clientData, Tcl_Interp *interp,
		      int argc, char *argv[])
{
 int mops_status = MOPS_OK;
 mops_list_object *sel = mops_current_selection;
 mops_nurbcurve_object *curve = NULL;
 int i, j, a, b, clamp_me;
 double u, *Uh = NULL, *Qw = NULL, *realQw = NULL, *realUh = NULL;
 int t = 1, nh = 0;
 char fname[] = "elevate_curve_tcmd";

  if(!sel)
    {
      mops_error(interp, MOPS_NOSEL, fname, NULL);
      return TCL_OK;
    }

  if(argc >= 2)
    Tcl_GetInt(interp, argv[1], &t);

  if(sel->object)
    {
      if(sel->object->type == MOPS_OTNURBCURVE)
	{
	  if(sel->object->selp)
	    mops_clear_point_selection(sel->object);

	  curve = (mops_nurbcurve_object *)sel->object->object;

	  /* clamp the curve? */
	  clamp_me = MOPS_FALSE;

	  if(curve->knot_type == MOPS_KTBSPLINE)
	    {
	      clamp_me = MOPS_TRUE;
	    }
	  else
	    {
	      if(curve->knot_type == MOPS_KTCUSTOM)
		{
		  a = 1;
		  u = curve->knotv[0];
		  for(i = 1; i < curve->order; i++)
		    if(u == curve->knotv[i])
		      a++;

		  j = curve->length+curve->order-1;
		  b = 1;
		  u = curve->knotv[j];
		  for(i = j; i >= curve->length; i--)
		    if(u == curve->knotv[i])
		      b++;

		  if((a < curve->order) || (b < curve->order))
		    {
		      clamp_me = MOPS_TRUE;
		    }
		}
	    }
	  
	  if(clamp_me)
	    mops_status = mops_clamp_curve(curve);

	  /* alloc new knotv & new controlv */
	  if(!(Uh = calloc((curve->length + curve->length*t +
			    curve->order + t), 
			     sizeof(double))))
	    {
	      mops_error(interp, MOPS_OUTOFMEM, fname, NULL);
	      return TCL_OK;
	    }
	  if(!(Qw = calloc((curve->length + curve->length*t)*
			   (curve->has_weights?4:3), sizeof(double))))
	    {
	      free(Uh);
	      mops_error(interp, MOPS_OUTOFMEM, fname, NULL);
	      return TCL_OK;
	    }

	  /* fill Uh & Qw */
	  mops_status = mops_nb_DegreeElevateCurve(curve->has_weights?4:3,
			      curve->length-1, curve->order-1, curve->knotv,
		              curve->controlv, t, &nh, Uh, Qw);

	  if(mops_status)
	    {
	      mops_error(interp,mops_status,fname,"degree elevation failed");
	      free(Uh); free(Qw); return TCL_OK;
	    }

	  if(!(realQw = realloc(Qw, nh*(curve->has_weights?4:3)*
				sizeof(double))))
	    {
	      mops_error(interp, MOPS_ERROR, fname, "Memory may have leaked!");
	      mops_error(interp, MOPS_OUTOFMEM, fname, NULL);
	      return TCL_OK;
	    }

	  if(!(realUh = realloc(Uh, (nh+curve->order+t)*sizeof(double))))
	    {
	      mops_error(interp, MOPS_ERROR, fname, "Memory may have leaked!");
	      mops_error(interp, MOPS_OUTOFMEM, fname, NULL);
	      return TCL_OK;
	    }

	  free(curve->knotv);
	  curve->knotv = realUh;

	  free(curve->controlv);
	  curve->controlv = realQw;

	  curve->knot_type = MOPS_KTCUSTOM;
	  
	  curve->order += t;
	  
	  curve->length = nh;

	  Qw = NULL;
	  Uh = NULL;
	  realQw = NULL;
	  realUh = NULL;

	  mops_nurbcurve_recreatemp(curve);

	}
      else
	{
	  mops_error(interp, MOPS_ERROR, fname,
		     "object is not a NURBCurve");
	  return TCL_OK;
	}
    }

  mops_notify_parent();

 return TCL_OK;
}


int
mops_clamp_curve_tcmd(ClientData clientData, Tcl_Interp *interp,
		      int argc, char *argv[])
{
 int mops_status = MOPS_OK;
 mops_list_object *sel = mops_current_selection;
 mops_nurbcurve_object *curve = NULL;
 double *knotv, u;
 int count_start, count_end, i, j;
 char fname[] = "clamp_curve_tcmd";

  if(!sel)
    {
      mops_error(interp, MOPS_NOSEL, fname, NULL);
      return TCL_OK;
    }

  /*
  if(argc >= 2)
    Tcl_GetInt(interp, argv[1], &t);
  */
  while(sel)
  {

    if(sel->object)
      {
	if(sel->object->type == MOPS_OTNURBCURVE)
	  {
	    if(sel->object->selp)
	      mops_clear_point_selection(sel->object);

	    curve = (mops_nurbcurve_object *)sel->object->object;
	    knotv = curve->knotv;

	    if((curve->knot_type == MOPS_KTNURB) ||
	       (curve->knot_type == MOPS_KTBEZIER))
	      return TCL_OK;
	    
	    if(curve->knot_type == MOPS_KTCUSTOM)
	      {
		count_start = 0;
		count_end = 0;
		u = knotv[0];
		for(i = 1; i < curve->order; i++)
		 if(u == knotv[i])
		   count_start++;

		j = curve->length+curve->order-1;
		u = knotv[j];
		for(i = j-1; i >= curve->length; i--)
		 if(u == knotv[i])
		   count_end++;

		/*
		if(count_start || count_end)
		  {
		    mops_error(interp, MOPS_ERROR, fname,
			       "cannot clamp this curve");
		    return TCL_OK;
		  }
		*/
	      }


	    mops_status = mops_clamp_curve(curve);


	    if(mops_status)
	      {
		mops_error(interp, mops_status, fname, "clamping failed");
		return TCL_OK;
	      }
	    curve->knot_type = MOPS_KTCUSTOM;

	    mops_nurbcurve_recreatemp(curve);

	  }
	else
	  {
	    mops_error(interp, MOPS_ERROR, fname,
		       "object is not a NURBCurve");
	    return TCL_OK;
	  }
    }

  sel = sel->next;
  }

  mops_notify_parent();

 return TCL_OK;
}


int
mops_reparam_curve_tcmd(ClientData clientData, Tcl_Interp *interp,
		      int argc, char *argv[])
{
 int mops_status = MOPS_OK;
 mops_list_object *sel = mops_current_selection;
 mops_nurbcurve_object *curve = NULL;
 double *knotv = NULL, *controlv = NULL, d, *lengths = NULL, knot = 0.0;
 int a, i, j, stride, mode = 0;
 char fname[] = "reparam_curve_tcmd";

  if(!sel)
    {
      mops_error(interp, MOPS_NOSEL, fname, NULL);
      return TCL_OK;
    }

  if(argc >= 2)
    {
      if(!strcmp(argv[1], "-rs"))
	mode = 1;
      if(!strcmp(argv[1], "-cl"))
	mode = 2;

    }

  while(sel)
  {

    if(sel->object)
      {
	if(sel->object->type == MOPS_OTNURBCURVE)
	  {

	    curve = (mops_nurbcurve_object *)sel->object->object;
	    stride = curve->has_weights?4:3;

	    controlv = curve->controlv;

	    switch(mode)
	      {
	      case 1:
		if(curve->knot_type == MOPS_KTCUSTOM)
		  {
		    mops_status = mops_rescale_knotv(curve->length+
						     curve->order,
						     curve->knotv);
		    if(mops_status)
		      {
			mops_error(interp, mops_status, fname,
				   "rescaling failed");
			return TCL_OK;
		      }
		  }
		break;
	      case 2:
		knotv = NULL;
		if(!(knotv = calloc(curve->length+curve->order,
				    sizeof(double))))
		  {
		    mops_error(interp, MOPS_OUTOFMEM, fname, NULL);
		    return TCL_OK;
		  }

		/* get chordlength of curve */
		d = 0.0;
		lengths = NULL;
		if(!(lengths = calloc(curve->length, sizeof(double))))
		  return MOPS_OUTOFMEM;
		a = 0;
		for(i = 0; i < (curve->length-1); i++)
		  {
		    lengths[i] = MOPS_VLEN((controlv[a+3] - controlv[a]),
					   (controlv[a+4] - controlv[a+1]),
					   (controlv[a+5] - controlv[a+2]));

		    d += lengths[i];
		
		    a += stride;
		  }

		
		/* fill new knotv */
		for(i = 0; i < curve->order; i++)
		  knotv[i] = 0.0;
		for(i = curve->length; i < curve->length+curve->order; i++)
		  knotv[i] = 1.0;

		if(d > 0)
		  {
		    j = 0;
		    for(i = curve->order; i < curve->length; i++)
		      {
			knot += lengths[j]/d;
			knotv[i] = knot;
			fprintf(stderr,"knotv[%d]=%g\n",i,knotv[i]);
			j++;
		      }
		  }
		else
		  {
		    for(i = curve->order; i < curve->length; i++)
		      {
			knotv[i] = i/(curve->length-2);
		      }
		  }

		/* reparam curve */
		/*
		mops_status = mops_reparam_curve(curve, knotv);
		
		if(mops_status)
		  {
		    mops_error(interp, mops_status, fname,
			       "reparameterization failed");
		    return TCL_OK;
		  }
		*/
		free(curve->knotv);
		curve->knotv = knotv;
		curve->knot_type = MOPS_KTCUSTOM;
		free(lengths);
		/*		free(knotv);*/
		break;
	      } /* switch */
	  }
	else
	  {
	    mops_error(interp, MOPS_ERROR, fname,
		       "object is not a NURBCurve");
	    return TCL_OK;
	  }
    }

  sel = sel->next;
  }

  mops_notify_parent();

 return TCL_OK;
}


int
mops_collapse_tcmd(ClientData clientData, Tcl_Interp *interp,
		      int argc, char *argv[])
{
 int mops_status = MOPS_OK;
 mops_list_object *sel = mops_current_selection;
 char fname[] = "collapse_tcmd";

  if(!sel)
    {
      mops_error(interp, MOPS_NOSEL, fname, NULL);
      return TCL_OK;
    }

  while(sel)
  {

    if(sel->object)
      {
	if(sel->object->type == MOPS_OTNURBCURVE)
	  {
	    mops_status = mops_nurbcurve_collapseselp(sel->object);
	    MOPS_CRERROR

 	    if(sel->object->selp)
	      mops_clear_point_selection(sel->object);

	  }
	else
	  {
	    mops_error(interp, MOPS_ERROR, fname,
		       "object is not a NURBCurve");
	    return TCL_OK;
	  }

    }

   sel = sel->next;
  }

  mops_notify_parent();

 return TCL_OK;
}


int
mops_explode_tcmd(ClientData clientData, Tcl_Interp *interp,
		      int argc, char *argv[])
{
 int mops_status = MOPS_OK;
 mops_list_object *sel = mops_current_selection;
 char fname[] = "explode_tcmd";

  if(!sel)
    {
      mops_error(interp, MOPS_NOSEL, fname, NULL);
      return TCL_OK;
    }

  while(sel)
  {

    if(sel->object)
      {
	if(sel->object->type == MOPS_OTNURBCURVE)
	  {
	    mops_status = mops_nurbcurve_explodemp(sel->object);
	    MOPS_CRERROR

 	    if(sel->object->selp)
	      mops_clear_point_selection(sel->object);

	  }
	else
	  {
	    mops_error(interp, MOPS_ERROR, fname,
		       "object is not a NURBCurve");
	    return TCL_OK;
	  }

    }

   sel = sel->next;
  }

  mops_notify_parent();

 return TCL_OK;
}


