/*
 * The Mops, a free 3D Tcl/Tk/OpenGL/RenderMan modeler
 *
 * The Mops is copyrighted (C) 1998-1999 by Randolf Schultz
 * (rschultz@informatik.uni-rostock.de)
 *
 * See the file License for details.
 *
 */

#include "mops.h"

/* Quaternions */

/*
  Quaternion code adapted from various articles
  from c.g.a.opengl and from GLUT example code
 */


/*
 * This function computes a quaternion based on an axis (defined by
 * the given vector) and an angle about which to rotate.  The angle is
 * expressed in radians.  The result is put into the third argument.
 */
void
mops_axis_to_quat(float a[3], float phi, float q[4])
{
 double len = 0.0;

  len = MOPS_V3LEN(a);
  a[0] *= 1.0/len;
  a[1] *= 1.0/len;
  a[2] *= 1.0/len;

  q[0]=a[0];
  q[1]=a[1];
  q[2]=a[2];
  MOPS_V3SCAL(q, sin(phi/2.0))

  q[3] = cos(phi/2.0);

 return;
}


/*
 * Given two rotations, e1 and e2, expressed as quaternion rotations,
 * figure out the equivalent single rotation and stuff it into dest.
 *
 * This routine also normalizes the result every RENORMCOUNT times it is
 * called, to keep error from creeping in.
 *
 * NOTE: This routine is written so that q1 or q2 may be the same
 * as dest (or each other).
 */
void
mops_add_quat(float q1[4], float q2[4], float dest[4])
{
 static int count=0;
 float t1[4], t2[4], t3[4];
 float tf[4];

 t1[0]=q1[0]; t1[1]=q1[1]; t1[2]=q1[2];
 MOPS_V3SCAL(t1,q2[3]);

 t2[0]=q2[0]; t2[1]=q2[1]; t2[2]=q2[2];
 MOPS_V3SCAL(t2,q1[3]);

 MOPS_V3CROSS(t3,q2,q1);
 MOPS_V3ADD(tf,t1,t2);
 MOPS_V3ADD(tf,t3,tf);

 tf[3] = q1[3] * q2[3] - (MOPS_V3DOT(q1,q2));

 dest[0] = tf[0];
 dest[1] = tf[1];
 dest[2] = tf[2];
 dest[3] = tf[3];

 if (++count > MOPS_QRENORMCOUNT) {
   count = 0;
   mops_norm_quat(dest);
 }

 return;
}


/*
 * Quaternions always obey:  a^2 + b^2 + c^2 + d^2 = 1.0
 * If they don't add up to 1.0, dividing by their magnitued will
 * renormalize them.
 *
 * Note: See the following for more information on quaternions:
 *
 * - Shoemake, K., Animating rotation with quaternion curves, Computer
 *   Graphics 19, No 3 (Proc. SIGGRAPH'85), 245-254, 1985.
 * - Pletinckx, D., Quaternion calculus as a basic tool in computer
 *   graphics, The Visual Computer 5, 2-13, 1989.
 */
void
mops_norm_quat(float q[4])
{
 int i;
 float mag;

  mag = (q[0]*q[0] + q[1]*q[1] + q[2]*q[2] + q[3]*q[3]);
  for (i = 0; i < 4; i++) q[i] /= mag;

 return;
}

/*
 * Build a rotation matrix, given a quaternion rotation.
 *
 */
void
mops_build_rotmatrix(float m[4][4], float q[4])
{

  m[0][0] = 1.0 - 2.0 * (q[1] * q[1] + q[2] * q[2]);
  m[0][1] = 2.0 * (q[0] * q[1] - q[2] * q[3]);
  m[0][2] = 2.0 * (q[2] * q[0] + q[1] * q[3]);
  m[0][3] = 0.0;

  m[1][0] = 2.0 * (q[0] * q[1] + q[2] * q[3]);
  m[1][1]= 1.0 - 2.0 * (q[2] * q[2] + q[0] * q[0]);
  m[1][2] = 2.0 * (q[1] * q[2] - q[0] * q[3]);
  m[1][3] = 0.0;

  m[2][0] = 2.0 * (q[2] * q[0] - q[1] * q[3]);
  m[2][1] = 2.0 * (q[1] * q[2] + q[0] * q[3]);
  m[2][2] = 1.0 - 2.0 * (q[1] * q[1] + q[0] * q[0]);
  m[2][3] = 0.0;

  m[3][0] = 0.0;
  m[3][1] = 0.0;
  m[3][2] = 0.0;
  m[3][3] = 1.0;

 return;
}


/*
 *  Convert the quaternion q to euler[psi, theta, phi].
 *  Angles are defined in radians.
 */

#define EPSILON (1.0e-05)

void
mops_quat_to_euler(float q[4], float euler[3])
{
 double sin_theta;      /* sin of pitch */

 double R00, R01, R02;  /* rotation matrix elements */
 double R11, R12;
 double R21, R22;

  R02 = 2.0*(q[0]*q[2] - q[3]*q[1]);

  sin_theta = -R02;

  if (fabs(sin_theta) > (1.0 - EPSILON))
    {
      /* very close to +- 90 degrees pitch, set heading to zero
         and compute the roll angle */
      euler[0] = 0.0;

      if (sin_theta > 0)
        euler[1] = MOPS_PI/2.0;
      else
        euler[1] = -MOPS_PI/2.0;

      R11 = 2.0*(q[3]*q[3] + q[1]*q[1]) - 1.0;
      R21 = 2.0*(q[1]*q[2] - q[3]*q[0]);

      euler[2] = atan2(-R21, R11);
    }
  else
    {
      R00 = 2.0*(q[3]*q[3] + q[0]*q[0]) - 1.0;
      R01 = 2.0*(q[0]*q[1] + q[3]*q[2]);
      R12 = 2.0*(q[1]*q[2] + q[3]*q[0]);
      R22 = 2.0*(q[3]*q[3] + q[2]*q[2]) - 1.0;

      /* heading */
      euler[0] = atan2(R01, R00);

      /* pitch */
      euler[1] = asin(sin_theta);

      /* roll */
      euler[2] = atan2(R12, R22);
    }

 return;
}

