/*
 * The Mops, a free 3D Tcl/Tk/OpenGL/RenderMan modeler
 *
 * The Mops is copyrighted (C) 1998-1999 by Randolf Schultz
 * (rschultz@informatik.uni-rostock.de)
 *
 * See the file License for details.
 *
 */

#include "mops.h"

#ifdef MOPSIDR
#include "idr.h"
#endif

#include <limits.h>
#include <float.h>

/* Revolve */

/* The Name */
static char *Revolve_name = "Revolve";

static char *Revolve_customid = NULL;

/* prototypes */
typedef struct Revolve_s
{
 mops_object *revolution;
 int has_upper_cap;
 mops_object *upper_cap;
 int has_lower_cap;
 mops_object *lower_cap;
 double ThetaMax;
 double glu_sampling_tolerance;
 int glu_display_mode;
} Revolve;

int Revolve_create_cb(Tcl_Interp *interp, int argc, char *argv[], void **c);
int Revolve_delete_cb(void *c);
int Revolve_copy_cb(void *source, void **dest);
int Revolve_draww_cb(struct Togl *togl, mops_object *o);
int Revolve_draws_cb(struct Togl *togl, mops_object *o);
int Revolve_wrib_cb(mops_object *o);
int Revolve_read_cb(FILE *fileptr, mops_custom_object *custom);
int Revolve_write_cb(FILE *fileptr, mops_custom_object *custom);
int Revolve_getprop_cb(Tcl_Interp *interp);
int Revolve_set_attrprop_tcmd(ClientData clientData, Tcl_Interp *interp,
			     int argc, char *argv[]);
int Revolve_get_attrprop_tcmd(ClientData clientData, Tcl_Interp *interp,
			     int argc, char *argv[]);
int Revolve_notify_cb(mops_object *o);
int Revolve_convert(Tcl_Interp *interp, mops_object *o);
int Revolve_convertNP_tcmd(ClientData clientData, Tcl_Interp *interp,
			 int argc, char *argv[]);

int Revolve_provide_cb(mops_object *o, void **target);


#ifdef MOPSIDR
void Revolve_countbmms_cb(mops_object *o);
#endif

/* functions */
int
Revolve_Init(Tcl_Interp *interp)
{
 int mops_status = MOPS_OK;
 Tcl_HashEntry *entry;
 char fname[] = "Revolve_Init";

  mops_status = mops_register_custom_objecttype(
				     Revolve_name,
				     Revolve_create_cb,
				     Revolve_draws_cb, 
				     NULL,
				     Revolve_draww_cb,
				     Revolve_delete_cb,
				     Revolve_copy_cb,
				     Revolve_getprop_cb,
				     NULL, /* getpoint */ /* no single point
					                     editing! */
				     Revolve_wrib_cb,
				     Revolve_read_cb,
				     Revolve_write_cb,
				     Revolve_notify_cb
				     );

  if(mops_status)
    {
      mops_error(interp, mops_status, fname, NULL);
      return TCL_ERROR;
    }

  /* get custom-id from name */
  entry = Tcl_FindHashEntry(&mops_custom_id_ht, Revolve_name);

  if (entry != NULL)
    {
      Revolve_customid = Tcl_GetHashValue(entry);
    }
  else
    {
      mops_error(interp, MOPS_HASH, fname, NULL);
      return TCL_OK;
    }

  /* register converter */
  mops_status = mops_register_converter(interp,Revolve_name,Revolve_convert);

  /* register some C-functions as Tcl-Commands */
  /* they deal with the Revolve-Attributes Property GUI */
  Tcl_CreateCommand (interp, "Revolve_getRevolveAttr",
		     Revolve_get_attrprop_tcmd,
		     (ClientData) NULL, (Tcl_CmdDeleteProc *) NULL);
  Tcl_CreateCommand (interp, "Revolve_setRevolveAttr",
		     Revolve_set_attrprop_tcmd,
		     (ClientData) NULL, (Tcl_CmdDeleteProc *) NULL);

  /* another command, convert to a NURBPatch */
  Tcl_CreateCommand (interp, "revolveCNP", Revolve_convertNP_tcmd,
		     (ClientData) NULL, (Tcl_CmdDeleteProc *) NULL);

  /* register provider */
  mops_status = mops_register_provider(interp, Revolve_name,
				       MOPS_OTNURBPATCH, Revolve_provide_cb);


#ifdef MOPSIDR
  mops_status = idr_register_countbmmcb(Revolve_customid,
					Revolve_countbmms_cb);
#endif

 return TCL_OK;
}


int
Revolve_create_cb(Tcl_Interp *interp, int argc, char *argv[], void **c)
{
 int mops_status = MOPS_OK;
 int got_object = MOPS_FALSE;
 mops_list_object *sel = mops_current_selection;
 mops_object *curve = NULL, *newo = NULL;
 Revolve *new = NULL;
 char fname[] = "Revolve_create";

  if(!sel)
    {
      return MOPS_NOSEL;
    }

  if(sel->object->type != MOPS_OTNURBCURVE)
    {
      mops_status = mops_get_object(sel->object, MOPS_OTNURBCURVE, &curve);
      if(curve)
	got_object = MOPS_TRUE;
    }
  else
    {
      curve = sel->object;
    }


  if(!curve)
    {
      mops_error(interp, MOPS_ERROR, fname, "object is not a NURBCurve");
      return MOPS_ERROR;
    }

  if(!(new = calloc(1, sizeof(Revolve))))
    {
      return MOPS_OUTOFMEM;
    }

  if(!(newo = calloc(1, sizeof(mops_object))))
    {
      return MOPS_OUTOFMEM;
    }

  mops_default_attributes(newo);
  newo->type = MOPS_OTNURBPATCH;

  /* create no caps */
  new->has_upper_cap = 0;
  new->has_lower_cap = 0;
  new->ThetaMax = 360.0;
  new->glu_sampling_tolerance = 30.0;
  new->glu_display_mode = 0;

  new->revolution = newo;

  mops_status = mops_revolve_nurb(curve, new->ThetaMax,
				(mops_nurbpatch_object **)&(newo->object));

  if(mops_status)
    {
      free(new);
      return mops_status;
    }

  ((mops_nurbpatch_object *)newo->object)->glu_sampling_tolerance = 30.0;
  ((mops_nurbpatch_object *)newo->object)->glu_display_mode = 0;

  *c = new;

  if(got_object)
    mops_delete_object(curve);

 return mops_status;
}


int
Revolve_delete_cb(void *c)
{
 int mops_status = MOPS_OK;
 Revolve *del = NULL;

  if(!c)
    return MOPS_OK;

  del = (Revolve *)c;

  if(del->revolution)
    mops_delete_object(del->revolution);
  
  if(del->upper_cap)
    mops_delete_object(del->upper_cap);

  if(del->lower_cap)
    mops_delete_object(del->lower_cap);

  free(c);

 return mops_status;
}


int
Revolve_copy_cb(void *source, void **dest)
{
 int mops_status = MOPS_OK;
 Revolve *new = NULL, *srcrevolve = NULL;
 

  /* copy custom object */
  if(!(new = calloc(1,sizeof(Revolve))))
    return MOPS_OUTOFMEM;
  
  memcpy(new, source, sizeof(Revolve));
  /* danger, links point to original hierarchy! */

  srcrevolve = (Revolve *)source;

  if(srcrevolve->revolution)
    mops_status = mops_copy_object(srcrevolve->revolution,
				   &(new->revolution));

  if(mops_status)
    { free(new); return mops_status; }

  if(srcrevolve->upper_cap)
    mops_status = mops_copy_object(srcrevolve->upper_cap,
				   &(new->upper_cap));
  if(mops_status)
    { free(new); return mops_status; }

  if(srcrevolve->lower_cap)
    mops_status = mops_copy_object(srcrevolve->lower_cap,
				   &(new->lower_cap));
  if(mops_status)
    { free(new); return mops_status; }

  *dest = new;

 return mops_status;
}


int
Revolve_draww_cb(struct Togl *togl, mops_object *o)
{
 int mops_status = MOPS_OK;
 int instance = MOPS_FALSE;
 GLfloat color[4] = {0.8, 0.75, 0.65, 1.0};
 Revolve *rev = NULL;
 mops_custom_object *custom = NULL;
 int reset_color = 0;

  if(!o)
    return MOPS_OK;

  if(o->type == MOPS_OTINSTANCE)
    {
      o = (mops_object *)o->object;

      instance = MOPS_TRUE;
    }

  if(!o->editor)
    return MOPS_OK;

  custom = o->object;
  if(!custom)
    return MOPS_OK;    

  rev = custom->object;
  if(!rev)
    return MOPS_OK;

  glPushMatrix();

  if(!instance)
    {
      mops_render_attributes(o);

      if(o->selected)
	{
	  glColor3f MOPS_CLSEL;
	  reset_color = 1;
	}
    }
  else
    {
      if(o->colr != -1 && o->colg != -1 && o->colb != -1)
	{
	  color[0] = o->colr/255.0;
	  color[1] = o->colg/255.0;
	  color[2] = o->colb/255.0;
	  color[3] = o->cola/255.0; 
	  glMaterialfv (GL_FRONT_AND_BACK, GL_AMBIENT_AND_DIFFUSE, color);
	}
    }

  /* draw revolution */
  if(rev->revolution)
    mops_status = mops_draw_nurbpatch_c(togl, rev->revolution);

  /* draw caps */
  if(rev->upper_cap)
    mops_status = mops_draw_nurbpatch_c(togl, rev->upper_cap);

  if(rev->lower_cap)
    mops_status = mops_draw_nurbpatch_c(togl, rev->lower_cap);

  if(reset_color)
    glColor3f MOPS_CLUSEL;

  glPopMatrix();

 return mops_status;
}


int
Revolve_draws_cb(struct Togl *togl, mops_object *o)
{
 int mops_status = MOPS_OK;
 int instance = MOPS_FALSE;
 GLfloat color[4] = {0.8, 0.75, 0.65, 1.0};
 Revolve *rev = NULL;
 mops_custom_object *custom = NULL;

  if(!o)
    return MOPS_OK;

  if(o->type == MOPS_OTINSTANCE)
    {
      o = (mops_object *)o->object;

      instance = MOPS_TRUE;
    }

  if(!o->editor)
    return MOPS_OK;

  custom = o->object;
  if(!custom)
    return MOPS_OK;    

  rev = custom->object;
  if(!rev)
    return MOPS_OK;


  glPushMatrix();

  if(!instance)
    mops_render_attributes(o);
  else
    {
      if(o->colr != -1 && o->colg != -1 && o->colb != -1)
	{
	  color[0] = o->colr/255.0;
	  color[1] = o->colg/255.0;
	  color[2] = o->colb/255.0;
	  color[3] = o->cola/255.0; 
	  glMaterialfv (GL_FRONT_AND_BACK, GL_AMBIENT_AND_DIFFUSE, color);
	}
    }

  /* draw revolution */
  if(rev->revolution)
    mops_status = mops_draw_nurbpatch(togl, rev->revolution);

  /* draw caps */
  if(rev->upper_cap)
    mops_status = mops_draw_nurbpatch(togl, rev->upper_cap);

  if(rev->lower_cap)
    mops_status = mops_draw_nurbpatch(togl, rev->lower_cap);

  glPopMatrix();

 return mops_status;
}


int
Revolve_wrib_cb(mops_object *o)
{
 int mops_status = MOPS_OK;
 char dsoname[] = "revolve_rmdso.so";
 int a, i, j;
 Revolve *rev = NULL;
 mops_custom_object *custom = NULL;
 mops_nurbcurve_object *curve = NULL;
 mops_object *down = NULL;
 double bb[24] = {0};
 RtString *data[2];
 RtBound bound;
 Tcl_DString ds;
 Tcl_Obj *tod = NULL, *toi = NULL;

  if(!o)
    return MOPS_OK;

  custom = o->object;
  if(!custom)
    return MOPS_OK;    

  rev = custom->object;
  if(!rev)
    return MOPS_OK;

  RiAttributeBegin();
  RiTransformBegin();
  mops_wrib_attributes(o);
  mops_wrib_shaders(o);

  /*
  down = o->down;
  curve = (mops_nurbcurve_object *)down->object;
  mops_status = mops_get_sub_bb(o, bb);

  bound[0] = DBL_MAX; bound[2] = DBL_MAX; bound[4] = DBL_MAX;
  bound[1] = -DBL_MAX; bound[3] = -DBL_MAX; bound[5] = -DBL_MAX;

  a = 0;
  for(i = 0; i < 8; i++)
    {
      if(bb[a] < bound[0])
	bound[0] = (RtFloat)bb[a];
      if(bb[a] > bound[1])
	 bound[1] = (RtFloat)bb[a];
      a += 3;
    }

  a = 1;
  for(i = 0; i < 8; i++)
    {
      if(bb[a] < bound[2])
	bound[2] = (RtFloat)bb[a];
      if(bb[a] > bound[3])
	bound[3] = (RtFloat)bb[a];
      a += 3;
    }

  a = 2;
  for(i = 0; i < 8; i++)
    {
      if(bb[a] < bound[4])
	bound[4] = (RtFloat)bb[a];
      if(bb[a] > bound[5])
	bound[5] = (RtFloat)bb[a];
      a += 3;
    }

  data[0] = (RtString *)dsoname;
  Tcl_DStringInit(&ds);
  tod = Tcl_NewDoubleObj(rev->ThetaMax);
  Tcl_DStringAppend(&ds, Tcl_GetStringFromObj(tod, NULL), -1);
  Tcl_DStringAppend(&ds, ",", -1);
  toi = Tcl_NewIntObj(curve->length);
  Tcl_DStringAppend(&ds, Tcl_GetStringFromObj(toi, NULL), -1);
  Tcl_DStringAppend(&ds, ",", -1);
  Tcl_SetIntObj(toi, curve->order);
  Tcl_DStringAppend(&ds, Tcl_GetStringFromObj(toi, NULL), -1);
  Tcl_DStringAppend(&ds, ",", -1);
  if(curve->has_weights)
    Tcl_DStringAppend(&ds, "3,1,", -1);
  else
    Tcl_DStringAppend(&ds, "2,0,", -1);

  j = 0;
  for(i = 0; i < curve->length; i++)
    {
      Tcl_SetDoubleObj(tod, curve->controlv[j]);      
      Tcl_DStringAppend(&ds, Tcl_GetStringFromObj(tod, NULL), -1);
      Tcl_DStringAppend(&ds, " ", -1);
      Tcl_SetDoubleObj(tod, curve->controlv[j+1]);      
      Tcl_DStringAppend(&ds, Tcl_GetStringFromObj(tod, NULL), -1);

      if(curve->has_weights)
	{
	  Tcl_DStringAppend(&ds, " ", -1);
	  Tcl_SetDoubleObj(tod, curve->controlv[j+3]);      
	  Tcl_DStringAppend(&ds, Tcl_GetStringFromObj(tod, NULL), -1);
	  j += 4;
	}
      else
	j += 3;

      Tcl_DStringAppend(&ds, ",", -1);
    }

  for(i = 0; i < (curve->length+curve->order); i++)
    {
      Tcl_SetDoubleObj(tod, curve->knotv[i]);      
      Tcl_DStringAppend(&ds, Tcl_GetStringFromObj(tod, NULL), -1);
      Tcl_DStringAppend(&ds, ",", -1);
    }

  data[1] = (RtString *)Tcl_DStringValue(&ds);

  RiProcedural((RtPointer)data, bound, &RiProcDynamicLoad, NULL);

  Tcl_DStringFree(&ds);
  Tcl_IncrRefCount(tod);Tcl_DecrRefCount(tod);
  Tcl_IncrRefCount(toi);Tcl_DecrRefCount(toi);
  */  
  
  mops_wrib_npatch(rev->revolution);
  
  if(rev->upper_cap)
    {
      mops_wrib_trimcurves(rev->upper_cap);
      mops_wrib_npatch(rev->upper_cap);
    }
  if(rev->lower_cap)
    {
      mops_wrib_trimcurves(rev->lower_cap);
      mops_wrib_npatch(rev->lower_cap);
    }
  
  RiTransformEnd();
  RiAttributeEnd(); 

 return mops_status;
}


int
Revolve_write_cb(FILE *fileptr, mops_custom_object *custom)
{
 int mops_status = MOPS_OK;
 Revolve *rev = NULL;

  if(!custom)
    return MOPS_OK;    
  rev = (Revolve *)(custom->object);
  if(!rev)
    return MOPS_OK;

  fprintf(fileptr,"%d\n",rev->has_upper_cap);
  fprintf(fileptr,"%d\n",rev->has_lower_cap);
  fprintf(fileptr,"%g\n",rev->ThetaMax);
  fprintf(fileptr,"%g\n",rev->glu_sampling_tolerance);
  fprintf(fileptr,"%d\n",rev->glu_display_mode);

 return mops_status;
}

int
Revolve_read_cb(FILE *fileptr, mops_custom_object *custom)
{
 int mops_status = MOPS_OK;
 Revolve *rev = NULL;

  if(!custom)
    return MOPS_OK;
    
  if(!(custom->object = calloc(1,sizeof(Revolve))))
    return MOPS_OUTOFMEM;

  rev = (Revolve *)(custom->object);

  fscanf(fileptr,"%d\n",&rev->has_upper_cap);
  fscanf(fileptr,"%d\n",&rev->has_lower_cap);
  rev->ThetaMax = 360.0;
  rev->glu_sampling_tolerance = 30.0;

  if(mops_read_version > 4)
    {
      fscanf(fileptr,"%lg\n",&rev->ThetaMax);
    }

  if(mops_read_version > 5)
    {
      fscanf(fileptr,"%lg\n",&rev->glu_sampling_tolerance);
      fscanf(fileptr,"%d\n",&rev->glu_display_mode);
    }

 return mops_status;
}

int
Revolve_notify_cb(mops_object *o)
{
 int mops_status = MOPS_OK;
 int got_object = MOPS_FALSE;
 Revolve *rev = NULL;
 mops_object *c = NULL, *cap = NULL, *trim = NULL, *newo = NULL;
 mops_nurbcurve_object *curve = NULL;
 mops_custom_object *custom = NULL;
 double tolerance = 30.0;
 int mode = 0, i;
 double P1[4] = {0}, P2[4] = {0}, *controlv = NULL;
 GLdouble m[16] = {0};
 GLfloat mr[4][4];

  if(!o)
    return MOPS_OK;

  custom = o->object;
  if(!custom)
    return MOPS_OK;    

  rev = custom->object;
  if(!rev)
    return MOPS_OK;

  /* ok, something changed below, we have to rebuild the revolution */

  if(rev->revolution)
    mops_delete_object(rev->revolution);
  rev->revolution = NULL;

  if(rev->upper_cap)
    mops_delete_object(rev->upper_cap);
  rev->upper_cap = NULL;

  if(rev->lower_cap)
    mops_delete_object(rev->lower_cap);
  rev->lower_cap = NULL;

  tolerance = rev->glu_sampling_tolerance;
  mode = rev->glu_display_mode;

  c = NULL;

  if(o->down)
    {
      if(o->down->type == MOPS_OTNURBCURVE)
	{
	  c = o->down;
	}
      else
	{
	  mops_status = mops_get_object(o->down, MOPS_OTNURBCURVE, &c);
	  if(c)
	    got_object = MOPS_TRUE;
	}
    }

  if(c)
    {
      /* get curves transformation matrix */
      glMatrixMode (GL_MODELVIEW);
      glPushMatrix();
      glLoadIdentity ();
      glTranslated(c->movx, c->movy, c->movz);

      mops_build_rotmatrix(mr, c->quat);
      glMultMatrixf(&mr[0][0]);

      glScaled (c->scalx, c->scaly, c->scalz);
      glGetDoublev(GL_MODELVIEW_MATRIX, m);
      glPopMatrix();

      if(!(newo = calloc(1, sizeof(mops_object))))
	{
	  return MOPS_OUTOFMEM;
	}
      mops_default_attributes(newo);
      newo->type = MOPS_OTNURBPATCH;
      rev->revolution = newo;

      mops_revolve_nurb(c, rev->ThetaMax,
			(mops_nurbpatch_object **)&(newo->object));

      ((mops_nurbpatch_object *)newo->object)->glu_sampling_tolerance =
	tolerance;
      ((mops_nurbpatch_object *)newo->object)->glu_display_mode = mode;


      if(rev->has_upper_cap)
	{
	  curve = (mops_nurbcurve_object *)c->object;
	  if(curve->has_weights)
	    mops_nb_CurvePoint4D(curve->length-1, curve->order-1,
				 curve->knotv, curve->controlv,
				 curve->knotv[curve->length], P1);
	  else
	    mops_nb_CurvePoint3D(curve->length-1, curve->order-1,
				 curve->knotv, curve->controlv,
				 curve->knotv[curve->length], P1);

	  /* apply transform */
	  MOPS_APTRAN3(P2,P1,m)	    

	    if(!(cap = calloc(1, sizeof(mops_object))))
	      return MOPS_OUTOFMEM;
	  cap->type = MOPS_OTNURBPATCH;

	  if(!(controlv = calloc(4*3, sizeof(double))))
	    {free(cap); return MOPS_OUTOFMEM;}
	  controlv[0] = P2[0];
	  controlv[2] = P2[0];

	  controlv[3] = -P2[0];
	  controlv[5] = P2[0];

	  controlv[6] = P2[0];
	  controlv[8] = -P2[0];

	  controlv[9] = -P2[0];
	  controlv[11] = -P2[0];

	  for(i=1;i<=10;i+=3)
	    controlv[i] = P2[1];

	  mops_status = mops_create_nurbpatch(2, 2, 2, 2,
				MOPS_KTBEZIER, MOPS_KTBEZIER, MOPS_FALSE,
				controlv, NULL, NULL,
				(mops_nurbpatch_object **)(&(cap->object)));

	  ((mops_nurbpatch_object *)cap->object)->glu_sampling_tolerance =
	    tolerance;
	  ((mops_nurbpatch_object *)cap->object)->glu_display_mode = mode;

	  rev->upper_cap = cap;

	  /* create trimcurve */
	  if(!(trim = calloc(1, sizeof(mops_object))))
	    return MOPS_OUTOFMEM;
	  mops_default_attributes(trim);
	  trim->type = MOPS_OTNURBCURVE;
	  mops_create_nurbcircle((mops_nurbcurve_object **)
				 (&(trim->object)));
	  trim->scalx = 0.5;
	  trim->scaly = 0.5;
	  trim->movx = 0.5;
	  trim->movy = 0.5;

	  mops_revert_ncurve(trim->object);

	  cap->down = trim;
	   
	}

      if(rev->has_lower_cap)
	{
	  curve = (mops_nurbcurve_object *)c->object;
	  if(curve->has_weights)
	    mops_nb_CurvePoint4D(curve->length-1, curve->order-1,
				 curve->knotv, curve->controlv,
				 curve->knotv[curve->order-1], P1);
	  else
	    mops_nb_CurvePoint3D(curve->length-1, curve->order-1,
				 curve->knotv, curve->controlv,
				 curve->knotv[curve->order-1], P1);

	  /* apply transform */
	  MOPS_APTRAN3(P2,P1,m)	    

	    if(!(cap = calloc(1, sizeof(mops_object))))
	      return MOPS_OUTOFMEM;
	  cap->type = MOPS_OTNURBPATCH;

	  if(!(controlv = calloc(4*3, sizeof(double))))
	    {free(cap); return MOPS_OUTOFMEM;}
	  controlv[0] = P2[0];
	  controlv[2] = P2[0];

	  controlv[3] = -P2[0];
	  controlv[5] = P2[0];

	  controlv[6] = P2[0];
	  controlv[8] = -P2[0];

	  controlv[9] = -P2[0];
	  controlv[11] = -P2[0];

	  for(i=1;i<=10;i+=3)
	    controlv[i] = P2[1];

	  mops_status = mops_create_nurbpatch(2, 2, 2, 2,
				MOPS_KTBEZIER, MOPS_KTBEZIER, MOPS_FALSE,
				controlv, NULL, NULL,
				(mops_nurbpatch_object **)(&(cap->object)));

	  ((mops_nurbpatch_object *)cap->object)->glu_sampling_tolerance =
	    tolerance;
	  ((mops_nurbpatch_object *)cap->object)->glu_display_mode = mode;
	  rev->lower_cap = cap;

	  /* create trimcurve */
	  if(!(trim = calloc(1, sizeof(mops_object))))
	    return MOPS_OUTOFMEM;
	  mops_default_attributes(trim);
	  trim->type = MOPS_OTNURBCURVE;
	  mops_create_nurbcircle((mops_nurbcurve_object **)
				 (&(trim->object)));
	  trim->scalx = 0.5;
	  trim->scaly = 0.5;
	  trim->movx = 0.5;
	  trim->movy = 0.5;

	  mops_revert_ncurve(trim->object);

	  cap->down = trim;
	   
	}
    }

  if(got_object)
    mops_delete_object(c);

 return mops_status;
}


/************************/
/************************/
/************************/

int
Revolve_getprop_cb(Tcl_Interp *interp)
{
 int mops_status = MOPS_OK;

  /* Surface-Shader */
  mops_status = mops_add_surfshader_property(interp);

  /* Displacement-Shader */
  mops_status = mops_add_dispshader_property(interp);

  /* Interior-Shader */
  mops_status = mops_add_intshader_property(interp);

  /* Exterior-Shader */
  mops_status = mops_add_extshader_property(interp);

  /* Revolve-Parameters */
  Tcl_SetVar(interp,"propertyList","RevolveAttr", TCL_APPEND_VALUE |\
	     TCL_LIST_ELEMENT | TCL_LEAVE_ERR_MSG | TCL_GLOBAL_ONLY);

  Tcl_SetVar(interp,"propertyProcList","Revolve_buildRevolveAttrGUI",
	     TCL_APPEND_VALUE | TCL_LIST_ELEMENT |\
	     TCL_LEAVE_ERR_MSG | TCL_GLOBAL_ONLY);

  Tcl_SetVar(interp,"propertyGetProcList","Revolve_getRevolveAttr",
	     TCL_APPEND_VALUE | TCL_LIST_ELEMENT |\
	     TCL_LEAVE_ERR_MSG | TCL_GLOBAL_ONLY);

  Tcl_SetVar(interp,"propertySetProcList","Revolve_setRevolveAttr",
	     TCL_APPEND_VALUE | TCL_LIST_ELEMENT |\
	     TCL_LEAVE_ERR_MSG | TCL_GLOBAL_ONLY);

  Tcl_SetVar(interp,"propertyDataList","RevolveAttrData",
	     TCL_APPEND_VALUE | TCL_LIST_ELEMENT |\
	     TCL_LEAVE_ERR_MSG | TCL_GLOBAL_ONLY);

 return TCL_OK;
}


/* Tcl -> C! */
int
Revolve_set_attrprop_tcmd(ClientData clientData, Tcl_Interp *interp,
			 int argc, char *argv[])
{
 mops_object *o = NULL;
 mops_list_object *sel = mops_current_selection;
 mops_custom_object *custom = NULL;
 Revolve *rev = NULL;
 char *n1="RevolveAttrData";
 Tcl_Obj *to = NULL, *toa = NULL, *ton = NULL;

  if(!sel) /* oops? */
    return TCL_OK;

  o = sel->object; 
  if(!o)
    return TCL_OK;

  custom = o->object;
  if(!custom)
    return TCL_OK;

  rev = (Revolve *) custom->object;

  toa = Tcl_NewStringObj(n1,-1);
  ton = Tcl_NewStringObj("HasUpperCap",-1);
  to = Tcl_ObjGetVar2(interp,toa,ton,TCL_LEAVE_ERR_MSG | TCL_GLOBAL_ONLY);
  Tcl_GetIntFromObj(interp,to, &rev->has_upper_cap);

  Tcl_SetStringObj(ton,"HasLowerCap",-1);
  to = Tcl_ObjGetVar2(interp,toa,ton,TCL_LEAVE_ERR_MSG | TCL_GLOBAL_ONLY);
  Tcl_GetIntFromObj(interp,to, &rev->has_lower_cap);

  Tcl_SetStringObj(ton,"ThetaMax",-1);
  to = Tcl_ObjGetVar2(interp,toa,ton,TCL_LEAVE_ERR_MSG | TCL_GLOBAL_ONLY);
  Tcl_GetDoubleFromObj(interp,to, &(rev->ThetaMax));

  Tcl_SetStringObj(ton,"Tolerance",-1);
  to = Tcl_ObjGetVar2(interp,toa,ton,TCL_LEAVE_ERR_MSG | TCL_GLOBAL_ONLY);
  Tcl_GetDoubleFromObj(interp,to, &(rev->glu_sampling_tolerance));

  Tcl_SetStringObj(ton,"Display",-1);
  to = Tcl_ObjGetVar2(interp,toa,ton,TCL_LEAVE_ERR_MSG | TCL_GLOBAL_ONLY);
  Tcl_GetIntFromObj(interp,to, &(rev->glu_display_mode));

  Tcl_IncrRefCount(toa);Tcl_DecrRefCount(toa);
  Tcl_IncrRefCount(ton);Tcl_DecrRefCount(ton);

  Revolve_notify_cb(o);

 return TCL_OK;
}

/* C -> Tcl! */
int
Revolve_get_attrprop_tcmd(ClientData clientData, Tcl_Interp *interp,
			 int argc, char *argv[])
{
 mops_object *o = NULL;
 mops_list_object *sel = mops_current_selection;
 mops_custom_object *custom = NULL;
 Revolve *rev = NULL;
 char *n1="RevolveAttrData";
 Tcl_Obj *to = NULL, *toa = NULL, *ton = NULL;

  if(!sel) /* oops? */
    return TCL_OK;

  o = sel->object; 
  if(!o)
    return TCL_OK;

  custom = o->object;
  if(!custom)
    return TCL_OK;

  rev = (Revolve *) custom->object;

  toa = Tcl_NewStringObj(n1,-1);
  ton = Tcl_NewStringObj("HasUpperCap",-1);
  to = Tcl_NewIntObj(rev->has_upper_cap);
  Tcl_ObjSetVar2(interp,toa,ton,to,TCL_LEAVE_ERR_MSG |
		 TCL_GLOBAL_ONLY);

  Tcl_SetStringObj(ton,"HasLowerCap",-1);
  to = Tcl_NewIntObj(rev->has_lower_cap);
  Tcl_ObjSetVar2(interp,toa,ton,to,TCL_LEAVE_ERR_MSG |
		 TCL_GLOBAL_ONLY);

  Tcl_SetStringObj(ton,"ThetaMax",-1);
  to = Tcl_NewDoubleObj(rev->ThetaMax);
  Tcl_ObjSetVar2(interp,toa,ton,to,TCL_LEAVE_ERR_MSG | TCL_GLOBAL_ONLY);

  Tcl_SetStringObj(ton,"Tolerance",-1);
  to = Tcl_NewDoubleObj(rev->glu_sampling_tolerance);
  Tcl_ObjSetVar2(interp,toa,ton,to,TCL_LEAVE_ERR_MSG | TCL_GLOBAL_ONLY);

  Tcl_SetStringObj(ton,"Display",-1);
  to = Tcl_NewIntObj(rev->glu_display_mode);
  Tcl_ObjSetVar2(interp,toa,ton,to,TCL_LEAVE_ERR_MSG | TCL_GLOBAL_ONLY);

  Tcl_IncrRefCount(toa);Tcl_DecrRefCount(toa);
  Tcl_IncrRefCount(ton);Tcl_DecrRefCount(ton);

 return TCL_OK;
}


int
Revolve_convert(Tcl_Interp *interp, mops_object *o)
{
 int mops_status = MOPS_OK;
 mops_custom_object *custom = NULL;
 Revolve *rev = NULL;
 mops_nurbpatch_object *patch = NULL;
 mops_object *new = NULL;
 char fname[] = "Revolve_convert";


  custom = (mops_custom_object*)o->object;
  if(!custom)
    return MOPS_ERROR;

  if(custom->type != Revolve_customid)
    return MOPS_ERROR;

  rev = (Revolve *) custom->object;

  patch = (mops_nurbpatch_object *)rev->revolution->object;

  if(!patch)
   return MOPS_ERROR;

  /* convert revolution */
  if(!(new = calloc(1,sizeof(mops_object))))
    return MOPS_OUTOFMEM;

  mops_status = mops_copy_attributes(o, new);

  new->type = MOPS_OTNURBPATCH;

  mops_status = mops_copy_nurbpatch(patch,
				  (mops_nurbpatch_object**)&(new->object));

  if(mops_status)
    {
      mops_error(interp,mops_status,fname,NULL);
      free(new);
      return MOPS_ERROR;
    }

  mops_status = mops_link_new_object(new);

  /* convert upper cap */
  if(rev->upper_cap)
    {
      mops_status = mops_copy_object(rev->upper_cap, &new);

      if(mops_status)
	{
	  mops_error(interp,mops_status,fname,NULL);
	  return mops_status;
	}

      mops_status = mops_copy_attributes(o, new);
      mops_status = mops_link_new_object(new);
    }
  /* convert lower cap */
  if(rev->lower_cap)
    {
      mops_status = mops_copy_object(rev->lower_cap, &new);

      if(mops_status)
	{
	  mops_error(interp,mops_status,fname,NULL);
	  return mops_status;
	}

      mops_status = mops_copy_attributes(o, new);
      mops_status = mops_link_new_object(new);
    }

 return mops_status;
}


int
Revolve_convertNP_tcmd(ClientData clientData, Tcl_Interp *interp,
		     int argc, char *argv[])
{
 int mops_status = MOPS_OK;
 mops_object *o = NULL;
 mops_list_object *sel = mops_current_selection;
 mops_custom_object *custom = NULL;
 char fname[] = "Revolve_convertNP";

  if(!sel) /* oops? */
    return TCL_OK;

  o = sel->object; 
  if(!o)
    return TCL_OK;

  if(o->type != MOPS_OTCUSTOM)
    return TCL_OK;

  custom = (mops_custom_object*)o->object;
  if(!custom)
    return TCL_OK;

  if(custom->type != Revolve_customid)
    return TCL_OK;

  mops_status = Revolve_convert(interp, o);
  if(mops_status)
    {
      mops_error(interp,MOPS_ERROR,fname,"converter failed");
      return TCL_OK;
    }

 return TCL_OK;
}

int
Revolve_provide_cb(mops_object *o, void **target)
{
 int mops_status = MOPS_OK;
 Revolve *rev = NULL;
 mops_custom_object *custom = NULL;

  custom = (mops_custom_object *)o->object;

  rev = (Revolve *)custom->object;

  mops_status = Revolve_notify_cb(o);

  if(mops_status)
    return mops_status;

  mops_copy_attributes(o, rev->revolution);

  *target = rev->revolution;

 return mops_status;
}


#ifdef MOPSIDR
void
Revolve_countbmms_cb(mops_object *o)
{
 mops_custom_object *custom = NULL;
 Revolve *rev = NULL;

 if(!o)
   return;

 custom = (mops_custom_object *)o->object;

 if(!custom)
   return;

 rev = (Revolve *)custom->object;

 if(!rev)
   return;

 if((rev->has_upper_cap) || (rev->has_lower_cap))
   idr_bmm.trim++;

 return;
}
#endif
