/*
 * The Mops, a free 3D Tcl/Tk/OpenGL/RenderMan modeler
 *
 * The Mops is copyrighted (C) 1998-1999 by Randolf Schultz
 * (rschultz@informatik.uni-rostock.de)
 *
 * See the file License for details.
 *
 */

#include "mops.h"

/* Skin */

/* The Name */
static char *Skin_name = "Skin";

static char *Skin_customid = NULL;

/* prototypes */
typedef struct Skin_s
{
 mops_object *skin;
 int uorder;
 int uknot_type;
 int interpolate;
 double glu_sampling_tolerance;
 int glu_display_mode;
} Skin;

int Skin_create_cb(Tcl_Interp *interp, int argc, char *argv[], void **c);
int Skin_delete_cb(void *c);
int Skin_copy_cb(void *source, void **dest);
int Skin_draww_cb(struct Togl *togl, mops_object *o);
int Skin_draws_cb(struct Togl *togl, mops_object *o);
int Skin_wrib_cb(mops_object *o);
int Skin_read_cb(FILE *fileptr, mops_custom_object *custom);
int Skin_write_cb(FILE *fileptr, mops_custom_object *custom);
int Skin_getprop_cb(Tcl_Interp *interp);
int Skin_set_attrprop_tcmd(ClientData clientData, Tcl_Interp *interp,
			     int argc, char *argv[]);
int Skin_get_attrprop_tcmd(ClientData clientData, Tcl_Interp *interp,
			     int argc, char *argv[]);

int Skin_notify_cb(mops_object *o);
int Skin_convert(Tcl_Interp *interp, mops_object *o);
int Skin_convertNP_tcmd(ClientData clientData, Tcl_Interp *interp,
			 int argc, char *argv[]);

int Skin_provide_cb(mops_object *o, void **target);

/* functions */
int
Skin_Init(Tcl_Interp *interp)
{
 int mops_status = MOPS_OK;
 Tcl_HashEntry *entry;
 char fname[] = "Skin_Init";

  mops_status = mops_register_custom_objecttype(
				     Skin_name,
				     Skin_create_cb,
				     Skin_draws_cb, 
				     NULL, /* drawh */ /* no handles! */
				     Skin_draww_cb,
				     Skin_delete_cb,
				     Skin_copy_cb,
				     Skin_getprop_cb,
				     NULL, /* getpoint */ /* no single point
					                     editing! */
				     Skin_wrib_cb,
				     Skin_read_cb,
				     Skin_write_cb,
				     Skin_notify_cb
				     );

  if(mops_status)
    {
      mops_error(interp, mops_status, fname, NULL);
      return TCL_ERROR;
    }

  /* get custom-id from name */
  entry = Tcl_FindHashEntry(&mops_custom_id_ht, Skin_name);

  if (entry != NULL)
    {
      Skin_customid = Tcl_GetHashValue(entry);
    }
  else
    {
      mops_error(interp, MOPS_HASH, fname, NULL);
      return TCL_OK;
    }

  /* register converter */
  mops_status = mops_register_converter(interp,Skin_name,Skin_convert);

  /* register some C-functions as Tcl-Commands */
  /* they deal with the Skin-Attributes Property GUI */
  Tcl_CreateCommand (interp, "Skin_getSkinAttr",
		     Skin_get_attrprop_tcmd,
		     (ClientData) NULL, (Tcl_CmdDeleteProc *) NULL);
  Tcl_CreateCommand (interp, "Skin_setSkinAttr",
		     Skin_set_attrprop_tcmd,
		     (ClientData) NULL, (Tcl_CmdDeleteProc *) NULL);

  /* another command, convert to a NURBPatch */
  Tcl_CreateCommand (interp, "skinCNP", Skin_convertNP_tcmd,
		     (ClientData) NULL, (Tcl_CmdDeleteProc *) NULL);

  /* register provider */
  mops_status = mops_register_provider(interp, Skin_name,
				       MOPS_OTNURBPATCH, Skin_provide_cb);

 return TCL_OK;
}


int
Skin_create_cb(Tcl_Interp *interp, int argc, char *argv[], void **c)
{
 int mops_status = MOPS_OK;
 mops_list_object *sel = mops_current_selection;
 mops_custom_object *custom = NULL;
 int count = 0;
 Skin *new = NULL;
 char fname[] = "Skin_create";
    
  if(!sel)
    {
      return MOPS_NOSEL;
    }


  while(sel)
    {
      if(sel->object->type != MOPS_OTNURBCURVE)
	{
	  if(sel->object->type == MOPS_OTINSTANCE)
	    {
	      if(((mops_object *)sel->object->object)->type !=
		 MOPS_OTNURBCURVE)
		mops_error(interp, MOPS_ERROR, fname,
			   "object is not a NURBCurve");
	      else
		count++;
	    }
	  else
	    {
	      if(sel->object->type == MOPS_OTCUSTOM)
		{
		  custom = (mops_custom_object *)(sel->object->object);
		  if(custom->provides != MOPS_OTNURBCURVE)
		    mops_error(interp, MOPS_ERROR, fname,
			       "object is not a NURBCurve");
		  else
		    count++;
		}
	      else
		{
		  mops_error(interp, MOPS_ERROR, fname,
			     "object is not a NURBCurve");
		}
	    }

	}
      else
	count++;

      sel = sel->next;
    }


  if(count < 2)
    {
      mops_error(interp,MOPS_ERROR,fname,"select atleast two NURB curves!");
      return MOPS_ERROR;
    }

  if(!(new = calloc(1, sizeof(Skin))))
    {
      return MOPS_OUTOFMEM;
    }

  new->uorder = 4;
  new->uknot_type = MOPS_KTNURB;
  new->glu_sampling_tolerance = 30.0;

  /*
  if(!(newo = calloc(1, sizeof(mops_object))))
    {
      free(new); return MOPS_OUTOFMEM;
    }
  new->skin = newo;
  */

  if(argc >= 4)
    {
      Tcl_GetInt(interp, argv[3], &(new->uorder));
    }

  *c = new;

 return mops_status;
}


int
Skin_delete_cb(void *c)
{
 int mops_status = MOPS_OK;
 Skin *del = NULL;

  if(!c)
    return MOPS_OK;

  del = (Skin *)c;

  if(del->skin)
    mops_delete_object(del->skin);
  
  free(c);

 return mops_status;
}


int
Skin_copy_cb(void *source, void **dest)
{
 int mops_status = MOPS_OK;
 Skin *new = NULL, *srcskin = NULL;
 
  /* copy custom object */
  if(!(new = calloc(1,sizeof(Skin))))
    return MOPS_OUTOFMEM;
  
  memcpy(new, source, sizeof(Skin));
  /* danger, links point to original hierarchy! */

  srcskin = (Skin *)source;
  mops_status = mops_copy_object(srcskin->skin,
				    &(new->skin));

  if(mops_status)
    { free(new); return mops_status; }

  *dest = new;

 return mops_status;
}


int
Skin_draww_cb(struct Togl *togl, mops_object *o)
{
 int mops_status = MOPS_OK;
 int instance = MOPS_FALSE;
 GLfloat color[4] = {0.8, 0.75, 0.65, 1.0};
 Skin *sk = NULL;
 mops_custom_object *custom = NULL;
 int reset_color = 0;

  if(!o)
    return MOPS_OK;

  if(o->type == MOPS_OTINSTANCE)
    {
      o = (mops_object *)o->object;

      instance = MOPS_TRUE;
    }

  if(!o->editor)
    return MOPS_OK;

  custom = o->object;
  if(!custom)
    return MOPS_OK;    

  sk = custom->object;
  if(!sk)
    return MOPS_OK;

  glPushMatrix();

  if(!instance)
    {
      mops_render_attributes(o);

      if(o->selected)
	{
	  glColor3f MOPS_CLSEL;
	  reset_color = 1;
	}
    }
  else
    {
      if(o->colr != -1 && o->colg != -1 && o->colb != -1)
	{
	  color[0] = o->colr/255.0;
	  color[1] = o->colg/255.0;
	  color[2] = o->colb/255.0;
	  color[3] = o->cola/255.0; 
	  glMaterialfv (GL_FRONT_AND_BACK, GL_AMBIENT_AND_DIFFUSE, color);
	}
    }

 
  /* draw skin */
  if(sk->skin)
    {
      mops_status = mops_draw_nurbpatch_c(togl, sk->skin);
    }

  if(reset_color)
    glColor3f MOPS_CLUSEL;

  glPopMatrix();

 return mops_status;
}


int
Skin_draws_cb(struct Togl *togl, mops_object *o)
{
 int mops_status = MOPS_OK;
 int instance = MOPS_FALSE;
 GLfloat color[4] = {0.8, 0.75, 0.65, 1.0};
 Skin *sk = NULL;
 mops_custom_object *custom = NULL;

  if(!o)
    return MOPS_OK;

  if(o->type == MOPS_OTINSTANCE)
    {
      o = (mops_object *)o->object;

      instance = MOPS_TRUE;
    }

  if(!o->editor)
    return MOPS_OK;

  custom = o->object;
  if(!custom)
    return MOPS_OK;    

  sk = custom->object;
  if(!sk)
    return MOPS_OK;


  glPushMatrix();

  if(!instance)
    mops_render_attributes(o);
  else
    {
      if(o->colr != -1 && o->colg != -1 && o->colb != -1)
	{
	  color[0] = o->colr/255.0;
	  color[1] = o->colg/255.0;
	  color[2] = o->colb/255.0;
	  color[3] = o->cola/255.0; 
	  glMaterialfv (GL_FRONT_AND_BACK, GL_AMBIENT_AND_DIFFUSE, color);
	}
    }

   /* draw skin */
  if(sk->skin)
    {
      mops_status = mops_draw_nurbpatch(togl, sk->skin);
    }

  glPopMatrix();

 return mops_status;
}


int
Skin_wrib_cb(mops_object *o)
{
 int mops_status = MOPS_OK;
 Skin *sk = NULL;
 mops_custom_object *custom = NULL;

  if(!o)
    return MOPS_OK;

  custom = o->object;
  if(!custom)
    return MOPS_OK;    

  sk = custom->object;
  if(!sk)
    return MOPS_OK;

  RiAttributeBegin();
  RiTransformBegin();
   mops_wrib_attributes(o);
   mops_wrib_shaders(o);
       
   mops_wrib_npatch(sk->skin);

  RiTransformEnd();
  RiAttributeEnd(); 

 return mops_status;
}


int
Skin_write_cb(FILE *fileptr, mops_custom_object *custom)
{
 int mops_status = MOPS_OK;
 Skin *sk = NULL;

  if(!custom)
    return MOPS_OK;    
  sk = (Skin *)(custom->object);
  if(!sk)
    return MOPS_OK;
  fprintf(fileptr,"%d\n",sk->interpolate);
  fprintf(fileptr,"%d\n",sk->uorder);
  fprintf(fileptr,"%d\n",sk->uknot_type);
  fprintf(fileptr,"%g\n",sk->glu_sampling_tolerance);
  fprintf(fileptr,"%d\n",sk->glu_display_mode);

 return mops_status;
}

int
Skin_read_cb(FILE *fileptr, mops_custom_object *custom)
{
 int mops_status = MOPS_OK;
 Skin *sk = NULL;

  if(!custom)
    return MOPS_OK;
    
  if(!(custom->object = calloc(1,sizeof(Skin))))
    return MOPS_OUTOFMEM;

  sk = (Skin *)(custom->object);

  fscanf(fileptr,"%d\n",&sk->interpolate);
  fscanf(fileptr,"%d\n",&sk->uorder);
  fscanf(fileptr,"%d\n",&sk->uknot_type);
  fscanf(fileptr,"%lg\n",&sk->glu_sampling_tolerance);
  fscanf(fileptr,"%d\n",&sk->glu_display_mode);

 return mops_status;
}


int
Skin_notify_cb(mops_object *o)
{
 int mops_status = MOPS_OK;
 Skin *sk = NULL;
 mops_object *down = NULL, *c = NULL, *last = NULL, *all_curves = NULL;
 mops_object *newo = NULL;
 mops_custom_object *custom = NULL;
 double tolerance = 30.0;
 int mode = 0, count = 0;

  if(!o)
    return MOPS_OK;

  custom = o->object;
  if(!custom)
    return MOPS_OK;    

  sk = custom->object;
  if(!sk)
    return MOPS_OK;

  /* ok, something changed below, we have to rebuild the skin */
  tolerance = sk->glu_sampling_tolerance;
  mode = sk->glu_display_mode;


  if(sk->skin)
    {
      mops_delete_object(sk->skin);
      sk->skin = NULL;
    }

  down = o->down;

  if(!down)
    return MOPS_OK;

  while(down)
    {
      c = NULL;

      if(down->type == MOPS_OTNURBCURVE)
	{
	  mops_copy_object(down, &c);
	}
      else
	{
	  mops_status = mops_get_object(down, MOPS_OTNURBCURVE, &c);
	}

      if(c == NULL)
	{
	  /*	  mops_error(interp, MOPS_ERROR, fname,
		     "object is not a NURBCurve");*/
	}
      else
	{
	  if(last)
	    last->next = c;
	  else
	    all_curves = c;

	  last = c;

	  count++;
	}

      down = down->next;
    }

  if(count < 2)
    return MOPS_ERROR;

  if(!(newo = calloc(1, sizeof(mops_object))))
    {
      return MOPS_OUTOFMEM;
    }

  mops_default_attributes(newo);

  newo->type = MOPS_OTNURBPATCH;
      
  sk->skin = newo;

  mops_status = mops_skin_nurbcurves(all_curves, sk->uorder,
		 sk->uknot_type, sk->interpolate,
		 (mops_nurbpatch_object **)&newo->object);


  if(sk->skin)
    {
      ((mops_nurbpatch_object *)newo->object)->glu_sampling_tolerance =
	tolerance;
      ((mops_nurbpatch_object *)newo->object)->glu_display_mode = mode;

    }

  /* clean up the mess */
  mops_status = mops_delete_multi_objects(all_curves);

 return mops_status;
}


/************************/
/************************/
/************************/

int
Skin_getprop_cb(Tcl_Interp *interp)
{
 int mops_status = MOPS_OK;

  /* Surface-Shader */
  mops_status = mops_add_surfshader_property(interp);

  /* Displacement-Shader */
  mops_status = mops_add_dispshader_property(interp);

  /* Interior-Shader */
  mops_status = mops_add_intshader_property(interp);

  /* Exterior-Shader */
  mops_status = mops_add_extshader_property(interp);

  /* Skin-Parameters */
  Tcl_SetVar(interp,"propertyList","SkinAttr", TCL_APPEND_VALUE |\
	     TCL_LIST_ELEMENT | TCL_LEAVE_ERR_MSG | TCL_GLOBAL_ONLY);

  Tcl_SetVar(interp,"propertyProcList","Skin_buildSkinAttrGUI",
	     TCL_APPEND_VALUE | TCL_LIST_ELEMENT |\
	     TCL_LEAVE_ERR_MSG | TCL_GLOBAL_ONLY);

  Tcl_SetVar(interp,"propertyGetProcList","Skin_getSkinAttr",
	     TCL_APPEND_VALUE | TCL_LIST_ELEMENT |\
	     TCL_LEAVE_ERR_MSG | TCL_GLOBAL_ONLY);

  Tcl_SetVar(interp,"propertySetProcList","Skin_setSkinAttr",
	     TCL_APPEND_VALUE | TCL_LIST_ELEMENT |\
	     TCL_LEAVE_ERR_MSG | TCL_GLOBAL_ONLY);

  Tcl_SetVar(interp,"propertyDataList","SkinAttrData",
	     TCL_APPEND_VALUE | TCL_LIST_ELEMENT |\
	     TCL_LEAVE_ERR_MSG | TCL_GLOBAL_ONLY);

 return TCL_OK;
}


/* Tcl -> C! */
int
Skin_set_attrprop_tcmd(ClientData clientData, Tcl_Interp *interp,
			 int argc, char *argv[])
{
 mops_object *o = NULL;
 mops_list_object *sel = mops_current_selection;
 mops_custom_object *custom = NULL;
 Skin *sk = NULL;
 char *n1="SkinAttrData";
 Tcl_Obj *to = NULL, *toa = NULL, *ton = NULL;

  if(!sel) /* oops? */
    return TCL_OK;

  o = sel->object; 
  if(!o)
    return TCL_OK;

  custom = o->object;
  if(!custom)
    return TCL_OK;

  sk = (Skin *) custom->object;

  toa = Tcl_NewStringObj(n1,-1);
  ton = Tcl_NewStringObj("Interpolate",-1);
  to = Tcl_ObjGetVar2(interp,toa,ton,TCL_LEAVE_ERR_MSG | TCL_GLOBAL_ONLY);
  Tcl_GetIntFromObj(interp,to, &sk->interpolate);

  Tcl_SetStringObj(ton,"Order_U",-1);
  to = Tcl_ObjGetVar2(interp,toa,ton,TCL_LEAVE_ERR_MSG | TCL_GLOBAL_ONLY);
  Tcl_GetIntFromObj(interp,to, &sk->uorder);

  Tcl_SetStringObj(ton,"Knot-Type_U",-1);
  to = Tcl_ObjGetVar2(interp,toa,ton,TCL_LEAVE_ERR_MSG | TCL_GLOBAL_ONLY);
  Tcl_GetIntFromObj(interp,to, &sk->uknot_type);

  Tcl_SetStringObj(ton,"Tolerance",-1);
  to = Tcl_ObjGetVar2(interp,toa,ton,TCL_LEAVE_ERR_MSG | TCL_GLOBAL_ONLY);
  Tcl_GetDoubleFromObj(interp,to, &(sk->glu_sampling_tolerance));

  Tcl_SetStringObj(ton,"Display",-1);
  to = Tcl_ObjGetVar2(interp,toa,ton,TCL_LEAVE_ERR_MSG | TCL_GLOBAL_ONLY);
  Tcl_GetIntFromObj(interp,to, &(sk->glu_display_mode));

  Tcl_IncrRefCount(toa);Tcl_DecrRefCount(toa);
  Tcl_IncrRefCount(ton);Tcl_DecrRefCount(ton);

  Skin_notify_cb(o);

 return TCL_OK;
}

/* C -> Tcl! */
int
Skin_get_attrprop_tcmd(ClientData clientData, Tcl_Interp *interp,
			 int argc, char *argv[])
{
 mops_object *o = NULL;
 mops_list_object *sel = mops_current_selection;
 mops_custom_object *custom = NULL;
 Skin *sk = NULL;
 char *n1="SkinAttrData";
 Tcl_Obj *to = NULL, *toa = NULL, *ton = NULL;

  if(!sel) /* oops? */
    return TCL_OK;

  o = sel->object; 
  if(!o)
    return TCL_OK;

  custom = o->object;
  if(!custom)
    return TCL_OK;

  sk = (Skin *) custom->object;

  toa = Tcl_NewStringObj(n1,-1);
  ton = Tcl_NewStringObj("Interpolate",-1);
  to = Tcl_NewIntObj(sk->interpolate);
  Tcl_ObjSetVar2(interp,toa,ton,to,TCL_LEAVE_ERR_MSG |
		 TCL_GLOBAL_ONLY);

  Tcl_SetStringObj(ton,"Order_U",-1);
  to = Tcl_NewIntObj(sk->uorder);
  Tcl_ObjSetVar2(interp,toa,ton,to,TCL_LEAVE_ERR_MSG |
		 TCL_GLOBAL_ONLY);

  Tcl_SetStringObj(ton,"Knot-Type_U",-1);
  to = Tcl_NewIntObj(sk->uknot_type);
  Tcl_ObjSetVar2(interp,toa,ton,to,TCL_LEAVE_ERR_MSG |
		 TCL_GLOBAL_ONLY);

  Tcl_SetStringObj(ton,"Tolerance",-1);
  to = Tcl_NewDoubleObj(sk->glu_sampling_tolerance);
  Tcl_ObjSetVar2(interp,toa,ton,to,TCL_LEAVE_ERR_MSG | TCL_GLOBAL_ONLY);

  Tcl_SetStringObj(ton,"Display",-1);
  to = Tcl_NewIntObj(sk->glu_display_mode);
  Tcl_ObjSetVar2(interp,toa,ton,to,TCL_LEAVE_ERR_MSG | TCL_GLOBAL_ONLY);

  Tcl_IncrRefCount(toa);Tcl_DecrRefCount(toa);
  Tcl_IncrRefCount(ton);Tcl_DecrRefCount(ton);

 return TCL_OK;
}


int
Skin_convert(Tcl_Interp *interp, mops_object *o)
{
 int mops_status = MOPS_OK;
 mops_custom_object *custom = NULL;
 Skin *sk = NULL;
 mops_object *new = NULL;
 char fname[] = "skin_convertNP";

  if(o->type != MOPS_OTCUSTOM)
    return MOPS_ERROR;

  custom = (mops_custom_object*)o->object;
  if(!custom)
    return MOPS_ERROR;

  if(custom->type != Skin_customid)
    return MOPS_ERROR;

  sk = (Skin *) custom->object;

  if(!sk || !sk->skin)
    return MOPS_ERROR;

  mops_status = mops_copy_object(sk->skin, &new);

  if(mops_status)
    {
      mops_error(interp,mops_status,fname,NULL);
      free(new);
      return mops_status;
    }

  mops_status = mops_default_attributes(new);
  mops_status = mops_copy_attributes(o, new);
  mops_status = mops_link_new_object(new);

 return mops_status;
}


int
Skin_convertNP_tcmd(ClientData clientData, Tcl_Interp *interp,
		     int argc, char *argv[])
{
 int mops_status = MOPS_OK;
 mops_object *o = NULL;
 mops_list_object *sel = mops_current_selection;
 mops_custom_object *custom = NULL;
 char fname[] = "skin_convertNP";

  if(!sel) /* oops? */
    return TCL_OK;

  o = sel->object; 
  if(!o)
    return TCL_OK;

  if(o->type != MOPS_OTCUSTOM)
    return TCL_OK;

  custom = (mops_custom_object*)o->object;
  if(!custom)
    return TCL_OK;

  if(custom->type != Skin_customid)
    return TCL_OK;

  mops_status = Skin_convert(interp, o);
  if(mops_status)
    {
      mops_error(interp,MOPS_ERROR,fname,"converter failed");
      return TCL_OK;
    }


 return TCL_OK;
}

int
Skin_provide_cb(mops_object *o, void **target)
{
 int mops_status = MOPS_OK;
 Skin *sk = NULL;
 mops_custom_object *custom = NULL;

  custom = (mops_custom_object *)o->object;

  sk = (Skin *)custom->object;

  mops_status = Skin_notify_cb(o);

  if(mops_status)
    return mops_status;

  mops_copy_attributes(o, sk->skin);

  *target = sk->skin;

 return mops_status;
}
