/*
 * The Mops, a free 3D Tcl/Tk/OpenGL/RenderMan modeler
 *
 * The Mops is copyrighted (C) 1998-1999 by Randolf Schultz
 * (rschultz@informatik.uni-rostock.de)
 *
 * See the file License for details.
 *
 */

#include "mops.h"

/* Sweep */

/* The Name */
static char *Sweep_name = "Sweep";

static char *Sweep_customid = NULL;

/* prototypes */
typedef struct Sweep_s
{
 mops_object *sweep;
 int rotate;
 int interpolate;
 int sections;
 double glu_sampling_tolerance;
 int glu_display_mode;
} Sweep;

int Sweep_create_cb(Tcl_Interp *interp, int argc, char *argv[], void **c);
int Sweep_delete_cb(void *c);
int Sweep_copy_cb(void *source, void **dest);
int Sweep_draww_cb(struct Togl *togl, mops_object *o);
int Sweep_draws_cb(struct Togl *togl, mops_object *o);
int Sweep_wrib_cb(mops_object *o);
int Sweep_read_cb(FILE *fileptr, mops_custom_object *custom);
int Sweep_write_cb(FILE *fileptr, mops_custom_object *custom);
int Sweep_getprop_cb(Tcl_Interp *interp);
int Sweep_set_attrprop_tcmd(ClientData clientData, Tcl_Interp *interp,
			     int argc, char *argv[]);
int Sweep_get_attrprop_tcmd(ClientData clientData, Tcl_Interp *interp,
			     int argc, char *argv[]);

int Sweep_notify_cb(mops_object *o);
int Sweep_convert(Tcl_Interp *interp, mops_object *o);
int Sweep_convertNP_tcmd(ClientData clientData, Tcl_Interp *interp,
			 int argc, char *argv[]);

int Sweep_provide_cb(mops_object *o, void **target);


/* functions */
int
Sweep_Init(Tcl_Interp *interp)
{
 int mops_status = MOPS_OK;
 Tcl_HashEntry *entry;
 char fname[] = "Sweep_Init";

  mops_status = mops_register_custom_objecttype(
				     Sweep_name,
				     Sweep_create_cb,
				     Sweep_draws_cb, 
				     NULL, /* drawh */ /* no handles! */
				     Sweep_draww_cb,
				     Sweep_delete_cb,
				     Sweep_copy_cb,
				     Sweep_getprop_cb,
				     NULL, /* getpoint */ /* no single point
					                     editing! */
				     Sweep_wrib_cb,
				     Sweep_read_cb,
				     Sweep_write_cb,
				     Sweep_notify_cb
				     );

  if(mops_status)
    {
      mops_error(interp, mops_status, fname, NULL);
      return TCL_ERROR;
    }

  /* get custom-id from name */
  entry = Tcl_FindHashEntry(&mops_custom_id_ht, Sweep_name);

  if (entry != NULL)
    {
      Sweep_customid = Tcl_GetHashValue(entry);
    }
  else
    {
      mops_error(interp, MOPS_HASH, fname, NULL);
      return TCL_OK;
    }

  /* register converter */
  mops_status = mops_register_converter(interp,Sweep_name,Sweep_convert);

  /* register some C-functions as Tcl-Commands */
  /* they deal with the Sweep-Attributes Property GUI */
  Tcl_CreateCommand (interp, "Sweep_getSweepAttr",
		     Sweep_get_attrprop_tcmd,
		     (ClientData) NULL, (Tcl_CmdDeleteProc *) NULL);
  Tcl_CreateCommand (interp, "Sweep_setSweepAttr",
		     Sweep_set_attrprop_tcmd,
		     (ClientData) NULL, (Tcl_CmdDeleteProc *) NULL);

  /* another command, convert to a NURBPatch */
  Tcl_CreateCommand (interp, "sweepCNP", Sweep_convertNP_tcmd,
		     (ClientData) NULL, (Tcl_CmdDeleteProc *) NULL);

  /* register provider */
  mops_status = mops_register_provider(interp, Sweep_name,
				       MOPS_OTNURBPATCH, Sweep_provide_cb);

 return TCL_OK;
}


int
Sweep_create_cb(Tcl_Interp *interp, int argc, char *argv[], void **c)
{
 int mops_status = MOPS_OK;
 mops_list_object *sel = mops_current_selection;
 Sweep *new = NULL;
 mops_object *curve = NULL, *curve2 = NULL, *newo = NULL;
 int got_object = MOPS_FALSE, got_object2 = MOPS_FALSE;
 char fname[] = "Sweep_create";

  if(argc<4)
    {
      mops_error(interp, MOPS_ARGS, fname, "numsections");
      return MOPS_ERROR;
    }

  if(!sel)
    {
      return MOPS_NOSEL;
    }

  if(!sel->next)
    {
      mops_error(interp,MOPS_ERROR,fname,"select two NURB curves!");
      return MOPS_ERROR;
    }

  if(sel->object->type != MOPS_OTNURBCURVE)
    {
      mops_status = mops_get_object(sel->object, MOPS_OTNURBCURVE, &curve);
      if(curve)
	got_object = MOPS_TRUE;
    }
  else
    {
      curve = sel->object;
    }

  sel = sel->next;

  if(sel->object->type != MOPS_OTNURBCURVE)
    {
      mops_status = mops_get_object(sel->object, MOPS_OTNURBCURVE, &curve2);
      if(curve2)
	got_object2 = MOPS_TRUE;
    }
  else
    {
      curve2 = sel->object;
    }

  if(!curve || !curve2)
    {
      mops_error(interp, MOPS_ERROR, fname, "object is not a NURBCurve");
      if(got_object) mops_delete_object(curve);
      if(got_object2) mops_delete_object(curve2);

      return MOPS_ERROR;
    }


  if(!(new = calloc(1, sizeof(Sweep))))
    {
      if(got_object) mops_delete_object(curve);
      if(got_object2) mops_delete_object(curve2);

      return MOPS_OUTOFMEM;
    }

  new->glu_sampling_tolerance = 30.0;

  if(!(newo = calloc(1, sizeof(mops_object))))
    {
      if(got_object) mops_delete_object(curve);
      if(got_object2) mops_delete_object(curve2);

      return MOPS_OUTOFMEM;
    }

  mops_default_attributes(newo);
  newo->type = MOPS_OTNURBPATCH;
  new->sweep = newo;
  Tcl_GetInt(interp,argv[3], &(new->sections));
  new->rotate = MOPS_TRUE;

  mops_status = mops_sweep_nurb(curve, curve2, new->sections, MOPS_TRUE,
				(mops_nurbpatch_object **)&(newo->object));

  if(mops_status)
    {
      free(new); free(newo);
      if(got_object) mops_delete_object(curve);
      if(got_object2) mops_delete_object(curve2);

      return mops_status;
    }

  if(new->interpolate)
    {
      mops_status = mops_interpolate_npatchu((mops_nurbpatch_object *)
					     newo->object);

      if(mops_status)
	{
	  free(new); free(newo);
	  if(got_object) mops_delete_object(curve);
	  if(got_object2) mops_delete_object(curve2);

	  return mops_status;
	}
    }

  *c = new;

  if(got_object)
    mops_delete_object(curve);

  if(got_object2)
    mops_delete_object(curve2);

 return mops_status;
}


int
Sweep_delete_cb(void *c)
{
 int mops_status = MOPS_OK;
 Sweep *del = NULL;

  if(!c)
    return MOPS_OK;

  del = (Sweep *)c;

  if(del->sweep)
    mops_delete_object(del->sweep);
  
  free(c);

 return mops_status;
}


int
Sweep_copy_cb(void *source, void **dest)
{
 int mops_status = MOPS_OK;
 Sweep *new = NULL, *srcsweep = NULL;
 

  /* copy custom object */
  if(!(new = calloc(1,sizeof(Sweep))))
    return MOPS_OUTOFMEM;
  
  memcpy(new, source, sizeof(Sweep));
  /* danger, links point to original hierarchy! */

  srcsweep = (Sweep *)source;
  mops_status = mops_copy_object(srcsweep->sweep,
				    &(new->sweep));

  if(mops_status)
    { free(new); return mops_status; }

  *dest = new;

 return mops_status;
}


int
Sweep_draww_cb(struct Togl *togl, mops_object *o)
{
 int mops_status = MOPS_OK;
 int instance = MOPS_FALSE;
 GLfloat color[4] = {0.8, 0.75, 0.65, 1.0};
 Sweep *swp = NULL;
 mops_custom_object *custom = NULL;
 int reset_color = 0;

  if(!o)
    return MOPS_OK;

  if(o->type == MOPS_OTINSTANCE)
    {
      o = (mops_object *)o->object;

      instance = MOPS_TRUE;
    }

  if(!o->editor)
    return MOPS_OK;

  custom = o->object;
  if(!custom)
    return MOPS_OK;    

  swp = custom->object;
  if(!swp)
    return MOPS_OK;

  glPushMatrix();

  if(!instance)
    {
      mops_render_attributes(o);

      if(o->selected)
	{
	  glColor3f MOPS_CLSEL;
	  reset_color = 1;
	}
    }
  else
    {
      if(o->colr != -1 && o->colg != -1 && o->colb != -1)
	{
	  color[0] = o->colr/255.0;
	  color[1] = o->colg/255.0;
	  color[2] = o->colb/255.0;
	  color[3] = o->cola/255.0; 
	  glMaterialfv (GL_FRONT_AND_BACK, GL_AMBIENT_AND_DIFFUSE, color);
	}
    }

 
  /* draw sweep */
  if(swp->sweep)
    {
      mops_status = mops_draw_nurbpatch_c(togl, swp->sweep);
    }

  if(reset_color)
    glColor3f MOPS_CLUSEL;

  glPopMatrix();

 return mops_status;
}


int
Sweep_draws_cb(struct Togl *togl, mops_object *o)
{
 int mops_status = MOPS_OK;
 int instance = MOPS_FALSE;
 GLfloat color[4] = {0.8, 0.75, 0.65, 1.0};
 Sweep *swp = NULL;
 mops_custom_object *custom = NULL;

  if(!o)
    return MOPS_OK;

  if(o->type == MOPS_OTINSTANCE)
    {
      o = (mops_object *)o->object;

      instance = MOPS_TRUE;
    }

  if(!o->editor)
    return MOPS_OK;

  custom = o->object;
  if(!custom)
    return MOPS_OK;    

  swp = custom->object;
  if(!swp)
    return MOPS_OK;


  glPushMatrix();

  if(!instance)
    mops_render_attributes(o);
  else
    {
      if(o->colr != -1 && o->colg != -1 && o->colb != -1)
	{
	  color[0] = o->colr/255.0;
	  color[1] = o->colg/255.0;
	  color[2] = o->colb/255.0;
	  color[3] = o->cola/255.0; 
	  glMaterialfv (GL_FRONT_AND_BACK, GL_AMBIENT_AND_DIFFUSE, color);
	}
    }

   /* draw sweep */
  if(swp->sweep)
    {
      mops_status = mops_draw_nurbpatch(togl, swp->sweep);
    }

  glPopMatrix();

 return mops_status;
}


int
Sweep_wrib_cb(mops_object *o)
{
 int mops_status = MOPS_OK;
 Sweep *swp = NULL;
 mops_custom_object *custom = NULL;

  if(!o)
    return MOPS_OK;

  custom = o->object;
  if(!custom)
    return MOPS_OK;    

  swp = custom->object;
  if(!swp)
    return MOPS_OK;

  RiAttributeBegin();
  RiTransformBegin();
   mops_wrib_attributes(o);
   mops_wrib_shaders(o);
       
   mops_wrib_npatch(swp->sweep);

  RiTransformEnd();
  RiAttributeEnd(); 

 return mops_status;
}


int
Sweep_write_cb(FILE *fileptr, mops_custom_object *custom)
{
 int mops_status = MOPS_OK;
 Sweep *swp = NULL;

  if(!custom)
    return MOPS_OK;    
  swp = (Sweep *)(custom->object);
  if(!swp)
    return MOPS_OK;

  fprintf(fileptr,"%d\n",swp->sections);
  fprintf(fileptr,"%d\n",swp->rotate);
  fprintf(fileptr,"%g\n",swp->glu_sampling_tolerance);
  fprintf(fileptr,"%d\n",swp->glu_display_mode);
  fprintf(fileptr,"%d\n",swp->interpolate);

 return mops_status;
}

int
Sweep_read_cb(FILE *fileptr, mops_custom_object *custom)
{
 int mops_status = MOPS_OK;
 Sweep *swp = NULL;

  if(!custom)
    return MOPS_OK;
    
  if(!(custom->object = calloc(1,sizeof(Sweep))))
    return MOPS_OUTOFMEM;

  swp = (Sweep *)(custom->object);

  fscanf(fileptr,"%d\n",&swp->sections);
  fscanf(fileptr,"%d\n",&swp->rotate);

  if(mops_read_version > 5)
    {
      fscanf(fileptr,"%lg\n",&swp->glu_sampling_tolerance);
      fscanf(fileptr,"%d\n",&swp->glu_display_mode);
    }

  if(mops_read_version > 7)
    {
      fscanf(fileptr,"%d\n",&swp->interpolate);
    }

 return mops_status;
}

int
Sweep_notify_cb(mops_object *o)
{
 int mops_status = MOPS_OK;
 Sweep *swp = NULL;
 mops_object *down = NULL, *c = NULL, *c2 = NULL, *newo = NULL;
 mops_custom_object *custom = NULL;
 double tolerance = 30.0;
 int mode = 0;
 int got_object = MOPS_FALSE, got_object2 = MOPS_FALSE;

  if(!o)
    return MOPS_OK;

  custom = o->object;
  if(!custom)
    return MOPS_OK;    

  swp = custom->object;
  if(!swp)
    return MOPS_OK;

  /* ok, something changed below, we have to rebuild the sweep */

  /* save some GLU parameters */
  tolerance = swp->glu_sampling_tolerance;
  mode = swp->glu_display_mode;

  if(swp->sweep)
    {
      mops_delete_object(swp->sweep);
      swp->sweep = NULL;
    }

  down = o->down;

  if(!down)
    return MOPS_OK;

  if(!down->next)
    return MOPS_OK;

  if(down->type == MOPS_OTNURBCURVE)
    {
      c = down;
    }
  else
    {
      mops_status = mops_get_object(down, MOPS_OTNURBCURVE, &c);
      if(c)
	got_object = MOPS_TRUE;
    }

  down = down->next;
  if(down->type == MOPS_OTNURBCURVE)
    {
      c2 = down;
    }
  else
    {
      mops_status = mops_get_object(down, MOPS_OTNURBCURVE, &c2);
      if(c2)
	got_object2 = MOPS_TRUE;
    }

  if(!c || !c2)
    return MOPS_OK;

  if(!(newo = calloc(1, sizeof(mops_object))))
    {
      if(got_object) mops_delete_object(c);
      if(got_object2) mops_delete_object(c2);

      return MOPS_OUTOFMEM;
    }

  mops_default_attributes(newo);
  newo->type = MOPS_OTNURBPATCH;
  swp->sweep = newo;

  mops_sweep_nurb(c, c2, swp->sections, swp->rotate,
		  (mops_nurbpatch_object **)&(newo->object));
  
  if(swp->sweep)
    {
      ((mops_nurbpatch_object *)swp->sweep->object)->
	glu_sampling_tolerance = tolerance;
      ((mops_nurbpatch_object *)swp->sweep->object)->
	glu_display_mode = mode;

      if(swp->interpolate)
	mops_status = mops_interpolate_npatchu((mops_nurbpatch_object *)
					       newo->object);

    }

  if(got_object)
    mops_delete_object(c);

  if(got_object2)
    mops_delete_object(c2);

 return mops_status;
}


/************************/
/************************/
/************************/

int
Sweep_getprop_cb(Tcl_Interp *interp)
{
 int mops_status = MOPS_OK;

  /* Surface-Shader */
  mops_status = mops_add_surfshader_property(interp);

  /* Displacement-Shader */
  mops_status = mops_add_dispshader_property(interp);

  /* Interior-Shader */
  mops_status = mops_add_intshader_property(interp);

  /* Exterior-Shader */
  mops_status = mops_add_extshader_property(interp);

  /* Sweep-Parameters */
  Tcl_SetVar(interp,"propertyList","SweepAttr", TCL_APPEND_VALUE |\
	     TCL_LIST_ELEMENT | TCL_LEAVE_ERR_MSG | TCL_GLOBAL_ONLY);

  Tcl_SetVar(interp,"propertyProcList","Sweep_buildSweepAttrGUI",
	     TCL_APPEND_VALUE | TCL_LIST_ELEMENT |\
	     TCL_LEAVE_ERR_MSG | TCL_GLOBAL_ONLY);

  Tcl_SetVar(interp,"propertyGetProcList","Sweep_getSweepAttr",
	     TCL_APPEND_VALUE | TCL_LIST_ELEMENT |\
	     TCL_LEAVE_ERR_MSG | TCL_GLOBAL_ONLY);

  Tcl_SetVar(interp,"propertySetProcList","Sweep_setSweepAttr",
	     TCL_APPEND_VALUE | TCL_LIST_ELEMENT |\
	     TCL_LEAVE_ERR_MSG | TCL_GLOBAL_ONLY);

  Tcl_SetVar(interp,"propertyDataList","SweepAttrData",
	     TCL_APPEND_VALUE | TCL_LIST_ELEMENT |\
	     TCL_LEAVE_ERR_MSG | TCL_GLOBAL_ONLY);

 return TCL_OK;
}


/* Tcl -> C! */
int
Sweep_set_attrprop_tcmd(ClientData clientData, Tcl_Interp *interp,
			 int argc, char *argv[])
{
 mops_object *o = NULL;
 mops_list_object *sel = mops_current_selection;
 mops_custom_object *custom = NULL;
 Sweep *swp = NULL;
 char *n1="SweepAttrData";
 Tcl_Obj *to = NULL, *toa = NULL, *ton = NULL;

  if(!sel) /* oops? */
    return TCL_OK;

  o = sel->object; 
  if(!o)
    return TCL_OK;

  custom = o->object;
  if(!custom)
    return TCL_OK;

  swp = (Sweep *) custom->object;


  toa = Tcl_NewStringObj(n1,-1);
  ton = Tcl_NewStringObj("Sections",-1);
  to = Tcl_ObjGetVar2(interp,toa,ton,TCL_LEAVE_ERR_MSG | TCL_GLOBAL_ONLY);
  Tcl_GetIntFromObj(interp,to, &swp->sections);

  Tcl_SetStringObj(ton,"Rotate",-1);
  to = Tcl_ObjGetVar2(interp,toa,ton,TCL_LEAVE_ERR_MSG | TCL_GLOBAL_ONLY);
  Tcl_GetIntFromObj(interp,to, &swp->rotate);

  Tcl_SetStringObj(ton,"Interpolate",-1);
  to = Tcl_ObjGetVar2(interp,toa,ton,TCL_LEAVE_ERR_MSG | TCL_GLOBAL_ONLY);
  Tcl_GetIntFromObj(interp,to, &swp->interpolate);

  Tcl_SetStringObj(ton,"Tolerance",-1);
  to = Tcl_ObjGetVar2(interp,toa,ton,TCL_LEAVE_ERR_MSG | TCL_GLOBAL_ONLY);
  Tcl_GetDoubleFromObj(interp,to, &(swp->glu_sampling_tolerance));

  Tcl_SetStringObj(ton,"Display",-1);
  to = Tcl_ObjGetVar2(interp,toa,ton,TCL_LEAVE_ERR_MSG | TCL_GLOBAL_ONLY);
  Tcl_GetIntFromObj(interp,to, &(swp->glu_display_mode));

  Tcl_IncrRefCount(toa);Tcl_DecrRefCount(toa);
  Tcl_IncrRefCount(ton);Tcl_DecrRefCount(ton);

  Sweep_notify_cb(o);

 return TCL_OK;
}

/* C -> Tcl! */
int
Sweep_get_attrprop_tcmd(ClientData clientData, Tcl_Interp *interp,
			 int argc, char *argv[])
{
 mops_object *o = NULL;
 mops_list_object *sel = mops_current_selection;
 mops_custom_object *custom = NULL;
 Sweep *swp = NULL;
 char *n1="SweepAttrData";
 Tcl_Obj *to = NULL, *toa = NULL, *ton = NULL;

  if(!sel) /* oops? */
    return TCL_OK;

  o = sel->object; 
  if(!o)
    return TCL_OK;

  custom = o->object;
  if(!custom)
    return TCL_OK;

  swp = (Sweep *) custom->object;

  toa = Tcl_NewStringObj(n1,-1);
  ton = Tcl_NewStringObj("Sections",-1);
  to = Tcl_NewIntObj(swp->sections);
  Tcl_ObjSetVar2(interp,toa,ton,to,TCL_LEAVE_ERR_MSG |
		 TCL_GLOBAL_ONLY);

  Tcl_SetStringObj(ton,"Rotate",-1);
  to = Tcl_NewIntObj(swp->rotate);
  Tcl_ObjSetVar2(interp,toa,ton,to,TCL_LEAVE_ERR_MSG |
		 TCL_GLOBAL_ONLY);

  Tcl_SetStringObj(ton,"Interpolate",-1);
  to = Tcl_NewIntObj(swp->interpolate);
  Tcl_ObjSetVar2(interp,toa,ton,to,TCL_LEAVE_ERR_MSG |
		 TCL_GLOBAL_ONLY);

  Tcl_SetStringObj(ton,"Tolerance",-1);
  to = Tcl_NewDoubleObj(swp->glu_sampling_tolerance);
  Tcl_ObjSetVar2(interp,toa,ton,to,TCL_LEAVE_ERR_MSG | TCL_GLOBAL_ONLY);

  Tcl_SetStringObj(ton,"Display",-1);
  to = Tcl_NewIntObj(swp->glu_display_mode);
  Tcl_ObjSetVar2(interp,toa,ton,to,TCL_LEAVE_ERR_MSG | TCL_GLOBAL_ONLY);

  Tcl_IncrRefCount(toa);Tcl_DecrRefCount(toa);
  Tcl_IncrRefCount(ton);Tcl_DecrRefCount(ton);

 return TCL_OK;
}


int
Sweep_convert(Tcl_Interp *interp, mops_object *o)
{
 int mops_status = MOPS_OK;
 mops_custom_object *custom = NULL;
 Sweep *swp = NULL;
 mops_object *new = NULL;
 char fname[] = "sweep_convertNP";

  if(o->type != MOPS_OTCUSTOM)
    return MOPS_ERROR;

  custom = (mops_custom_object*)o->object;
  if(!custom)
    return MOPS_ERROR;

  if(custom->type != Sweep_customid)
    return MOPS_ERROR;

  swp = (Sweep *) custom->object;

  mops_status = mops_copy_object(swp->sweep, &new);

  if(mops_status)
    {
      mops_error(interp,mops_status,fname,NULL);
      free(new);
      return mops_status;
    }

  mops_status = mops_copy_attributes(o, new);
  mops_status = mops_link_new_object(new);

 return mops_status;
}


int
Sweep_convertNP_tcmd(ClientData clientData, Tcl_Interp *interp,
		     int argc, char *argv[])
{
 int mops_status = MOPS_OK;
 mops_object *o = NULL;
 mops_list_object *sel = mops_current_selection;
 mops_custom_object *custom = NULL;
 char fname[] = "sweep_convertNP";

  if(!sel) /* oops? */
    return TCL_OK;

  o = sel->object; 
  if(!o)
    return TCL_OK;

  if(o->type != MOPS_OTCUSTOM)
    return TCL_OK;

  custom = (mops_custom_object*)o->object;
  if(!custom)
    return TCL_OK;

  if(custom->type != Sweep_customid)
    return TCL_OK;

  mops_status = Sweep_convert(interp, o);
  if(mops_status)
    {
      mops_error(interp,MOPS_ERROR,fname,"converter failed");
      return TCL_OK;
    }


 return TCL_OK;
}


int
Sweep_provide_cb(mops_object *o, void **target)
{
 int mops_status = MOPS_OK;
 Sweep *swp = NULL;
 mops_custom_object *custom = NULL;

  custom = (mops_custom_object *)o->object;

  swp = (Sweep *)custom->object;

  mops_status = Sweep_notify_cb(o);

  if(mops_status)
    return mops_status;

  mops_copy_attributes(o, swp->sweep);

  *target = swp->sweep;

 return mops_status;
}
